////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		LayerSeparate.cpp
// File Function:	API of layer separation for Arabic documents
//
//				Developed by: Yefeng Zheng
//			   First created: April 2004
//			University of Maryland, College Park
//					All Right Reserved 
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "Tools.h"
#include "ArabicTextIDAPI.h"
#include "WordSegment.h"
#include "Output.h"
#include "Tree.h"

int TextLineIdentify( WORD_ZONE *pWord, int nWord, int &nTextLineHeight, Tree DecisionTree_Short, Tree DecisionTree_Long );
int ExtractWordSizeFeature( WORD_ZONE *pWord, int nWord, int nCurWord, int nAverHeight, double *Feature, int &nFeature );

CWordSegment	WordSegment;	//Word segmentation class
CImage			m_Img;			//Image class

CArabicLayerSeparate::CArabicLayerSeparate()
{ 
	m_pWord = NULL;
	m_nWord = 0;
	
	Tree_Line_Short = NULL;
	Tree_Line_Long = NULL;
	Tree_Word_Size = NULL;
	return; 
}

CArabicLayerSeparate::~CArabicLayerSeparate()
{
	if( m_pWord != NULL )
	{
		free( m_pWord );
		m_pWord = NULL;
	}
	m_nWord = 0;

	if( Tree_Line_Short != NULL )
	{
		FreeTree( (Tree)Tree_Line_Short );
		Tree_Line_Short = NULL;
	}
	if( Tree_Line_Long != NULL )
	{
		FreeTree( (Tree)Tree_Line_Long );
		Tree_Line_Long = NULL;
	}
	if( Tree_Word_Size != NULL )
	{
		FreeTree( (Tree)Tree_Word_Size );
		Tree_Word_Size = NULL;
	}
	return;
}

/******************************************************************************
/*	Name:		TISetImage
/*	Function:	Set image
/*	Parameter:	fnImg -- Image file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISetImage(char *fnImg)
{
	if( m_Img.Read( fnImg ) != 0 )
		return -1;
	return 0;
}

/******************************************************************************
/*	Name:		TISetImage
/*	Function:	Set image from memory
/*	Parameter:	pImgData	-- Image data
/*				w			-- Image width
/*				h			-- Image height
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISetImage( BYTE *pImgData, int w, int h )
{
	if( m_Img.Initialize( pImgData, w, (w+7)/8, h, DIB_1BIT, FALSE ) !=0 )
		return -1;
	return 0;
}

/******************************************************************************
/*	Name:		TISetClassifier
/*	Function:	Set classifier
/*	Parameter:	fnShortTextLineClassifier -- Trained classifier for short text lines
/*				fnLongTextLineClassifier  -- Trained classifier for long text lines
/*				fnWordClassifier		  -- Trained classifier for words
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISetClassifier( char *fnShortTextLineClassifier, char *fnLongTextLineClassifier,
									  char *fnWordClassifier)
{
	if( ReadTree( fnShortTextLineClassifier, (Tree&)Tree_Line_Short ) != 0 )
		return -1;
	if( ReadTree( fnLongTextLineClassifier, (Tree&)Tree_Line_Long ) != 0 )
		return -1;
	if( ReadTree( fnWordClassifier, (Tree&)Tree_Word_Size ) != 0 )
		return -1;

	return 0;
}

/******************************************************************************
/*	Name:		TIWordSegmentation
/*	Function:	Word segmentation
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TIWordSegmentation( )
{
	WordSegment.SegmentWord(m_Img);

	if( m_pWord != NULL )
		free( m_pWord );

	m_nWord = WordSegment.m_nWord;
	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*m_nWord );
	for( int i=0; i<WordSegment.m_nWord; i++ )
	{
		m_pWord[i].x = WordSegment.m_pWord[i].x;
		m_pWord[i].y = WordSegment.m_pWord[i].y;
		m_pWord[i].w = WordSegment.m_pWord[i].w;
		m_pWord[i].h = WordSegment.m_pWord[i].h;
		m_pWord[i].nType = WordSegment.m_pWord[i].nType;
		m_pWord[i].nIndex = i;
	}
	return 0;
}

/******************************************************************************
/*	Name:		TIClassification
/*	Function:	Classification
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TIClassification( )
{
	int i;
	double	Feature[200];
	int		nFeature;
	int		nTextLineHeight;

	TextLineIdentify( m_pWord, m_nWord, nTextLineHeight, (Tree)Tree_Line_Short, (Tree)Tree_Line_Long );

	for( i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_UNDETERMINED )
		{
			if( ExtractWordSizeFeature( m_pWord, m_nWord, i, nTextLineHeight, Feature, nFeature ) != 0 )
				return -1;
			m_pWord[i].nType = ClassTree( Feature, nFeature, (Tree)Tree_Word_Size ) + 1;
		}
		else if( m_pWord[i].nType == TYPE_OTHERS )
			m_pWord[i].nType = TYPE_HANDWRITING;
		
	}
	//Re-set classification results
	for( i=0; i<m_nWord; i++ )
	{
		for( int j=0; j<WordSegment.m_nWord; j++ )
		{ 
			if( m_pWord[i].x == WordSegment.m_pWord[j].x &&
			  m_pWord[i].y == WordSegment.m_pWord[j].y &&
			  m_pWord[i].w == WordSegment.m_pWord[j].w &&
			  m_pWord[i].h == WordSegment.m_pWord[j].h )
			{
				WordSegment.m_pWord[j].nType = m_pWord[i].nType;
				break;
			}
		}
	}
				
	return 0;
}

/******************************************************************************
/*	Name:		TISetWord
/*	Function:	This function provides a way for users can use his/her own module
/*				to perform segmentation, classification of logo, figure, etc., and 
/*				then fill information to the structure of WORD_ZONE.
/*	Parameter:	pWord   -- Segmented word zone
/*				nWord   -- Number of word zone
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISetWord(WORD_ZONE *pWord, int nWord)
{
	if( m_pWord != NULL )
		free( m_pWord );
	m_nWord = 0;

	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*nWord );
	for( int i=0; i<nWord; i++ )
	{
		m_pWord[i] = pWord[i];
		m_pWord[i].nIndex = -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		TISetWord
/*	Function:	This function provides a way for users can use his/her own module
/*				to perform segmentation, classification of logo, figure, etc., and 
/*				then fill information to the structure of WORD_ZONE.
/*				Similar to the previous function, but the custom's segmentation
/*				result is stored in ZONE file.
/*	Parameter:	fnSegment -- File name of the segmenation result
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISetWord(char *fnCustom)
{
	if( m_pWord != NULL )
		free( m_pWord );
	m_nWord = 0;

	ZONE *pZone;
	int  nZone;
	ReadZone( pZone, nZone, fnCustom );
	m_nWord = nZone;
	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*nZone);
	for(int i=0; i<nZone; i++ )
	{
		m_pWord[i].x = pZone[i].x;
		m_pWord[i].y = pZone[i].y;
		m_pWord[i].w = pZone[i].w;
		m_pWord[i].h = pZone[i].h;
		m_pWord[i].nType = pZone[i].nType;
	}
	free( pZone );
	return 0;
}

/******************************************************************************
/*	Name:		TIGetWord
/*	Function:	Get the segmented and classified word zones.
/*	Parameter:	pWord   -- Segmented word zone
/*				nWord   -- Number of word zone
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TIGetWord(WORD_ZONE *&pWord, int &nWord )
{
	pWord = m_pWord;
	nWord = m_nWord;
	return 0;
}

/******************************************************************************
/*	Name:		TIOutputZone
/*	Function:	Output result to ZONE file
/*	Parameter:	fnZone -- Output file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TIOutputZone( char *fnZone )
{
	FILE *fpZone = fopen( fnZone, "wt" );
	if( fpZone == NULL )
		return -1;
	fprintf( fpZone, "Total Zones: %d\n", m_nWord );
	for( int i=0; i<m_nWord; i++ )
	{
		fprintf( fpZone, "%5d %5d %5d %5d    %d ", m_pWord[i].x, m_pWord[i].y, m_pWord[i].w, m_pWord[i].h, m_pWord[i].nType );
		fprintf( fpZone, "\n" );
	}
	fclose( fpZone);
	return 0;
}

/******************************************************************************
/*	Name:		TIGetLayer
/*	Function:	Get the specified content to an image
/*	Parameter:	pImgData	-- Image of the layer
/*				w			-- Image width
/*				h			-- Image height
/*				nType       -- Layer type, 1 -- Printed text, 2 -- Handwriting, 3 -- Noise
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TIGetLayer( BYTE *&pImgData, int &w, int &h, int nType )
{
	CImage	ImgOut;
	w = m_Img.GetWidth();
	h = m_Img.GetHeight();
	ImgOut.Initialize( w, h, 1, 1 );
	switch( nType )
	{
		case TYPE_PRINT_TEXT:
			WordSegment.OutputImage( ImgOut, TYPE_PRINT_TEXT );
			break;
		case TYPE_HANDWRITING:
			WordSegment.OutputImage( ImgOut, TYPE_HANDWRITING );
			break;
		case TYPE_NOISE:
			WordSegment.OutputImage( ImgOut, TYPE_NOISE );
			break;
		default:
			return -1;
	}
	int wb = ImgOut.GetLineSizeInBytes();
	BYTE *pImg = ImgOut.LockRawImg();
	int ww = (w+7)/8;
	pImgData = (BYTE*)malloc( sizeof(BYTE)*ww*h);
	for( int i=0; i<h; i++ )
		memcpy( pImgData+i*ww, pImg+(h-1-i)*wb, ww );

/*	CImage TestImg;
	TestImg.Initialize( pImgData, w, ww, h, DIB_1BIT, FALSE );
	TestImg.Write( "ttt.tif" );
*/	return 0;
}

/******************************************************************************
/*	Name:		TISaveLayer
/*	Function:	Output the specified content to an image
/*	Parameter:	fnLayerImg	-- Image file name of the image
/*				nType		-- Layer type
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CArabicLayerSeparate::TISaveLayer( char *fnLayerImg, int nType )
{
	CImage	ImgOut;
	int w = m_Img.GetWidth();
	int h = m_Img.GetHeight();
	ImgOut.Initialize( w, h, 1, 1 );
	
	switch( nType )
	{
		case TYPE_PRINT_TEXT:
			WordSegment.OutputImage( ImgOut, TYPE_PRINT_TEXT );
			break;
		case TYPE_HANDWRITING:
			WordSegment.OutputImage( ImgOut, TYPE_HANDWRITING );
			break;
		case TYPE_NOISE:
			WordSegment.OutputImage( ImgOut, TYPE_NOISE );
			break;
		default:
			return -1;
	}
	ImgOut.Write( fnLayerImg );
	return 0;
}
