////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		TextLine.cpp
// File Function:	Handwriting/machine printed text identification on text line level
//
//				Developed by: Zheng Yefeng
//			   First created: March 2004
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ArabicTextIDAPI.h"
#include "Tools.h"
#include "Tree.h"
#include <math.h>

#define	MIN_WORDS_SHORT_LINE	3
#define	MIN_WORDS_LONG_LINE		5

//Structure of a text line
typedef	struct TEXT_LINE
{
	#define		MAX_WORD_PER_LINE	200
	RECT	rect;			//Bounding box of a text line
	int		WordIndex[MAX_WORD_PER_LINE];
	int		nWord;			//Number of words inside the text line
	int		nHandWord;		//Number of handwritten word (get from groundtruth)
	int		nPrintWord;		//Number of printed word (get from groundtruth)
	int		nTrueType;		//Type of a text line. It may be
							//TYPE_PRINT_TEXT, TYPE_HANDWRITING, TYPE_NOISE (containing both handwriting and printed text)
	int		nType;			//Text line classification result. It may be
							//TYPE_PRINT_TEXT, TYPE_HANDWRITING, TYPE_NOISE (filtered by size without being sent to a classifier)
}TEXT_LINE;

/******************************************************************************************
/*	Name:		SameTextLine
/*	Function:	Test if two rectangles (representing words, or sub text lines) 
/*				may be lying at the same text line
/*				We do not know the type of these two rectangles.
/*	Parameter:	p1  -- Rectangle one
/*				p2  -- Rectangle two
/*	Return:		TRUE -- They can be merged.
/*******************************************************************************************/
BOOL SameTextLine( RECT p1, RECT p2 )
{
	int TH_Hor = 35;
	int nMaxLeft = max( p1.left, p2.left );
	int nMinRight = min( p1.right, p2.right );
	int nHorDist = nMaxLeft-nMinRight;
	if( nHorDist > TH_Hor ||		//Too far away in horizontal
		nHorDist < 0 )				//Overlap in horizontal
		return FALSE;

	int nDeltaTop = abs( p1.top - p2.top );
	int nDeltaBottom = abs( p1.bottom - p2.bottom );

	int nMaxTop = max(p1.top, p2.top );
	int nMinBottom = min( p1.bottom, p2.bottom );
	int nOverlap = nMinBottom - nMaxTop;

	//Set threshold of vertical overlapping
	int TH_Ver;
	if( min( nDeltaTop, nDeltaBottom ) <= 1 )
		TH_Ver = (int)( 0.3*max( p1.bottom-p1.top, p2.bottom-p2.top ) );
	else
		TH_Ver = (int)( 0.5*max( p1.bottom-p1.top, p2.bottom-p2.top ) );

	if( nOverlap <= TH_Ver )
		return FALSE;
	return TRUE;
}

/******************************************************************************************
/*	Name:		SamePrintTextLine
/*	Function:	Test if two rectangles (representing words, or sub text lines) 
/*				may be lying at the same text line.
/*				It is used after text line classification. Only two printed text line, or
/*				a printed text line with a unknown type text line should be considered.
/*				The mergeing thresholds are much lower than the previous function
/*	Parameter:	L1  -- Text line one
/*				L2  -- Text line two
/*	Return:		TRUE -- They can be merged.
/*******************************************************************************************/
BOOL SamePrintTextLine( TEXT_LINE L1, TEXT_LINE L2, int nTextLineHeight )
{
	if( L1.nType == TYPE_HANDWRITING || L2.nType == TYPE_HANDWRITING )
		return FALSE;
	if( L1.nType != TYPE_PRINT_TEXT && L2.nType != TYPE_PRINT_TEXT )
		return FALSE;

	int TH_Hor = 3*nTextLineHeight;
	int nMaxLeft = max( L1.rect.left, L2.rect.left );
	int nMinRight = min( L1.rect.right, L2.rect.right );
	int nHorDist = nMaxLeft-nMinRight;
	if( nHorDist > TH_Hor )	//Too far away in horizontal
		return FALSE;

	int nMaxTop = max(L1.rect.top, L2.rect.top );
	int nMinBottom = min( L1.rect.bottom, L2.rect.bottom );
	int nOverlap = nMinBottom - nMaxTop;
	int TH_Ver;
	if( L1.nType == TYPE_PRINT_TEXT && L2.nType == TYPE_OTHERS )
		TH_Ver =  L2.rect.bottom - L2.rect.top - 2;
	else if( L1.nType == TYPE_OTHERS && L2.nType == TYPE_PRINT_TEXT )
		TH_Ver = L1.rect.bottom - L1.rect.top - 2;
	else
		TH_Ver = (int)( 0.8*min(L1.rect.bottom - L1.rect.top, L2.rect.bottom - L2.rect.top) );

	if( nOverlap < TH_Ver ) //Overlap in horizontal is too small
		return FALSE;
	return TRUE;
}

/******************************************************************************************
/*	Name:		RemoveTextLine
/*	Function:	Remove a text line
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*				nRemove   -- The current one needs removal
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int RemoveTextLine( TEXT_LINE *pTextLine, int &nTextLine, int nRemove )
{
	for( int k=nRemove; k<nTextLine-1; k++ )
		pTextLine[k] = pTextLine[k+1];
	nTextLine--;
	return 0;
}

/******************************************************************************************
/*	Name:		MergeTextLine
/*	Function:	Merge text lines, and update the parameters of the result text line
/*	Parameter:	L1  -- Text line one, containing the result after merging
/*				L2  -- Text line two
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int	MergeTextLine( TEXT_LINE &L1, TEXT_LINE L2 )
{
	//Update the list of words contained by the text line
	if( L1.nWord + L2.nWord > MAX_WORD_PER_LINE )
		return -1;
	for( int i=0; i<L2.nWord; i++ )
		L1.WordIndex[i+L1.nWord] = L2.WordIndex[i];
	L1.nWord += L2.nWord;

	//Update the bounding box
	L1.rect.left = min( L1.rect.left, L2.rect.left );
	L1.rect.right = max( L1.rect.right, L2.rect.right );
	L1.rect.top = min( L1.rect.top, L2.rect.top );
	L1.rect.bottom = max( L1.rect.bottom, L2.rect.bottom );

	return 0;
}

/******************************************************************************************
/*	Name:		MergeTextLine
/*	Function:	Merge text lines
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int MergeTextLine( TEXT_LINE *pTextLine, int &nTextLine )
{
	int i, j;

	//Merging text lines
	i=0;
	while( i<nTextLine )
	{
		for( j=i+1; j<nTextLine; j++ )
		{
			if( SameTextLine( pTextLine[i].rect, pTextLine[j].rect ) )
				break;
		}
		if( j<nTextLine )
		{
			MergeTextLine( pTextLine[i], pTextLine[j] );
			RemoveTextLine( pTextLine, nTextLine, j );
		}
		else
			i++;
	}


/*	i=0;
	while( i<nTextLine )
	{
		for( j=0; j<nTextLine; j++ )
		{
			if( j==i )	continue;
			if( SameTextLine( pTextLine[i].rect, pTextLine[j].rect ) )
				break;
		}
		if( j<nTextLine )
		{
			MergeTextLine( pTextLine[i], pTextLine[j] );
			for( k=j; k<nTextLine-1; k++ )
				pTextLine[k] = pTextLine[k+1];
			nTextLine--;
			if( j<i )	i--;
		}
		else
			i++;
	}
*/
	return 0;
}

/******************************************************************************************
/*	Name:		MergeTextLineInSide
/*	Function:	If a small text line is bouned by a large one, merge it into the large text line
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int MergeTextLineInSide( TEXT_LINE *pTextLine, int &nTextLine )
{
	//Filter text lines with one or two words
	int	i, j;
	i = 0;
	while( i<nTextLine )
	{
		if( pTextLine[i].nWord !=1 )	//not a small text line
		{
			i++;
			continue;
		}
		//Test if it is bounded inside another text line
		for( j=0; j<nTextLine; j++ )
		{
			if( j == i )	continue;
			if( pTextLine[j].nWord <= 2 )	continue;
			if( !(pTextLine[i].rect.left >= pTextLine[j].rect.left && 
				pTextLine[i].rect.right <= pTextLine[j].rect.right) )
				continue;
			
			if( !(pTextLine[i].rect.top >= pTextLine[j].rect.top &&
				 pTextLine[i].rect.bottom <= pTextLine[j].rect.bottom) )
				 continue;

			break;
		}
		if( j<nTextLine )
		{
			MergeTextLine( pTextLine[j], pTextLine[i] );
			RemoveTextLine( pTextLine, nTextLine, i );
		}
		else
			i++;
	}
	return 0;
}

/******************************************************************************************
/*	Name:		SplitTextLine
/*	Function:	Split text lines
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*				pWord	  -- Segmented words
/*				nCutLine  -- The text line need to be cut
/*				nCutWord  -- The cutting point inside the text line
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int SplitTextLine( TEXT_LINE *pTextLine, int &nTextLine, WORD_ZONE *pWord, int nCutLine, int nCutWord )
{
	RECT rcLeft, rcRight;
	rcLeft.left = rcLeft.top = rcRight.left = rcRight.top = 100000;
	rcLeft.right = rcLeft.bottom = rcRight.right = rcRight.bottom = -1;
	
	//The left sub text line
	pTextLine[nTextLine].nWord = nCutWord+1;
	for( int i=0; i<=nCutWord; i++ )
	{
		int nIndex = pTextLine[nCutLine].WordIndex[i];

		rcLeft.left = min( rcLeft.left, pWord[nIndex].x );
		rcLeft.top = min( rcLeft.top, pWord[nIndex].y );
		rcLeft.right = max( rcLeft.right, pWord[nIndex].x + pWord[nIndex].w );
		rcLeft.bottom = max( rcLeft.bottom, pWord[nIndex].y + pWord[nIndex].h );

		pTextLine[nTextLine].WordIndex[i] = nIndex;
	}
	pTextLine[nTextLine].rect = rcLeft;

	//The right sub text line
	pTextLine[nTextLine+1].nWord = pTextLine[nCutLine].nWord - pTextLine[nTextLine].nWord;
	for( i=nCutWord+1; i<pTextLine[nCutLine].nWord; i++ )
	{
		int nIndex = pTextLine[nCutLine].WordIndex[i];

		rcRight.left = min( rcRight.left, pWord[nIndex].x );
		rcRight.top = min( rcRight.top, pWord[nIndex].y );
		rcRight.right = max( rcRight.right, pWord[nIndex].x + pWord[nIndex].w );
		rcRight.bottom = max( rcRight.bottom, pWord[nIndex].y + pWord[nIndex].h );

		pTextLine[nTextLine+1].WordIndex[i-nCutWord] = nIndex;
	}
	pTextLine[nTextLine+1].rect = rcRight;

	nTextLine += 2;
	RemoveTextLine( pTextLine, nTextLine, nCutLine );

	return 0;
}

/******************************************************************************************
/*	Name:		SplitTextLine
/*	Function:	Split text lines
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*				pWord	  -- Segmented words
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int	SplitTextLine( TEXT_LINE *pTextLine, int &nTextLine, WORD_ZONE *pWord, int nTLHeight )
{
	int nLeftTop[1000], nLeftBottom[1000];
	int	nRightTop[1000], nRightBottom[1000];
	int i, j;
	i = 0;
	while( i<nTextLine )
	{
		//Do not cut short text lines
		if( pTextLine[i].nWord <= MIN_WORDS_LONG_LINE )
		{
			i++;
			continue;
		}
		//Only cut text lines with large height
		int nTextLineHeight = pTextLine[i].rect.bottom - pTextLine[i].rect.top;
		if( nTextLineHeight < nTLHeight+10 )	
		{
			i++;
			continue;
		}
		int nWord = pTextLine[i].nWord;
		if( nWord >= 1000 )	//Too many words inside the text line. It is impossible.
		{
			i++;
			continue;
		}

		//Calculate the bounding box of the left and right sub text lines.
		int nIndex = pTextLine[i].WordIndex[0];
		nLeftTop[0] = pWord[nIndex].y;
		nLeftBottom[0] = pWord[nIndex].y + pWord[nIndex].h;

		nIndex = pTextLine[i].WordIndex[nWord-1];
		nRightTop[0] = pWord[nIndex].y;
		nRightBottom[0] = pWord[nIndex].y + pWord[nIndex].h;

		for( j=1; j<nWord; j++ )
		{
			nIndex = pTextLine[i].WordIndex[j];
			nLeftTop[j] = min( nLeftTop[j-1], pWord[nIndex].y );
			nLeftBottom[j] = max( nLeftBottom[j-1], pWord[nIndex].y + pWord[nIndex].h );
			
			nIndex = pTextLine[i].WordIndex[nWord-1-j];
			nRightTop[j] = min( nRightTop[j-1], pWord[nIndex].y );
			nRightBottom[j] = max( nRightBottom[j-1], pWord[nIndex].y + pWord[nIndex].h );
		}

		//Searching the cut point that after cutting the homogenity of two sub text lines increase
		int	nGap[1000];
		int nTGap, nBGap;
		for( j=0; j<nWord-1; j++ )
		{
			//Calculate the homogenity of the left sub text line
			nTGap =	nBGap = 0;
			for( int k=0; k<=j; k++ )
			{
				nIndex = pTextLine[i].WordIndex[k];
				nTGap += pWord[nIndex].y - nLeftTop[j];
				nBGap += nLeftBottom[j] - (pWord[nIndex].y + pWord[nIndex].h);
			}
			nGap[j] = nTGap+nBGap;

			//Calculate the homogenity of the right sub text line
			nTGap = nBGap = 0;
			for( k=j+1; k<nWord; k++ )
			{
				nIndex = pTextLine[i].WordIndex[k];
				nTGap += pWord[nIndex].y - nRightTop[nWord-2-j];
				nBGap += nRightBottom[nWord-2-j] - (pWord[nIndex].y + pWord[nIndex].h);
			}
			nGap[j] += nTGap + nBGap;
		}
		nTGap =	nBGap = 0;
		for( int k=0; k<nWord; k++ )
		{
			nIndex = pTextLine[i].WordIndex[k];
			nTGap += pWord[nIndex].y - pTextLine[i].rect.top;
			nBGap += pTextLine[i].rect.bottom - (pWord[nIndex].y + pWord[nIndex].h);
		}
		int nTextLineGap = nTGap + nBGap;

		int nCut;
		int nMinGap = nTextLineGap;
		//Short text lines
		for( j=0; j<nWord-1; j++ )
		{
			if( nGap[j] < nMinGap  )
			{
				nMinGap = nGap[j];
				nCut = j;
			}
		}
		
		//Decide to cut the text line or not
		if( nTextLineGap - nMinGap >= nTextLineGap/3 )
			SplitTextLine( pTextLine, nTextLine, pWord, i, nCut );
		i++;
	}
	return 0;
}
/******************************************************************************************
/*	Name:		LabelSmallTextLine
/*	Function:	Label small text lines as TYPE_OTHERS
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int LabelSmallTextLine( TEXT_LINE *pTextLine, int &nTextLine )
{
	for( int i=0; i<nTextLine; i++ )
	{
		if( pTextLine[i].nWord < MIN_WORDS_SHORT_LINE)
			pTextLine[i].nType = TYPE_OTHERS;
	}
	return 0;
}

/******************************************************************************************
/*	Name:		EstimateTextLineHeight
/*	Function:	Estimate dominant height of text lines.
/*				This is a very imporatant parameter for text line classification.
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int EstimateTextLineHeight( TEXT_LINE *pTextLine, int nTextLine )
{
	//Estimate dominant text line height
#define	MAX_TEXT_LINE_HEIGHT	128
	int nHistHeight[MAX_TEXT_LINE_HEIGHT];
	int	nSmoothHistHeight[MAX_TEXT_LINE_HEIGHT];
	ZeroMemory( nHistHeight, sizeof(int)*MAX_TEXT_LINE_HEIGHT );
	ZeroMemory( nSmoothHistHeight, sizeof(int)*MAX_TEXT_LINE_HEIGHT );
	for( int i=0; i<nTextLine; i++ )
	{
		if( pTextLine[i].nType == TYPE_OTHERS )	continue;
		int nHeight = pTextLine[i].rect.bottom - pTextLine[i].rect.top + 1;
		if( nHeight >= MAX_TEXT_LINE_HEIGHT )
			continue;
		//Weighted by the width of the text line
//		nHistHeight[nHeight] += sqrt(pTextLine[i].rect.right - pTextLine[i].rect.left);
		nHistHeight[nHeight] ++;
	}
	//Smoothing the histogram
	for( i=3; i<MAX_TEXT_LINE_HEIGHT-3; i++ )
		for( int j=-1; j<=1; j++ )
			nSmoothHistHeight[i] += nHistHeight[i+j];

	int nTextLineHeight = 0;
	for( i=0; i<MAX_TEXT_LINE_HEIGHT; i++ )
	{
		if( nSmoothHistHeight[i] > nSmoothHistHeight[nTextLineHeight] )
			nTextLineHeight = i;
	}
	return nTextLineHeight;
}

/******************************************************************************************
/*	Name:		LabelLargeTextLine
/*	Function:	Label a text line with height larger than a threshold as TYPE_OTHERS.
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int LabelLargeTextLine( TEXT_LINE *pTextLine, int &nTextLine, int nTextLineHeight )
{
	for( int i=0; i<nTextLine; i++ )
	{
		int h = pTextLine[i].rect.bottom - pTextLine[i].rect.top;
		if( h > nTextLineHeight+10)	//Too high
			pTextLine[i].nType = TYPE_OTHERS;
	}
	return 0;
}	

/******************************************************************************************
/*	Name:		SortWordInTextLine
/*	Function:	Sort words inside a text line from left to right. This is a preprocedure to
/*				extract the feature of averge horizontal distance between neighboring words.
/*	Parameter:	pTextLine -- List of text lines
/*				nTextLine -- Number of text lines
/*				pWord	  -- All word blocks
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int SortWordInTextLine( TEXT_LINE *pTextLine, int nTextLine, WORD_ZONE *pWord )
{
	int i, j, k;
	for( i=0; i<nTextLine; i++ )
	{
		for( j=0; j<pTextLine[i].nWord-1; j++ )
		{
			int nLeftPos = j;
			int nLeftWordIndex = pTextLine[i].WordIndex[j];
			for( k=j+1; k<pTextLine[i].nWord; k++ )
			{
				int nWordIndex = pTextLine[i].WordIndex[k];
				if( pWord[nWordIndex].x < pWord[nLeftWordIndex].x )
				{
					nLeftPos = k;
					nLeftWordIndex = nWordIndex;
				}
			}
			if( j != nLeftPos )
			{
				int tmp = pTextLine[i].WordIndex[j];
				pTextLine[i].WordIndex[j] = pTextLine[i].WordIndex[nLeftPos];
				pTextLine[i].WordIndex[nLeftPos] = tmp;
			}
		}
	}
	return 0;
}


/******************************************************************************************
/*	Name:		ExtractFeatureTextLine
/*	Function:	Extract features for text line level classification
/*	Parameter:	TextLine  -- A text lines
/*				pWord	  -- All word blocks
/*				Feature   -- Extracted features
/*				nFeature  -- Number of extracted features
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int ExtractFeatureTextLine( TEXT_LINE TextLine, WORD_ZONE *pWord, int nTextLineHeight, double *Feature, int &nFeature )
{
	int i;
	//Global size feature
	nFeature = 0;
	int w = TextLine.rect.right - TextLine.rect.left + 1;
	int h = TextLine.rect.bottom - TextLine.rect.top + 1;
	Feature[nFeature++] = 1.0*w/nTextLineHeight;
	Feature[nFeature++] = 1.0*h/nTextLineHeight;
	Feature[nFeature++] = 1.0*w/h;

	//Number of words inside the text line
	int		nWord = TextLine.nWord;
	Feature[nFeature++] = nWord;

	//Calculate average top, and bottom gaps between a word and the text line
	double	nMeanTopGap = 0;
	double	nVarTopGap = 0;
	double	nMeanBottomGap = 0;
	double	nVarBottomGap = 0;
	for( i=0; i<nWord; i++ )
	{
		int nWordIndex = TextLine.WordIndex[i];
		int nTopGap = pWord[nWordIndex].y - TextLine.rect.top;
		nMeanTopGap += nTopGap;
		nVarTopGap += nTopGap*nTopGap;
		int nBottomGap = TextLine.rect.bottom - pWord[nWordIndex].y - pWord[nWordIndex].h;
		nMeanBottomGap += nBottomGap;
		nVarBottomGap += nBottomGap * nBottomGap;
	}
	nMeanTopGap /= nWord;
	nVarTopGap = sqrt( nVarTopGap/nWord - nMeanTopGap*nMeanTopGap );
	nMeanBottomGap /= nWord;
	nVarBottomGap = sqrt( nVarBottomGap/nWord - nMeanBottomGap*nMeanBottomGap );
	
	Feature[nFeature++] = nMeanTopGap;
	Feature[nFeature++] = nVarTopGap;
	Feature[nFeature++] = nMeanBottomGap;
	Feature[nFeature++] = nVarBottomGap;

	//Calculate horizontal gaps between neighboring words
	double	nMeanGap = 0;
	double	nVarGap = 0;
	int		nCountWord = 0;
	for( i=0; i<nWord-1; i++ )
	{
		int nGap = pWord[TextLine.WordIndex[i+1]].x - (pWord[TextLine.WordIndex[i]].x+pWord[TextLine.WordIndex[i]].w);
		if( nGap <= 0 )
			continue;
		nMeanGap += nGap;
		nVarGap += nGap*nGap;
		nCountWord++;
	}
	if( nCountWord == 0 )
		nCountWord = 1;
	nMeanGap /= nCountWord;
	nVarGap = sqrt( nVarGap/nCountWord - nMeanGap*nMeanGap );

	Feature[nFeature++] = nMeanGap;
	Feature[nFeature++] = nVarGap;
	return 0;
}

/******************************************************************************************
/*	Name:		ExtractFeature3TextLine
/*	Function:	Extract features for text lines containing 3 words
/*	Parameter:	TextLine  -- A text lines
/*				pWord	  -- All word blocks
/*				Feature   -- Extracted features
/*				nFeature  -- Number of extracted features
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int ExtractFeature3TextLine( TEXT_LINE TextLine, WORD_ZONE *pWord, int nTextLineHeight, double *Feature, int &nFeature )
{
	int i;
	//Global size feature
	nFeature = 0;
	double w = TextLine.rect.right - TextLine.rect.left + 1;
	double h = TextLine.rect.bottom - TextLine.rect.top + 1;
	Feature[nFeature++] = 1.0*w/nTextLineHeight;
	Feature[nFeature++] = 1.0*h/nTextLineHeight;
	Feature[nFeature++] = 1.0*w/h;

	//Maximum and minimum word width and height
	int nWord = TextLine.nWord;
	int nMaxWidth = 0;
	int nMinWidth = 10000;
	int nMinHeight1 = 10000;
	int nMinHeight2 = 10000;
	for( i=0; i<nWord; i++ )
	{
		int nWordIndex = TextLine.WordIndex[i];
		int ww = pWord[nWordIndex].w;
		nMaxWidth = max( nMaxWidth, ww );
		nMinWidth = min( nMinWidth, ww );

		int wh = pWord[nWordIndex].h;
		if( wh < nMinHeight1 )
		{ 
			nMinHeight2 = nMinHeight1;
			nMinHeight1 = wh;
		}
		else if( wh < nMinHeight2 )
			nMinHeight2 = wh;
	}
	Feature[nFeature++] = nMaxWidth/h;
	Feature[nFeature++] = nMinWidth/h;
	Feature[nFeature++] = nMinHeight1/h;
	Feature[nFeature++] = nMinHeight2/h;

	//Calculate average top, and bottom gaps between a word and the text line
	double	nMaxTopGap1 = 0;
	double	nMaxTopGap2 = 0;
	double	nMaxBottomGap1 = 0;
	double	nMaxBottomGap2 = 0;
	for( i=0; i<nWord; i++ )
	{
		int nWordIndex = TextLine.WordIndex[i];
		int nTopGap = pWord[nWordIndex].y - TextLine.rect.top;
		if( nTopGap >= nMaxTopGap1 )
		{
			nMaxTopGap2 = nMaxTopGap1;
			nMaxTopGap1 = nTopGap;
		}
		else if( nTopGap >= nMaxTopGap2 )
			nMaxTopGap2 = nTopGap;

		int nBottomGap = TextLine.rect.bottom - pWord[nWordIndex].y - pWord[nWordIndex].h;
		if( nBottomGap >= nMaxBottomGap1 )
		{
			nMaxBottomGap2 = nMaxBottomGap1;
			nMaxBottomGap1 = nBottomGap;
		}
		else if( nBottomGap >= nMaxBottomGap2 )
			nMaxBottomGap2 = nBottomGap;
	}
	Feature[nFeature++] = nMaxTopGap1/h;
	Feature[nFeature++] = nMaxTopGap2/h;
	Feature[nFeature++] = nMaxBottomGap1/h;
	Feature[nFeature++] = nMaxBottomGap2/h;

	//Calculate horizontal gaps between neighboring words
	double	nMaxGap = -10000;
	double	nMinGap = 10000;
	for( i=0; i<nWord-1; i++ )
	{
		int nGap = pWord[TextLine.WordIndex[i+1]].x - (pWord[TextLine.WordIndex[i]].x+pWord[TextLine.WordIndex[i]].w);
		nMaxGap = max( nMaxGap, nGap );
		nMinGap = min( nMinGap, nGap );
	}
	Feature[nFeature++] = nMaxGap/h;
	Feature[nFeature++] = nMinGap/h;
	return 0;
}

/******************************************************************************************
/*	Name:		ClassifyTextLine
/*	Function:	Classify a text line into printed, or handwriting
/*	Parameter:	pTextLine -- Text lines
/*				nTextLine -- Number of text lines
/*				pWord	  -- All word blocks
/*				nTextLineHeight  -- Dominant text line height, used for normalization
/*				Tree_Line_Short  -- Classification tree for short text lines
/*				Tree_Line_Long	 -- Classification tree for long text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int ClassifyTextLine( TEXT_LINE *pTextLine, int nTextLine, WORD_ZONE *pWord, int nTextLineHeight, 
					 Tree Tree_Line_Short,  Tree Tree_Line_Long)
{
	//Sort words inside a text line from left to right
	SortWordInTextLine( pTextLine, nTextLine, pWord );

	int nErr = 0;
	//Extract features
	double	Feature[100];
	int		nFeature;
	for( int i=0; i<nTextLine; i++ )
	{
		if( pTextLine[i].nType == TYPE_OTHERS ) //This text line is filtered as NOISE.
			continue;

		//Extract features
		ExtractFeatureTextLine( pTextLine[i], pWord, nTextLineHeight, Feature, nFeature );
		
		//Classification with C4.5 classification tree
		if( pTextLine[i].nWord < MIN_WORDS_LONG_LINE )
			pTextLine[i].nType = ClassTree( Feature, nFeature, Tree_Line_Short );
		else
			pTextLine[i].nType = ClassTree( Feature, nFeature, Tree_Line_Long );

		switch( pTextLine[i].nType )
		{
		case 0:
			pTextLine[i].nType = TYPE_HANDWRITING;
			break;
		case 1:
			pTextLine[i].nType = TYPE_PRINT_TEXT;
			break;
		}
	}
	return 0;
}

/******************************************************************************************
/*	Name:		ExtendPrintTextLine
/*	Function:	Extend and merge printed text lines after classification
/*	Parameter:	pTextLine -- Text lines
/*				nTextLine -- Number of text lines
/*				pWord     -- Word blocks
/*				Tree_Line_Short  -- Classification tree for short text lines
/*				Tree_Line_Long	 -- Classification tree for long text lines
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int ExtendPrintTextLine( TEXT_LINE *pTextLine, int &nTextLine, WORD_ZONE *pWord, int nTextLineHeight )
{
	int i, j;
	i=0;
	while( i<nTextLine )
	{
		if( pTextLine[i].nType != TYPE_PRINT_TEXT )
		{
			i++;
			continue;
		}
		for( j=0; j<nTextLine; j++ )
		{
			if( j==i )	continue;
			if( pTextLine[j].nType == TYPE_HANDWRITING )
				continue;
			if( SamePrintTextLine( pTextLine[i], pTextLine[j], nTextLineHeight ) )
				break;
		}
		if( j<nTextLine )
		{
			MergeTextLine( pTextLine[i], pTextLine[j] );
			RemoveTextLine( pTextLine, nTextLine, j );
			if( j<i )	i--;
		}
		else
			i++;
	}
	return 0;
}

/******************************************************************************************
/*	Name:		LabelWordType
/*	Function:	Label word type using the text line classification results
/*	Parameter:	pTextLine	-- Text lines
/*				nTextLine	-- Number of text lines
/*				pWord		-- Word blocks
/*				nWord		-- Number of word blocks
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int	LabelWordType( TEXT_LINE *pTextLine, int nTextLine, WORD_ZONE *pWord, int nWord )
{
	int i, j;
/*	for( i=0; i<nWord; i++ )
	{
		if( pWord[i].nType != TYPE_OTHERS )
			pWord[i].nType = TYPE_NOISE;
	}
*/	for( i=0; i<nTextLine; i++ )
	{
		//This text line is not classified
		if( pTextLine[i].nType == TYPE_OTHERS )
			continue;

		int nWord = pTextLine[i].nWord;
		for( j=0; j<nWord; j++ )
		{
			int	nIndex = pTextLine[i].WordIndex[j];
			pWord[nIndex].nType = pTextLine[i].nType;
		}
	}
	return 0;
}

/******************************************************************************************
/*	Name:		TextLineIdentify
/*	Function:	Handwriting/printed text identification at text line level
/*	Parameter:	pWord -- Word blocks
/*				nWord -- Number of word blocks
/*				nTextLineHeight -- Estimated text line height
/*	Return:		0 -- Succeed
/*******************************************************************************************/
int TextLineIdentify( WORD_ZONE *pWord, int nWord, int &nTextLineHeight, Tree Tree_Line_Short, Tree Tree_Line_Long )
{
	int i;
	TEXT_LINE	*pTextLine = new TEXT_LINE[nWord];
	//Each word form a text line
	for( i=0; i<nWord; i++ )
	{
		pTextLine[i].rect.left = pWord[i].x;
		pTextLine[i].rect.right = pWord[i].x + pWord[i].w;
		pTextLine[i].rect.top = pWord[i].y;
		pTextLine[i].rect.bottom = pWord[i].y + pWord[i].h;
		pTextLine[i].nWord = 1;
		pTextLine[i].WordIndex[0] = i;
		pTextLine[i].nTrueType = -1;
		pTextLine[i].nType = -1;
	}
	int nTextLine = nWord;

	MergeTextLine( pTextLine, nTextLine);
	MergeTextLine( pTextLine, nTextLine);

	MergeTextLineInSide( pTextLine, nTextLine );
	//Sort words inside a text line from left to right
	SortWordInTextLine( pTextLine, nTextLine, pWord );

	//Label short text lines with only one or two words as TYPE_OTHERS
	LabelSmallTextLine( pTextLine, nTextLine );
	nTextLineHeight = EstimateTextLineHeight( pTextLine, nTextLine );
	SplitTextLine( pTextLine, nTextLine, pWord, nTextLineHeight );
	//Label short text lines with only one or two words as TYPE_OTHERS
	LabelSmallTextLine( pTextLine, nTextLine );

	LabelLargeTextLine( pTextLine, nTextLine, nTextLineHeight );

	ClassifyTextLine( pTextLine, nTextLine, pWord, nTextLineHeight, Tree_Line_Short, Tree_Line_Long );

	ExtendPrintTextLine( pTextLine, nTextLine, pWord, nTextLineHeight );

	LabelWordType( pTextLine, nTextLine, pWord, nWord);

	delete[] pTextLine;
	return 0;
}
