///////////////////////////////////////////////////////////////////////
// File Name:		WordClass.cpp
// File Function:	Extract features for machine printed text/handwriting
//					classification at the word level.
//
//				Developed by: Yefeng Zheng
//			   First created: April, 2004
//			University of Maryland, College Park
//					All Right Reserved
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ArabicTextIDAPI.h"

/**************************************************************************************************
/*	Name:		GetLeftRightNeighbor
/*	Function:	Get left and right neighboring blocks
/*	Parameter:	zone	 -- Zone blocks
/*				nCurZone -- Current processing zone block
/*				nZone    -- Number of zone blocks
/*				LeftIndex  -- Left neigboring zone index
/*				LeftDist   -- Left distance to the nearset zone
/*				RightIndex -- Right neighboring zone index
/*				RightDist  -- Right distance to the nearest zone
/*				TH_Hor   -- Threshold for horizontal distance
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetLeftRightNeighbor( WORD_ZONE *zone, int nCurZone, int nZone, int &LeftIndex, int &LeftDist, int &RightIndex, int &RightDist, int TH_Hor )
{
	LeftDist = 10000;
	RightDist = 10000;
	LeftIndex = -1;
	RightIndex = -1;

	//Search LeftIndex and RightIndex
	for( int j=0; j<nZone; j++ )
	{
		if( j == nCurZone )						continue;
		if( zone[j].nType == TYPE_OTHERS ||
			zone[j].nType == TYPE_NOISE )		continue;

		int MaxTop = max( zone[nCurZone].y, zone[j].y );
		int MinBottom = min( zone[nCurZone].y+zone[nCurZone].h, zone[j].y+zone[j].h );
		if( MaxTop >= MinBottom )//Not overlap in vertical
			continue;

		int TH_V = min( zone[nCurZone].h/3, zone[nCurZone].h/3 );
		if( MinBottom - MaxTop < TH_V )//Overlap rate is too small 
			continue;

		int MaxLeft = max( zone[nCurZone].x, zone[j].x );
		int MinRight = min( zone[nCurZone].x+zone[nCurZone].w, zone[j].x+zone[j].w );
		if( MaxLeft <= MinRight )//Overlap in horizontal
			continue;
		if( MaxLeft > MinRight+TH_Hor )//Too far away
			continue;

		int nDist = MaxLeft - MinRight + abs( zone[nCurZone].h-zone[j].h ) + 
					abs( (zone[nCurZone].y+zone[nCurZone].h)/2 - (zone[j].y+zone[j].h)/2 );
		if( zone[nCurZone].x > zone[j].x )//Left
		{
			if( LeftDist > nDist )
			{
				LeftDist = nDist;
				LeftIndex = j;
			}
		}
		else//Right
		{
			if( RightDist > nDist )
			{
				RightDist = nDist;
				RightIndex = j;
			}
		}
	}
	return 0;
}

/**************************************************************************************************
/*	Name:		ExtractWordSizeFeature
/*	Function:	Extract word size features for classification
/*	Parameter:	pWord		-- Word blocks
/*				nWord		-- Number of word blocks
/*				nCurWord	-- Current processing word block
/*				nAverHeight -- Average text line height
/*				Feature		-- Extracted features
/*				nFeature	-- Number of extracted features
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int ExtractWordSizeFeature( WORD_ZONE *pWord, int nWord, int nCurWord, int nAverHeight, double *Feature, int &nFeature )
{
	int		LeftIndex, RightIndex;
	int		LeftDist, RightDist;
	int		TH_HOR = 30;
	GetLeftRightNeighbor( pWord, nCurWord, nWord, LeftIndex, LeftDist, RightIndex, RightDist, TH_HOR );

	nFeature = 0;
	Feature[nFeature++] = 1.0*pWord[nCurWord].w/nAverHeight;
	Feature[nFeature++] = 1.0*pWord[nCurWord].h/nAverHeight;

	//Features of the left neighbor
	if( LeftIndex >= 0 )
	{
		Feature[nFeature++] = 1.0;
		Feature[nFeature++] = 1.0*LeftDist/nAverHeight;
		Feature[nFeature++] = 1.0*pWord[LeftIndex].w/nAverHeight;
		Feature[nFeature++] = 1.0*pWord[LeftIndex].h/nAverHeight;

		int nMinTop = min( pWord[nCurWord].y, pWord[LeftIndex].y );
		int nMaxBottom = max( pWord[nCurWord].y+pWord[nCurWord].h, pWord[LeftIndex].y+pWord[LeftIndex].h );
		Feature[nFeature++] = 1.0*(nMaxBottom - nMinTop)/nAverHeight;
	}
	else
	{
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
	}
	
	//Features of the right neighbor
	if( RightIndex >= 0 )
	{
		Feature[nFeature++] = 1.0;
		Feature[nFeature++] = 1.0*RightDist/nAverHeight;
		Feature[nFeature++] = 1.0*pWord[RightIndex].w/nAverHeight;
		Feature[nFeature++] = 1.0*pWord[RightIndex].h/nAverHeight;

		int nMinTop = min( pWord[nCurWord].y, pWord[RightIndex].y );
		int nMaxBottom = max( pWord[nCurWord].y+pWord[nCurWord].h, pWord[RightIndex].y+pWord[RightIndex].h );
		Feature[nFeature++] = 1.0*(nMaxBottom - nMinTop)/nAverHeight;
	}
	else
	{
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
		Feature[nFeature++] = -1;
	}
	return 0;
}
