////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		WordSegment.cpp
// File Function:	Word segmentation
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "WordSegment.h"
#include "define.h"
#include "math.h"
#include "ImageObj.h"
#include "DirLine.h"
 
/*****************************************************************************
/*	Name:		CWordSegment
/*	Function:	Construct function. Initialize several memebers
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
CWordSegment::CWordSegment()
{
	knn_dis = NULL;
	m_pWord = NULL;
	m_nWord = 0;
	m_pComp = NULL;
	m_nComp = 0;
	est_para.ch = 0;
	est_para.cs = 0;
	est_para.cw = 0;
	est_para.or = 0;
	est_para.ts = 0;

	para.hatol = 30;
	para.vatol = 30;
	para.hsize = 500;
	para.lsize = 2;
	para.knn = 5;
}

/*****************************************************************************
/*	Name:		~CWordSegment
/*	Function:	Deconstruct function. Free memory.
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
CWordSegment::~CWordSegment()
{
	FreeMem();
}

/*****************************************************************************************
/*	Name:		FreeMem
/*	Function:	Free memory
/*	Parameter:	NULL
/*	Return:		NULL
/***************************************************************************************/
int CWordSegment::FreeMem()
{
	if( m_pComp != NULL )
	{
		free( m_pComp );
		m_pComp = NULL;
	}

	int i;
	if( knn_dis != NULL )
	{
		for(i=0; i<m_nComp; i++ )		
			free( knn_dis[i] );
		free( knn_dis );
		knn_dis = NULL;
	}
	if( m_pWord != NULL )
	{
		for( i=0; i<m_nWord; i++ )
			free( m_pWord[i].child );
		free( m_pWord );
		m_pWord = NULL;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		sort2
/*	Function:	Sort arrays. Written by Song Mao. It is much faster than other
/*				algorithms.
/*	Parameter:	n	-- Number of members of the array
/*				ra  -- Array to be sorted
/*				rb  -- Array of index, adjust rb with ra
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::sort2(int n, float *ra, int *rb)
{
	int l,j,ir,i;
	float rra;
    int   rrb;

	if (n == 1) return;
	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1) {
			rra=ra[--l];
            rrb=rb[l];
                }
		else {
			rra=ra[ir];
			rrb=rb[ir];
			ra[ir]=ra[1];
			rb[ir]=rb[1];
			if (--ir == 1) {
				ra[1]=rra;
				rb[1]=rrb;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) ++j;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				rb[i]=rb[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra[i]=rra;
 		rb[i]=rrb;
	}
}

/*****************************************************************************
/*	Name:		dis
/*	Function:	Get Euclidean distance between centers of two rectangles
/*	Parameter:	a   -- Rectangle one
/*				b   -- Rectangle two
/*	Return:		Euclidean distance
/*****************************************************************************/
float CWordSegment::dis(ZONE a, ZONE b)
{
  	float d;
  	float x0, y0, x1, y1;
  
  	x0 = (float) (a.x + 0.5 * a.w );
  	y0 = (float) (a.y + 0.5 * a.h );
  	x1 = (float) (b.x + 0.5 * b.w );
  	y1 = (float) (b.y + 0.5 * b.h );
  
  	d = (float) (sqrt((x0 - x1) * (x0 - x1) + (y0 - y1) * (y0 - y1)) );
  	return(d);
}

/*****************************************************************************
/*	Name:		dis2
/*	Function:	Get Euclidean distance between two rectangles (NOT centers)
/*	Parameter:	a   -- Rectangle one
/*				b   -- Rectangle two
/*	Return:		Manhantan distance
/*****************************************************************************/
double  CWordSegment::dis2(ZONE a, ZONE b)
{  
	int nMaxLeft = max( a.x, b.x );
	int nMinRight = min( a.x+a.w, b.x+b.w );
	int nMaxTop = max( a.y, b.y );
	int nMinBottom = min( a.y+a.h, b.y+b.h );	
	if( nMaxTop-nMinBottom <= 0 || nMaxLeft-nMinRight <= 0 )
		return max( nMaxTop-nMinBottom, nMaxLeft-nMinRight );
	else
		return sqrt( (nMaxTop-nMinBottom)*(nMaxTop-nMinBottom) + (nMaxLeft-nMinRight)*(nMaxLeft-nMinRight) );
}

/*****************************************************************************
/*	Name:		dis_manhantan
/*	Function:	Get Manhantan distance between two rectangles
/*	Parameter:	a   -- Rectangle one
/*				b   -- Rectangle two
/*	Return:		Manhantan distance
/*****************************************************************************/
double  CWordSegment::dis_manhantan(ZONE a, ZONE b)
{  
	int nMaxLeft = max( a.x, b.x );
	int nMinRight = min( a.x+a.w, b.x+b.w );
	int nMaxTop = max( a.y, b.y );
	int nMinBottom = min( a.y+a.h, b.y+b.h );	
	return max( nMaxTop-nMinBottom, nMaxLeft-nMinRight );
}

/*****************************************************************************
/*	Name:		ang
/*	Function:	Get angle of the line joining centers of two rectangles
/*	Parameter:	a   -- Rectangle one
/*				b   -- Rectangle two
/*	Return:		Angle of the line
/*****************************************************************************/
float CWordSegment::ang(ZONE a, ZONE b)
{  
  	float al;
  	float x0, y0, x1, y1;
  
  	x0 = (float)( a.x + 0.5 * a.w );
  	y0 = (float)( a.y + 0.5 * a.h );
  	x1 = (float) (b.x + 0.5 * b.w );
  	y1 = (float) (b.y + 0.5 * b.h );
  
  	if(x1 == x0) {
    		if(y0 > y1)
      			al = (float) M_PI_2;
    		else
      			al = (float) -M_PI_2;
  	}
  	else 
    		al = (float) (atan((y1 - y0) / (x1 - x0)) );
  	return(al);
}

/*****************************************************************************
/*	Name:		selectK
/*	Function:	Written by Song Mao. I am not clear about it.
/*	Parameter:	a   -- 
/*				b   -- 
/*				n	--
/*				k   -- 
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::selectK(float a[], int b[], int n, int k)
{
  	float v,t;
  	int   p, i, j, l, r;
  
  	l = 1; r = n;
  	while(r > l) {
    		v = a[r];
		i = l - 1;
		j = r;
    		for(;;) {
      			while(a[++i] < v);
      			while(a[--j] > v && j > 1);
      			if(i >= j)
				break;
      			t = a[i];
				a[i] = a[j];
				a[j] = t;
      			p = b[i];
				b[i] = b[j];
				b[j] = p;
    		}
    		t = a[i];
			a[i] = a[r];
			a[r] = t;
    		p = b[i];
			b[i] = b[r];
			b[r] = p;
    		if(i >= k)
				r = i - 1;
    		if(i <= k)
				l = i + 1;
  	}
}

/*****************************************************************************
/*	Name:		EstPara
/*	Function:	Estimate parameters of character width and height, character 
/*				space, word space etc.
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::EstPara( )
{
  	int    i, j, m;
  	int    *a_h,*w_h, *b_h;
  	int    a_index , max_a = 0;
  	int    d_tmp = 0;
  	int    max_w = 0, max_b = 0;
  	int    size = 10000;
  	float  l_h_a, s_h_a;
  	float  l_v_a, s_v_a;

  	float *comp_x;
  	int   *comp_i;
  	float *d_array;
  	int   *i_array;
  	int   s_pos, e_pos;
  	int   pre_s_pos = 0, pre_e_pos = 0;
  	int   flg = 0;
  	int   n;
  	int   x_dis;
  	float l_x, r_x;

  	int   k;

  	a_h = (int *) malloc (181 * sizeof(int));
  	knn_dis = (DIS **) malloc (m_nComp * sizeof(DIS *));
  	for(i = 0; i < m_nComp; i++)
    		knn_dis[i] = (DIS *) malloc (para.knn * sizeof(DIS));
  	for (i = 0; i < 181; i++) 
    		a_h[i] = 0;
  	/* sort connected component by there x coordinates */
  	comp_x = (float *) malloc (m_nComp * sizeof(float));
  	comp_i = (int *) malloc (m_nComp * sizeof(int));

  	for(i = 0; i < m_nComp; i++) {
    		comp_x[i] = (float)( m_pComp[i].x + 0.5 * m_pComp[i].w );
    		comp_i[i] = i;
 	}

 	sort2(m_nComp, --comp_x, --comp_i);
 	comp_x++;
 	comp_i++;

	d_array = (float *) malloc (m_nComp * sizeof(float));
  	i_array = (int *)   malloc (m_nComp * sizeof(int));

  	if(m_nComp <= para.knn)
    		k = m_nComp - 1;
  	else 
    		k = para.knn;

  	for(i = 0; i < m_nComp; i++) {
    		x_dis = 1;
    		while(1) {
      			l_x = comp_x[i] - x_dis;
      			r_x = comp_x[i] + x_dis;
      			if(l_x <= comp_x[0])
        			s_pos = 0;
      			else {
        			m = 0;
       				while(!(comp_x[i - m - 1] < l_x && comp_x[i - m] >= l_x))
          			m++;
        			s_pos = i - m;
      			}
      			if(r_x >= comp_x[m_nComp - 1])
        			e_pos = m_nComp - 1;
      			else {
        			m = 0;
        			while(!(comp_x[i + m + 1] > r_x && comp_x[i + m] <= r_x))
         				m++;
        			e_pos = i + m;
      			}
      			if(e_pos - s_pos >= k) {
        			if(flg == 0) {
          				n = 0;
          				for(m = s_pos; m <= e_pos; m++) {
            					if(m != i) {
              						if(dis(m_pComp[comp_i[i]],	m_pComp[comp_i[m]]) == -1) {   
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]); 
              						}
              						else
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]); 
              						i_array[n] = comp_i[m];
              						n++;
            					}
          				}
          				selectK(--d_array, --i_array, n, k);
          				d_array++;
          				i_array++;
          				pre_s_pos = s_pos;
          				pre_e_pos = e_pos;
          				flg = 1;
          				if(d_array[k - 1] > x_dis) {
            					x_dis++;
          				}
          				else {
            					flg = 0;
            					break;
          				}
        			}
        			else if(flg == 1) {
          				n = k;
          				if(pre_s_pos != 0) {
            					for(m = s_pos; m < pre_s_pos; m++) {
              						if(dis(m_pComp[comp_i[i]],
								m_pComp[comp_i[m]]) == -1) {   
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]); 
              						}
              						else
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]);  
              						i_array[n] = comp_i[m];
              						n++;
            					}
          				}
          				if(pre_e_pos < m_nComp - 1) {
            					for(m = pre_e_pos + 1; m <= e_pos; m++) {
              						if(dis(m_pComp[comp_i[i]],
								m_pComp[comp_i[m]]) == -1) {   
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]);  
              						}
              						else
                						d_array[n] =
									dis(m_pComp[comp_i[i]],
									m_pComp[comp_i[m]]);  
              						i_array[n] = comp_i[m];
              						n++;
            					}
          				}
          				selectK(--d_array, --i_array, n, k);
          				d_array++;
          				i_array++;
          				pre_s_pos = s_pos;
          				pre_e_pos = e_pos;
 
 	         			if(d_array[k - 1] > x_dis) {
  	          				x_dis++;
   		      			}
         	 			else {
          	  				flg = 0;
           	 				break;
          				}
        			}
      			}
      			else
        			x_dis++;
    		}

    		for(j = 0; j < k; j++) {
      			knn_dis[comp_i[i]][j].d = d_array[j];
      			knn_dis[comp_i[i]][j].a = ang(m_pComp[comp_i[i]],m_pComp[i_array[j]]);
      			knn_dis[comp_i[i]][j].i = i_array[j];

      			a_index = (int) ((knn_dis[comp_i[i]][j].a) * 180/M_PI);
      			a_h[a_index + 90]++;
      			if(a_h[a_index + 90] > max_a) {
        			max_a = a_h[a_index + 90];
        			est_para.or = a_index;
      			}
    		}
  	}

  	l_h_a = (float) ((est_para.or + para.hatol) * M_PI / 180 ); 
  	s_h_a = (float) ((est_para.or - para.hatol) * M_PI / 180 ); 
  	l_v_a = (float) ((est_para.or + 90 + para.vatol) * M_PI / 180 );
  	s_v_a = (float) ((est_para.or + 90 - para.vatol) * M_PI / 180 );

  	free(d_array);
  	free(i_array);
  	free(comp_x);
  	free(comp_i);

    w_h = (int *) malloc (size * sizeof(int));
    b_h = (int *) malloc (size * sizeof(int));
    for (i = 0; i < size; i++) {
      	w_h[i] = 0;
      	b_h[i] = 0;
    }
	/**----------estimate character size---------------------------------**/
	int	width[128];
	int	height[128];
	for( i=0; i<128; i++ )
	{
		width[i] = 0;
		height[i] = 0;
	}
	max_w = 0;
	int max_h = 0;
	for( i=0; i<m_nComp; i++ )
	{
		int w = m_pComp[i].w + 1;
		if( w>=15 && w<128 )
		{
			width[w]++;
			if( width[w] > width[max_w] )
			{
				max_w = w;
				est_para.cw = w;
			}
		}
		
		int h = m_pComp[i].h + 1;
		if( h>=15 && h < 128 )
		{
			height[h]++;
			if( height[h] > height[max_h] )
			{
				max_h = h;
				est_para.ch = h;
			}
		}
	}
	//Too few connected components, take default values
	if( height[max_h] <= 20 || width[max_w] <= 20 )
	{
		est_para.cw = 20;
		est_para.ch = 20;
	}
	if( est_para.cw > est_para.ch*2 )//To many touch characters
		est_para.cw = est_para.cw;
    /**----------make within-line\between-line histogram-----------------**/
	max_w = 0;
	max_b = 0;
	for(i = 0; i < m_nComp; i++) 
	{
      	for(m = 0; m < para.knn; m++)  
		{
        	d_tmp = (int)(knn_dis[i][m].d);
			if( d_tmp > size )
				exit( -1 );
        	if (knn_dis[i][m].a < l_h_a && knn_dis[i][m].a > s_h_a ) 
			{//within-line
          		w_h[d_tmp]++;
          		if(w_h[d_tmp] > max_w) 
				{
	    			max_w = w_h[d_tmp];
	    			est_para.cs = d_tmp;
          		}
        	}      
        	else 
			{//between-line
          		if (est_para.or < 0) 
				{
	    			if(l_v_a > 1.57) 
					{
	      				if( (knn_dis[i][m].a >= s_v_a && knn_dis[i][m].a <= M_PI_2 ) ||
	          				(knn_dis[i][m].a >= -M_PI_2 && knn_dis[i][m].a <= l_v_a - M_PI) ) 
						{
	        				b_h[d_tmp]++;
	        				if(b_h[d_tmp] > max_b) 
							{
	          					max_b = b_h[d_tmp];
	          					est_para.ts = d_tmp;
	        				}
	      				}
	    			}
	    			else 
					{
	      				if( (knn_dis[i][m].a >= s_v_a) && (knn_dis[i][m].a <= l_v_a) ) 
						{
	        				b_h[d_tmp]++;
	        				if(b_h[d_tmp] > max_b) 
							{
	          					max_b = b_h[d_tmp];
	          					est_para.ts = d_tmp;
	        				}
	      				}
	    			}
          		}
          		else if (est_para.or > 0) 
				{
	    			if(s_v_a < -1.57) 
					{
	      				if( (knn_dis[i][m].a <= l_v_a && knn_dis[i][m].a >= -M_PI_2 ) ||
	          				(knn_dis[i][m].a <= M_PI_2 && knn_dis[i][m].a >= s_v_a ) ) 
						{
	        				b_h[d_tmp]++;
	        				if(b_h[d_tmp] > max_b) 
							{
	          					max_b = b_h[d_tmp];
	          					est_para.ts = d_tmp;
	        				}
	      				}
					}
	    			else 
					{
	      				if( (knn_dis[i][m].a >= s_v_a) && (knn_dis[i][m].a <= l_v_a) )
						{
	        				b_h[d_tmp]++;
	        				if(b_h[d_tmp] > max_b) 
							{
	          					max_b = b_h[d_tmp];
	          					est_para.ts = d_tmp;
	        				}
	      				}
					}
				}
          		else if (est_para.or == 0) 
				{
	    			if( (knn_dis[i][m].a >= -1.57 && knn_dis[i][m].a <= (-90 + para.vatol) * M_PI / 180 ) ||
	        			(knn_dis[i][m].a >= (90 - para.vatol) * M_PI / 180 &&	knn_dis[i][m].a <= 1.57) ) 
					{
	      				b_h[d_tmp]++;
	      				if(b_h[d_tmp] > max_b) 
						{
	        				max_b = b_h[d_tmp];
	        				est_para.ts = d_tmp;	
	      				}
	    			}
         		}
        	}
      	}
    }
	//Too few components, take default values
	if( max_w < 20 )
		est_para.cs = (int)(1.2*est_para.cw);
	if( max_b < 20 )
		est_para.ts = 2*est_para.ch;

	if( est_para.cs >= est_para.cw*1.6 )//Too many touch characters
		est_para.cs = (int)(1.2*est_para.cw);
	else if( est_para.cs < est_para.cw )//Too many noise
		est_para.cs = (int)(1.2*est_para.cw);

    free (w_h);
    free (b_h);
    free (a_h);
}

/*****************************************************************************
/*	Name:		SizeCompatible
/*	Function:	Prevent handwritten characters merged with printed characters
/*	Parameter:	a -- Rectangle one
/*				b -- Rectangle two
/*	Return:		TRUE  -- Two rectangles can be merged
/*				FALSE -- Can not be merged
/*****************************************************************************/
BOOL CWordSegment::SizeCompatible(ZONE a, ZONE b)
{
	if( a.h < est_para.ch && b.h < est_para.ch )
		return TRUE;
	if( max( a.h, b.h ) >= 2*min(a.h, b.h ) )
		return FALSE;
	return TRUE;
}

/*****************************************************************************
/*	Name:		MakeWord
/*	Function:	Merge connected components into words
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::MakeWord( )
{
  	int  i, k;
  	int  *x, *y;
  	int  E = 0;
  	int  index_check;

  	int  pk;
  	index_check = m_nComp;
  	
	x = (int *) malloc ( index_check * sizeof(int));
  	y = (int *) malloc ( index_check * sizeof(int));

  	if(m_nComp <= para.knn)
    	pk = m_nComp - 1;
  	else 
    	pk = para.knn;
  	E = 0;
	
	double TH_Hdis, TH_Vdis;
	double TH_ang = 30;

	TH_Vdis = 3;
	TH_Hdis = est_para.cw*0.6;

  	for(i = 0; i < m_nComp; i++) {
    	for(k = 0; k < pk ; k++) {				
			if( SizeCompatible( m_pComp[i], m_pComp[knn_dis[i][k].i]) == FALSE )	continue;
			double	nHAngle = min( fabs(knn_dis[i][k].a * 180/M_PI), fabs(knn_dis[i][k].a * 180/M_PI - 180) );
			double	space = dis2(m_pComp[i], m_pComp[knn_dis[i][k].i] );
			double  MaxWidth = max( m_pComp[i].w, m_pComp[knn_dis[i][k].i].w );
			if( MaxWidth >= 3*est_para.cs )
				TH_ang = 10;
			else
				TH_ang = 20;
			
			BOOL	 bMerge = FALSE;
			if( space < 0 )		
				bMerge = TRUE;
			else
			{
				if( space < TH_Hdis && nHAngle < TH_ang )
					bMerge = TRUE;
			}
			if( bMerge == TRUE )
			{
				x[E] = i;
				y[E] = knn_dis[i][k].i;
				E++;

				if(E >= index_check) {
					index_check += m_nComp;
					x = (int *) realloc ( x, index_check * sizeof(int));
  					y = (int *) realloc ( y, index_check * sizeof(int));
				}
  
			}
    	}
  	}

	m_pWord = (ZONE*)malloc(sizeof(ZONE)*m_nComp);
	for(i = 0; i < m_nComp; i++) 
	{
   		m_pWord[i].x = m_pWord[i].y = 10000;
   		m_pWord[i].w = m_pWord[i].h = -10000;
		m_pWord[i].child = NULL;
		m_pWord[i].child_count = 0;
		m_pWord[i].nBlackPixels = 0;
		m_pWord[i].nType = 0;
		m_pWord[i].flg = 0;
		m_pWord[i].index = -1;
		m_pWord[i].nSize = 0;
		m_pWord[i].nConf = 0;
		for( int j=0; j<TYPE_TOTAL_CLASS; j++ )
			m_pWord[i].nClassConf[j] = 0;
   	}
	m_nWord = 0;
 	TransitiveClosure(E, x, y );
  	free(x);
  	free(y);
}

BOOL CWordSegment::CanMerge( ZONE Word1, ZONE Word2, int nHorDist, int nVerDist )
{
	//If one of them is very small, always merge them
	int	nSize1 = max( Word1.w, Word1.h );
	int nSize2 = max( Word2.w, Word2.h );
	if( nHorDist < 0 && nVerDist < 0 )
	{
		if( nSize1 < 15 || nSize2 < 15 )
			return TRUE;
	}
	else if( nSize1 < 10 || nSize2 < 10 )
		return TRUE;

	if( nHorDist < 0 && nVerDist > 0 )	//Vertical merging, using more strict condistions
	{
		//If both blocks 1 and 2 are large segments do not merge them
		if( (Word1.w > est_para.cw || Word1.h > est_para.ch) &&
			(Word2.w > est_para.cw || Word2.h > est_para.ch ) )
			return FALSE;
	}
	// Calculate the bounding box after merging
	int nLeft	= min( Word1.x, Word2.x );
	int nRight	= max( Word1.x+Word1.w, Word2.x+Word2.w );
	int nTop	= min( Word1.y, Word2.y );
	int nBottom = max( Word1.y+Word1.h, Word2.y+Word2.h );
	//Test if such merge will create a too large block
	if( nRight-nLeft+1 > est_para.cw * 10 || nBottom-nTop > est_para.ch * 4 )
		return FALSE;
	return TRUE;
}

/*****************************************************************************
/*	Name:		MergeSmallWord
/*	Function:	Merging small segments of a character with the main body
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
int	CWordSegment::MergeSmallWord( )
{
	int HTH = (int) min( 4, est_para.cw*0.6 );
	int	VTH = HTH;

	int i, j;
	i = 0;
	int nHorDist, nVerDist;
	while( i<m_nWord-1 )
	{
		//Find a word block j which is close to the word i
		for( j=i+1; j<m_nWord; j++ )
		{
			//test is block j is the main body of a character
			int nMaxTop = max( m_pWord[i].y, m_pWord[j].y );
			int nMinBottom = min( m_pWord[i].y+m_pWord[i].h, m_pWord[j].y+m_pWord[j].h );
			nVerDist = nMaxTop - nMinBottom;
			//close enough in the horizontal direction
			if( nVerDist > VTH )
				continue;		
			int nMaxLeft = max( m_pWord[i].x, m_pWord[j].x );
			int nMinRight = min( m_pWord[i].x+m_pWord[i].w, m_pWord[j].x+m_pWord[j].w );
			nHorDist = nMaxLeft - nMinRight;
			//close enough in the vertical direction
			if( nHorDist > HTH )
				continue;

			//test if two word blocks can be merged together
			if( CanMerge( m_pWord[i], m_pWord[j], nHorDist, nVerDist ) )
				break;
		}
		if( j==m_nWord )
		{
			i++;
			continue;
		}

		// Calculate the bounding box after merging
		int nLeft	= min( m_pWord[i].x, m_pWord[j].x );
		int nRight	= max( m_pWord[i].x+m_pWord[i].w, m_pWord[j].x+m_pWord[j].w );
		int nTop	= min( m_pWord[i].y, m_pWord[j].y );
		int nBottom = max( m_pWord[i].y+m_pWord[i].h, m_pWord[j].y+m_pWord[j].h );

		m_pWord[i].x = nLeft;
		m_pWord[i].w = nRight-nLeft;
		m_pWord[i].y = nTop;
		m_pWord[i].h = nBottom-nTop;

		m_pWord[i].child = (int*)realloc( m_pWord[i].child, sizeof(int)*(m_pWord[i].child_count+m_pWord[j].child_count) );
		for( int k=0; k<m_pWord[j].child_count; k++ )
			m_pWord[i].child[k+m_pWord[i].child_count] = m_pWord[j].child[k];
		m_pWord[i].child_count += m_pWord[j].child_count;
		m_pWord[i].nBlackPixels += m_pWord[j].nBlackPixels;

		free( m_pWord[j].child );
		for( k=j; k<m_nWord-1; k++ )
			m_pWord[k] = m_pWord[k+1];
		m_nWord--;
	}
	return 0;
} 

/*****************************************************************************
/*	Name:		SplitLargeWord
/*	Function:	Post processing after word segmentation. Split large word into
/*				several small ones.
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::SplitLargeWord( )
{
	int i=0, j, k;
	int nMaxGap = 0;
	int	TH_GAP;

	while( i<m_nWord )
	{
		int w = m_pWord[i].w;
		int h = m_pWord[i].h;
		if( h > est_para.ch * 4 || w < min( est_para.cs*8, est_para.cw*12) )	
		{
			i++;
			continue;
		}
		//Sort components from left to right
		for( j=0; j<m_pWord[i].child_count-1; j++ )
		{
			int nMinLeft = j;
			for( k=j+1; k<m_pWord[i].child_count; k++ )
			{
				int index1 = m_pWord[i].child[k];
				int index2 = m_pWord[i].child[nMinLeft];
				if( m_pComp[index1].x < m_pComp[index2].x )
					nMinLeft = k;
			}
			if( nMinLeft != j )
			{
				int nTemp = m_pWord[i].child[j];
				m_pWord[i].child[j] = m_pWord[i].child[nMinLeft];
				m_pWord[i].child[nMinLeft] = nTemp;
			}
		}
		nMaxGap = 0;
		int nMaxGapIndex = 0;
		int nLeft = m_pComp[m_pWord[i].child[0]].x+m_pComp[m_pWord[i].child[0]].w;
		int j=1;
		while( j<m_pWord[i].child_count )
		{
			int index = m_pWord[i].child[j];
			if( m_pComp[index].x+m_pComp[index].w < nLeft )
				j++;
			else if( m_pComp[index].x < nLeft )
			{
				j++;
				nLeft = m_pComp[index].x+m_pComp[index].w;
			}
			else
			{
				int nDist = m_pComp[index].x - nLeft;
				if( nDist > nMaxGap )
				{
					nMaxGap = nDist;
					nMaxGapIndex = j;
				}
				nLeft = m_pComp[index].x+m_pComp[index].w;
				j++;
			}
		}
		if( m_pWord[i].child_count > 20 )
			TH_GAP = (int)( est_para.cw*0.4 );
		else
			TH_GAP = (int)( est_para.cw*0.6 );

		if( nMaxGap >= TH_GAP )
		{
			CRect	rc;
			rc.left = 10000;
			rc.top = 10000;
			rc.right = -1;
			rc.bottom = -1;
			m_pWord[m_nWord].child_count = m_pWord[i].child_count - nMaxGapIndex;
			m_pWord[m_nWord].child = (int*)malloc( sizeof(int)*m_pWord[m_nWord].child_count );
			for( j=nMaxGapIndex; j<m_pWord[i].child_count; j++ )
			{
				rc.left = min( rc.left, m_pComp[m_pWord[i].child[j]].x );
				rc.right = max( rc.right, m_pComp[m_pWord[i].child[j]].x + m_pComp[m_pWord[i].child[j]].w);
				rc.top = min( rc.top, m_pComp[m_pWord[i].child[j]].y );
				rc.bottom = max( rc.bottom, m_pComp[m_pWord[i].child[j]].y + m_pComp[m_pWord[i].child[j]].h);
				m_pWord[m_nWord].child[j-nMaxGapIndex] = m_pWord[i].child[j];
			}
			m_pWord[m_nWord].x = rc.left;
			m_pWord[m_nWord].w = rc.right-rc.left+1;
			m_pWord[m_nWord].y = rc.top;
			m_pWord[m_nWord].h = rc.bottom-rc.top+1;

			int	index = m_pWord[i].child[nMaxGapIndex-1];
			m_pWord[i].w = m_pComp[index].w;
			m_pWord[i].child_count = nMaxGapIndex;

			rc.left = 10000;
			rc.top = 10000;
			rc.right = -1;
			rc.bottom = -1;
			for( j=0; j<m_pWord[i].child_count; j++ )
			{
				rc.left = min( rc.left, m_pComp[m_pWord[i].child[j]].x );
				rc.right = max( rc.right, m_pComp[m_pWord[i].child[j]].x + m_pComp[m_pWord[i].child[j]].w );
				rc.top = min( rc.top, m_pComp[m_pWord[i].child[j]].y );
				rc.bottom = max( rc.bottom, m_pComp[m_pWord[i].child[j]].y + m_pComp[m_pWord[i].child[j]].h );
			}
			m_pWord[i].x = rc.left;
			m_pWord[i].w = rc.right-rc.left;
			m_pWord[i].y = rc.top;
			m_pWord[i].h = rc.bottom-rc.top;
			m_nWord++;
		}
		else
			i++;
	}
	return ;
}

/*****************************************************************************
/*	Name:		TransitiveClosure
/*	Function:	Get words from connectness between connected components.
/*				Written by Song Mao. I am not clear.
/*	Parameter:	E --
/*				x --
/*				y --
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::TransitiveClosure(int E, int *x, int *y ) 
{
  	int i, j, k, e;
  	int ch_index = 0;
  	NODE_DOCSTRUM *t, *z;
  	NODE_DOCSTRUM **free_t;
  	NODE_DOCSTRUM **adj;
  	int  *val;
  	int  id = 0;
  	int l = 0;

  	adj = (NODE_DOCSTRUM **) malloc (m_nComp * sizeof(NODE_DOCSTRUM *));
  	val = (int *) malloc (m_nComp * sizeof(int));
  	z = (NODE_DOCSTRUM *) malloc (sizeof(NODE_DOCSTRUM));
  	z->next = z;

  	free_t = (NODE_DOCSTRUM **) malloc ( 2 * E * sizeof(NODE_DOCSTRUM *)); 

  	for(j = 0; j < m_nComp; j++) 
    		adj[j] = z;
  	for(e = 0; e < E; e++) {
    		t = (NODE_DOCSTRUM *) malloc (sizeof (NODE_DOCSTRUM));
    		free_t[l++] = t; 
    		t->v = x[e]; 
    		t->next = adj[y[e]];
    		adj[y[e]] = t;
    		t = (NODE_DOCSTRUM *) malloc (sizeof(NODE_DOCSTRUM));
    		free_t[l++] = t; 
    		t->v = y[e]; 
    		t->next = adj[x[e]]; 
    		adj[x[e]] = t;
  	}
  	for(i = 0; i < m_nComp; i++) 
    		val[i] = 0;
  	for(k = 0; k < m_nComp; k++)
    		if(val[k] == 0) { 
      			ch_index = 0;
      			visit(k, z, &adj, &id, &val, &ch_index);
        		m_pWord[m_nWord].flg = 0;
      			m_nWord++;
    		}
  	for(i = 0; i < l; i++)
    	free(free_t[i]);
  	free(free_t);
 	free(z);
  	free(adj);
  	free(val);
}

/*****************************************************************************
/*	Name:		visit
/*	Function:	Get bounding box and composite connected components of each word.
/*				Written by Song Mao. I am not clear.
/*	Parameter:	
/*	Return:		NULL
/*****************************************************************************/
void CWordSegment::visit(int k, NODE_DOCSTRUM *z, NODE_DOCSTRUM ***adj_p,
	int *id_p, int **val_p, int *ch_index_p)
{
   	NODE_DOCSTRUM *t;
   	(*val_p)[k] = ++(*id_p);
	int	left = m_pWord[m_nWord].x;
	int top = m_pWord[m_nWord].y;
	int right = m_pWord[m_nWord].x + m_pWord[m_nWord].w;
	int bottom = m_pWord[m_nWord].y + m_pWord[m_nWord].h;

   	m_pWord[m_nWord].x = min(left, m_pComp[k].x);
   	m_pWord[m_nWord].y = min(top, m_pComp[k].y);
   	m_pWord[m_nWord].w = max(right, m_pComp[k].x+m_pComp[k].w) - m_pWord[m_nWord].x;
   	m_pWord[m_nWord].h = max(bottom, m_pComp[k].y+m_pComp[k].h) - m_pWord[m_nWord].y;
	m_pWord[m_nWord].nBlackPixels += m_pComp[k].nBlackPixels;

	int	child_count = m_pWord[m_nWord].child_count;
	if( child_count == 0 )
		m_pWord[m_nWord].child = (int*)malloc( sizeof(int)*10 );
	else if( child_count%10 == 0 )
		m_pWord[m_nWord].child = (int*)realloc( m_pWord[m_nWord].child, sizeof(int)*(m_pWord[m_nWord].child_count+10 ) );
	m_pWord[m_nWord].child[ child_count ] = k;
	m_pWord[m_nWord].child_count++;
	
   	for(t = (*adj_p)[k]; t !=z; t = t->next)
     		if((*val_p)[t->v] == 0)
			visit(t->v, z, adj_p, id_p, val_p, ch_index_p);
}

/*****************************************************************************************
/*	Name:		ExtractConnComps
/*	Function:	Extracted connected components and store all detailed information
/*	Parameter:	Img  -- Input image
/*	Return:		Number of large noise blocks filtered
/***************************************************************************************/
int CWordSegment::ExtractConnComps( CImage &Img )
{
	ConnComp	*pConnComp;
	int			nConnComp;
	DirLine.GetConnComps( Img, pConnComp, nConnComp );

	m_pComp = (ZONE*)malloc( sizeof(ZONE)*nConnComp );
	m_nComp = 0;
	m_nLargeNoise = 0;
	for( int i=0; i<nConnComp; i++ )
	{
		RECT	rcBound = pConnComp[i].rcBound;
		m_pComp[m_nComp].x = rcBound.left;
		m_pComp[m_nComp].w = rcBound.right-rcBound.left+1;
		m_pComp[m_nComp].y = rcBound.top;
		m_pComp[m_nComp].h = rcBound.bottom-rcBound.top+1;
		m_pComp[m_nComp].index = i;
		m_pComp[m_nComp].nBlackPixels = pConnComp[i].nPixelsNum;
		m_nComp++;
	}	
	return m_nLargeNoise;
}

/*****************************************************************************************
/*	Name:		PreserveLargeNoiseBlock
/*	Function:	Preserve large noise block in order to output all noise blocks.
/*	Parameter:	NULL
/*	Return:		NULL
/***************************************************************************************/
int CWordSegment::PreserveLargeNoiseBlock( )
{
//Reserve large noise blocks
	int		nTotal = m_nWord+m_nLargeNoise;
 
	m_pWord = (ZONE*)realloc( m_pWord, sizeof( ZONE )*(m_nWord+m_nLargeNoise ) );
	
	ConnComp	*pConnComp;
	int			nConnComp;
	DirLine.GetConnComps( pConnComp, nConnComp);
	int			nLargeNoise = 0;
	for(int i=0; i<nConnComp; i++ )
	{
		BOOL	bNoise = FALSE;
		RECT	rcBound = pConnComp[i].rcBound;
		int		w = rcBound.right-rcBound.left;
		int		h = rcBound.bottom-rcBound.top;
		if( w<=para.lsize && h<=para.lsize )			continue;
		else if( w>=para.hsize && h>=para.hsize )		
			bNoise = TRUE;
		else if( w>=para.hsize*2 || h>=para.hsize*2)	
			bNoise = TRUE;
		if( bNoise == TRUE )
		{
			m_pComp[m_nComp+nLargeNoise].x = pConnComp[i].rcBound.left;
			m_pComp[m_nComp+nLargeNoise].w = pConnComp[i].rcBound.right-rcBound.left+1;
			m_pComp[m_nComp+nLargeNoise].y = pConnComp[i].rcBound.top;
			m_pComp[m_nComp+nLargeNoise].h = pConnComp[i].rcBound.bottom-rcBound.top+1;
			m_pComp[m_nComp+nLargeNoise].index = i;
			m_pComp[m_nComp+nLargeNoise].nBlackPixels = pConnComp[i].nPixelsNum;

			m_pWord[m_nWord].x = pConnComp[i].rcBound.left;
			m_pWord[m_nWord].y = pConnComp[i].rcBound.top;
			m_pWord[m_nWord].w = pConnComp[i].rcBound.right - pConnComp[i].rcBound.left + 1;
			m_pWord[m_nWord].h = pConnComp[i].rcBound.bottom - pConnComp[i].rcBound.top + 1;
			m_pWord[m_nWord].nType = 5;
			m_pWord[m_nWord].nConf = 1.0;
			m_pWord[m_nWord].nClassConf[0] = 0;
			m_pWord[m_nWord].nClassConf[1] = 0;
			m_pWord[m_nWord].nClassConf[2] = 1.0;
			m_pWord[m_nWord].nBlackPixels = pConnComp[i].nPixelsNum;
			m_pWord[m_nWord].flg = 0;
			m_pWord[m_nWord].index = m_nComp+nLargeNoise;
			m_pWord[m_nWord].child_count = 1;
			m_pWord[m_nWord].child = (int*)malloc( sizeof(int)*1 );
			m_pWord[m_nWord].child[0] = m_nComp+nLargeNoise;

			nLargeNoise++;
			m_nWord++;
			if( m_nWord > nTotal )
				return -1;
		}
	}
	return 0;
}

/*****************************************************************************
/*	Name:		SegmentWord
/*	Function:	Word segmentation.
/*	Parameter:	Img -- Input image.
/*	Return:		NULL
/*****************************************************************************/
int CWordSegment::SegmentWord(CImage &Img )
{
	if( m_pWord != NULL && m_nWord > 0 )
		FreeMem();
	
	ExtractConnComps( Img );
	EstPara( );
	//For Arabic script, the average character height estimation is more accurate than
	//the estimation of the character width.
	est_para.cw = est_para.ch;

	MakeWord( );
	MergeSmallWord( );

	//Label noise blocks
	LabelNoise();

	for( int i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_NOISE )
			continue;

		if( m_pWord[i].w <= 10 && m_pWord[i].h <= 10 )
		{
			m_pWord[i].nType = TYPE_OTHERS;
			continue;
		}
		if( m_pWord[i].h >= 60 )
		{
			m_pWord[i].nType = TYPE_OTHERS;
			continue;
		}
		m_pWord[i].nType = TYPE_UNDETERMINED;	//Un-determined!!!
	}
	return 0;
}

/*****************************************************************************
/*	Name:		LabelNoise
/*	Function:	Label noise blocks using some simple rules
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
int CWordSegment::LabelNoise( )
{
	for( int j=0; j<m_nWord; j++ )
	{
		//Filter noise blocks based on their sizes.
		BOOL	bNoise = FALSE;
		if( m_pWord[j].w <= 3 && m_pWord[j].h <= 3 )
			bNoise = TRUE;
		if( m_pWord[j].w >= 512 || m_pWord[j].h >= 256 )
			bNoise = TRUE;

		if( bNoise == TRUE )
			m_pWord[j].nType = TYPE_NOISE;
		else
			m_pWord[j].nType = TYPE_UNDETERMINED;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		OutputWordImage
/*	Function:	Output word image for feature extraction and classification
/*	Parameter:	ImgWord -- Output word image
/*				nWord   -- Serial number of the processing word block
/*	Return:		NULL
/*****************************************************************************/
int CWordSegment::OutputWordImage( CImage &ImgWord, int nWord )
{
	int w = m_pWord[nWord].w;
	int	h = m_pWord[nWord].h;
	ImgWord.FreeMem( );
	ImgWord.Initialize( w, h, 1, 1 );
	BYTE *pImgWord = ImgWord.LockRawImg();
	int		ImgWB = ImgWord.GetLineSizeInBytes();
	int		ImgHeight = ImgWord.GetHeight();
	for( int j=0; j<m_pWord[nWord].child_count; j++ )
		DirLine.AddCompImage( pImgWord, ImgWB, ImgHeight, CSize( m_pWord[nWord].x, m_pWord[nWord].y), m_pComp[m_pWord[nWord].child[j]].index );
	ImgWord.UnlockRawImg();

//	char	fnOutImg[MAX_PATH];
//	sprintf( fnOutImg, "c:\\word%05d.tif", nWord );
//	ImgWord.Write( fnOutImg );
	return 0;
}

/********************************************************************************************
/*	Name:		OutputImage
/*	Function:	Output separate images for printed, handwritten or noise
/*	Parameter:	ImgOut   -- Output image
/*				nOutType -- Word type. 1 -- Printed     2 -- Handwritten    3 -- Noise
/*	Return:		0 -- Succeed
/********************************************************************************************/
int CWordSegment::OutputImage( CImage &ImgOut, int nOutType )
{//Bite bits for noise, handwritten and printed
	BYTE *pImgOut = ImgOut.LockRawImg();
	int	wb = ImgOut.GetLineSizeInBytes();
	int w = ImgOut.GetWidth();
	int h = ImgOut.GetHeight();

	for( int i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType != nOutType )	continue;
		for( int j=0; j<m_pWord[i].child_count; j++ )
		{
			int CCIndex = m_pComp[m_pWord[i].child[j]].index ;
			DirLine.AddCompImage( pImgOut, wb, h, CSize(0,0), CCIndex);
		}
	}
	ImgOut.UnlockRawImg();
	return 0;
}

/*****************************************************************************
/*	Name:		SetWordSize
/*	Function:	Set word size
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
int	CWordSegment::SetWordSize( )
{
	for(int i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].w <= 12 && m_pWord[i].h <= 30 )
			m_pWord[i].nSize = SIZE_TINY;
		else if( m_pWord[i].w <= 60 && m_pWord[i].h < 40 )
			m_pWord[i].nSize = SIZE_SMALL;
		else
			m_pWord[i].nSize = SIZE_LARGE;
	}
	return 0;
}
