/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  pixconv.c
 *
 *      These functions convert between images of different types
 *      without scaling.
 *
 *      Colormap creation and addition
 *           PIXCMAP    *pixcmapCreate()
 *           PIXCMAP    *pixcmapCopy()
 *           void        pixcmapDestroy()
 *           INT32       pixcmapAddColor()
 *           INT32       pixcmapExtendArray()
 *           
 *      Conversion from colormap to full color or grayscale
 *           PIX        *pixRemoveColormap()
 *           INT32       colormapToArrays()
 *
 *      Conversion from RGB color to grayscale
 *           PIX        *pixConvertRGBToGray()
 *
 *      Conversion from RGB color to colormap (exact)
 *           PIX        *pixConvertRGBToColormap()
 *
 *      Conversion from grayscale to false color
 *           PIX        *pixConvertGrayToFalseColor()
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "allheaders.h"


    /* weighting factor for conversion of RGB to grayscale */
#define   RED_WEIGHT        0.25
#define   GREEN_WEIGHT      0.60
#define   BLUE_WEIGHT       0.15


#define  INITIAL_PTR_ARRAYSIZE      256    /* n'import quoi */


/*-------------------------------------------------------------*
 *                Colormap creation and addition               *
 *-------------------------------------------------------------*/
/*!
 *  pixcmapCreate()
 *
 *     Input:  n (initial allocated number of colors in colormap table)
 *     Return: cmap, or null on error
 */
PIXCMAP *
pixcmapCreate(INT32   n)
{
GCTE       *cta;
PIXCMAP    *cmap;

    PROCNAME("pixcmapCreate");

    if (n == 0)
        n = INITIAL_PTR_ARRAYSIZE;

    if ((cta = (GCTE *)CALLOC(n, sizeof(GCTE))) == NULL)
	return (PIXCMAP *)ERROR_PTR("cta not made", procName, NULL);

    if ((cmap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
	return (PIXCMAP *)ERROR_PTR("cmap not made", procName, NULL);
    cmap->type = CM_GIF;
    cmap->array = cta;
    cmap->size = 0;
    cmap->nalloc = n;
    cmap->n = 0;

    return cmap;
}


/*!
 *  pixcmapCopy()
 *
 *     Input:  cmaps 
 *     Return: cmapd, or null on error
 *
 *     Note: this copies both types of colormap
 */
PIXCMAP *
pixcmapCopy(PIXCMAP  *cmaps)
{
PIXCMAP    *cmapd;

    PROCNAME("pixcmapCopy");

    if (!cmaps)
	return (PIXCMAP *)ERROR_PTR("cmaps not defined", procName, NULL);

    if ((cmapd = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
	return (PIXCMAP *)ERROR_PTR("cmapd not made", procName, NULL);
    if ((cmapd->array = (void *)CALLOC(1, cmaps->size)) == NULL)
	return (PIXCMAP *)ERROR_PTR("colormap array not alloc", procName, NULL);
    cmapd->size = cmaps->size;
    cmapd->type = cmaps->type;
    memcpy(cmapd->array, cmaps->array, cmapd->size);
    cmapd->n = cmaps->n;
    cmapd->nalloc = cmaps->nalloc;

    return cmapd;
}


/*!
 *  pixcmapDestroy()
 *
 *      Input:  &cmap (<set to null>)
 *      Return: void
 */
void
pixcmapDestroy(PIXCMAP **pcmap)
{
PIXCMAP  *cmap;

    PROCNAME("pixcmapDestroy");

    if (pcmap == NULL) {
        WARNING("ptr address is null!", procName);
	return;
    }

    if ((cmap = *pcmap) == NULL)
        return;

    FREE((void *)cmap->array);
    FREE((void *)cmap);
    *pcmap = NULL;

    return;
}


/*!
 *  pixcmapAddColor()
 *
 *      Input:  cmap
 *              pixel (to be added)
 *      Return: 0 if OK, 1 on error
 */
INT32
pixcmapAddColor(PIXCMAP   *cmap,
                UINT32     pixel)
{
INT32   ncolors;
GCTE   *cta;

    PROCNAME("pixcmapAddColor");

    if (!cmap)
	return ERROR_INT("cmap not made", procName, 1);

    if (cmap->nalloc == 0) {  /* shouldn't happen, but just in case */
        if (cmap->type == CM_GIF)
	    ncolors = cmap->size / sizeof(GCTE);
	else  /* type == CM_BMP */
	    ncolors = cmap->size / sizeof(RGBQUAD);
        cmap->nalloc = cmap->n = ncolors;
    }

    if (cmap->n >= cmap->nalloc)
        pixcmapExtendArray(cmap);
    cta = (GCTE *)cmap->array;
    cta[cmap->n].cmRed = GET_DATA_BYTE(&pixel, COLOR_RED);
    cta[cmap->n].cmGreen = GET_DATA_BYTE(&pixel, COLOR_GREEN);
    cta[cmap->n].cmBlue = GET_DATA_BYTE(&pixel, COLOR_BLUE);
    cmap->size += sizeof(GCTE);
    cmap->n++;

    return 0;
}


/*!
 *  pixcmapExtendArray()
 *
 *     Input: cmap
 *     Return: 0 if OK, 1 on error
 */
INT32
pixcmapExtendArray(PIXCMAP   *cmap)
{

    PROCNAME("pixcmapExtendArray");

    if (!cmap)
	return ERROR_INT("cmap not made", procName, 1);

    if ((cmap->array = (GCTE **)reallocNew((void **)&cmap->array,
                                   4 * cmap->nalloc,
				   8 * cmap->nalloc)) == NULL)
	return ERROR_INT("new ptr array not returend", procName, 1);

    cmap->nalloc = 2 * cmap->nalloc;
    return 0;
}
    


/*-------------------------------------------------------------*
 *               Conversion from colormapped pix               *
 *-------------------------------------------------------------*/
/*!
 *   pixRemoveColormap()
 *
 *        Input:  pix
 *                type (CONVERT_TO_GRAYSCALE, CONVERT_TO_FULL_COLOR)
 *        Return: new pix, or null on error
 *
 *        Note:  The input pixs is restricted to 2, 4 or 8 bpp.
 *               For grayscale conversion, we use a weighted average
 *               of the squared RGB values.
 */
PIX *
pixRemoveColormap(PIX    *pixs,
                  INT32   type)
{
UINT8         sval, gval;
UINT8        *rmap, *gmap, *bmap;
INT32         i, j, numcolors, w, h, d, wpls, wpld;
INT32         mval;
UINT32       *datas, *lines, *datad, *lined;
PIXCMAP      *cmap;
PIX          *pixd;

    PROCNAME("pixRemoveColormap");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (type != CONVERT_TO_GRAYSCALE && type != CONVERT_TO_FULL_COLOR)
	return (PIX *)ERROR_PTR("invalid conversion type", procName, NULL);

    d = pixGetDepth(pixs);
    if (d != 8 && d != 4 && d != 2)
	return (PIX *)ERROR_PTR("pixs must be {8, 4, 2} bpp", procName, NULL);
    if ((cmap = pixGetColormap(pixs)) == NULL)
	return pixClone(pixs);

    if (colormapToArrays(cmap, &rmap, &gmap, &bmap, &numcolors))
	return (PIX *)ERROR_PTR("colormap arrays not made", procName, NULL);

    if (numcolors > (1 << d)) {
	fprintf(stderr,
	    "colormap table with %d colors is incompatible with pix depth %d\n",
	     numcolors, d);
	return pixClone(pixs);
    }
/*    fprintf(stderr, " removing colormap table w/ %d colors\n", numcolors); */

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);
    switch (type)
    {
    case CONVERT_TO_GRAYSCALE:
	if ((pixd = pixCreate(w, h, 8)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	datad = pixGetData(pixd);
	wpld = pixGetWpl(pixd);
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		switch (d)   /* depth test above; no default permitted */
		{
		case 8:
		    sval = GET_DATA_BYTE(lines, j);
		    break;
		case 4:
		    sval = GET_DATA_QBIT(lines, j);
		    break;
		case 2:
		    sval = GET_DATA_DIBIT(lines, j);
		    break;
                default:
		    return NULL;
		}

		mval = (rmap[sval] * rmap[sval] +
			2 * gmap[sval] * gmap[sval] +
			bmap[sval] * bmap[sval]) / 4;
		gval = (UINT8)sqrt((FLOAT64)mval);
		SET_DATA_BYTE(lined, j, gval);
	    }
	}
	break;

    case CONVERT_TO_FULL_COLOR:
	if ((pixd = pixCreate(w, h, 32)) == NULL)
	    return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
	datad = pixGetData(pixd);
	wpld = pixGetWpl(pixd);
	for (i = 0; i < h; i++) {
	    lines = datas + i * wpls;
	    lined = datad + i * wpld;
	    for (j = 0; j < w; j++) {
		if (d == 8)
		    sval = GET_DATA_BYTE(lines, j);
		else if (d == 4)
		    sval = GET_DATA_QBIT(lines, j);
		else   /* d == 2 */
		    sval = GET_DATA_DIBIT(lines, j);

		SET_DATA_BYTE(lined + j, COLOR_RED, rmap[sval]);
		SET_DATA_BYTE(lined + j, COLOR_GREEN, gmap[sval]);
		SET_DATA_BYTE(lined + j, COLOR_BLUE, bmap[sval]);
	    }
	}
	break;

    default:
        WARNING("invalid conversion type", procName);
	return pixClone(pixs);
    }

    FREE((void *)rmap);
    FREE((void *)gmap);
    FREE((void *)bmap);
    return pixd;
}


/*!
 *   colormapToArrays()
 *
 *      Input:  colormap
 *              &rmap, &gmap, &bmap  (<return> colormap arrays)
 *              &numcolors  (<return> size of colormap arrays)
 *      Return: 0 if OK; 1 on error
 */
INT32
colormapToArrays(PIXCMAP   *cmap,
                 UINT8    **prmap,
                 UINT8    **pgmap,
                 UINT8    **pbmap,
		 INT32     *pnumcolors)
{
UINT8                *rmap, *gmap, *bmap;
INT32                 i, numcolors;
GCTE        *pgcte;
RGBQUAD     *prgb;

    PROCNAME("colormapToArrays");

    if (!cmap)
	return ERROR_INT("cmap not defined", procName, 1);

    *prmap = *pgmap = *pbmap = NULL;
    *pnumcolors = 0;

    if (cmap->type == CM_BMP)
	numcolors = cmap->size / sizeof(RGBQUAD);
    else if (cmap->type == CM_GIF)
	numcolors = cmap->size / sizeof(GCTE);
    else
	return ERROR_INT("invalid colormap type", procName, 1);
    *pnumcolors = numcolors;

    if (((rmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL) ||
	((gmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL) ||
	((bmap = (UINT8 *)CALLOC(numcolors, sizeof(UINT8))) == NULL))
	    return ERROR_INT("calloc fail for *map", procName, 1);
    *prmap = rmap;
    *pgmap = gmap;
    *pbmap = bmap;

    switch (cmap->type)
    {
    case CM_BMP:
	prgb = (RGBQUAD *)cmap->array;
	for (i = 0; i < numcolors; i++, prgb++) {
	    rmap[i] = prgb->rgbRed;
	    gmap[i] = prgb->rgbGreen;
	    bmap[i] = prgb->rgbBlue;
	}
	break;

    case CM_GIF:
	pgcte = (GCTE *)cmap->array;
	for (i = 0; i < numcolors; i++, pgcte++) {
	    rmap[i] = pgcte->cmRed;
	    gmap[i] = pgcte->cmGreen;
	    bmap[i] = pgcte->cmBlue;
	}
	break;

    default:
	return ERROR_INT("shouldn't get here: invalid colormap type",
	                 procName, 1);
    }

    return 0;
}


/*-------------------------------------------------------------*
 *            Conversion from RGB color to grayscale           *
 *-------------------------------------------------------------*/
/*!
 *   pixConvertRGBToGray()
 *
 *        Input:  pix (32 bpp RGB)
 *        Return: 8 bpp pix, or null on error
 *
 *        Note:  we use a weighted average of the RGB values
 */
PIX *
pixConvertRGBToGray(PIX    *pixs)
{
UINT8       redbyte, greenbyte, bluebyte, graybyte;
INT32       i, j, w, h, wpls, wpld;
FLOAT32     norm;
UINT32     *datas, *words, *datad, *lined;
PIX        *pixd;

    PROCNAME("pixConvertRGBToGray");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
	return (PIX *)ERROR_PTR("pixs not 32 bpp", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    if ((pixd = pixCreate(w, h, 8)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixCopyResolution(pixd, pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);
    
    norm = 1. / (2.0 + GREEN_WEIGHT);
    for (i = 0; i < h; i++) {
	words = datas + i * wpls;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++, words++) {
	    redbyte = GET_DATA_BYTE(words, COLOR_RED);
	    greenbyte = GET_DATA_BYTE(words, COLOR_GREEN);
	    bluebyte = GET_DATA_BYTE(words, COLOR_BLUE);
	    graybyte = (UINT8)(RED_WEIGHT * redbyte +
		               GREEN_WEIGHT * greenbyte +
			       BLUE_WEIGHT * bluebyte + 0.5);
	    SET_DATA_BYTE(lined, j, graybyte);
	}
    }

    return pixd;
}


/*---------------------------------------------------------------------------*
 *               Conversion from RGB color to colormap (exact)
 *---------------------------------------------------------------------------*/
/*!
 *   pixConvertRGBToColormap()
 *
 *      Input:  pixs (32 bpp rgb)
 *              level (of octcube indexing, for histogram)
 *             &nerrors (number of improperly categorized pixels)
 *      Return: pixd (2, 4 or 8 bpp with colormap), or null on error
 *
 *      Notes:
 *         (1) This is appropriate for a color image, such as one made
 *             orthographically, that has a small number of colors. 
 *         (2) If there are more than 256 colors, this function will fail.
 *         (3) Calling with level = 4 or level = 5 should not get
 *             more than one color in each cube.  
 *         (4) The number of pixels whose color was not exactly 
 *             reproduced because more than 1 pixel of a given color
 *             was in the same octcube is returned in &nerrors.
 *         (5) These images are conveniently compressed losslessly with png.
 */     
PIX *
pixConvertRGBToColormap(PIX    *pixs,
                        INT32   level,
	                INT32  *pnerrors)
{
INT32           w, h, wpls, wpld, i, j, nerrors;
INT32           ncubes, ncolors, depth, cindex, val, oval;
INT32           rval, gval, bval;
INT32          *octarray;
UINT32          octindex;
UINT32         *rtab, *gtab, *btab;
UINT32         *lines, *lined, *datas, *datad, *ppixel;
UINT32         *colorarray;
GCTE           *gifcmap;
NUMA           *na;
PIX            *pixd;
PIXCMAP        *colormap;

    PROCNAME("pixConvertRGBToColormap");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)	
	return (PIX *)ERROR_PTR("pixs not 32 bpp", procName, NULL);
    if (!pnerrors)
	return (PIX *)ERROR_PTR("&nerrors not defined", procName, NULL);
    if (level < 2 || level > 5)
	return (PIX *)ERROR_PTR("level not in {2 ... 5}", procName, NULL);

        /* Get the histogram and count the number of occupied cubes.
	 * We don't yet know if this is the number of actual colors,
	 * but if it's not, we will make some approximation to pixel
	 * colors, because we only allow one color for all pixels in
	 * the same octcube. */
    na = pixOctcubeHistogram(pixs, level);
    ncubes = numaGetCount(na);
    ncolors = 0;
    for (i = 0; i < ncubes; i++) {
        numaGetIValue(na, i, &val);
	if (val > 0)
	    ncolors++;
    }
    if (ncolors > 256)
	return (PIX *)ERROR_PTR("too many colors", procName, NULL);
    fprintf(stderr, "ncubes = %d, ncolors = %d\n", ncubes, ncolors);

    if (makeRGBToIndexTables(&rtab, &gtab, &btab, level))
	return (PIX *)ERROR_PTR("tables not made", procName, NULL);

    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    datas = pixGetData(pixs);
    wpls = pixGetWpl(pixs);

    if (ncolors <= 4)
        depth = 2;
    else if (ncolors <= 16)
        depth = 4;
    else  /* ncolors <= 256 */
        depth = 8;
	
    if ((pixd = pixCreate(w, h, depth)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    pixCopyResolution(pixd, pixs);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

        /* octarray will give a ptr from the octcube to the colorarray */
    if ((octarray = (INT32 *)CALLOC(ncubes, sizeof(INT32))) == NULL)
	return (PIX *)ERROR_PTR("octarray not made", procName, NULL);
    if ((colorarray = (UINT32 *)CALLOC(ncolors + 1, sizeof(UINT32))) == NULL)
	return (PIX *)ERROR_PTR("colorarray not made", procName, NULL);

    cindex = 1;  /* start with 1 */ 
    nerrors = 0;
    for (i = 0; i < h; i++) {
        lines = datas + i * wpls;
        lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    ppixel = lines + j;
            rval = GET_DATA_BYTE(ppixel, COLOR_RED);
	    gval = GET_DATA_BYTE(ppixel, COLOR_GREEN);
	    bval = GET_DATA_BYTE(ppixel, COLOR_BLUE);
	    octindex = rtab[rval] | gtab[gval] | btab[bval];
	    oval = octarray[octindex];
	    if (oval == 0) {
	        octarray[octindex] = cindex;
		colorarray[cindex] = *ppixel;
		setPixelLow(lined, j, depth, cindex - 1);
		cindex++;
	    }
	    else {  /* already have seen this color; is it unique? */
		setPixelLow(lined, j, depth, oval - 1);
	        if (colorarray[oval] != *ppixel)
		    nerrors++;
	    }
	}
    }
    *pnerrors = nerrors;

    for (i = 0; i < ncolors; i++)
        fprintf(stderr, "color[%d] = %x\n", i, colorarray[i + 1]);

        /* make the colormap */
    if ((gifcmap = (GCTE *)CALLOC(ncolors, sizeof(GCTE))) == NULL)
	return (PIX *)ERROR_PTR("gifcmap not made", procName, NULL);
    for (i = 0; i < ncolors; i++) {
	ppixel = colorarray + i + 1;
	gifcmap[i].cmRed = GET_DATA_BYTE(ppixel, COLOR_RED);
	gifcmap[i].cmGreen = GET_DATA_BYTE(ppixel, COLOR_GREEN);
	gifcmap[i].cmBlue = GET_DATA_BYTE(ppixel, COLOR_BLUE);
    }

    if ((colormap = (PIXCMAP *)CALLOC(1, sizeof(PIXCMAP))) == NULL)
	return (PIX *)ERROR_PTR("colormap not made", procName, NULL);
    colormap->type = CM_GIF;
    colormap->array = gifcmap;
    colormap->size = ncolors * sizeof(GCTE);
    pixSetColormap(pixd, colormap);

    numaDestroy(&na);
    FREE((void *)octarray);
    FREE((void *)colorarray);
    FREE((void *)rtab);
    FREE((void *)gtab);
    FREE((void *)btab);

    return pixd;
}


/*---------------------------------------------------------------------------*
 *                Conversion from grayscale to false color
 *---------------------------------------------------------------------------*/
/*!
 *   pixConvertGrayToFalseColor()
 *
 *      Input:  pixs (8 or 16 bpp grayscale)
 *      Return: pixd (8 or 16 bpp with colormap), or null on error
 *
 *      Note: this simply adds a colormap to the input image
 *
 *      For 8 bpp, the mapping is:  0  --> blue
 *                                 128 --> green
 *                                 255 --> red
 *          From 0 to 127, we use only blue and green
 *          From 128 to 255, we use only green and red
 *      For 16 bpp, the mapping is similar, but scaled up by 256
 */     
PIX *
pixConvertGrayToFalseColor(PIX  *pixs)
{
INT32           d, i, ncolors, halfn, rval, bval, gval;
UINT32          pixel;
PIX            *pixd;
PIXCMAP        *cmap;

    PROCNAME("pixConvertGrayToFalseColor");

    if (!pixs)
	return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    d = pixGetDepth(pixs);
    if (d != 8 && d != 16)	
	return (PIX *)ERROR_PTR("pixs not 8 or 16 bpp", procName, NULL);

    if ((pixd = pixCopy(NULL, pixs)) == NULL)
	return (PIX *)ERROR_PTR("pixd not made", procName, NULL);

    if (d == 8)
	ncolors = 256;
    else  /* d == 16 */
	ncolors = 256 * 256;
    halfn = ncolors / 2;

    if ((cmap = pixcmapCreate(ncolors)) == NULL)
	return (PIX *)ERROR_PTR("cmap not made", procName, NULL);
    for (i = 0; i < ncolors; i++) {
	if (i < halfn) {
	    rval = 0;
	    bval = ncolors - 1 - 2 * i;
	    gval = 2 * i + 1;
	}
	else {
	    rval = 2 * (i - halfn) + 1;
	    bval = 0;
	    gval = ncolors - 1 - 2 * (i - halfn);
	}
	composeRGBPixel(rval, gval, bval, &pixel);
	pixcmapAddColor(cmap, pixel);
    }
    pixSetColormap(pixd, cmap);

    return pixd;
}


