/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  readfile.c:  reads image on file into memory
 *
 *      High level read functions
 *           PIX         *pixRead()
 *           PIX         *pixReadStream()
 *
 *      Format helper
 *           INT32        findFileFormat()
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include  "allheaders.h"


/*
 *  choose type of PIX to be generated
 */
#define  READ_24_BIT_COLOR        0    /* read in as 24 (really 32) bit pix */
#define  CONVERT_TO_PALETTE       1    /* convert to 8 bit colormapped pix  */

#ifndef  NO_CONSOLE_IO
#define  DEBUG     0
#endif  /* ~NO_CONSOLE_IO */

/*!
 *  pixReadStream()
 *
 *       Input:  file stream
 *       Return: pix if OK; NULL on error.
 */
PIX *
pixReadStreamName(char *filename)
{
INT32     format;
PIX      *pix;

    PROCNAME("pixReadStream");

	FILE *fp;
    if ((fp = fopenReadStream(filename)) == NULL)
		return (PIX *)ERROR_PTR("image file not found", procName, NULL);
    format = findFileFormat(fp);
#if DEBUG_FORMAT
    fprintf(stderr, "Read: file format is %d\n", format);
#endif  /* DEBUG_FORMAT */

    switch (format)
    {
    case IFF_BMP:
	if ((pix = pixReadStreamBmp(fp)) == NULL )
	    return (PIX *)ERROR_PTR( "bmp: no pix returned",
		    procName, NULL);
	break;

//  case IFF_JFIF_JPEG:
//	if ((pix = pixReadStreamJpeg(fp, READ_24_BIT_COLOR)) == NULL)
//	    return (PIX *)ERROR_PTR( "jpeg: no pix returned",
//		    procName, NULL);
//	break;

//  case IFF_PNG:
//	if ((pix = pixReadStreamPng(fp)) == NULL)
//	    return (PIX *)ERROR_PTR("png: no pix returned", procName, NULL);
//	break;

    case IFF_TIFF:
	if ((pix = pixReadStreamTiffName(filename)) == NULL)
	    return (PIX *)ERROR_PTR("tiff: no pix returned", procName, NULL);
	break;

    case IFF_PBM:
	return (PIX *)ERROR_PTR("pbm read not supported", procName, NULL);

    case IFF_PGM:
	return (PIX *)ERROR_PTR("pgm read not supported", procName, NULL);

    case IFF_PPM:
	return (PIX *)ERROR_PTR("ppm read not supported", procName, NULL);

    case IFF_UNKNOWN:
	return (PIX *)ERROR_PTR( "Unknown format: no pix returned",
		procName, NULL);
	break;
    }
	fclose( fp );
    return pix;
}


/*!
 *  pixRead()
 *         Input:  filename (with full pathname or in local directory)
 *         Return: pix, or null on error.
 */
PIX *
pixRead(const char  *filename)
{
FILE         *fp;
PIX          *pix;

    PROCNAME("pixRead");

    if (!filename)
	return (PIX *)ERROR_PTR("filename not defined", procName, NULL);

 /* if ((fp = fopenReadStream(filename)) == NULL)
	return (PIX *)ERROR_PTR("image file not found", procName, NULL);
    pix = pixReadStream(fp);
    fclose(fp);
*/
	pix = pixReadStreamName(filename);

    if (!pix)
	return (PIX *)ERROR_PTR("image not returned", procName, NULL);
    return pix;
}

/*!
 *  pixReadStream()
 *
 *       Input:  file stream
 *       Return: pix if OK; NULL on error.
 */
PIX *
pixReadStream(FILE   *fp)
{
INT32     format;
PIX      *pix;

    PROCNAME("pixReadStream");

    if (!fp)
	return (PIX *)ERROR_PTR("stream not defined", procName, NULL);

    format = findFileFormat(fp);
#if DEBUG_FORMAT
    fprintf(stderr, "Read: file format is %d\n", format);
#endif  /* DEBUG_FORMAT */

    switch (format)
    {
    case IFF_BMP:
	if ((pix = pixReadStreamBmp(fp)) == NULL )
	    return (PIX *)ERROR_PTR( "bmp: no pix returned",
		    procName, NULL);
	break;

//  case IFF_JFIF_JPEG:
//	if ((pix = pixReadStreamJpeg(fp, READ_24_BIT_COLOR)) == NULL)
//	    return (PIX *)ERROR_PTR( "jpeg: no pix returned",
//		    procName, NULL);
//	break;

//  case IFF_PNG:
//	if ((pix = pixReadStreamPng(fp)) == NULL)
//	    return (PIX *)ERROR_PTR("png: no pix returned", procName, NULL);
//	break;

    case IFF_TIFF:
	if ((pix = pixReadStreamTiff(fp)) == NULL)
	    return (PIX *)ERROR_PTR("tiff: no pix returned", procName, NULL);
	break;

    case IFF_PBM:
	return (PIX *)ERROR_PTR("pbm read not supported", procName, NULL);

    case IFF_PGM:
	return (PIX *)ERROR_PTR("pgm read not supported", procName, NULL);

    case IFF_PPM:
	return (PIX *)ERROR_PTR("ppm read not supported", procName, NULL);

    case IFF_UNKNOWN:
	return (PIX *)ERROR_PTR( "Unknown format: no pix returned",
		procName, NULL);
	break;
    }

    return pix;
}



/*!
 *  findFileFormat()
 *
 *       Input: fp
 *       Return: format integer; 0 on error or if format not recognized
 *       N.B.: this resets fp to BOF
 */
INT32
findFileFormat(FILE  *fp)
{
UINT16       twobytepw;
UINT8        jpegpw[12];
UINT8        pngpw[8];
char         pnmpw[2];

    PROCNAME("findFileFormat");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 0);

    rewind(fp);
    if (nbytesInFile(fp) < 12)
	return ERROR_INT("truncated file", procName, 0);

	/* check for the bmp and tiff 2-byte header ids */
    fread((char *)&twobytepw, 1, 2, fp);
    rewind(fp);

    if (convertLittleEnd16(twobytepw) == BMP_ID)
	return IFF_BMP;

    if (twobytepw == TIFF_BIGEND_ID || twobytepw == TIFF_LITTLEEND_ID)
        return IFF_TIFF;

        /* check for the p*m 2-byte header ids */
    fread(pnmpw, 1, 2, fp);
    rewind(fp);

    if ((pnmpw[0] == 'P' && pnmpw[1] == '4') || /* newer packed */
        (pnmpw[0] == 'P' && pnmpw[1] == '1'))   /* old format */
            return IFF_PBM;

    if ((pnmpw[0] == 'P' && pnmpw[1] == '5') || /* newer */
        (pnmpw[0] == 'P' && pnmpw[1] == '2'))   /* old */
            return IFF_PGM;

    if ((pnmpw[0] == 'P' && pnmpw[1] == '6') || /* newer */
        (pnmpw[0] == 'P' && pnmpw[1] == '3'))   /* old */
            return IFF_PPM;

        /*  check for an 11-byte JFIF JPEG header id:
	 *            standard:
	 *      {0xff, 0xd8, 0xff, 0xe0, 0xXX, 0xXX, 'J', 'F', 'I', 'F', 0x00}
	 *            some digital camera files:
	 *      {0xff, 0xd8, 0xff, 0xe1, 0xXX, 0xXX, 'E', 'x', 'i', 'f', 0x00}
         *         ("0xXX" means it can be anything)   */
    fread((char *)jpegpw, 1, 11, fp);
    rewind(fp);
#if DEBUG_FORMAT
    fprintf(stderr, "%x, %x, %x, %x\n",
    jpegpw[0], jpegpw[1], jpegpw[2], jpegpw[3]);
    fprintf(stderr, "%x, %x, %x, %x, %x\n",
    jpegpw[6], jpegpw[7], jpegpw[8], jpegpw[9], jpegpw[10]);
#endif  /* DEBUG_FORMAT */
    if (jpegpw[0] == 0xff && jpegpw[1] == 0xd8 && jpegpw[2] == 0xff &&
	jpegpw[3] == 0xe0 && jpegpw[6] == 0x4a && jpegpw[7] == 0x46 &&
	jpegpw[8] == 0x49 && jpegpw[9] == 0x46 && jpegpw[10] == 0x00)
	    return IFF_JFIF_JPEG;

    if (jpegpw[0] == 0xff && jpegpw[1] == 0xd8 && jpegpw[2] == 0xff &&
	jpegpw[3] == 0xe1 && jpegpw[6] == 0x45 && jpegpw[7] == 0x78 &&
	jpegpw[8] == 0x69 && jpegpw[9] == 0x66 && jpegpw[10] == 0x00)
	    return IFF_JFIF_JPEG;

	/*  check for the 8 byte PNG signature (png_signature in png.c):
	 *        {137, 80, 78, 71, 13, 10, 26, 10}      */
    fread((char *)pngpw, 1, 8, fp);
    rewind(fp);
    if (pngpw[0] == 137 && pngpw[1] == 80  && pngpw[2] == 78  &&
	pngpw[3] == 71  && pngpw[4] == 13  && pngpw[5] == 10  &&
	pngpw[6] == 26  && pngpw[7] == 10)
	    return IFF_PNG;

	/* format header not found */
    return IFF_UNKNOWN;
}


