/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/


/*
 *  rotateamlow.c
 *
 *         Grayscale and color rotation (area mapped)
 *
 *             Rotation about image center:
 *                  void        rotateAMLow()
 *
 *             Rotation about UL corner of image:
 *                  void        rotateAMCornerLow()
 *
 *             Fast color rotation (area mapped)
 *                  void        rotateAMColorFastLow()
 *
 */

#include <stdio.h>
#include <string.h>
#include <math.h>   /* required for sin and tan */

#include "allheaders.h"


/*------------------------------------------------------------------*
 *                     Rotation about the center                    *
 *------------------------------------------------------------------*/
void
rotateAMLow(UINT32  *datad,
            INT32    w,
	    INT32    h,
	    INT32    wpld,
	    UINT32  *datas,
	    INT32    wpls,
	    FLOAT32  angle,
	    UINT8    grayval)
{
INT32         i, j, xcen, ycen, wm2, hm2;
INT32         xdif, ydif, xpm, ypm, xp, yp, xf, yf;
INT32         v00, v01, v10, v11;
UINT8         val;
UINT32       *lines, *lined;
FLOAT32       sina, cosa;

    xcen = w / 2;
    wm2 = w - 2;
    ycen = h / 2;
    hm2 = h - 2;
    sina = 16. * sin(angle);
    cosa = 16. * cos(angle);

    for (i = 0; i < h; i++) {
	ydif = ycen - i;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xdif = xcen - j;
	    xpm = (INT32)(-xdif * cosa - ydif * sina + 0.5);
	    ypm = (INT32)(-ydif * cosa + xdif * sina + 0.5);
	    xp = xcen + (xpm >> 4);
	    yp = ycen + (ypm >> 4);
	    xf = xpm & 0x0f;
	    yf = ypm & 0x0f;

		/* if off the edge, write input grayval */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
		SET_DATA_BYTE(lined, j, grayval);
		continue;
	    }

	    lines = datas + yp * wpls;

		/* do area weighting.  Without this, we would
		 * simply do:
	         *   SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		 * which is faster but gives lousy results!
		 */
	    v00 = (16 - xf) * (16 - yf) * GET_DATA_BYTE(lines, xp);
	    v10 = xf * (16 - yf) * GET_DATA_BYTE(lines, xp + 1);
	    v01 = (16 - xf) * yf * GET_DATA_BYTE(lines + wpls, xp);
	    v11 = xf * yf * GET_DATA_BYTE(lines + wpls, xp + 1);
	    val = (UINT8)((v00 + v01 + v10 + v11 + 128) / 256);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}


/*------------------------------------------------------------------*
 *                     Rotation about the corner                    *
 *------------------------------------------------------------------*/
void
rotateAMCornerLow(UINT32  *datad,
                  INT32    w,
		  INT32    h,
		  INT32    wpld,
		  UINT32  *datas,
		  INT32    wpls,
		  FLOAT32  angle,
		  UINT8    grayval)
{
INT32         i, j, wm2, hm2;
INT32         xpm, ypm, xp, yp, xf, yf;
INT32         v00, v01, v10, v11;
UINT8         val;
UINT32       *lines, *lined;
FLOAT32       sina, cosa;

    wm2 = w - 2;
    hm2 = h - 2;
    sina = 16. * sin(angle);
    cosa = 16. * cos(angle);

    for (i = 0; i < h; i++) {
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xpm = (INT32)(j * cosa + i * sina + 0.5);
	    ypm = (INT32)(i * cosa - j * sina + 0.5);
	    xp = xpm >> 4;
	    yp = ypm >> 4;
	    xf = xpm & 0x0f;
	    yf = ypm & 0x0f;

		/* if off the edge, write input grayval */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
		SET_DATA_BYTE(lined, j, grayval);
		continue;
	    }

	    lines = datas + yp * wpls;

		/* do area weighting.  Without this, we would
		 * simply do:
	         *   SET_DATA_BYTE(lined, j, GET_DATA_BYTE(lines, xp));
		 * which is faster but gives lousy results!
		 */
	    v00 = (16 - xf) * (16 - yf) * GET_DATA_BYTE(lines, xp);
	    v10 = xf * (16 - yf) * GET_DATA_BYTE(lines, xp + 1);
	    v01 = (16 - xf) * yf * GET_DATA_BYTE(lines + wpls, xp);
	    v11 = xf * yf * GET_DATA_BYTE(lines + wpls, xp + 1);
	    val = (UINT8)((v00 + v01 + v10 + v11 + 128) / 256);
	    SET_DATA_BYTE(lined, j, val);
	}
    }

    return;
}


/*------------------------------------------------------------------*
 *                  Fast color rotation (area mapped)
 *------------------------------------------------------------------*/
/*
 *  rotateAMColorFastLow()
 *
 *     This is a special simplification of area mapping with division
 *     of each pixel into 16 sub-pixels.  The exact coefficients that
 *     should be used are the same as for the 4x linear interpolation
 *     scaling case, and are given there.  I tried to approximate these
 *     as weighted coefficients with a maximum sum of 4, which 
 *     allows us to do the arithmetic in parallel for the R, G and B
 *     components in a 32 bit pixel.  However, there are three reasons
 *     for not doing that:
 *        (1) the loss of accuracy in the parallel implementation
 *            is visually significant
 *        (2) the parallel implementation (described below) is slower
 *        (3) the parallel implementation requires allocation of
 *            a temporary color image
 *
 *     There are 16 cases for the choice of the subpixel, and
 *     for each, the mapping to the relevant source
 *     pixels is as follows:
 *
 *      subpixel      src pixel weights
 *      --------      -----------------
 *         0          sp1
 *         1          (3 * sp1 + sp2) / 4
 *         2          (sp1 + sp2) / 2
 *         3          (sp1 + 3 * sp2) / 4
 *         4          (3 * sp1 + sp3) / 4
 *         5          (9 * sp1 + 3 * sp2 + 3 * sp3 + sp4) / 16
 *         6          (3 * sp1 + 3 * sp2 + sp3 + sp4) / 8
 *         7          (3 * sp1 + 9 * sp2 + sp3 + 3 * sp4) / 16
 *         8          (sp1 + sp3) / 2
 *         9          (3 * sp1 + sp2 + 3 * sp3 + sp4) / 8
 *         10         (sp1 + sp2 + sp3 + sp4) / 4 
 *         11         (sp1 + 3 * sp2 + sp3 + 3 * sp4) / 8
 *         12         (sp1 + 3 * sp3) / 4
 *         13         (3 * sp1 + sp2 + 9 * sp3 + 3 * sp4) / 16
 *         14         (sp1 + sp2 + 3 * sp3 + 3 * sp4) / 8
 *         15         (sp1 + 3 * sp2 + 3 * sp3 + 9 * sp4) / 16
 *
 *     Another way to visualize this is to consider the area mapping
 *     (or linear interpolation) coefficients  for the pixel sp1.
 *     Expressed in fourths, they can be written as asymmetric matrix:
 *
 *           4      3      2      1
 *           3      2.25   1.5    0.75
 *           2      1.5    1      0.5
 *           1      0.75   0.5    0.25
 *
 *     The coefficients for the three neighboring pixels can be
 *     similarly written.
 *
 *     This is implemented here, where, for each color component,
 *     we inline its extraction from each participating word,
 *     construct the linear combination, and combine the results
 *     into the destination 32 bit RGB pixel, using the appropriate shifts.
 *     
 *     It is interesting to note that the alternative method, where
 *     we do the arithmetic on the 32 bit pixels directly (after
 *     shifting the components so they won't overflow into each other)
 *     is significantly inferior.  Because we have only 8 bits for
 *     internal overflows, which can be distributed as 2, 3, 3, it
 *     is impossible to add these with the correct linear 
 *     interpolation coefficients, which require a sum of up to 16.
 *     We'd need to round off to a sum of 4, and this causes
 *     appreciable visual artifacts in the rotated image.
 *     (If you want to try it out, all the code for the inferior
 *     method is #if 0'd at the end of this file, for reference.)
 */
void
rotateAMColorFastLow(UINT32  *datad,
                     INT32    w,
	             INT32    h,
	             INT32    wpld,
	             UINT32  *datas,
	             INT32    wpls,
	             FLOAT32  angle,
		     UINT8    grayval)
{
INT32         i, j, xcen, ycen, wm2, hm2;
INT32         xdif, ydif, xpm, ypm, xp, yp, xf, yf;
UINT32        edgeval, word1, word2, word3, word4, red, blue, green;
UINT32       *pword, *lines, *lined;
FLOAT32       sina, cosa;

    xcen = w / 2;
    wm2 = w - 2;
    ycen = h / 2;
    hm2 = h - 2;
    sina = 4. * sin(angle);
    cosa = 4. * cos(angle);

    edgeval = (grayval << 24) | (grayval << 16) | (grayval << 8);
    for (i = 0; i < h; i++) {
	ydif = ycen - i;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xdif = xcen - j;
	    xpm = (INT32)(-xdif * cosa - ydif * sina + 0.5);
	    ypm = (INT32)(-ydif * cosa + xdif * sina + 0.5);
	    xp = xcen + (xpm >> 2);
	    yp = ycen + (ypm >> 2);
	    xf = xpm & 0x03;
	    yf = ypm & 0x03;

		/* if off the edge, write input grayval */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
	        *(lined + j) = edgeval;
		continue;
	    }

	    lines = datas + yp * wpls;
	    pword = lines + xp;

	    switch (xf + 4 * yf)
	    {
            case 0:
	        *(lined + j) = *pword;
	        break;
            case 1:
		word1 = *pword;
		word2 = *(pword + 1);
		red = 3 * (word1 >> 24) + (word2 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) +
		            ((word2 >> 16) & 0xff);
		blue = 3 * ((word1 >> 8) & 0xff) +
		            ((word2 >> 8) & 0xff);
		*(lined + j) = ((red << 22) & 0xff000000) |
		               ((green << 14) & 0x00ff0000) |
			       ((blue << 6) & 0x0000ff00);
                break;
            case 2:
		word1 = *pword;
		word2 = *(pword + 1);
		red = (word1 >> 24) + (word2 >> 24);
		green = ((word1 >> 16) & 0xff) + ((word2 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + ((word2 >> 8) & 0xff);
		*(lined + j) = ((red << 23) & 0xff000000) |
		               ((green << 15) & 0x00ff0000) |
			       ((blue << 7) & 0x0000ff00);
                break;
            case 3:
		word1 = *pword;
		word2 = *(pword + 1);
		red = (word1 >> 24) + 3 * (word2 >> 24);
		green = ((word1 >> 16) & 0xff) +
		          3 * ((word2 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) +
		          3 * ((word2 >> 8) & 0xff);
		*(lined + j) = ((red << 22) & 0xff000000) |
		               ((green << 14) & 0x00ff0000) |
			       ((blue << 6) & 0x0000ff00);
                break;
            case 4:
		word1 = *pword;
		word3 = *(pword + wpls);
		red = 3 * (word1 >> 24) + (word3 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) +
		            ((word3 >> 16) & 0xff);
		blue = 3 * ((word1 >> 8) & 0xff) +
		            ((word3 >> 8) & 0xff);
		*(lined + j) = ((red << 22) & 0xff000000) |
		               ((green << 14) & 0x00ff0000) |
			       ((blue << 6) & 0x0000ff00);
                break;
            case 5:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = 9 * (word1 >> 24) + 3 * (word2 >> 24) +
		      3 * (word3 >> 24) + (word4 >> 24);
		green = 9 * ((word1 >> 16) & 0xff) +
		        3 * ((word2 >> 16) & 0xff) +
		        3 * ((word3 >> 16) & 0xff) +
		        ((word4 >> 16) & 0xff);
		blue = 9 * ((word1 >> 8) & 0xff) + 
		       3 * ((word2 >> 8) & 0xff) +
		       3 * ((word3 >> 8) & 0xff) +
		       ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 20) & 0xff000000) |
		               ((green << 12) & 0x00ff0000) |
			       ((blue << 4) & 0x0000ff00);
                break;
            case 6:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = 3 * (word1 >> 24) +  3 * (word2 >> 24) +
		      (word3 >> 24) + (word4 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) + 
		        3 * ((word2 >> 16) & 0xff) +
		        ((word3 >> 16) & 0xff) +
		        ((word4 >> 16) & 0xff);
		blue = 3 * ((word1 >> 8) & 0xff) +
		       3 * ((word2 >> 8) & 0xff) +
		       ((word3 >> 8) & 0xff) +
		       ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 21) & 0xff000000) |
		               ((green << 13) & 0x00ff0000) |
			       ((blue << 5) & 0x0000ff00);
                break;
            case 7:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = 3 * (word1 >> 24) + 9 * (word2 >> 24) +
		      (word3 >> 24) + 3 * (word4 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) + 
		        9 * ((word2 >> 16) & 0xff) +
		        ((word3 >> 16) & 0xff) +
		        3 * ((word4 >> 16) & 0xff);
		blue = 3 * ((word1 >> 8) & 0xff) +
		       9 * ((word2 >> 8) & 0xff) +
	  	       ((word3 >> 8) & 0xff) +
	  	       3 * ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 20) & 0xff000000) |
		               ((green << 12) & 0x00ff0000) |
			       ((blue << 4) & 0x0000ff00);
                break;
            case 8:
		word1 = *pword;
		word3 = *(pword + wpls);
		red = (word1 >> 24) + (word3 >> 24);
		green = ((word1 >> 16) & 0xff) + ((word3 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + ((word3 >> 8) & 0xff);
		*(lined + j) = ((red << 23) & 0xff000000) |
		               ((green << 15) & 0x00ff0000) |
			       ((blue << 7) & 0x0000ff00);
                break;
            case 9:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = 3 * (word1 >> 24) + (word2 >> 24) +
		      3 * (word3 >> 24) + (word4 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) + ((word2 >> 16) & 0xff) +
		        3 * ((word3 >> 16) & 0xff) + ((word4 >> 16) & 0xff);
		blue = 3 * ((word1 >> 8) & 0xff) + ((word2 >> 8) & 0xff) +
		       3 * ((word3 >> 8) & 0xff) + ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 21) & 0xff000000) |
		               ((green << 13) & 0x00ff0000) |
			       ((blue << 5) & 0x0000ff00);
                break;
            case 10:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = (word1 >> 24) + (word2 >> 24) +
		      (word3 >> 24) + (word4 >> 24);
		green = ((word1 >> 16) & 0xff) + ((word2 >> 16) & 0xff) +
		        ((word3 >> 16) & 0xff) + ((word4 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + ((word2 >> 8) & 0xff) +
		       ((word3 >> 8) & 0xff) + ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 22) & 0xff000000) |
		               ((green << 14) & 0x00ff0000) |
			       ((blue << 6) & 0x0000ff00);
                break;
            case 11:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = (word1 >> 24) + 3 * (word2 >> 24) +
		      (word3 >> 24) + 3 * (word4 >> 24);
		green = ((word1 >> 16) & 0xff) + 3 * ((word2 >> 16) & 0xff) +
		        ((word3 >> 16) & 0xff) + 3 * ((word4 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + 3 * ((word2 >> 8) & 0xff) +
		       ((word3 >> 8) & 0xff) + 3 * ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 21) & 0xff000000) |
		               ((green << 13) & 0x00ff0000) |
			       ((blue << 5) & 0x0000ff00);
                break;
            case 12:
		word1 = *pword;
		word3 = *(pword + wpls);
		red = (word1 >> 24) + 3 * (word3 >> 24);
		green = ((word1 >> 16) & 0xff) +
		          3 * ((word3 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) +
		          3 * ((word3 >> 8) & 0xff);
		*(lined + j) = ((red << 22) & 0xff000000) |
		               ((green << 14) & 0x00ff0000) |
			       ((blue << 6) & 0x0000ff00);
                break;
            case 13:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = 3 * (word1 >> 24) + (word2 >> 24) +
		      9 * (word3 >> 24) + 3 * (word4 >> 24);
		green = 3 * ((word1 >> 16) & 0xff) + ((word2 >> 16) & 0xff) +
		        9 * ((word3 >> 16) & 0xff) + 3 * ((word4 >> 16) & 0xff);
		blue = 3 *((word1 >> 8) & 0xff) + ((word2 >> 8) & 0xff) +
		       9 * ((word3 >> 8) & 0xff) + 3 * ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 20) & 0xff000000) |
		               ((green << 12) & 0x00ff0000) |
			       ((blue << 4) & 0x0000ff00);
                break;
            case 14:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = (word1 >> 24) + (word2 >> 24) +
		      3 * (word3 >> 24) + 3 * (word4 >> 24);
		green = ((word1 >> 16) & 0xff) +((word2 >> 16) & 0xff) +
		        3 * ((word3 >> 16) & 0xff) + 3 * ((word4 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + ((word2 >> 8) & 0xff) +
		       3 * ((word3 >> 8) & 0xff) + 3 * ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 21) & 0xff000000) |
		               ((green << 13) & 0x00ff0000) |
			       ((blue << 5) & 0x0000ff00);
                break;
            case 15:
		word1 = *pword;
		word2 = *(pword + 1);
		word3 = *(pword + wpls);
		word4 = *(pword + wpls + 1);
		red = (word1 >> 24) + 3 * (word2 >> 24) +
		      3 * (word3 >> 24) + 9 * (word4 >> 24);
		green = ((word1 >> 16) & 0xff) + 3 * ((word2 >> 16) & 0xff) +
		        3 * ((word3 >> 16) & 0xff) + 9 * ((word4 >> 16) & 0xff);
		blue = ((word1 >> 8) & 0xff) + 3 * ((word2 >> 8) & 0xff) +
		       3 * ((word3 >> 8) & 0xff) + 9 * ((word4 >> 8) & 0xff);
		*(lined + j) = ((red << 20) & 0xff000000) |
		               ((green << 12) & 0x00ff0000) |
			       ((blue << 4) & 0x0000ff00);
                break;
            default:
	        fprintf(stderr, "shouldn't get here\n");
		break;
            }
	}
    }

    return;
}


#if 0
/*!
 *  pixRotateAMColorFast2()
 *
 *         Input: pixs
 *                angle (radians; clockwise is positive)
 *                grayval (0 to bring in BLACK, 255 for WHITE)
 *         Return:  pixd, or null on error
 *
 *         Notes:
 *             - This rotates a color image about the image center.
 *               A positive angle gives a clockwise rotation.
 *             - It uses area mapping, dividing each pixel into
 *               16 subpixels.
 *             - It creates a temporary 32-bit color image.
 *             - It is slightly slower than pixRotateAMColorFast(),
 *               which uses less memory because it does not create
 *               a temporary image.
 */
PIX *
pixRotateAMColorFast2(PIX     *pixs,
	              FLOAT32  angle,
		      UINT8    grayval)
{
INT32      w, h, wpls, wpld;
UINT32    *datas, *datad;
PIX    	  *pixshft, *pixd;

    PROCNAME("pixRotateAMColorFast2");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("pixs must be 32 bpp", procName, NULL);

    if (ABS(angle) < VERY_SMALL_ANGLE)
	return pixClone(pixs);

    if ((pixshft = pixShiftRGB258(pixs)) == NULL)
        return (PIX *)ERROR_PTR("pixshft not defined", procName, NULL);

    w = pixGetWidth(pixshft);
    h = pixGetHeight(pixshft);
    datas = pixGetData(pixshft);
    wpls = pixGetWpl(pixshft);
    pixd = pixCreateTemplate(pixshft);
    datad = pixGetData(pixd);
    wpld = pixGetWpl(pixd);

    rotateAMColorFastLow2(datad, w, h, wpld, datas, wpls, angle, grayval);

    pixDestroy(&pixshft);
    return pixd;
}


/*!
 *   pixShiftRGB256()
 *
 *      Makes a new 32 bpp image with the R, G and B components
 *      right-shifted by 2, 5 and 8 bits, respectively.
 */
PIX *
pixShiftRGB258(PIX  *pixs)
{
INT32     w, h, wpls, wpld, i, j;
UINT32    word;
UINT32   *datas, *datad, *lines, *lined;
PIX      *pixd;

    PROCNAME("pixShift258");

    if (!pixs)
        return (PIX *)ERROR_PTR("pixs not defined", procName, NULL);
    if (pixGetDepth(pixs) != 32)
        return (PIX *)ERROR_PTR("depth not 32 bpp", procName, NULL);
    w = pixGetWidth(pixs);
    h = pixGetHeight(pixs);
    wpls = pixGetWpl(pixs);
    datas = pixGetData(pixs);

    if ((pixd = pixCreate(w, h, 32)) == NULL)
        return (PIX *)ERROR_PTR("pixd not made", procName, NULL);
    wpld = pixGetWpl(pixd);
    datad = pixGetData(pixd);
      
    for (i = 0; i < h; i++) {
        lines = datas + i * wpls;
        lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    word = *(lines + j);
	    *(lined + j) = ((word & 0xff000000) >> 2) |
		           ((word & 0x00ff0000) >> 5) |
		           ((word & 0x0000ff00) >> 8);
        }
    }

    return pixd;
}


/*
 *  Alternative version for fast color rotation
 */
void
rotateAMColorFastLow2(UINT32  *datad,
                      INT32    w,
	              INT32    h,
	              INT32    wpld,
	              UINT32  *datas,
	              INT32    wpls,
	              FLOAT32  angle,
		      UINT8    grayval)
{
INT32         i, j, xcen, ycen, wm2, hm2;
INT32         xdif, ydif, xpm, ypm, xp, yp, xf, yf;
UINT32        edgeval, word;
UINT32       *pword, *lines, *lined;
FLOAT32       sina, cosa;

    xcen = w / 2;
    wm2 = w - 2;
    ycen = h / 2;
    hm2 = h - 2;
    sina = 4. * sin(angle);
    cosa = 4. * cos(angle);

    edgeval = (grayval << 24) | (grayval << 16) | (grayval << 8);
    for (i = 0; i < h; i++) {
	ydif = ycen - i;
	lined = datad + i * wpld;
	for (j = 0; j < w; j++) {
	    xdif = xcen - j;
	    xpm = (INT32)(-xdif * cosa - ydif * sina + 0.5);
	    ypm = (INT32)(-ydif * cosa + xdif * sina + 0.5);
	    xp = xcen + (xpm >> 2);
	    yp = ycen + (ypm >> 2);
	    xf = xpm & 0x03;
	    yf = ypm & 0x03;

		/* if off the edge, write the input grayval */
	    if (xp < 0 || yp < 0 || xp > wm2 || yp > hm2) {
	        *(lined + j) = edgeval;
		continue;
	    }

	    lines = datas + yp * wpls;
	    pword = lines + xp;

	    switch (xf + 4 * yf)
	    {
            case 0:
	        word = *pword;
	        *(lined + j) = ((word & 0x3fc00000) << 2) |
		               ((word & 0x0007f800) << 5) |
			       ((word & 0x000000ff) << 8);
	        break;
            case 1:
	        word = 3 * (*pword) + *(pword + 1);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 2:
	        word = *pword + *(pword + 1);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 3:
	        word = *pword + 3 * (*(pword + 1));
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 4:
	        word = 3 * (*pword) + *(pword + wpls);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 5:
	        word = 2 * (*pword) + *(pword + 1) + *(pword + wpls);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 6:
	        word = *pword + *(pword + 1);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 7:
	        word = *pword + 2 * (*(pword + 1)) + *(pword + wpls + 1);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 8:
	        word = *pword + *(pword + wpls);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 9:
	        word = *pword + *(pword + wpls);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 10:
	        word = *pword + *(pword + 1) + *(pword + wpls) +
		       *(pword + wpls + 1);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 11:
	        word = *(pword + 1) + *(pword + wpls + 1);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 12:
	        word = *pword + 3 * (*(pword + wpls));
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 13:
	        word = *pword + 2 * (*(pword + wpls)) + *(pword + wpls + 1);
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            case 14:
	        word = *(pword + wpls) + *(pword + wpls + 1);
	        *(lined + j) = ((word & 0x7f800000) << 1) |
		               ((word & 0x000ff000) << 4) |
		               ((word & 0x000001fe) << 7);
                break;
            case 15:
	        word = *(pword + 1) + *(pword + wpls) +
		           2 * (*(pword + wpls + 1));
	        *(lined + j) = (word & 0xff000000) |
		               ((word & 0x001fe000) << 3) |
		               ((word & 0x000003fc) << 6);
                break;
            default:  /* for testing only; no interpolation, no shift */
	        fprintf(stderr, "shouldn't get here\n");
                *(lined + j) = *pword;
		break;
            }
	}
    }

    return;
}
#endif

