/*====================================================================*
 -  Copyright (C) 2001 Leptonica.  All rights reserved.
 -  This software is distributed in the hope that it will be
 -  useful, but with NO WARRANTY OF ANY KIND.
 -  No author or distributor accepts responsibility to anyone for the
 -  consequences of using this software, or for whether it serves any
 -  particular purpose or works at all, unless he or she says so in
 -  writing.  Everyone is granted permission to copy, modify and
 -  redistribute this source code, for commercial or non-commercial
 -  purposes, with the following restrictions: (1) the origin of this
 -  source code must not be misrepresented; (2) modified versions must
 -  be plainly marked as such; and (3) this notice may not be removed
 -  or altered from any source or modified source distribution.
 *====================================================================*/

/*
 *  tiffio.c
 *                     
 *        Reading tiff:
 *                PIX       *pixReadStreamTiff()
 *
 *        Writing tiff:
 *                INT32      pixWriteStreamTiff()
 *         static INT32      pixWriteToTiffStream()
 *
 *        Information about tiff file
 *                INT32      fprintTiffInfo()
 *
 *        Open tiff stream from file stream
 *         static TIFF       *fopenTiff()
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
//#include <unistd.h>
#include <fcntl.h>

#include "allheaders.h"


#define   DEFAULT_RESOLUTION     300

    /* static functions with TIFF interfaces */
static TIFF    *fopenTiff(FILE  *fp, const char  *modestr); 
static INT32    pixWriteToTiffStream(TIFF  *tif, PIX  *pix, INT32  type);

/*--------------------------------------------------------------*
 *                     Reading from stream                      *
 *--------------------------------------------------------------*/
/*!
 * pixReadStreamTiff()
 *
 *     Input:  stream
 *     Return: pix, or null on error
 */
PIX *
pixReadStreamTiffName(char *filename)
{
UINT8       *linebuf, *data;
UINT16       spp, bps, bpp, tiffbpl, photometry;
UINT16      *redmap, *greenmap, *bluemap;
INT32        d, wpl, bpl, i, j, k, numcolors;
UINT32       w, h, res, pixel;
UINT32      *line, *ppixel;
FLOAT32      fres;
PIX         *pix;
PIXCMAP     *cmap;
TIFF        *tif;

    PROCNAME("pixReadStreamTiff");

    if ((tif = TIFFOpen(filename, "r") ) == NULL)
	return (PIX *)ERROR_PTR("tif not opened", procName, NULL);
    
	/* use default fields for bps and spp */
    TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bps);
    if (bps > 8)
	return (PIX *)ERROR_PTR("bps > 8", procName, NULL);
    TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
    bpp = bps * spp;
    if (spp == 1)
	d = bps;
    else if (spp == 3)
        d = 32;
    else
	return (PIX *)ERROR_PTR("spp not in set {1,3}", procName, NULL);

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    tiffbpl = TIFFScanlineSize(tif);

    if ((linebuf = (UINT8 *)CALLOC(tiffbpl + 1, sizeof(UINT8))) == NULL)
	return (PIX *)ERROR_PTR("calloc fail for linebuf", procName, NULL);
	
    if ((pix = pixCreate(w, h, d)) == NULL)
	return (PIX *)ERROR_PTR("pix not made", procName, NULL);
    data = (UINT8 *)pixGetData(pix);
    wpl = pixGetWpl(pix);
    bpl = 4 * wpl;

	/* read the data */
    if (spp == 1) {
        for (i = 0 ; i < h ; i++) {
	    if (TIFFReadScanline(tif, linebuf, i, 0) < 0)
	        return (PIX *)ERROR_PTR("line read fail", procName, NULL);
	    memcpy((char *)data, (char *)linebuf, tiffbpl);
	    data += bpl;
        }
        pixEndianByteSwap(pix);
    }
    else {
        line = pixGetData(pix);
        for (i = 0 ; i < h ; i++, line += wpl)
	{
	    if (TIFFReadScanline(tif, linebuf, i, 0) < 0)
	        return (PIX *)ERROR_PTR("line read fail", procName, NULL);
	    for (j = 0, k = 0, ppixel = line; j < w; j++) {
		SET_DATA_BYTE(ppixel, COLOR_RED, linebuf[k++]);
		SET_DATA_BYTE(ppixel, COLOR_GREEN, linebuf[k++]);
		SET_DATA_BYTE(ppixel, COLOR_BLUE, linebuf[k++]);
		ppixel++;
	    } 
        }
    }

    if (TIFFGetField(tif, TIFFTAG_XRESOLUTION, &fres)) {
	res = (UINT32)fres;
	pixSetXRes(pix, res);
    }
    if (TIFFGetField(tif, TIFFTAG_YRESOLUTION, &fres)) {
	res = (UINT32)fres;
	pixSetYRes(pix, res);
    }

    if (TIFFGetField(tif, TIFFTAG_COLORMAP, &redmap, &greenmap, &bluemap))
    {     /* save the colormap */
        numcolors = 1 << bps;
	if ((cmap = pixcmapCreate(numcolors)) == NULL)
	    return (PIX *)ERROR_PTR("cmap not made", procName, NULL);
        for (i = 0; i < numcolors; i++) {
	    composeRGBPixel(redmap[i], greenmap[i], bluemap[i], &pixel);
	    pixcmapAddColor(cmap, pixel);
	}
        pixSetColormap(pix, cmap);
    }
    else {   /* no colormap; check photometry and invert if necessary */
        TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometry);
        if ((d == 1 && photometry == PHOTOMETRIC_MINISBLACK) ||
            (d == 8 && photometry == PHOTOMETRIC_MINISWHITE))
	    pixInvert(pix, pix);
    }

    FREE((char *)linebuf);
    TIFFClose(tif);

    return pix;
}


/*--------------------------------------------------------------*
 *                     Reading from stream                      *
 *--------------------------------------------------------------*/
/*!
 * pixReadStreamTiff()
 *
 *     Input:  stream
 *     Return: pix, or null on error
 */
PIX *
pixReadStreamTiff(FILE   *fp)
{
UINT8       *linebuf, *data;
UINT16       spp, bps, bpp, tiffbpl, photometry;
UINT16      *redmap, *greenmap, *bluemap;
INT32        d, wpl, bpl, i, j, k, numcolors;
UINT32       w, h, res, pixel;
UINT32      *line, *ppixel;
FLOAT32      fres;
PIX         *pix;
PIXCMAP     *cmap;
TIFF        *tif;

    PROCNAME("pixReadStreamTiff");

    if (!fp)
	return (PIX *)ERROR_PTR("stream not defined", procName, NULL);

    if ((tif = fopenTiff(fp, "rb")) == NULL)
	return (PIX *)ERROR_PTR("tif not opened", procName, NULL);
    
	/* use default fields for bps and spp */
    TIFFGetFieldDefaulted(tif, TIFFTAG_BITSPERSAMPLE, &bps);
    if (bps > 8)
	return (PIX *)ERROR_PTR("bps > 8", procName, NULL);
    TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);
    bpp = bps * spp;
    if (spp == 1)
	d = bps;
    else if (spp == 3)
        d = 32;
    else
	return (PIX *)ERROR_PTR("spp not in set {1,3}", procName, NULL);

    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    tiffbpl = TIFFScanlineSize(tif);

    if ((linebuf = (UINT8 *)CALLOC(tiffbpl + 1, sizeof(UINT8))) == NULL)
	return (PIX *)ERROR_PTR("calloc fail for linebuf", procName, NULL);
	
    if ((pix = pixCreate(w, h, d)) == NULL)
	return (PIX *)ERROR_PTR("pix not made", procName, NULL);
    data = (UINT8 *)pixGetData(pix);
    wpl = pixGetWpl(pix);
    bpl = 4 * wpl;

	/* read the data */
    if (spp == 1) {
        for (i = 0 ; i < h ; i++) {
	    if (TIFFReadScanline(tif, linebuf, i, 0) < 0)
	        return (PIX *)ERROR_PTR("line read fail", procName, NULL);
	    memcpy((char *)data, (char *)linebuf, tiffbpl);
	    data += bpl;
        }
        pixEndianByteSwap(pix);
    }
    else {
        line = pixGetData(pix);
        for (i = 0 ; i < h ; i++, line += wpl)
	{
	    if (TIFFReadScanline(tif, linebuf, i, 0) < 0)
	        return (PIX *)ERROR_PTR("line read fail", procName, NULL);
	    for (j = 0, k = 0, ppixel = line; j < w; j++) {
		SET_DATA_BYTE(ppixel, COLOR_RED, linebuf[k++]);
		SET_DATA_BYTE(ppixel, COLOR_GREEN, linebuf[k++]);
		SET_DATA_BYTE(ppixel, COLOR_BLUE, linebuf[k++]);
		ppixel++;
	    } 
        }
    }

    if (TIFFGetField(tif, TIFFTAG_XRESOLUTION, &fres)) {
	res = (UINT32)fres;
	pixSetXRes(pix, res);
    }
    if (TIFFGetField(tif, TIFFTAG_YRESOLUTION, &fres)) {
	res = (UINT32)fres;
	pixSetYRes(pix, res);
    }

    if (TIFFGetField(tif, TIFFTAG_COLORMAP, &redmap, &greenmap, &bluemap))
    {     /* save the colormap */
        numcolors = 1 << bps;
	if ((cmap = pixcmapCreate(numcolors)) == NULL)
	    return (PIX *)ERROR_PTR("cmap not made", procName, NULL);
        for (i = 0; i < numcolors; i++) {
	    composeRGBPixel(redmap[i], greenmap[i], bluemap[i], &pixel);
	    pixcmapAddColor(cmap, pixel);
	}
        pixSetColormap(pix, cmap);
    }
    else {   /* no colormap; check photometry and invert if necessary */
        TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometry);
        if ((d == 1 && photometry == PHOTOMETRIC_MINISBLACK) ||
            (d == 8 && photometry == PHOTOMETRIC_MINISWHITE))
	    pixInvert(pix, pix);
    }

    FREE((char *)linebuf);
    TIFFClose(tif);

    return pix;
}


/*--------------------------------------------------------------*
 *                       Writing to stream                      *
 *--------------------------------------------------------------*/
/*!
 *
 *  pixWriteStreamTiff()
 *
 *      Input:  stream (opened for append or write)
 *              pix
 *              type (IFF_TIFF, IFF_TIFF_PACKBITS,
 *                    IFF_TIFF_G3, IFF_TIFF_G4)
 *      Return: 0 if OK, 1 on error
 *
 *      Notes: For images with bpp > 1, this resets the type, if
 *             necessary, to write uncompressed data.  G3 and G4
 *             are only defined for 1 bpp; PACKBITS works if
 *             bpp > 1 but typically expands images that are not 
 *             synthetically generated.  TIFF G4 is good for binary
 *             compression of text/line-art, but terrible for
 *             halftones and dithered patterns.
 */
INT32
pixWriteStreamTiff(FILE  *fp,
                   PIX   *pix,
		   INT32  type)
{
TIFF	*tif;

    PROCNAME("pixWriteStreamTiff");

    if (!fp)
	return ERROR_INT("stream not defined", procName, 1 );
    if (!pix)
	return ERROR_INT("pix not defined", procName, 1 );

	/* only use compression on binary images */
    if (pixGetDepth(pix) != 1)
	type = IFF_TIFF;

    if ((tif = fopenTiff(fp, "w")) == NULL)
	return ERROR_INT("tif not opened", procName, 1);

    pixWriteToTiffStream(tif, pix, type);
    
    TIFFClose(tif);
    return 0;
}


INT32
pixWriteStreamTiffName(const char *filename,
           PIX   *pix,
		   INT32  type)
{
TIFF	*tif;

    PROCNAME("pixWriteStreamTiff");

    if (!pix)
	return ERROR_INT("pix not defined", procName, 1 );

	/* only use compression on binary images */
    if (pixGetDepth(pix) != 1)
	type = IFF_TIFF;

    if ((tif = TIFFOpen(filename, "w")) == NULL)
	return ERROR_INT("tif not opened", procName, 1);

    pixWriteToTiffStream(tif, pix, type);
    
    TIFFClose(tif);
    return 0;
}

/*!
 *
 *  pixWriteToTiffStream()
 *
 *      Input:  tif (data structure, opened to a file)
 *              pix
 *              type  (IFF_TIFF: for any image; no compression
 *                     IFF_TIFF_PACKBITS: for any image except 32 bpp
 *                     IFF_TIFF_G4 and IFF_TIFF_G3: for binary only)
 *      Return: 0 if OK, 1 on error
 *
 *      Note: no compression on 32 bpp
 *
 */
static INT32
pixWriteToTiffStream(TIFF   *tif,
                     PIX    *pix,
                     INT32   type)
{
UINT8       *rmap, *gmap, *bmap;
UINT16       redmap[256], greenmap[256], bluemap[256];
UINT32       xres, yres;
INT32        w, h, d, i, j, k, wpl, bpl, tiffbpl, numcolors;
UINT8       *linebuf, *data;
UINT32      *line, *ppixel;
PIXCMAP     *cmap;

    PROCNAME("pixWriteToTiffStream");

    if (!tif)
	return ERROR_INT("tif stream not defined", procName, 1);
    if (!pix)
	return ERROR_INT( "pix not defined", procName, 1 );

    w = pixGetWidth(pix);
    h = pixGetHeight(pix);
    d = pixGetDepth(pix);
    xres = pixGetXRes(pix);
    yres = pixGetYRes(pix);
    if (xres == 0) xres = DEFAULT_RESOLUTION;
    if (yres == 0) yres = DEFAULT_RESOLUTION;

	/* ------------------ write out the header -------------  */
    TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, (UINT32)RESUNIT_INCH);
    TIFFSetField(tif, TIFFTAG_XRESOLUTION, (FLOAT64)xres);
    TIFFSetField(tif, TIFFTAG_YRESOLUTION, (FLOAT64)yres);

    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, (UINT32)w);
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH, (UINT32)h);
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);

    if (d == 1)
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISWHITE);
    else if (d == 32) {
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
        TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,
	               (UINT16)8, (UINT16)8, (UINT16)8);
        TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (UINT16)3);
    }
    else if ((cmap = pixGetColormap(pix)) == NULL)
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    else {  /* colormap */
        if (colormapToArrays(cmap, &rmap, &gmap, &bmap, &numcolors))
	    return ERROR_INT( "undefined cmap type", procName, 1 );
        if (numcolors > 256) {
	    WARNING("number of colors > 256", procName);
	    numcolors = 256;
	}
	for (i = 0; i < numcolors; i++) {
	    redmap[i] = (rmap[i] << 8) | rmap[i];
	    greenmap[i] = (gmap[i] << 8) | gmap[i];
	    bluemap[i] = (bmap[i] << 8) | bmap[i];
	}
	FREE((void *)rmap);
	FREE((void *)gmap);
	FREE((void *)bmap);

	TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
	TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (UINT16)1);
	TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, (UINT16)d);
	TIFFSetField(tif, TIFFTAG_COLORMAP, redmap, greenmap, bluemap);
    }

    if (d != 32) {
        TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, (UINT16)d);
        TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (UINT16)1);
    }

    TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG); 
    if (type == IFF_TIFF)  /* no compression */
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    else if (type == IFF_TIFF_G4)
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX4);
    else if (type == IFF_TIFF_G3)
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_CCITTFAX3);
    else if (type == IFF_TIFF_PACKBITS)
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
    else { 
	WARNING("unknown tiff compression; using none", procName);
	TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
    }

	/* ------------- write out the image data -------------  */
    tiffbpl = TIFFScanlineSize(tif);
    wpl = pixGetWpl(pix);
    bpl = 4 * wpl;
    if (tiffbpl > bpl)
	fprintf(stderr, "Big trouble: tiffbpl = %d, bpl = %d\n", tiffbpl, bpl);
    if ((linebuf = (UINT8 *)CALLOC(1, bpl)) == NULL)
	return ERROR_INT("calloc fail for linebuf", procName, 1);

	/* use single strip for image */
    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, h);

    if (d != 32) {
        pixEndianByteSwap(pix);
        data = (UINT8 *)pixGetData(pix);
        for (i = 0 ; i < h; i++, data += bpl) {
            memcpy((char *)linebuf, (char *)data, tiffbpl);
	    if (TIFFWriteScanline(tif, linebuf, i, 0) < 0)
	        break;
        }
        pixEndianByteSwap(pix);
    }
    else {
	line = pixGetData(pix);
        for (i = 0 ; i < h; i++, line += wpl) {
	    line = pixGetData(pix) + i * wpl;
	    for (j = 0, k = 0, ppixel = line; j < w; j++) {
		linebuf[k++] = GET_DATA_BYTE(ppixel, COLOR_RED);
		linebuf[k++] = GET_DATA_BYTE(ppixel, COLOR_GREEN);
		linebuf[k++] = GET_DATA_BYTE(ppixel, COLOR_BLUE);
		ppixel++;
	    } 
	    if (TIFFWriteScanline(tif, linebuf, i, 0) < 0)
	        break;
        }
    }

    FREE((char *)linebuf);

    return 0;
}



/*--------------------------------------------------------------*
 *                   Print info to stream                       *
 *--------------------------------------------------------------*/
/*
 *  fprintTiffInfo()
 * 
 *    Input:  stream (for output of tag data)
 *            tiffile (input)
 *    Return: 0 if OK; 1 on error
 */
INT32
fprintTiffInfo(FILE        *fpout,
	       const char  *tiffile)
{
TIFF	*tif;

    PROCNAME("fprintTiffInfo");

    if (!tiffile)
	return ERROR_INT("tiffile not defined", procName, 1);
    if (!fpout)
	return ERROR_INT("stream out not defined", procName, 1);

    if ((tif = TIFFOpen(tiffile, "rb")) == NULL)
	return ERROR_INT("tif not open for read", procName, 1);

    TIFFPrintDirectory(tif, fpout, 0);
    TIFFClose(tif);

    return 0;
}



/*--------------------------------------------------------------*
 *               Open tiff stream from file stream              *
 *--------------------------------------------------------------*/
/*!
 *  fopenTiff()
 *
 *     Input:  stream
 *             modestring ("r", "w", ...)
 *     Return: tiff (data structure, opened for a file descriptor)
 *
 *     Why is this here?  Leffler did not provide a function that
 *     takes a stream and gives a TIFF.  He only gave one that
 *     generates a TIFF starting with a file descriptor.  So we
 *     need to make it here, because it is useful to have functions
 *     that take a stream as input.
 *
 *     Note: requires lseek to rewind to BOF; fseek won't hack it.
 *    
 */
static TIFF *
fopenTiff(FILE        *fp,
	  const char  *modestring)
{
INT32    fd;

    PROCNAME("fopenTiff");

    if (!fp)
        return (TIFF *)ERROR_PTR("stream not opened", procName, NULL);
    if (!modestring)
        return (TIFF *)ERROR_PTR("modestring not defined", procName, NULL);

    if ((fd = fileno(fp)) < 0)
        return (TIFF *)ERROR_PTR("invalid file descriptor", procName, NULL);
    lseek(fd, 0, SEEK_SET);

    return TIFFFdOpen(fd, "TIFFstream", modestring);
}

