#include "stdafx.h"
#include "Classtif.h"

//Image type
enum { DIB_1BIT, DIB_4BIT, DIB_8BIT, DIB_16BIT, DIB_24BIT, DIB_32BIT, DIB_UNKNOWN };
//Palette type
enum{ WHITEISZERO, BLACKISZERO, RGB, PALETTE };


/************************************************************************************
/*	Name:		FreeMem
/*	Function:	Free memory
/*	Parameter:	none
/*	return:		0
/***********************************************************************************/
int FreeMem( HGLOBAL &m_hImg )
{
    if ( m_hImg != (HGLOBAL)NULL )
		GlobalFree(m_hImg) ;
    m_hImg = (HGLOBAL)NULL ;
    return 0 ;
}

/************************************************************************************
/*	Name:		GetPaletteSize
/*	Function:	Get palette size
/*	Parameter:	None
/*	return:		Palette size
/*
/***********************************************************************************/
int GetPaletteSize ( LPBITMAPINFO lpbi )
{
    switch ( lpbi->bmiHeader.biBitCount )
    {
        case 1 :
			return 2*sizeof(RGBQUAD) ;
		case 4 :
			return 16*sizeof(RGBQUAD) ;
        case 8 :
            return 256*sizeof(RGBQUAD) ;
        case 16 :
        case 24 :
        case 32 :
            return 0 ;
        default :
            return -1 ;
    }
}

/************************************************************************************
/*	Name:		GetLineSizeInByte
/*	Function:	Get the number of bytes used to represent an image line
/*	Parameter:	lpbi -- Pointer to BITMAPINFO
/*	return:		Number of bytes used to represent an image line
/*
/***********************************************************************************/
int GetLineSizeInBytes(LPBITMAPINFO lpbi)
{
    int nWidth  = lpbi->bmiHeader.biWidth ;
    switch ( lpbi->bmiHeader.biBitCount )
    {
        case 1 :
            return (nWidth+31)/32*4 ;
		case 4 :
			return ((nWidth+1)/2+3)/4*4 ;
        case 8 :
            return (nWidth+3)/4*4 ;
        case 16 :
            return (nWidth*2+3)/4*4 ;
        case 24 :
            return (nWidth*3+3)/4*4 ;
        case 32 :
            return nWidth*4 ;
        default :
            return -1 ;
    }
}

/************************************************************************************
/*	Name:		GetType
/*	Function:	Get image type
/*	Parameter:	None
/*	return:		Image type
/*
/***********************************************************************************/
int GetType(HGLOBAL m_hImg)
{
    if ( m_hImg == (HGLOBAL)NULL )	return -1;
    LPBITMAPINFO lpbi = (LPBITMAPINFO)GlobalLock(m_hImg) ;
	int m_nBitCount  = lpbi->bmiHeader.biBitCount ;
    GlobalUnlock(m_hImg) ;
    switch ( m_nBitCount )
    {
        case 1 :
            return DIB_1BIT ;
		case 4:
			return DIB_4BIT ;
        case 8 :
            return DIB_8BIT ;
        case 16 :
            return DIB_16BIT ;
        case 24 :
            return DIB_24BIT ;
        case 32 :
            return DIB_32BIT ;
        default :
            return DIB_UNKNOWN ;
    }
}

/************************************************************************************
/*	Name:		GetWidth
/*	Function:	Get image width in pixels
/*	Parameter:	None
/*	return:		Image width
/*
/***********************************************************************************/
int GetWidth(HGLOBAL m_hImg)
{
    if ( m_hImg != (HGLOBAL)NULL )
    {
        LPBITMAPINFO lpbi = (LPBITMAPINFO)GlobalLock(m_hImg) ;
	    int nWidth  = lpbi->bmiHeader.biWidth ;
        GlobalUnlock(m_hImg) ;
        return nWidth ;
    }
    else
        return -1;
}

/************************************************************************************
/*	Name:		GetHeight
/*	Function:	Get image height in pixels
/*	Parameter:	None
/*	return:		Image height
/*
/***********************************************************************************/
int GetHeight(HGLOBAL m_hImg)
{
    if ( m_hImg != (HGLOBAL)NULL )
    {
        LPBITMAPINFO lpbi = (LPBITMAPINFO)GlobalLock(m_hImg) ;
	    int nHeight  = lpbi->bmiHeader.biHeight ;
        GlobalUnlock(m_hImg) ;
        return nHeight ;
    }
    else
        return -1;
}

/******************************************************************************
/*	Name:		GetFileExt
/*	Function:	Get file extension
/*	Parameter:	FileName -- File name
/*	Return:		File extension
/*
/*****************************************************************************/
CString GetFileExt ( CString FileName )
{
    int pos ;
    if ( (pos=FileName.ReverseFind('.')) >= 0 )
        return FileName.Right(FileName.GetLength()-pos-1) ;
    else
        return "" ;
}

/******************************************************************************
/*	Name:		IsColor
/*	Function:	Retrieval the palette to see if the image is color
/*	Parameter:	void
/*	Return:		TRUE -- Color image
/*				FALSE -- Not color image
/*****************************************************************************/
BOOL IsColor( HGLOBAL m_hImg )
{
	LPBITMAPINFO lpbi = (LPBITMAPINFO) GlobalLock ( m_hImg ) ;
	//If not gray image, return FALSE, do not invert pallete.
	int	nPalette;
	switch( GetType( m_hImg ) )
	{
	case DIB_1BIT:
	case DIB_UNKNOWN:
		return FALSE;
	case DIB_4BIT:
		nPalette = 1<<4;
		break;
	case DIB_8BIT:
		nPalette = 1<<8;
		break;
	case DIB_16BIT:
	case DIB_24BIT:
	case DIB_32BIT:
		return TRUE;
	default:
		return FALSE;
	}

	for( int i=0; i<nPalette; i++ )
		if( lpbi->bmiColors[i].rgbBlue != lpbi->bmiColors[i].rgbGreen || lpbi->bmiColors[i].rgbBlue != lpbi->bmiColors[i].rgbRed )
			return TRUE;
	return FALSE;
}
/************************************************************************************
/*	Name:		Read
/*	Function:	Read image
/*	Parameter:	csFileName -- Image file name;
/*	return:		0 -- Succeed
/***********************************************************************************/
int ReadImage(char* csFileName,  HGLOBAL &m_hImg)
{
    CString csExt = GetFileExt( csFileName ) ;
    csExt.MakeLower() ;
	
TRY
{
    if ( csExt=="tif" || csExt=="tiff" )
    {
        CTiff Tiff (csFileName, CFile::modeRead|CFile::shareDenyWrite ) ;
        HGLOBAL hTiff = Tiff.readimage() ;
       	if ( hTiff != (HGLOBAL)NULL )
        {
            m_hImg = Tiff.convertToDIB ( hTiff ) ;
            Tiff.release(hTiff);
        }
		else
			return -1 ;
    }
	else 
		return -1;

	return 0 ;
}
CATCH ( CFileException, e )
{
    return -1 ;
}
END_CATCH
}

/************************************************************************************
/*	Name:		WriteBMP
/*	Function:	Write Bitmap image file
/*	Parameter:	theFile -- Image file
/*	return:		0  -- Succeed
/*				-1 -- Failed
/*
/***********************************************************************************/
int WriteBMP(CFile *theFile,  HGLOBAL m_hImg)
{
    ASSERT_VALID ( theFile ) ;
TRY
{
	if ( m_hImg != (HGLOBAL)NULL )
    {
        BITMAPFILEHEADER FileHeader ;

        LPBITMAPINFO lpbi = (LPBITMAPINFO)GlobalLock( m_hImg ) ;
        int nDIBSize = sizeof(BITMAPINFOHEADER) + GetPaletteSize(lpbi) + GetLineSizeInBytes(lpbi)*lpbi->bmiHeader.biHeight ;
        FileHeader.bfType = 0x4d42 ;
        FileHeader.bfSize = sizeof(FileHeader) + nDIBSize ;
        FileHeader.bfReserved1 = 0 ;
        FileHeader.bfReserved2 = 0 ;
        FileHeader.bfOffBits = sizeof(FileHeader) + sizeof(BITMAPINFOHEADER) + GetPaletteSize(lpbi) ;
        theFile->Write(&FileHeader, sizeof(FileHeader));
		
		//BMP file has different definition of resolution
		//In BMP the resolution is pixels per meter, instead of pixels per inch.
		long 	biXPelsPerMeter = lpbi->bmiHeader.biXPelsPerMeter;
		long	biYPelsPerMeter = lpbi->bmiHeader.biYPelsPerMeter;

		lpbi->bmiHeader.biXPelsPerMeter = (long)(lpbi->bmiHeader.biXPelsPerMeter/0.0254);
		lpbi->bmiHeader.biYPelsPerMeter = (long)(lpbi->bmiHeader.biYPelsPerMeter/0.0254);
        theFile->WriteHuge(lpbi, nDIBSize) ;
		lpbi->bmiHeader.biXPelsPerMeter = biXPelsPerMeter;
		lpbi->bmiHeader.biYPelsPerMeter = biYPelsPerMeter;
        GlobalUnlock (m_hImg) ;
        return 0 ;
    }
    else
        return -1 ;
}
CATCH ( CFileException, e )
{
    return -1 ;
}
END_CATCH
}

/************************************************************************************
/*	Name:		Write
/*	Function:	Write image file
/*	Parameter:	csFileName -- Image file name
/*	return:		0  -- Succeed
/*				-1 -- Failed
/*
/***********************************************************************************/
int WriteImage(char* csFileName,  HGLOBAL m_hImg)
{
    CString csExt = GetFileExt(csFileName) ;
    csExt.MakeLower() ;
	int	option;
 	switch( GetType( m_hImg ) )
	{
	case	DIB_1BIT:
				option = 0x0f04;	//G4 compression
//				option = 0x0704;	//No compression
//				option = 0x0B04;	//LZW compression
				break;
	case	DIB_4BIT:
				option = 0x0b10;
				break;
	case	DIB_8BIT:
				option = 0x0b20;
				break;
	case	DIB_24BIT:
				option = 0x0860;
				break;
	default:
			return -1;
	}

    if ( csExt=="tif" || csExt=="tiff" )
    {
		if ( m_hImg != (HGLOBAL)NULL )
		{
			CTiff Tiff (csFileName, CFile::modeCreate | CFile::modeWrite |CFile::shareDenyRead) ;
			HANDLE handle=Tiff.convertFromDIB(m_hImg);
 			long metric = (GetHeight(m_hImg)<<16) + GetWidth( m_hImg );
			Tiff.writeimage(metric , option ,handle);
		}
		else 
			return -1;
    }
    if ( csExt == "bmp" )
    {
		CFile *theFile ;
		TRY
		{
			theFile = new CFile ( csFileName, CFile::modeCreate|CFile::modeWrite|CFile::shareDenyRead ) ;
		}
		CATCH( CFileException, e )
		{
		  CString sErrMsg = "File " + CString(csFileName) + " could not be opened " ;
		  return -1 ;
		}
		END_CATCH    
		ASSERT_VALID ( theFile ) ;
		int rt = WriteBMP ( theFile, m_hImg) ;
		delete theFile ;
		return rt ;
    }
	else 
		return -1;
    return 0 ;
}

