/////////////////////////////////////////////////////////////////////////
// File Name:		DirLineDSCC.cpp
// File Function:	Build Directional Single Connected Chain
//
//				Developed by: Zheng Yefeng
//			(c)copyright to Tsinghua THOCR Lab, 2001
//					All Right Reserved
//////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DirLine.h"

IMPLEMENT_SERIAL(CDirLine, CObject, 0)

CDirLine::CDirLine()
{		
	m_nCurTree = 0;
	m_pTree = NULL;
	ZeroMemory(CharWidth, sizeof(int)*128);
	ZeroMemory(CharHeight, sizeof(int)*128);
	m_nStrip = 1;
}

CDirLine::~CDirLine()
{
    FreeMem();
}

/****************************************************************************
/*	Name:		FreeMem
/*	Function:	Free memory
/*	Parameter:	void
/*	return:		0 -- Succeed
/*				1 -- Error
/*
/***************************************************************************/
int CDirLine::FreeMem()
{
	if( m_pTree != NULL)
		for(int i = 0; i <= m_nCurTree; i++) 
			delete m_pTree[i]; 
	free(m_pTree);
	m_pTree = NULL;
	m_nCurTree = 0;
	ZeroMemory(CharWidth, sizeof(int)*128);
	ZeroMemory(CharHeight, sizeof(int)*128);
	m_nCharHeight = 0;
	m_nCharWidth = 0;
    return 0 ;
}
/***************************************************************************
/*	Name:		BuildConnTree
/*	Function:	Extract runlengths 
/*	Parameter:	Img -- Form image
/8				rRange -- Image region to analysed
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::BuildConnTree ( CImage& Img, RECT rRange )
{
    int  ImgType = Img.GetType() ;
    if ( ImgType!=DIB_1BIT && ImgType!=DIB_8BIT && ImgType!= DIB_24BIT)
        return -2 ;
    int w = Img.GetWidth() ;
    int h = Img.GetHeight() ;
    int wb = Img.GetLineSizeInBytes() ;
	BOOL	bUpSideDown = Img.GetOrientation();
    if ( rRange.top<0 || rRange.top>=h || rRange.bottom<0 || rRange.bottom>=h || rRange.bottom<rRange.top ||
         rRange.left<0 || rRange.left>=w || rRange.right<0 || rRange.right>=w || rRange.right<rRange.left )
        return -3 ;

    int j, k, st, ed, span, valleys ;

    if ( m_bIsHorLine )
    {
        st = rRange.left ;
        ed = rRange.right ;
        span = rRange.bottom-rRange.top+1 ;
    }
    else
    {
        st = rRange.top ;
        ed = rRange.bottom ;
        span = rRange.right-rRange.left+1 ;
    }

    if ( !m_pTree[m_nCurTree]->Initialize(rRange) )
        return -1 ;

    int *t = (int *)malloc(sizeof(int)*span+16) ;
    if ( t == (int *)NULL ) return -1;
    int *s = (int *)malloc(sizeof(int)*span+16) ;
    if ( s == (int *)NULL ) { free(t); return -1 ; }
    LineValley *v = (LineValley *)malloc(sizeof(LineValley)*span*2+16) ;
    if ( v == (LineValley *)NULL ) { free(s); free(t); return -1; }

    BYTE *p = Img.LockRawImg() ;
    if ( p==(BYTE *)NULL || wb<=0 )
	{ free(v); free(s); free(t); return -1; }
	
    for ( j=st ; j<=ed ; j++ )
    {
		if ( ImgType == DIB_1BIT )
        {
            if ( m_bIsHorLine )
				valleys = GetColumnRunLength ( p, wb, h, j, rRange.top, rRange.bottom, v, bUpSideDown ) ;
			else
			   valleys = GetRowRunLength ( p, wb, h, j, rRange.left, rRange.right, v, bUpSideDown) ;
		} 
	    else
            break ;

        for ( k=0 ; k<valleys ; k++ )
		{
			v[k].x = j ;
			int vx = j;
			if( ImgType != DIB_1BIT )	continue;
			if( m_bIsHorLine )
			{
				if( v[k].ye == rRange.bottom && v[k].ye < h-1 )
				{
					BYTE mask = 128>>(vx%8) ;
					while(v[k].ye < h-1 )
					{
						BYTE	DownLineByte = *( p+(h-1-v[k].ye-1)*wb+vx/8 );
						if ( (DownLineByte & mask) == 0 )
								break;
						v[k].ye++;
					}
				}
				if( v[k].ys == rRange.top && v[k].ys > 0 )
				{
					BYTE mask = 128>>(vx%8) ;
					while(v[k].ys > 0 )
					{
						BYTE	UpLineByte = *( p+(h-1-v[k].ys+1)*wb+vx/8 );
						if( (UpLineByte & mask) == 0 )	break;
						v[k].ys --;
					}
				}
			}
			else
			{ 
				if( v[k].ye == rRange.right && v[k].ye < w-1 )
				{
					BYTE	mask = 128>>( (v[k].ye+1)%8 );
					while( v[k].ye < w-1 )
					{
						BYTE	RightPixelByte = *( p+(h-1-vx)*wb+(v[k].ye+1)/8 );
						if ( (RightPixelByte & mask) == 0 )	break;
						v[k].ye++;
					}
				}
				if( v[k].ys == rRange.left && v[k].ys > 0 )
				{
					BYTE	mask = 128>>( (v[k].ys-1)%8 );
					while(v[k].ys > 0 )
					{
						BYTE	LeftPixelByte = *( p+(h-1-vx)*wb+(v[k].ys-1)/8 );
						if( (LeftPixelByte & mask) == 0 )		break;
						v[k].ys--;
					}
				}
			}
		}
		m_pTree[m_nCurTree]->AddNewCol ( v, valleys, j ) ;
	}
     
	free(v); 
	free(s); 
	free(t); 
	Img.UnlockRawImg();
    return 0 ;
}

/***************************************************************************
/*	Name:		AquireHorLineData
/*	Function:	Read a vertical image line data to build horizontal DSCC
*	Parameter:	p -- Pointer to image
/*				w -- Width of the whole image
/*				h -- Height of the whole image
/*				n -- X offset
/*				start -- Start line
/*				end -- End line
/*				buf -- Buffer for read data
/*				ImgType -- Image type
/*				Channel -- Channel for Color image
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::AquireHorLineData ( BYTE *p, int w, int h, int n, int start, int end, int *buf, int ImgType, int Channel, BOOL bUpSideDown )
{
    int i ;
	if ( ImgType == DIB_1BIT )
	{
		BYTE mask = 128>>(n%8) ;
		if( bUpSideDown )
		{
			p += w*(h-1-start) + n/8 ;   // DIB: upside down
			for ( i=start ; i<=end ; i++, p-=w )
				if ( *p & mask )
					*buf++ = 0 ;
				else
					*buf++ = 255 ;
		}
		else
		{
			p += w*start + n/8;
			for ( i=start ; i<=end ; i++, p+=w )
				if ( *p & mask )
					*buf++ = 0 ;
				else
					*buf++ = 255 ;
		}
	}  
	else if ( ImgType == DIB_8BIT )
    {
		if( bUpSideDown )
		{
			p += w*(h-1-start) + n ;   // DIB: upside down
			for ( i=start ; i<=end ; i++, p-=w )
				*buf++ = *p ;
		}
		else
		{//Not upside down
			p += w*start + n;
			for( i=start; i<=end ; i++, p+=w )
				*buf++ = *p;
		}
    }
    else if ( ImgType == DIB_24BIT)
	{
		if( bUpSideDown )
		{
			p += w*(h-1-start) + 3*n;
			for ( i=start; i<=end ; i++, p-=w )
				*buf++ = *(p+Channel);
		}
		else
		{
			p += w*start + 3*n;
			for ( i=start; i<=end ; i++, p+=w )
				*buf++ = *(p+Channel);
		}
	}
	else
        return -1 ;
    return 0 ;
}

/***************************************************************************
/*	Name:		AquireVerLineData
/*	Function:	Read a horizontal image line data to build vertical DSCC
/*	Parameter:	p -- Pointer to image
/*				w -- Width of the whole image
/*				h -- Height of the whole image
/*				n -- Line
/*				start -- Start X ordinate
/*				end -- End X ordinate
/*				buf -- Buffer for read data
/*				ImgType -- Image type
/*				Channel -- Channel for Color image
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::AquireVerLineData ( BYTE *p, int w, int h, int n, int start, int end, int *buf, int ImgType, int Channel, BOOL bUpSideDown )
{
    int i, j ;
	if ( ImgType == DIB_1BIT )
    {
        BYTE mask[8] = {128,64,32,16,8,4,2,1} ;
        BYTE tmp ;
		if( bUpSideDown )
			p += w*(h-1-n) + start/8 ;
		else
			p += start/8 ;
		for ( i=start, j=start%8, tmp=*p++ ; i<=end ; i++ )
			{
				if ( tmp & mask[j] )
					*buf++ = 0 ;
				else
					*buf++ = 255 ;
				j++ ;
				if ( j == 8 )
				{
					tmp= *p++ ;
					j = 0 ;
				}
			}
    }
	else if ( ImgType == DIB_8BIT )
    {
		if( bUpSideDown )
			p += w*(h-1-n) + start ;   // DIB: upside down
		else 
			p += start;
        for ( i=start ; i<=end ; i++ )
            *buf++ = *p++ ;
    }
	else if( ImgType == DIB_24BIT )
	{
		if( bUpSideDown )
			p += w*(h-1-n) + 3*start;
		else
			p += 3*start;
		for ( i=start; i<=end; i++)
		{
			*buf++ = *(p+Channel);
			p += 3;
		}
	}
    else
        return -1 ;

    return 0 ;
}

/***************************************************************************
/*	Name:		GetColumnRunLength
/*	Function:	Build column runlength
/*	Parameter:
/*	Return:		0  -- Succeed
/*				-1 -- Error
/***************************************************************************/
int CDirLine::GetColumnRunLength ( BYTE *p, int w, int h, int column, int start, int end, LineValley *valley, BOOL bUpSideDown )
{
    int i, Valleys=0 ;
    BOOL bInValley ;

    BYTE mask = 128>>(column%8) ;
	if( bUpSideDown )
		p += w*(h-1-start) + column/8 ;   // DIB: upside down
	else
		p += w*start + column/8;

    for ( i=start, bInValley=FALSE ; i<=end ; i++ )
    {
        if ( *p & mask )
        {
            if ( !bInValley )
            {
                valley->ys = valley->yvs = i ;
                valley->EdgeGray = 255 ;
                valley->gray = 0 ;
                bInValley = TRUE ;
            }
        }
        else
        {
            if ( bInValley )
            {
                valley->ye = valley->yve = i-1 ;
                valley++ ;
                Valleys++ ;
                bInValley = FALSE ;
            }
        }
		if( i == end )	break;
		if( bUpSideDown )	p-=w;
		else				p+=w;
    }
    if ( bInValley )
    {
        valley->ye = valley->yve = i ;
        Valleys++ ;
    }
    return Valleys ;
}

/***************************************************************************
/*	Name:		GetRowRunLength
/*	Function:	Build row runlength
/*	Parameter:
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::GetRowRunLength ( BYTE *p, int w, int h, int row, int start, int end, LineValley *valley, BOOL bUpSideDown )
{
    int i, j, Valleys=0 ;
    BOOL bInValley ;

    BYTE mask[8] = {128,64,32,16,8,4,2,1} ;
    BYTE tmp ;
	if( bUpSideDown )
		p += w*(h-1-row) + start/8 ;
	else
		p += w*row + start/8 ;

    for ( i=start, j=start%8, bInValley=FALSE, tmp=*p++ ; i<=end ; i++ )
    {
        if ( tmp & mask[j] )
        {
            if ( !bInValley )
            {
                valley->ys = valley->yvs = i ;
                valley->EdgeGray = 255 ;
                valley->gray = 0 ;
                bInValley = TRUE ;
            }
        }
        else
        {
            if ( bInValley )
            {
                valley->ye = valley->yve = i-1 ;
                valley++ ;
                Valleys++ ;
                bInValley = FALSE ;
            }
        }
        j++ ;
        if ( j == 8 )
        {
            tmp= *p++ ;
            j = 0 ;
        }
    }
    if ( bInValley )
    {
        valley->ye = valley->yve = i-1 ;
        Valleys++ ;
    }
    return Valleys ;
}

/***************************************************************************
/*	Name:		ColRunLenSmooth
/*	Function:	Column runlength smoothing
/*	Parameter:
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::ColRunLenSmooth(BYTE *p, int wb, int h, int col, LineValley *valley, int& Valleys)
{
	int T1 = 3; 
	int T2 = 6;
	int MinLen = 20;

    BYTE mask = 128>>(col%8) ;
	int i=0, j;
	while(i<Valleys-1)
	{
		if(valley[i+1].ys-valley[i].ye <= T1 || 
		(valley[i+1].ys-valley[i].ye<=T2 && valley[i].ye-valley[i].ys > MinLen && valley[i+1].ye-valley[i+1].ys > MinLen) )
		{
			for(j=valley[i].ye+1; j<=valley[i+1].ys-1; j++)
				*(p+wb*(h-1-j) + col/8) |= mask;
			valley[i].ye = valley[i+1].ye;
			valley[i].yve = valley[i+1].yve;
			for(j=i+1; j<Valleys-1; j++)
				valley[j] = valley[j+1];
			Valleys--;
			continue;
		}
		i++;
	}
	return 0;
}

/***************************************************************************
/*	Name:		RowRunLenSmooth
/*	Function:	Row runlength smoothing
/*	Parameter:
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::RowRunLenSmooth(BYTE *p, int wb, int h, int row,LineValley *valley, int& Valleys)
{
	int T1 = 3; 
	int T2 = 6;
	int MinLen = 20;

	BYTE mask[8] = {128,64,32,16,8,4,2,1} ;
	p+=wb*(h-1-row);
	int i=0, j;
	while(i<Valleys-1)
	{
		if(valley[i+1].ys-valley[i].ye <= T1 || 
		(valley[i+1].ys-valley[i].ye<=T2 && valley[i].ye-valley[i].ys > MinLen && valley[i+1].ye-valley[i+1].ys > MinLen) )
		{
//			for(j=valley[i].ye; j<=valley[i+1].ys-1; j++) //Change the raw imgae.
//				*(p+j/8) |= mask[j%8];
			valley[i].ye = valley[i+1].ye;
			valley[i].yve = valley[i+1].yve;
			for(j=i+1; j<Valleys-1; j++)
				valley[j] = valley[j+1];
			Valleys--;
			continue;
		}
		i++;
	}
	return 0;
}

/***************************************************************************
/*	Name:		ValleyDetect
/*	Function:	Detect valley on a image line for binary/gray image
/*	Parameter:
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*
/***************************************************************************/
int CDirLine::ValleyDetect ( int *p, int start, int end, LineValley *valley, int *s, int Depth, int MaxLen )
{//0 --- black             255 --- white
    BOOL dire, InValley ;
    int i, j, nadir_pos, vsp, vep, vs, nadir, mid, t, l ;
    int Valleys = 0 ;
    int sp = 0 ;

    nadir = vs = l = *p ;
    nadir_pos = vsp = start ;
    dire = FALSE ;
    InValley = FALSE ;

    for ( i=start ; i<=end ; i++, p++ )  // DIB: upside down
    {
        t = *p ;
        if ( InValley )
            s[sp++] = t ;
        if ( dire )     // downward
        {
            if ( t > l )
            {
                if ( l < nadir )
                {
                    nadir = l ;
                    nadir_pos = i-1 ;
                }
                if ( !InValley )
                {
                    if ( vs-l > Depth )
                    {
                        InValley = TRUE ;
                        for ( j=i-vsp ; j>=0 ; j-- )
                            s[sp++] = *(p-j) ;
                    }
                }
                dire = FALSE ;
            }
            else
            {
                if ( !InValley )
                    if ( i-vsp > MaxLen  )
                        if ( vs-t < Depth )
                        {
                            vsp = i-1 ;
                            vs = l ;
                            nadir_pos= i ;
                            nadir = t ;
                        }
            }
        }
        else    // upward
        {
            if ( t < l )
            {
                if ( !InValley )
                {
                    mid = (nadir+vs*2)/3 ;
                    if ( l>=mid )
                    {
                        vsp = i-1 ;
                        vs = l ;
                        nadir = t ;
                        nadir_pos = i ;
                    }
                }
                else
                {
                    mid = (nadir+vs*2)/3 ;
                    if ( l>=mid )
                    {
                        InValley = FALSE ;
                        valley->ys = vsp ;
                        valley->ye = i-1 ;
                        int count=0 ;
                        mid = max(l,vs)-Depth ;
                        for ( j=sp-2 ; j>=0 ; j-- )
                            if ( s[j] <= mid )
                            {
                                vep = vsp+j ;
                                break ;
                            }
                            else
                            {
                                count++ ;
                                if ( count == MaxLen )
                                {
                                    valley->ye -= MaxLen ;
                                    count = 0 ;
                                }
                            }
                        mid = vs-Depth ;
                        for ( j=0 ; j<sp ; j++ )
                            if ( s[j] <= mid )
                            {
                                vsp += j ;
                                break ;
                            }
                        valley->yvs = vsp ;
                        valley->yve = vep ;
                        valley->EdgeGray = max(l,vs) ;
                        valley->gray = nadir ;
                        valley++ ;

                        Valleys ++ ;
                        vsp = i-1 ;
                        vs = l ;
                        nadir = t ;
                        nadir_pos = i ;
                        sp = 0 ;
                    }
                }
                dire = TRUE ;
            }
        }
        l = t ;
    }

    if ( InValley )
    {
        valley->ys = vsp ;
        valley->ye = i-1 ;
        mid = max(l,vs)-Depth ;
        for ( j=sp-2 ; j>=0 ; j-- )
            if ( s[j] <= mid )
            {
                vep = vsp+j ;
                break ;
            }
        mid = vs-Depth ;
        for ( j=0 ; j<sp ; j++ )
            if ( s[j] <= mid )
            {
                vsp += j ;
                break ;
            }
        valley->yvs = vsp ;
        valley->yve = vep ;
        valley->EdgeGray = max(l,vs) ;
        valley->gray = nadir ;
        Valleys ++ ;
    }
    return Valleys ;
}
/****************************************************************************
/*	Name:		EraseConnComp
/*	Function:	Erase too large connected components
/*	Parameter:	Image		--	Image
/*				nMinWidth	--	Minimum width
/*				nMaxWidth	--	Maximum width
/*				nMinHeight	--	Minimum height
/*				nMaxHeight	--	Maximum height
/*	return:		0 -- Succeed
/*				1 -- Error
/***************************************************************************/
int CDirLine::EraseConnComp(CImage &Image, int nMinWidth, int nMaxWidth, int nMinHeight, int nMaxHeight)
{
    FreeMem() ;
	RECT rBoundRange;
	rBoundRange.left = 0;
	rBoundRange.right = Image.GetWidth()-1;
	rBoundRange.top = 0;
	rBoundRange.bottom = Image.GetHeight()-1;

	m_bIsHorLine = TRUE;
	m_nStrip = 1;
	m_pTree = (CConnTree**)malloc(sizeof(CConnTree*));
	m_nCurTree = 0;
	m_pTree[0] = new CConnTree();
	m_pTree[0]->m_IsHorConn = m_bIsHorLine;
	if ( BuildConnTree(Image, rBoundRange) != 0 )
		return -1 ;
	m_pTree[0]->GetAllConnComps();
	m_pTree[0]->EraseConnComp( Image, nMinWidth, nMaxWidth, nMinHeight, nMaxHeight );
	return 0 ;
}

/**********************************************************************************
/*	Name:		GetAverCharWidth
/*	Function:	Estimate the character width
/*	Parameter:	bLastStrip -- TRUE  Make the final decision no matter if there are 
/*									enough connected components
/*							  FALSE Do not make the decision if lacking enough 
/*									connected components
/*	return:		0  -- Succeed
/*				-1 -- Error
/*
/*********************************************************************************/
int CDirLine::GetAverCharWidth(BOOL bLastStrip)
{
	if(m_nCharWidth > 0)	return 0;
	m_pTree[m_nCurTree]->GetAllConnComps();
	for(int j=0; j<m_pTree[m_nCurTree]->m_nTotalConnComps; j++)
	{//Calculate histogram	
		RECT rcBound = m_pTree[m_nCurTree]->m_pConnComp[j].rcBound;
		int  w = rcBound.right - rcBound.left;
		int  h = rcBound.bottom - rcBound.top;
		if(w>=128 || h >= 128)	continue;
		if(w>15)	CharWidth[w]++;
		if(h>15)	CharHeight[h]++;
	}
	int W[5], H[5], numW, numH, MaxW, MaxH;
	j=125;		numW=0;		MaxW=125;
	while(j>15)
	{//Find peaks in width histogram	
		if(CharWidth[j]<10)		
		{	j--;		continue;}
		while(CharWidth[j-1] > CharWidth[j] && j>15)	j--;
		W[numW] = j;
		if(CharWidth[W[numW]] > CharWidth[MaxW])		MaxW = j;
		while(CharWidth[j-1] <= CharWidth[j] && j> 15)	j--;
		numW++;
		if(numW>=5)	break;
	}
	j=125;		numH=0;		MaxH = 125;
	while(j>15)
	{//Find peaks in height histogram
		if(CharHeight[j]<10)	
		{j--;			continue;}
		while(CharHeight[j-1] > CharHeight[j] && j>15)	j--;
		H[numH]=j;
		if(CharHeight[H[numH]] > CharHeight[MaxH])		MaxH = j;
		while(CharHeight[j-1] <= CharHeight[j] && j>15)	j--;
		numH++;
		if(numH>=5)	break;
	}

	if( m_nCurTree >= 2 && CharWidth[MaxW]>15 && CharHeight[MaxH]>15)
	{//Peak is high enough
		for(j=0; j<numW; j++)
		{//Multi-Peak select the peak with large value
			if(CharWidth[W[j]] > CharWidth[MaxW]/3)		break;
		}
		m_nCharWidth = W[j];
		for(j=0; j<numH; j++)
		{
			if(CharHeight[H[j]] > CharHeight[MaxH]/3)	break;
		}
		m_nCharHeight = H[j];
	}
	else if(bLastStrip)
	{//Peak is not high enough 
		if(CharWidth[MaxW-2]+CharWidth[MaxW-1]+CharWidth[MaxW]+CharWidth[MaxW+1]+CharWidth[MaxW+2] > 30
		 && CharHeight[MaxH-2]+CharHeight[MaxH-1]+CharHeight[MaxH]+CharHeight[MaxH+1]+CharHeight[MaxH+2] > 30)
		{
			m_nCharWidth = MaxW;
			m_nCharHeight = MaxH;
		}
		else
		{	m_nCharWidth = 40;
			m_nCharHeight = 20;
		}
	}
	return 0;
}
