////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		RemoveNoise.cpp
// File Function:	This program remove two kinds of noise:
//					1. Too large or too small connected components on images
//					2. Marginal black strips around the images
//
// NOTE: Modified by Yefeng Zheng to output the noise layer on Nov. 12, 2004
//
//				Developed by: Yefeng Zheng
//			   First created: March, 2004
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RemoveNoise.h"
#include "ImageObj.h"
#include "DirLine.h"
#include "tools.h"
#include "ImageTool.h"

CWinApp theApp;

/******************************************************************************
/*	Name:		Usage
/*	Function:	Print help
/*	Parameter:	NULL
/*	Return:		NULL
/*****************************************************************************/
void Usage( )
{
	printf( "Usage: RemoveNoise [-m] [-w MinWidth] [-W MaxWidth] [-h MinHeight] [-H MaxHeight] [-o CleanOut] [-n NoiseOut] [-?] filename\n" );
	printf( "This program remove two kinds of noise:\n" );
	printf( "1. Too large or two small connected components on images\n" );
	printf( "2. Marginal black strips around the images\n" );
	printf( "Options:\n" );
	printf( "     -m            -- Remove marginal black strips.\n" );
	printf( "     -w MinWidth   -- Connected components with widths equal or less than MinWidth will be removed. Default value is 0.\n" );
	printf( "     -W MaxWidth   -- Connected components with widths equal or more than MaxWidth will be removed. Default value is infinity.\n" );
	printf( "     -h MinHeight  -- Connected components with heights equal or less than MinHeight will be removed. Default value is 0.\n" );
	printf( "     -H MaxHeight  -- Connected components with heights equal or more than MaxHeight will be removed. Default value is infinity.\n" );
	printf( "     -o CleanOut   -- Output file name of the cleaned image. Default is Clean.tif\n" );
	printf( "     -n NoiseOut   -- Output file name of the noise image. Default is Noise.tif\n" );
	printf( "     -?            -- Help\n" );
	return;
}

/******************************************************************************
/*	Name:		_tmain
/*	Function:	Main function
/*	Parameter:	
/*	Return:		0 -- Succeed
/*****************************************************************************/
int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
#define	INFINITY	100000

	int		i, j;
	char	fnImg[256], fnOutImg[256] , fnNoiseImg[256];
	bool	bBlackStrip = FALSE;	//Remove black strips or not.
	int		nMinWidth = 0;		//Connected components with widths equal or less than MinWidth will be removed.
	int		nMaxWidth = INFINITY; //Connected components with widths equal or more than MaxWidth will be removed.
	int		nMinHeight = 0;     //Connected components with heights equal or less than MinHeight will be removed.
	int		nMaxHeight = INFINITY; //Connected components with heights equal or more than MaxHeight will be removed.

//Parse command line
	strcpy( fnImg, "" );
	strcpy( fnOutImg, "Clean.tif" );
	strcpy( fnNoiseImg, "Noise.tif" );

	for(i=1;i<argc;i++)
	{
		if(argv[i][0] != '-') break;
		++i;
		switch( argv[i-1][1])
		{
		case 'm':
			bBlackStrip = TRUE;
			i--;
			break;
		case 'w':
			nMinWidth = atoi( argv[i] );
			break;
		case 'W':
			nMaxWidth = atoi( argv[i] );
			break;
		case 'h':
			nMinHeight = atoi( argv[i] );
			break;
		case 'H':
			nMaxHeight = atoi( argv[i] );
			break;
		case 'o':
			strcpy( fnOutImg, argv[i] );
			break;
		case 'n':
			strcpy( fnNoiseImg, argv[i] );
			break;
		case '?':
			Usage();
			return 0;
		default:
			printf( "Input parameters error!\n" );
			Usage();
			return -1;
		}
	}

	if( i != argc-1 )
	{
		printf( "Input parameters error!\n" );
		Usage( );
		return -1;
	}
	strcpy( fnImg, argv[i] );

//Process all files
	CImage  Img, WholeImg;
	if( Img.Read( fnImg, FALSE ) != 0 )		return -1;
	WholeImg.Initialize( Img );

	if( bBlackStrip )
		CImageTool::EraseBlackStrip( Img );
	if( nMinWidth != 0 || nMaxWidth != INFINITY ||
		nMinHeight != 0 || nMaxHeight != INFINITY )
	{
		CDirLine m_DirLine;
		m_DirLine.EraseConnComp( Img, nMinWidth, nMaxWidth, nMinHeight, nMaxHeight );
	}	
	Img.Write( fnOutImg );

//Output the noise layer
//Added by Yefeng Zheng 11/12/2004 
	int w = Img.GetWidth();
	int h = Img.GetHeight();
	int wb = Img.GetLineSizeInBytes();
	BYTE *pCleanImg = Img.LockRawImg();
	BYTE *pWholeImg = WholeImg.LockRawImg();
		
	for( i=0; i<h; i++ )
	{
		for( j=0; j<wb; j++ )
		{
			*pWholeImg -= *pCleanImg;
			pWholeImg++;
			pCleanImg++;
		}
	}

	WholeImg.Write( fnNoiseImg );
	return 0;
}


