////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		GroupRunLength.cpp
// File Function:	1. Grouping vertical runlength belonging to a line
//					2. Creating polyline representation of a real line
//
//				Developed by: Yefeng Zheng
//			   First created: Feb. 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "Tools.h"
#include "ParalLine.h"
#include "PolyLine.h"
#include <math.h>
#include "GroupRunLength.h"

#define		MIN_PIXEL	20

/******************************************************************************
/*	Name:		GetPolyLine
/*	Function:	Get polyline representation from a idel straight line
/*	Parameter:	Img		-- Image
/*				Line	-- Lines detected
/*				nLine	-- The number of lines
/*				PolyLine -- Polyline representation of lines
/*				nPolyLine-- Number of polylines. nPolyLine <= nLine
/*							Some false alarm lines will be removed.
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	GetPolyLine( CImage &Img, LINE *Line, int nLine, POLYLINE *&PolyLine, int &nPolyLine)
{
	PolyLine = (POLYLINE*)malloc( sizeof(POLYLINE)*nLine );
	nPolyLine = 0;
	if( PolyLine == NULL )
	{
		printf( "Memory allocation failed!\n" );
		return -1;
	}

	int		w = Img.GetWidth();
	int		h = Img.GetHeight();
	int		wb = Img.GetLineSizeInBytes( );
	BYTE*	pImg = Img.LockRawImg( );
	COLRUNLENGTH *ColRun;
	int			 nRun;

	int	nMaxSearchRange = 10;
	for( int i=0; i<nLine; i++ )
	{
		GetColRunLength( Img, Line[i], nMaxSearchRange, ColRun, nRun);
		AllocRunLength( Line[i], ColRun, nRun );
		LocalFilter( Line[i], ColRun, nRun );

		GetEndPoint( Line[i], ColRun, nRun );
		if( Line[i].StPnt.x == -1 || Line[i].StPnt.y == -1 )//not virtual line
		{
			PolyLine[nPolyLine].nPnt = 2;
			PolyLine[nPolyLine].Pnt[0] = Line[i].StPnt;
			PolyLine[nPolyLine].Pnt[1] = Line[i].EdPnt;
	//		Line2PolyLine( ColRun, nRun, PolyLine.Pnt[0].x, PolyLine.Pnt[1].x, PolyLine );
			if( PolyLine[nPolyLine].nPnt > 0 )
				nPolyLine ++;
		}
		free( ColRun );
	}
	return 0;
}

/******************************************************************************
/*	Name:		GetLineParam
/*	Function:	Get MMSE estimation of line parameters from the run length 
/*				composing the line. y = a*x + b.
/*	Parameter:	ColRun	-- Column (vertical) run length
/*				nRun	-- Number of run length
/*				a		-- Parameter a
/*				b		-- Parameter b
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int GetLineParam( COLRUNLENGTH *ColRun, int nRun, double &a, double &b )
{
	double	SumX, SumY, SumXX, SumYY, SumXY;
	int		N;
//Get MMSE approximate of a line
	N = 0;
	SumX = SumY = SumXX = SumYY = SumXY = 0;
	for(int i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		SumX += ColRun[i].x;
		SumY += (ColRun[i].ys+ColRun[i].ye)*0.5;
		N++;
	}
	if( N<=0 )	return 0;

	double	ax = SumX/N;
	double  ay = SumY/N;

	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		SumXX += (ColRun[i].x - ax)*(ColRun[i].x - ax);
		SumXY += (ColRun[i].x - ax)*( 0.5*(ColRun[i].ys + ColRun[i].ye ) - ay);
	}
	b = SumXY / SumXX;
	a = ay - b*ax;
	return 0;
}

/******************************************************************************
/*	Name:		GetEndPoint
/*	Function:	Get two end points of a line. If the line is a false alarm with
/*				two few black pixels on it, then the StPnt of the line is 
/*				CPoint(-1, -1).
/*	Parameter:	Line	-- A line
/*				ColRun	-- Column runlength
/*				nRun	-- Number of runlength
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int GetEndPoint( LINE &Line, COLRUNLENGTH *ColRun, int nRun )
{
	int nLeftCount=0;
	int nRightCount=0; 
	int	nLineRun = 0;

	int	binwidth[20];
	int	nMaxBin=0;
	ZeroMemory( binwidth, sizeof(int)*20);
//Test if the line is a virtual line
	int i, j;
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		int	nBin = 20*(ColRun[i].x-1) / Line.EdPnt.x;
		binwidth[nBin]++;
		nLineRun ++;
	}
	if( nLineRun <= MIN_PIXEL )
	{
		Line.StPnt = CPoint( -1, -1 );
		Line.EdPnt = CPoint( -1, -1 );
		return -1;
	}
	else if( nLineRun <= MIN_PIXEL*5 )
	{
		int nMaxBin = 0;
		for( i=0; i<20; i++ )
			nMaxBin = (int)max( nMaxBin, binwidth[i] );
		if( nMaxBin >= 0.5*nLineRun )
		{
			Line.StPnt = CPoint( -1, -1 );
			Line.EdPnt = CPoint( -1, -1 );
			return -1;
		}
	}
	Line.nPixel = nLineRun;

//Get MMSE approximate esimate of line parameters
	double a, b;
	GetLineParam( ColRun, nRun, a, b );
	double nSkewVar = fabs(atan(b) - atan2( Line.EdPnt.y-Line.StPnt.y, Line.EdPnt.x - Line.StPnt.x ) ) * 180/PI;
	if( nSkewVar > 2.5 )
	{//Skew variance is too large
		Line.StPnt = CPoint( -1, -1 );
		Line.EdPnt = CPoint( -1, -1 );
		return -1;
	}
//Search left end point
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		nLeftCount++;
		if( nLeftCount >= 3 )
		{
			for(j=i; j>=0; j--)
			{
				if( ColRun[i].nUsed == FALSE )
					continue;
				if( abs(ColRun[j].x - ColRun[i].x)>6 )
					break;
			}
			j = max( 0, j );
			Line.StPnt.x = ColRun[j].x;
			Line.StPnt.y = (int)(a + b*ColRun[j].x );
			break;
		}
	}

//Search right end point
	for( i=nRun-1; i>=0; i-- )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		nRightCount++;
		if( nRightCount >= 3 )
		{
			for( j=i; j<nRun; j++ )
			{
				if( ColRun[j].nUsed == FALSE )	
					continue;
				if( abs(ColRun[j].x - ColRun[i].x ) > 6 )
					break;
			}
			j = min( nRun-1, j);
			Line.EdPnt.x = ColRun[j].x;
			Line.EdPnt.y = (int)(a + b*ColRun[j].x );
			break;
		}
	}
	return 0;
}

/******************************************************************************
/*	Name:		GetColRunLength
/*	Function:	Extract column runlength
/*	Parameter:	Img		-- Image
/*				Line	-- A line
/*				nMaxSearchRange	-- Maximum search range above and below the line
/*				ColRun	-- Column runlength
/*				nRun	-- Number of runlength
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	GetColRunLength( CImage &Img, LINE Line, int nMaxSearchRange, COLRUNLENGTH *&ColRun, int &nRun)
{
	int i, j;
	int		w = Img.GetWidth();
	int		h = Img.GetHeight();
	int		wb = Img.GetLineSizeInBytes();
	BYTE	*pImg = Img.LockRawImg();
	CPoint	StPnt = Line.StPnt;
	CPoint	EdPnt = Line.EdPnt;

    BYTE	mask[8] = {128,64,32,16,8,4,2,1} ;
    BYTE*	tmp ;

	nRun = 0;
	int		nMaxRun = 100;
	ColRun = (COLRUNLENGTH *)malloc( sizeof(COLRUNLENGTH)*nMaxRun);
	if( ColRun == NULL )		return -1;
	for( i=StPnt.x; i<=EdPnt.x; i++ )
	{
		int nByte = i/8;
		int nBit = i%8;
		BOOL bTracked = FALSE;
		int y = StPnt.y + (i-StPnt.x)*(EdPnt.y-StPnt.y)/(EdPnt.x-StPnt.x);
		int ys = max(0, y-nMaxSearchRange );
		int ye = min(h-1, y+nMaxSearchRange );
		tmp = pImg + (h-1-ys)*wb + nByte;
		for( j=ys; j<=ye; j++ )
		{
			if( (*tmp & mask[nBit]) != 0 )//Black pixel
			{
				if( bTracked == FALSE )
				{
					ColRun[nRun].ys = j;
					bTracked = TRUE;
				}
			}
			else
			{//White pixel
				if( bTracked )
				{
					bTracked = FALSE;
					ColRun[nRun].ye = j-1;
					ColRun[nRun].x = i;
					nRun++;
					if( nRun >= nMaxRun )
					{
						nMaxRun += 100;
						ColRun = (COLRUNLENGTH*)realloc( ColRun, sizeof(COLRUNLENGTH)*nMaxRun);
					}
				}
			}
			tmp-=wb;
		}
		if( bTracked )
		{
			bTracked = FALSE;
			ColRun[nRun].ye = ye;
			ColRun[nRun].x = i;
			nRun++;
			if( nRun >= nMaxRun )
			{
				nMaxRun += 100;
				ColRun = (COLRUNLENGTH*)realloc( ColRun, sizeof(COLRUNLENGTH)*nMaxRun);
			}
		}
	}
	Img.UnlockRawImg();
    return 0 ;
}

/******************************************************************************
/*	Name:		LocalFilter
/*	Function:	Using local filter to exclude some candidate run length from 
/*				composing a line.
/*	Parameter:	Line	-- A line
/*				ColRun	-- Column runlength
/*				nRun	-- Number of runlength
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int LocalFilter( LINE Line, COLRUNLENGTH *ColRun, int nRun )
{
	int		i;
	int		DIST_TH = 3;

	double	a, b;
	GetLineParam( ColRun, nRun, a, b );
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;

		int y = (int)(a + b*ColRun[i].x );
		int	dys = ColRun[i].ys - y;
		int dye = ColRun[i].ye - y;
		if( dys * dye < 0 )		continue;
		else if( min( abs(dys), abs(dye) ) > DIST_TH )
			ColRun[i].nUsed = FALSE;
	}
	return 0;
}

/******************************************************************************
/*	Name:		AllocRunLength
/*	Function:	Label some run length as candidates to compose a line
/*	Parameter:	Line	-- A line
/*				ColRun	-- Column runlength
/*				nRun	-- Number of runlength
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int AllocRunLength( LINE Line, COLRUNLENGTH *ColRun, int nRun )
{
	CPoint	StPnt = Line.StPnt;
	CPoint	EdPnt = Line.EdPnt;
	int nDist;

	int i;
	for( i=0; i<nRun; i++ )
		ColRun[i].nUsed = FALSE;

	i = 0;
	while( i<nRun )
	{
		int nCurrentX = ColRun[i].x;
		int	y = StPnt.y + (nCurrentX-StPnt.x)*(EdPnt.y-StPnt.y)/(EdPnt.x-StPnt.x);
		
		int j=i;
		int	nMinDist = 1000;
		int nMinDistIndex = -1;
		while( ColRun[j].x == nCurrentX )
		{
			int	dys = ColRun[j].ys - y;
			int dye = ColRun[j].ye - y;
			if( dys * dye < 0 )//Overlaping with the line on the vertical direction. It is a candidate.
				nDist = min(dys, dye);
			else//Otherwise find the candidate with minimum vertical distance to the line
				nDist = min( abs(dys), abs(dye) );

			if( nDist < nMinDist )
			{
				nMinDist = nDist;
				nMinDistIndex = j;
			}
			j++;
			if( j >= nRun )		break;
		}
		ColRun[nMinDistIndex].nUsed = TRUE;
		i = j;
	}
	return 0;
}