////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		LineDetect.cpp
// File Function:	Parallel background line detection using HMM model
//
//				Developed by: Yefeng Zheng
//			   First created: Feb. 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "Tools.h"
#include "math.h"
#include "ParalLine.h"
#include "DirLine.h"
#include "RuleLineAPI.h"
#include "LineDetect.h"
#include "GroupRunlength.h"


/*****************************************************************************
/*	Name:		DetectLine
/*	Function:	Detect parallel background lines.
/*				It is a "main" function to call all other functions.
/*	Parameter:	Img		  -- Image data
/*				model	  -- HMM model parameters
/*				bFilter	  -- Performing DSCC filtering or not
/*				pLine	  -- Detected rule lines
/*				nLine	  -- Number of detected rule lines
/*				bLine     -- TRUE, there are rule lines on the document image
/*							 FALSE, there may be no rule lines.
/*							 No matter which value is return for bRuleLine, 
/*							 this function will try to detect some rule lines.
/*	Return:		0 -- Succeed
/*****************************************************************************/
int DetectLine( CImage &Img, HMM_MODEL model, BOOL bFilter, LINE *&pLine, int &nLine, bool &bLine )
{
	// Clear return values
	pLine = NULL;
	nLine = 0;
	bLine = false;

	double	nSkewAngle;
	int		nLineGap;
	double	*HorProj;
	int		nHorProj;
	double	*HorProjPeak;

	int w = Img.GetWidth();
	int h = Img.GetHeight();
	int	wb = Img.GetLineSizeInBytes( );

	CImage	FilteredImg;
	CDirLine	DirLine;
	if( bFilter )
		DirLine.DSCCFiltering( Img, FilteredImg, TRUE );
	else
		FilteredImg.Initialize( Img );

	DirLine.EstimateSkew( FilteredImg, nSkewAngle );
	DirLine.RefineSkewEstimation( FilteredImg, nSkewAngle );
	CImageTool::DoSkewHorProject(FilteredImg, nSkewAngle, HorProj, nHorProj );

	EstimateLineGap( HorProj, nHorProj, nLineGap );

	FindProjPeak( HorProj, nHorProj, HorProjPeak );
	Quantize( HorProjPeak, nHorProj, w );

	int		*LinePos;
	double	Prob;

	LineDTW( HorProjPeak, nHorProj, nLineGap, LinePos, nLine, Prob, model);
	GetParalLine( LinePos, nLine, pLine, nSkewAngle, w, h );	

	int		nLine2 = nLine;
	// Determine the average variation from the average line gap.
	// We will use it to judge if there are rule lines presented.
	int		i;
	double	nLineGapVar = 0;
	if( nLine > 1 )
	{
		for( i=0; i<nLine-1; i++ )
			nLineGapVar += fabs( LinePos[i+1] - LinePos[i] - nLineGap );
		// Average variation
		nLineGapVar /= (nLine-1);
		// Relative variation
		nLineGapVar /= nLineGap;
	}

	// Determine the end points of lines.
	// If there are too few black pixels on a line, it is removed.
	COLRUNLENGTH *ColRun;
	int			 nRun;
	int	nMaxSearchRange = 10;
	for( i=0; i<nLine; i++ )
	{
		GetColRunLength( FilteredImg, pLine[i], nMaxSearchRange, ColRun, nRun);
		AllocRunLength( pLine[i], ColRun, nRun );
		LocalFilter( pLine[i], ColRun, nRun );
		GetEndPoint( pLine[i], ColRun, nRun );
		free( ColRun );
	}
	
	i=0;
	while( i<nLine )
	{
		if( pLine[i].StPnt.x == -1 || pLine[i].StPnt.y == -1 )//virtual line
		{
			for( int j=i; j<nLine-1; j++ )
				pLine[j] = pLine[j+1];
			nLine--;
		}
		else
			i++;
	}
	free( LinePos );
	free( HorProj );
	free( HorProjPeak );

// Automatic judge if there are rule lines on the document
	// Determine the degradtion of rule lines
	double	nBlackRatio = 0;
	if( nLine > 0 )
	{
		for( i=0; i<nLine; i++ )
			nBlackRatio += pLine[i].nPixel;
		nBlackRatio /= (nLine*w);
	}
	
/*	FILE *fp = fopen( "C:\\RuleLine.txt", "at" );
	fprintf( fp, "%f,%f,%f,Y\n", 1.0*nLine/nLine2, nLineGapVar, nBlackRatio);
	fclose( fp );
*/
	// Decision rules learned using the C4.5 classification tree.
	double	nLineRatio = 1.0*nLine/nLine2;
	if( nBlackRatio > 0.198833 )
		bLine = true;
	else if( nLineRatio > 0.710526 && nLineGapVar <= 0.006087 )
		bLine = true;
	if( nLine <= 5 )	// Too few rule lines
		bLine = false;

	return 0;
}

/*****************************************************************************
/*	Name:		Quantize
/*	Function:	Quantize the peaks into four levels
/*	Parameter:	Proj		-- Projection
/*				nProj		-- Size of the projection array
/*				w           -- Image width, used for quantization
/*	Return:		0 -- Succeed
/*****************************************************************************/
int Quantize( double *Proj, int nProj, int w )
{
	for( int i=0; i<nProj; i++ )
	{
		if( Proj[i] <= 0 )
			Proj[i] = 0;
		else if( Proj[i] < w/16 )
			Proj[i] = 1;
		else if( Proj[i] < w/8 )
			Proj[i] = 2;
		else if( Proj[i] < w/4 )
			Proj[i] = 3;
		else 
			Proj[i] = 4;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		FindProjPeak
/*	Function:	1. Search peaks on skewed horizontal projection
/*				2. Quantize the height of peaks to 5 levels (0 for non-peaks).
/*	Parameter:	Proj		-- Projection
/*				nProj		-- Size of the projection array
/*				ProjPeak	-- Quantization result
/*	Return:		0 -- Succeed
/*****************************************************************************/
int FindProjPeak( double *Proj, int nProj, double *&ProjPeak )
{
	int	TH_PEAK = 20;
	
	double *tmpProj;
	tmpProj = (double*)malloc( sizeof(double)*nProj );
	CopyMemory( tmpProj, Proj, sizeof(double)*nProj );
	CImageTool::SmoothProject( tmpProj, nProj );

//Search peaks and valleys on the projection
	int *nLeftPeak, *nRightPeak, nPeak;
	SearchPeak( tmpProj, nProj, nLeftPeak, nRightPeak, nPeak );
	int *nLeftValley, *nRightValley, nValley;
	SearchValley( tmpProj, nProj, nLeftValley, nRightValley, nValley );

	ProjPeak = (double*)malloc( sizeof(double)*nProj );
	ZeroMemory( ProjPeak, sizeof(double)*nProj );
	for( int i=0; i<nPeak; i++ )
	{
		//Remove some small peak;
		double	nLeftPeakHeight = 0;
		double	nRightPeakHeight = 0;
		if( nLeftPeak[i] == 0 )
		{
			nLeftPeakHeight = tmpProj[0];
			nRightPeakHeight = tmpProj[0]-tmpProj[nLeftValley[0]];
		}
		else if( nRightPeak[i] == nProj-1 )
		{
			nLeftPeakHeight = tmpProj[nLeftPeak[i]]-tmpProj[nRightValley[nValley-1]];
			nRightPeakHeight = tmpProj[nRightPeak[i]];
		}
		else
		{
			for( int j=0; j<nValley-1; j++ )
				if( nRightValley[j] < nLeftPeak[i] && nLeftValley[j+1] > nRightPeak[i] )
					break;
			if( j >= nValley-1 )
			{
				printf( "Search peak and valley error!\n" );
				return -1;
			}
			nLeftPeakHeight = tmpProj[nLeftPeak[i]] - tmpProj[nRightValley[j]];
			nRightPeakHeight = tmpProj[nRightPeak[i]] - tmpProj[nLeftValley[j+1]];
		}

		if( max( nLeftPeakHeight, nRightPeakHeight ) < TH_PEAK )
			continue;

		//Quantize peaks
		int nPos = (nLeftPeak[i]+nRightPeak[i])/2;
		ProjPeak[nPos] = Proj[nPos];
	}

	free( nLeftPeak );
	free( nRightPeak );
	free( nLeftValley );
	free( nRightValley );
	free( tmpProj );
	return 0;
}

/****************************************************************************
/*	Name:		EstimateLineGap
/*	Function:	Estimate gap between neighboring lines
/*	Parameter:	HorProj		-- Horizontal projection
/*				nHorProj	-- Dimension of the horizontal projection
/*				nLineGap	-- Estimated line gap
/*	return:		0 -- Succeed
/*				1 -- Error
/***************************************************************************/
int EstimateLineGap( double *HorProj, int nHorProj, int &nLineGap )
{
	#define	MIN_LINE_GAP	15
	#define	MAX_LINE_GAP	300

	int	i, j;
	double AutoCorr[MAX_LINE_GAP];
	ZeroMemory( AutoCorr, sizeof(double)*MAX_LINE_GAP );
	for( i=MIN_LINE_GAP; i<MAX_LINE_GAP; i++ )
	{
		AutoCorr[i] = 0;
		for( j=0; j<nHorProj-i; j++ )
		{
			AutoCorr[i] += HorProj[j]*HorProj[j+i];
		}
	}

//Smooth the projection
	CImageTool::SmoothProject( AutoCorr, MAX_LINE_GAP );

/*	int	AutoCorr2[MAX_LINE_GAP];
	for( i=0; i<MAX_LINE_GAP; i++ )
		AutoCorr2[i] = AutoCorr[i]/50000;
	CImageTool::DumpProject( "C:\\AutoCorrelation.tif", AutoCorr2, MAX_LINE_GAP, TRUE );
*/
//Find first large enough peak
	int *LeftPeak, *RightPeak, *PeakHeight, nPeak;
	SearchPeak( AutoCorr, MAX_LINE_GAP, LeftPeak, RightPeak, nPeak );
	if( nPeak <= 1 )
	{
		free( LeftPeak );
		free( RightPeak );
		return -1;
	}
	PeakHeight = (int*)malloc( sizeof(int)*nPeak );
	int	nMaxPeak = 1;
	for( i=1; i<nPeak; i++ )
	{
		PeakHeight[i] = GetPeakHeight( AutoCorr, MAX_LINE_GAP, LeftPeak[i] );
		if( PeakHeight[i] > PeakHeight[nMaxPeak] )
			nMaxPeak = i;
	}
	for( i=1; i<nPeak; i++ )
	{
		if( PeakHeight[i] > PeakHeight[nMaxPeak]*0.8 )
		{
			nMaxPeak = i;
			break;
		}
	}
	
	int nPeakWidth = GetPeakWidth( AutoCorr, MAX_LINE_GAP, LeftPeak[nMaxPeak] );

	if( nMaxPeak != 1 )
	{
		for( i=nMaxPeak-1; i>=1; i-- )
		{
			if( nPeakWidth <= 20 &&
				PeakHeight[i] < PeakHeight[nMaxPeak]/3 )
				continue;

			for(int nLeftValley=LeftPeak[i]-1; nLeftValley>1; nLeftValley-- )
				if( AutoCorr[nLeftValley] < AutoCorr[nLeftValley-1] && 
					AutoCorr[nLeftValley] < AutoCorr[nLeftValley-2] )
					break;
			for(int nRightValley=RightPeak[i]+1; nRightValley<MAX_LINE_GAP-2; nRightValley++ )
				if( AutoCorr[nRightValley] < AutoCorr[nRightValley+1] &&
					AutoCorr[nRightValley] < AutoCorr[nRightValley+2] )
					break;
//			if( AutoCorr[LeftPeak[i]] > 1.5*max(AutoCorr[nLeftValley], AutoCorr[nRightValley] ) )
			{
				if( fabs( (LeftPeak[i]+RightPeak[i]) - (LeftPeak[nMaxPeak]+RightPeak[nMaxPeak])/2 ) <= 4 )
					break;
			}
		}
		if( i >= 1 )
			nMaxPeak = i;
	}
	nLineGap = (LeftPeak[nMaxPeak]+RightPeak[nMaxPeak])/2;
	
	free( LeftPeak );
	free( RightPeak );
	free( PeakHeight );
	return 0;
}

/****************************************************************************************************
/*	Name:		GetParalLine
/*	Function:	Get parallel lines using the output of HMM model
/*	Parameter:	LinePos  -- Projection position of detected lines. Output of HMM model
/*				nLine    -- Number of lines
/*				pLine	 -- Detection lines
/*				nSkewAngle  -- Skew angle
/*				w		 -- Image width
/*				h        -- Image height
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	GetParalLine( int *LinePos, int &nLine, LINE *&pLine, double nSkewAngle, int w, int h )
{
	pLine = (LINE*)malloc( sizeof(LINE)*nLine );
	int	dh = (int)( w*tan( nSkewAngle ) );
	for( int i=0; i<nLine; i++ )
	{
		pLine[i].StPnt.x = 0;
		pLine[i].EdPnt.x = w-1;
		if( nSkewAngle >=0 )
			pLine[i].StPnt.y = LinePos[i];
		else
			pLine[i].StPnt.y = LinePos[i]+dh;
		pLine[i].EdPnt.y = pLine[i].StPnt.y - dh;
	}
//Remove lines outside of the image
	i=0;
	while( i<nLine )
	{
		if( (pLine[i].StPnt.y <= 0 && pLine[i].EdPnt.y <= 0 ) ||
			(pLine[i].StPnt.y >= h-1 && pLine[i].EdPnt.y >= h-1 ) )
		{
			for( int j=i; j<nLine-1; j++ )
				pLine[j] = pLine[j+1];
			nLine--;
		}
		else
			i++;
	}
	AdjustLinePosition( w, h, pLine, nLine );
	return 0;
}

/****************************************************************************************************
/*	Name:		LineDTW
/*	Function:	Using DTW to search optimal state sequence of HMM model
/*	Parameter:	Proj	-- Horizontal projection (Quantized)
/*				nProj   -- Number of the projection
/*				LineGap -- Estimated vertical line gap
/*				LinePos -- DTW searching result. The positions of the detected lines.
/*				nLine	-- Number of lines
/*				TotalProb -- Probability to observe such sequence given the model
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int LineDTW( double *Proj, int nProj, int nLineGap, int *&LinePos, int &nLine, double &TotalProb, HMM_MODEL model)
{
	int		i, j;
	int	nLocalMinLineGap = model.nMinLineGap + nLineGap;
	int	nLocalMaxLineGap = model.nMaxLineGap + nLineGap;

	//Alloc memory
	int		nMaxLine = nProj/nLocalMinLineGap+10;
	int		**LineList = (int**)malloc( sizeof(int*)*nProj );
	for( i=0; i<nProj; i++ )
		LineList[i] = (int*)malloc( sizeof(int)*nMaxLine );
	int		*LineNumList = (int*)malloc( sizeof(int)*nProj );
	ZeroMemory( LineNumList, sizeof(int)*nProj );
	double  *Probability = (double*)malloc( sizeof(double)*nProj );
	for( i=0; i<nProj; i++ )
		Probability[i] = -1e+10;

	//Set top gap
	double	nNewProb;
	double	nAllTextProb = 0;
	for( i=0; i<nLocalMaxLineGap; i++ )
	{
		LineNumList[i] = 1;
		LineList[i][0] = i;
		Probability[i] = nAllTextProb + model.LinePeakProb[(int)Proj[i]];
		nAllTextProb += model.TextPeakProb[(int)Proj[i]];
	}

    for( i=0; i<=nProj-1; i++ )
	{
		if( LineNumList[i] == 0 )	continue;//Do not need proprogate

		nAllTextProb = 0;
		for( j=i+1; j<min( nProj, i+nLocalMinLineGap); j++ )
			nAllTextProb += model.TextPeakProb[(int)Proj[j]];

		for( j=min( nProj-1, i+nLocalMinLineGap); j<=min( nProj-1, i+nLocalMaxLineGap); j++ )
		{
			int nGap = j-i-nLineGap-model.nMinLineGap;
			if( nGap >= 0 && nGap <= model.nMaxLineGap-model.nMinLineGap )
			{
				nNewProb = Probability[i] + nAllTextProb + model.LineGapDist[nGap];
				nNewProb += model.LinePeakProb[(int)Proj[j]];

				if( nNewProb > Probability[j] )//Update decoding result of j
				{
					Probability[j] = nNewProb;
					LineNumList[j] = LineNumList[i]+1;
					CopyMemory( LineList[j], LineList[i], sizeof(int)*(LineNumList[i]) );
					LineList[j][LineNumList[j]-1] = j;
				}
			}
			else//Last row
			{
				nNewProb = Probability[i] + nAllTextProb;
				nNewProb += model.TextPeakProb[(int)Proj[j]];

				if( nNewProb > Probability[j] )//Update decoding result of j
				{
					Probability[j] = nNewProb;
					LineNumList[j] = LineNumList[i];
					CopyMemory( LineList[j], LineList[i], sizeof(int)*(LineNumList[i]) );
				}
			}
			nAllTextProb += model.TextPeakProb[(int)Proj[j]];
		}
	}
	nLine = LineNumList[nProj-1];
	LinePos = (int*)malloc( sizeof(int)*nLine );
	CopyMemory( LinePos, LineList[nProj-1], sizeof(int)*(nLine) );
	TotalProb = Probability[nProj-1];

	for( i=0; i<nProj; i++ )
		free( LineList[i] );
	free( LineList );
	free( LineNumList );
	free( Probability );
	return 0;
}


