////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		LineRemoval.cpp
// File Function:	Remove lines
//
//				Developed by: Yefeng Zheng
//			   First created: March 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "Tools.h"
#include "math.h"
#include "ParalLine.h"
#include "DirLine.h"
#include "GroupRunLength.h"
#include "LineRemoval.h"
 
/******************************************************************************
/*	Name:		RemoveLine
/*	Function:	Remove lines
/*	Parameter:	fnImg	-- Input image file name
/*				fnLine	-- Lines information file name
/*				fnOut	-- Output image file name
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	RemoveLine( char *fnImg, char *fnLine, char *fnOut )
{
	LINE *pLine; 
	int  nLine;

	if( ReadDetectResult( fnLine, pLine, nLine ) != 0 )
		return -1;

	CImage Img;
	if( Img.Read( fnImg ) != 0 )
	{
		printf( "Read image: %s failed!\n", fnImg );
		return -1;
	}

	RemoveLineWithCharSeg( Img, pLine, nLine );
	Img.Write( fnOut );

	free( pLine );
	return 0;
}

/******************************************************************************
/*	Name:		RemoveLineWithCharSeg
/*	Function:	Remove lines with the character segmentation information
/*	Parameter:	Img		-- Image
/*				Line	-- Lines detected
/*				nLine	-- The number of lines
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	RemoveLineWithCharSeg( CImage &Img, LINE *Line, int nLine )
{
	int		w = Img.GetWidth();
	int		h = Img.GetHeight();
	int		wb = Img.GetLineSizeInBytes( );
	BYTE*	pImg = Img.LockRawImg( );
	COLRUNLENGTH *ColRun;
	int		nRun;

	int		TH_WIDTH = 3;
	int		nMaxSearchRange = 30;

	int		i, j;
	for( i=0; i<nLine; i++ )
	{
		//Group run-lengths, which belong to the line
		GetColRunLength( Img, Line[i], nMaxSearchRange, ColRun, nRun);
		AllocRunLength( Line[i], ColRun, nRun );
		LocalFilter( Line[i], ColRun, nRun );

/*		for( j=0; j<nRun; j++ )
		{
			if( ColRun[j].nUsed == FALSE )
				ColRun[j].nRemoveMode = UNREMOVE_OUTSIDE;
			else
				ColRun[j].nRemoveMode = REMOVE_OUTSIDE;
		}
*/
		//Classify each CC to be connected to the line or not
		//This is used to faciliate the character segmentation
		AnalysisCC( ColRun, nRun );

		//Character segmentation
		CHAR_ZONE *Char;
		int	nChar;
		CharSeg( ColRun, nRun, w, Char, nChar );

		//Remove line
		//Set remove mode for each column run length: INSIDE or OUTSIDE a character
		SetRemoveMode( ColRun, nRun, Char, nChar );
		//Estimate line width using run length outside the characters
		int nLineWidth = EstimateLineWidthFine( ColRun, nRun );
		//Set removal threshold
		int	nEraseThreshold = SetEraseThreshold( nLineWidth );
		//Label remove type for run length inside characters
		SetRemoveModeInsideChar( ColRun, nRun, nEraseThreshold );
		//Relabel some segments inside a character, which is connected to two unremovable segments on the left and right sides
		ReLabelSegment( ColRun, nRun, Char, nChar );
		//Set touch type for each character
		SetTouchType( ColRun, nRun, Char, nChar, nLineWidth );
		//Complete circle
//		CompleteCircle( ColRun, nRun, Char, nChar );

		for( j=0; j<nRun; j++ )
		{
			if( ColRun[j].nRemoveMode == REMOVE_INSIDE ||
				ColRun[j].nRemoveMode == REMOVE_OUTSIDE )
			{
				int		nWidth = ColRun[j].ye - ColRun[j].ys + 1;
				CImageTool::SetVerLineToWhite ( pImg, wb, h, ColRun[j].x, ColRun[j].ys, nWidth+1, TRUE ) ;
			}
		}

		for( j=0; j<nChar; j++ )
		{
			if( Char[j].StartSeg != NULL )
				free( Char[j].StartSeg );
			if( Char[j].EndSeg != NULL )
				free( Char[j].EndSeg );
		}
		free( Char );
		free( ColRun );
	}
	return 0;
}

/******************************************************************************
/*	Name:		RemoveLine
/*	Function:	Remove lines
/*	Parameter:	Img		-- Image
/*				Line	-- Lines detected
/*				nLine	-- The number of lines
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	RemoveLine( CImage &Img, LINE *Line, int nLine )
{
	int		w = Img.GetWidth();
	int		h = Img.GetHeight();
	int		wb = Img.GetLineSizeInBytes( );
	BYTE*	pImg = Img.LockRawImg( );
	COLRUNLENGTH *ColRun;
	int		nRun;

	int		TH_WIDTH = 3;
	int		nMaxSearchRange = 30;

	int		i, j;
	for( i=0; i<nLine; i++ )
	{
		//Group run-lengths, which belong to the line
		GetColRunLength( Img, Line[i], nMaxSearchRange, ColRun, nRun);
		AllocRunLength( Line[i], ColRun, nRun );
		LocalFilter( Line[i], ColRun, nRun );

		//Classify each CC to be connected to the line or not
		//This is used to faciliate the character segmentation
		AnalysisCC( ColRun, nRun );

		//Set remove mode for each column run length: INSIDE or OUTSIDE a character
		for( j=0; j<nRun; j++ )
		{
			if( ColRun[j].nUsed == FALSE )
				ColRun[j].nRemoveMode = UNREMOVE_OUTSIDE;
			else
				ColRun[j].nRemoveMode = REMOVE_INSIDE; //Inside
		}
		//Estimate line width using run length outside the characters
		int nLineWidth = EstimateLineWidth( ColRun, nRun );
		//Set removal threshold
		int	nEraseThreshold = SetEraseThreshold( nLineWidth );
		for( j=0; j<nRun; j++ )
		{
			if( ColRun[j].nRemoveMode == REMOVE_INSIDE ||
				ColRun[j].nRemoveMode == REMOVE_OUTSIDE )
			{
				int		nWidth = ColRun[j].ye - ColRun[j].ys + 1;
				CImageTool::SetVerLineToWhite ( pImg, wb, h, ColRun[j].x, ColRun[j].ys, nWidth+1, TRUE ) ;
			}
		}
		free( ColRun );
	}
	return 0;
}

/******************************************************************************
/*	Name:		EstimateLineWidthFine
/*	Function:	Esimation of line width with only line segments outside of the boudaries
/*				of characters, using the median value.
/*				Hope to increase the line width estimation accuracy.
/*	Parameter:	ColRun	-- Column run-lengths
/*				nRun	-- Number of column run-lengths
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int	EstimateLineWidthFine( COLRUNLENGTH *ColRun, int nRun )
{
	int i;
	int RunWidth[10000];

	int	nUsedRun = 0;
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE || ColRun[i].nRemoveMode != 0 )	
			continue;
		RunWidth[nUsedRun] = ColRun[i].ye - ColRun[i].ys + 1;
		nUsedRun ++;
	}
	int nLineWidth = GetMidValue( RunWidth, nUsedRun );
	return nLineWidth;
}

/******************************************************************************
/*	Name:		EstimateLineWidth
/*	Function:	Esimation of line width, using the median value
/*	Parameter:	ColRun	-- Column run-lengths
/*				nRun	-- Number of column run-lengths
/*	Return:		0  -- Succeed
/*				-1 -- Error
/*****************************************************************************/
int EstimateLineWidth( COLRUNLENGTH *ColRun, int nRun)
{
	int i;
	int RunWidth[10000];

	int	nUsedRun = 0;
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )	continue;
		RunWidth[nUsedRun] = ColRun[i].ye - ColRun[i].ys + 1;
		nUsedRun ++;
	}
	int nLineWidth = GetMidValue( RunWidth, nUsedRun );
	return nLineWidth;
}

/******************************************************************************
/*	Name:		AnalysisCC
/*	Function:	Classify each CC connected with a line or not
/*	Parameter:	ColRun	-- Column run-lengths
/*				nRun	-- Number of column run-lengths
/*	Return:		0  -- Succeed
/*****************************************************************************/
int	AnalysisCC( COLRUNLENGTH *ColRun, int nRun )
{
	int Connectness = 8;
	int	nCC = 0;

	int		i, j, k;
	COLRUNLENGTH	LastRun[1000];

	int		nLastRun=0;
	int		MaxTop, MinBottom;

	for( i=0; i<nRun; i++ )
		ColRun[i].nCC = -1;

	i=0;
	int		nCurrentX, nLastX;
	nLastX = 0;
	while( i<nRun )
	{
		//Merge runlength to connected components
		nCurrentX = ColRun[i].x;
		if( nCurrentX != nLastX+1 )
			nLastRun = 0;
		nLastX = nCurrentX;
		j=i;
		while( ColRun[j].x == nCurrentX )
		{

			for( k=0; k<nLastRun; k++ )
			{
				MaxTop = max( ColRun[j].ys, LastRun[k].ys );
				MinBottom = min( ColRun[j].ye, LastRun[k].ye );
				if( Connectness == 4 && MaxTop > MinBottom )
					continue;
				else if( Connectness == 8 && MaxTop > MinBottom+1 )
					continue;
				if( ColRun[j].nCC == -1 )//Merge runlength to connected component
					ColRun[j].nCC = LastRun[k].nCC;
				else
				{
					if( ColRun[j].nCC != LastRun[k].nCC )
					{//Mergt two connected componets
						int		nOldConn = LastRun[k].nCC;
						int		nNewConn = ColRun[j].nCC;
						for( int m=0; m<=j; m++ )
						{//Update all run lengths
							if( ColRun[m].nCC == nOldConn )
								ColRun[m].nCC = nNewConn;
						}
						for( m=0; m<nLastRun; m++ )
						{//Update run lengths of the left column
							if( LastRun[m].nCC == nOldConn )
								LastRun[m].nCC = nNewConn ;
						}
					}
				}
			}
			if( ColRun[j].nCC == - 1)//Create new CC
			{
				ColRun[j].nCC = nCC;
				nCC++;
			}
			j++;
			if( j>=nRun )	break;
		}
		//Update runlengths of the last line
		for( k=i; k<j; k++ )
			LastRun[k-i] = ColRun[k];
		nLastRun = j-i;
		i = j;
	}

	//Label all column runlengths connected with the line
	for( i=0; i<nCC; i++ )
	{
		//Test if the CC belonging to line or not
		BOOL	bLine = FALSE;
		for( j=0; j<nRun; j++ )
		{
			if( ColRun[j].nUsed == TRUE && ColRun[j].nCC == i )
			{
				bLine = TRUE;
				break;
			}
		}
		if( bLine )//The CC connected to the line, set all labels to negative
		{
			for( j=0; j<nRun; j++ )
				if( ColRun[j].nCC == i )
					ColRun[j].nCC = -i-1;
		}
	}
	return 0;
}

/******************************************************************************
/*	Name:		RemoveExtraCC
/*	Function:	Remove extra CC which do not connected to lines.
/*				Used to debug AnalysisCC() function.
/*	Parameter:	ImgOut  -- Image after extra CC removed
/*				ColRun	-- Column run-lengths
/*				nRun	-- Number of column run-lengths
/*	Return:		0  -- Succeed
/*****************************************************************************/
int	RemoveExtraCC( CImage &ImgOut, COLRUNLENGTH *ColRun, int nRun )
{
	AnalysisCC( ColRun, nRun );

	BYTE *pImgOut = ImgOut.LockRawImg();
	int	wb = ImgOut.GetLineSizeInBytes();
	int	h = ImgOut.GetHeight();
	for( int i=0; i<nRun; i++ )
	{
		if( ColRun[i].nCC >= 0 ) //Extra CC
			CImageTool::SetVerLineToWhite( pImgOut, wb, h, ColRun[i].x, ColRun[i].ys, ColRun[i].ye-ColRun[i].ys+1, TRUE );
	}
	return 0;
}

/******************************************************************************
/*	Name:		CharSeg
/*	Function:	Character segmentation
/*	Parameter:	ColRun	-- Column run-lengths
/*				nRun	-- Number of column run-lengths
/*				w       -- Width of the image
/*				Char  -- Bounding box of each character segmented
/*				nChar   -- Number of characteres segmented
/*	Return:		0  -- Succeed
/*****************************************************************************/
int	CharSeg( COLRUNLENGTH *ColRun, int nRun, int w, CHAR_ZONE *&Char, int &nChar)
{
	int i, j;
	int nLineWidth = EstimateLineWidth( ColRun, nRun );

	//Vertical projection
	int		TH_MAX = 1;
	double	*VerProj = (double*)malloc( sizeof(double)*w );
	if( VerProj == NULL )
	{
		printf( "Memory used up!\n" );
		return -1;
	}
	ZeroMemory( VerProj, sizeof(double)*w );

	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nCC >= 0 )	continue;//Extra CC
		int	x = ColRun[i].x;
		int hCC = ColRun[i].ye - ColRun[i].ys + 1;
		if( ColRun[i].nUsed == TRUE )
			VerProj[x] += max( 0,  hCC - 1.5*nLineWidth ); //Do not count line width
		else
			VerProj[x] += hCC;
	}

	CImageTool::SmoothProject( VerProj, w );
//	CImageTool::DumpProject( "C:\\VerProj.tif", VerProj, w, FALSE);

	//Character segmentation based on vertical projection
	int	*LeftValley, *RightValley;
	int nValley;
	SearchValley( VerProj, w, LeftValley, RightValley, nValley );
	
	nChar = 0;
	Char = (CHAR_ZONE*)malloc( sizeof(CHAR_ZONE)*nValley );

	i=0;
	while(i<nValley)
	{
		if( VerProj[LeftValley[i]] > TH_MAX )	
		{
			i++;
			continue;
		}
		Char[nChar].rc.left = RightValley[i]+1;
		Char[nChar].rc.top = 10000;
		Char[nChar].rc.bottom = 0;
		Char[nChar].nSegment = 0;
		Char[nChar].StartSeg = NULL;
		Char[nChar].EndSeg = NULL;

		for(int j=i+1; j<nValley; j++ )
		{
			if( VerProj[LeftValley[j]] <= TH_MAX )
			{
				Char[nChar].rc.right = LeftValley[j]-1;
				nChar ++;
				break;
			}
		}
		i = j;
	}
	//Free memory
	free( LeftValley );
	free( RightValley );
	free( VerProj );

	//Get top and bottom boundary of each character.
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nCC >= 0 )	continue;//Extra CC
		for( int j=0; j<nChar; j++ )
		{
			if( Char[j].rc.left <= ColRun[i].x && Char[j].rc.right >= ColRun[i].x )
			{
				Char[j].rc.top = min( Char[j].rc.top, ColRun[i].ys );
				Char[j].rc.bottom = max( Char[j].rc.bottom, ColRun[i].ye );
				break;
			}
		}
	}

	//Merging characters closing to each other
	i = 0;
	while( i < nChar-1 )
	{
		int nGap = Char[i+1].rc.left - Char[i].rc.right;
		int w1 = Char[i].rc.right - Char[i].rc.left + 1;
		int h1 = Char[i].rc.bottom - Char[i].rc.top + 1;
		int w2 = Char[i+1].rc.right - Char[i+1].rc.left + 1;
		int h2 = Char[i+1].rc.bottom - Char[i+1].rc.top + 1;
		int w  = Char[i+1].rc.right - Char[i].rc.left + 1;
		int h = max( Char[i+1].rc.bottom, Char[i].rc.bottom ) - min( Char[i+1].rc.top, Char[i].rc.top ) + 1;
		if( nGap <= 3 &&	//Two characters are close to each other enough
			(w1 < h1*0.5 || w2 < h2*0.5 || w < 1.3*h ) )//About half width of normal characters
		{
			Char[i].rc.right = Char[i+1].rc.right;
			Char[i].rc.top   = min( Char[i].rc.top, Char[i+1].rc.top );
			Char[i].rc.bottom = max( Char[i].rc.bottom, Char[i+1].rc.bottom );
			for( j=i+1; j<nChar-1; j++ )
				Char[j] = Char[j+1];
			nChar--;
		}
		else
			i++;
	}
	//Remove too small characters
	i = 0;
	while( i < nChar-1 )
	{
		int nWidth = Char[i].rc.right - Char[i].rc.left + 1;
		int nHeight = Char[i].rc.bottom - Char[i].rc.top + 1;
		if( nWidth <= 3 && nHeight <= 8 )
		{
			for( j=i; j<nChar-1; j++ )
				Char[j] = Char[j+1];
			nChar--;
		}
		else
			i++;
	}

	return 0;
}

/***********************************************************************
/*	Name:		SetEraseThreshold
/*	Function:	Set line erase threshold
/*	Parameter:	LineWidth -- Estimated average line width
/*	Return:		Erasion threshold
/************************************************************************/
int SetEraseThreshold(double LineWidth)
{
	int nWidth = (int)LineWidth;
	switch( nWidth )
	{
	case 0:
		return 2;
	case 1:
		return 2; 
	case 2:
		return 3;
	case 3:
		return 4;
	case 4:
		return 5;
	default:
		return (int)(1.3*LineWidth+0.5);
	}
	return -1;
}

int SetEraseThreshold2(double LineWidth)
{
	int nWidth = (int)LineWidth;
	switch( nWidth )
	{
	case 0:
		return 2;
	case 1:
		return 3; 
	case 2:
		return 4;
	case 3:
		return 5;
	case 4:
		return 6;
	default:
		return (int)(1.3*LineWidth+0.5);
	}
	return -1;
}

/******************************************************************************
/*	Name:		SetRemoveMode
/*	Function:	Set erase mode: INSIDE or OUTSIDE characters for each run length
/*	Parameter:	ColRun  -- Column run length
/*				nRun	-- Number of run length
/*				Char    -- Character segmentation results
/*				nChar   -- Number of characters
/*	Return:		0  -- Succeed
/*****************************************************************************/
int SetRemoveMode( COLRUNLENGTH *ColRun, int nRun, CHAR_ZONE *Char, int nChar )
{
	int i;
	int	j=0;
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nUsed == FALSE )
			ColRun[i].nRemoveMode = UNREMOVE_OUTSIDE;
/*		else 
			ColRun[i].nRemoveMode = REMOVE_INSIDE;
*/
		else
		{
			//Search a character boundary, which may cover the run-length
			while( j < nChar )
			{
				if( ColRun[i].x > Char[j].rc.right )
					j++;
				else break;
			}
			if( j == nChar )//Outside of the boundary of any character
				ColRun[i].nRemoveMode = REMOVE_OUTSIDE;
			else
			{
				if( ColRun[i].x < Char[j].rc.left )//Outside
					ColRun[i].nRemoveMode = REMOVE_OUTSIDE;
				else
					ColRun[i].nRemoveMode = REMOVE_INSIDE; //Inside
			}
		}
	}
	return 0;
}

/******************************************************************************
/*	Name:		SetRemoveModeInsideChar
/*	Function:	Label the remove mode for each run length
/*	Parameter:	ColRun			-- Column run length
/*				nRun			-- Number of run length
/*				nEraseThreshold	-- Erase threshold
/*	Return:		0  -- Succeed
/*****************************************************************************/
int SetRemoveModeInsideChar( COLRUNLENGTH *ColRun, int nRun, int nEraseThreshold )
{
	int i;
	for( i=0; i<nRun; i++ )
	{
		if( ColRun[i].nRemoveMode == REMOVE_INSIDE &&
			ColRun[i].ye - ColRun[i].ys + 1 >= nEraseThreshold )
				ColRun[i].nRemoveMode = UNREMOVE_INSIDE; //Do not remove
	}
	return 0;
}

/******************************************************************************
/*	Name:		ReLabelSegment
/*	Function:	Do not remove some segments inside characters
/*	Parameter:	ColRun			-- Column run length
/*				nRun			-- Number of run length
/*				Char    -- Character segmentation results
/*				nChar   -- Number of characters
/*	Return:		0  -- Succeed
/*****************************************************************************/
int ReLabelSegment( COLRUNLENGTH *ColRun, int nRun, CHAR_ZONE *Char, int nChar )
{
	int		i, j;
	j = 0;
	for( i=0; i<nChar; i++ )
	{
		while( j<nRun )
		{
			if( ColRun[j].nUsed == TRUE && ColRun[j].x >= Char[i].rc.left )
				break;
			j++;
		}
		if( j >= nRun )							break;
		if( ColRun[j].x > Char[i].rc.right )	break;

		int nCharWidth = Char[i].rc.right - Char[i].rc.left + 1;
		Char[i].StartSeg = (int*)malloc( sizeof(int)*(nCharWidth+1) );
		Char[i].EndSeg = (int*)malloc( sizeof(int)*(nCharWidth+1) );

		Char[i].StartSeg[0] = j;
		Char[i].EndSeg[0] = j;

		int		nLastX = ColRun[j].x;
		int		nLastRemoveMode = ColRun[j].nRemoveMode;

		while( j < nRun )
		{
			if( ColRun[j].x > Char[i].rc.right )
				break;
			if( ColRun[j].nUsed == FALSE )
			{//This run is not used to compose the line
				j++;
				continue;
			}
			if( ColRun[j].nRemoveMode == UNREMOVE_OUTSIDE || ColRun[j].nRemoveMode == REMOVE_OUTSIDE )
			{//Not inside the character region
				j++;
				continue;
			}

			if(	ColRun[j].x > nLastX+1 || ColRun[j].nRemoveMode != nLastRemoveMode )
			{//Find a new segment
				Char[i].nSegment ++;
				Char[i].StartSeg[Char[i].nSegment] = j;
				Char[i].EndSeg[Char[i].nSegment] = j;
				nLastRemoveMode = ColRun[j].nRemoveMode;
			}
			else
				Char[i].EndSeg[Char[i].nSegment] = j; //Upadate End Segments
			nLastX = ColRun[j].x;
			j++;
		}

		//Change remove mode of some small REMOVE_INSIDE segments
		for(int k=1; k<Char[i].nSegment-1; k++ )
		{
			int Start = Char[i].StartSeg[k];
			int End = Char[i].EndSeg[k];
			if( ColRun[Start].nRemoveMode == UNREMOVE_INSIDE ) 
				continue;
			if( ColRun[End].x - ColRun[Start].x + 1 >= 5 )
				continue;	//The segment is too large

			int nLeftRun = Char[i].EndSeg[k-1];
			if( ColRun[nLeftRun].x < ColRun[Start].x - 1 )
				continue;	//Not connected on the left border
			int nRightRun = Char[i].StartSeg[k+1];
			if( ColRun[nLeftRun].x > ColRun[End].x + 1 )
				continue;	//Not connected on the right border

			for( int m=Start; m<=End; m++ )	//Change the remove mode of small segments
				ColRun[m].nRemoveMode = UNREMOVE_INSIDE;
		}
	}
	return 0;
}

/******************************************************************************
/*	Name:		SetTouchType
/*	Function:	Set touch type for each character: UP_TOUCH, BOTTOM_TOUCH, INTERSECTION
/*	Parameter:	ColRun		-- Column run length
/*				nRun		-- Number of run length
/*				Char		-- Character segmentation results
/*				nChar		-- Number of characters
/*				nLineWidth	-- Estimated line width
/*	Return:		0  -- Succeed
/*****************************************************************************/
int SetTouchType( COLRUNLENGTH *ColRun, int nRun, CHAR_ZONE *Char, int nChar, int nLineWidth )
{
	int i, j;

	j = 0;
	for( i=0; i<nChar; i++ )
	{
		int		nLineRun = 0;
		double	ay = 0;
		Char[i].nTouchType = UNDETERMINATE;
		while( j<nRun )
		{
			if( ColRun[j].x >= Char[i].rc.left )
				break;
			j++;
		}
		if( j >= nRun )							break;
		if( ColRun[j].x > Char[i].rc.right )	break;

		while( ColRun[j].x <= Char[i].rc.right )
		{
			if( ColRun[j].nUsed )
			{
				ay += 0.5*(ColRun[j].ye + ColRun[j].ys );
				nLineRun++;
			}
			j++;
			if( j >= nRun )	break;
		}
		if( nLineRun == 0 )		break;
		ay /= nLineRun;
		if( ay <= Char[i].rc.top + 4 + nLineWidth/2)	
			Char[i].nTouchType = TOP_TOUCH;
		else if( ay >= Char[i].rc.bottom - 4 - nLineWidth/2)
			Char[i].nTouchType = BOTTOM_TOUCH;
		else
			Char[i].nTouchType = INTERSECTION;

	}
	return 0;
}

/******************************************************************************
/*	Name:		CompleteCircle
/*	Function:	Complete circel, special process of characters: "0", "8", etc.
/*	Parameter:	ColRun			-- Column run length
/*				nRun			-- Number of run length
/*				Char    -- Character segmentation results
/*				nChar   -- Number of characters
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CompleteCircle( COLRUNLENGTH *ColRun, int nRun, CHAR_ZONE *Char, int nChar )
{
	for( int k=0; k<nChar; k++ )
	{
		if( Char[k].nTouchType != TOP_TOUCH && Char[k].nTouchType != BOTTOM_TOUCH )
			continue;
		if( Char[k].rc.bottom - Char[k].rc.top + 1 <= 8 )	//The character is with too small height
			continue;

		int i, j;
		for( i=0; i<Char[k].nSegment; i++ )
		{
			int nStartRun = Char[k].StartSeg[i];
			if( ColRun[nStartRun].nRemoveMode != REMOVE_INSIDE )
				continue;
			int nEndRun = Char[k].EndSeg[i];
			int nSegWidth = ColRun[nEndRun].x - ColRun[nStartRun].x + 1;
			if( nSegWidth >= 8 )//The segment is too long
				continue;

			//Search two gap widths above or below the segments
			int y2, y3, w2, w3;
			int nLeftX2, nRightX2, nLeftX3, nRightX3, X;
			if( Char[k].nTouchType == TOP_TOUCH )
			{
				int nBottomY = 0;
				for( j=nStartRun; j<=nEndRun; j++ )
				{
					if( ColRun[j].nUsed == TRUE )
						nBottomY = max( ColRun[j].ye, nBottomY );
				}
				y2 = nBottomY + 3;
				y3 = nBottomY + 5;
			}
			else
			{
				int nTopY = 10000;
				for( j=nStartRun; j<=nEndRun; j++ )
				{
					if( ColRun[j].nUsed == TRUE )
						nTopY = min( ColRun[j].ys, nTopY );
				}
				y2 = nTopY - 3;
				y3 = nTopY - 5;
			}

			nLeftX2 = 0;
			nRightX2 = 10000;
			nLeftX3 = 0;
			nRightX3 = 10000;
			X = (ColRun[nStartRun].x + ColRun[nEndRun].x ) /2;

			for( j=Char[k].StartSeg[0]; j<=Char[k].EndSeg[Char[k].nSegment-1]; j++ )
			{
				if( ColRun[j].ys <= y2 && ColRun[j].ye >= y2 )
				{
					if( ColRun[j].x <= X )//Left side
						nLeftX2 = max( nLeftX2, ColRun[j].x );
					else
						nRightX2 = min( nRightX2, ColRun[j].x );
				}
				if( ColRun[j].ys <= y3 && ColRun[j].ye >= y3 )
				{
					if( ColRun[j].x <= X )//Left side
						nLeftX3 = max( nLeftX3, ColRun[j].x );
					else
						nRightX3 = min( nRightX3, ColRun[j].x );
				}
			}

			w2 = nRightX2 - nLeftX2 + 1;
			w3 = nRightX3 - nLeftX3 + 1;
			if( w2 > nSegWidth && w3 > w2 && w3 >= 2*nSegWidth )
			{//Do not remove this segment to complete the circle
				for( j=nStartRun; j<=nEndRun; j++ )
				{
					if( ColRun[j].nUsed == TRUE )
						ColRun[j].nRemoveMode = UNREMOVE_INSIDE;
				}
			}
		}
	}
	return 0;
}