////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		ParalLine.cpp
// File Function:	Common functions to processing paralle lines
//
//				Developed by: Yefeng Zheng
//			   First created: Sept. 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ParalLine.h"
#include "math.h"
#include "tools.h"

#ifndef	PI
#define PI  3.1415926535
#endif

/****************************************************************************************************
/*	Name:		ReadDetectResult
/*	Function:	Read ground groundtruth
/*	Parameter:	sFileName -- Groundtruth database file name
/*				pLine -- Pointer to reference lines 
/*				nLine -- Number of reference lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int ReadDetectResult(CString sFileName, LINE *&pLine, int &nLine )
{
	FILE *fp = fopen( sFileName, "rt" );
	if( fp == NULL )	return -1;
	char buf[1000];
	fgets( buf, 1000, fp );
	sscanf( buf, "Total Lines=%d", &nLine );
	if( nLine < 0 || nLine > 1000 )
	{
		printf( "Number of lines is wrong!\n" );
		return -1;
	}
	pLine = (LINE*)malloc(nLine*sizeof(LINE));
	if(pLine == NULL)		return -1;
	for(int i=0; i<nLine; i++)
	{
		fgets( buf, 1000, fp );
		sscanf( buf, "(%d%d)\t(%d%d)", &pLine[i].StPnt.x, &pLine[i].StPnt.y,
									  &pLine[i].EdPnt.x, &pLine[i].EdPnt.y );
	}
	fclose( fp );
	return 0;
}

/****************************************************************************************************
/*	Name:		ReadDetectResultBin
/*	Function:	Read ground groundtruth
/*	Parameter:	sFileName -- Groundtruth database file name
/*				pLine -- Pointer to reference lines 
/*				nLine -- Number of reference lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int ReadDetectResultBin(CString sFileName, LINE *&pLine, int &nLine )
{
	FILE	*fp = fopen(sFileName, "rb");
	if( fp == NULL)			return	-1; 
	fread(&nLine, sizeof(int), 1, fp);
	if( nLine < 0 || nLine > 1000 )
	{
		printf( "Number of lines is wrong!\n" );
		return -1;
	}
	pLine = (LINE*)malloc(nLine*sizeof(LINE));
	if(pLine == NULL)		return -1;
	for(int i=0; i<nLine; i++)
	{	
		fread( &pLine[i].StPnt, sizeof(CPoint), 1, fp);
		fread( &pLine[i].EdPnt, sizeof(CPoint), 1, fp);
		fread( &pLine[i].nType, sizeof(int), 1, fp);
	}
	fclose( fp );
	return 0;
}

/****************************************************************************************************
/*	Name:		WriteDetectResult
/*	Function:	Write detection result
/*	Parameter:	sFileName -- File name
/*				pLine -- Pointer to detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int WriteDetectResult(CString sFileName, LINE *pLine, int nLine )
{
	FILE *fpGt = fopen( sFileName, "wt" );
	if( fpGt == NULL )
		return -1;
	fprintf( fpGt, "Total Lines=%d\n", nLine );
	for( int i=0; i<nLine; i++ )
		fprintf( fpGt, "(%5d %5d)\t(%5d %5d)\n", pLine[i].StPnt.x, pLine[i].StPnt.y, 
											pLine[i].EdPnt.x, pLine[i].EdPnt.y );
	fclose( fpGt );
	return 0;
}

/****************************************************************************************************
/*	Name:		WriteDetectResultBin
/*	Function:	Write detection result
/*	Parameter:	sFileName -- File name
/*				pLine -- Pointer to detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int WriteDetectResultBin(CString sFileName, LINE *pLine, int nLine )
{	
	if(pLine == NULL )	return -1;
	FILE	*fp = fopen(sFileName, "wb");
	if(fp == NULL)						return -1;
	fwrite(&nLine, sizeof(int), 1, fp);
	for(int i=0; i<nLine; i++)
	{
		fwrite( &pLine[i].StPnt, sizeof(CPoint), 1, fp);
		fwrite( &pLine[i].EdPnt, sizeof(CPoint), 1, fp);
		fwrite( &pLine[i].nType, sizeof(int), 1, fp);
	}
	fclose(fp);
	return 0;
}

/****************************************************************************************************
/*	Name:		SortLine
/*	Function:	Sort parallel lines from top to bottom
/*	Parameter:	pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int SortLine( LINE *pLine, int nLine )
{
	for( int i=0; i<nLine-1; i++ )
	{
		int nMin = i;
		for( int j=i+1; j<nLine; j++ )
		{
			if( pLine[j].StPnt.y < pLine[nMin].StPnt.y )
				nMin = j;
		}
		if( nMin != i )
		{
			LINE	tmpLine = pLine[i];
			pLine[i] = pLine[nMin];
			pLine[nMin] = tmpLine;
		}
	}
	return 0;
}

/*************************************************************************************
/*	Name:		GetLineAngle
/*	Function:	Calculate the angle of the line formed by two points
/*	Parameter:	Start -- Starting point
/*				End -- Ending point
/*	Return:		Angle of the line
/**************************************************************************************/
double GetLineAngle ( POINT Start, POINT End )
{
    if ( End.x != Start.x )
    {
        double atg = atan ( ((double)(Start.y-End.y)) / ((double)(End.x-Start.x)) ) ;
        if ( End.x > Start.x )
            return atg ;
        else
        {
            if ( End.y < Start.y )
                return atg+PI ;
            else
                return atg-PI ;
        }
    }
    else
    {
        if ( End.y < Start.y )
            return PI/2 ;
        else if ( End.y > Start.y )
            return -PI/2 ;
        else
            return 0 ;
    }
}

/****************************************************************************************************
/*	Name:		GetLineSkew
/*	Function:	Get average line skew angle
/*	Parameter:	pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*				nSkew -- Average skew angle
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	GetLineSkew( LINE *pLine, int nLine, double &nSkew )
{
	nSkew = 0;
	int nNum = 0;
	for(int i=0; i<nLine; i++ )
	{
		if( pLine[i].nType == 1 )
		{
			nSkew += GetLineAngle( pLine[i].StPnt, pLine[i].EdPnt );
			nNum++;
		}
	}
	nSkew /= nNum;
	return 0;
}

/****************************************************************************************************
/*	Name:		GetLineGap
/*	Function:	Get average gap between neighboring lines
/*	Parameter:	pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*				nLineGap -- Average line gap between neighboring lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int GetLineGap( LINE *pLine, int nLine, double &nLineGap )
{
	double nLeftLineGap = 0;
	double nRightLineGap = 0;
	int		Num = 0;

	for( int i=0; i<nLine-1; i++ )
	{
		if( pLine[i].nType == 0 || pLine[i+1].nType == 0 )		break;
		nLeftLineGap += pLine[i+1].StPnt.y - pLine[i].StPnt.y;
		nRightLineGap += pLine[i+1].EdPnt.y - pLine[i].EdPnt.y;
		Num ++;
	}
	if( Num == 0 )	return -1;
	nLeftLineGap /= Num;
	nRightLineGap /= Num;
	if( fabs( nLeftLineGap-nRightLineGap ) >= 4 )
		printf( "Warning: lines are not parallel!" );
	nLineGap = (nLeftLineGap + nRightLineGap)/2;
	return 0;
}

/****************************************************************************************************
/*	Name:		RemoveVirtualLine
/*	Function:	Remove virtual lines
/*	Parameter:	pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	RemoveVirtualLine( LINE *pLine, int &nLine )
{
	int i=0;
	while( i<nLine )
	{
		if( pLine[i].nType == 0 )
		{
			for( int j=i; j<nLine-1; j++ )
				pLine[j] = pLine[j+1];
			nLine--;
		}
		else 
			i++;
	}
	return 0;
}

/****************************************************************************************************
/*	Name:		RemoveBorderLine
/*	Function:	Remove border lines
/*	Parameter:	pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	RemoveBorderLine( LINE *pLine, int &nLine )
{
	//Search for the first solid line
	int		nFirstLine =0;
	while( nFirstLine<nLine )
	{
		if( pLine[nFirstLine].nType == 0)	nFirstLine++;
		else		break;
	}
	//Remove virtual lines
	for( int j=nFirstLine; j<nLine; j++ )
		pLine[j-nFirstLine] = pLine[j];
	nLine -= nFirstLine;

	//Search for the last solid line
	int		nLastLine = nLine-1;
	while( nLastLine >= 0 )
	{
		if( pLine[nLastLine].nType == 0 )	nLastLine--;
		else			break;
	}
	nLine = nLastLine+1;
	return 0;
}

/****************************************************************************************************
/*	Name:		AdjustLinePosition
/*	Function:	Adjust starting and ending points of lines, extending them to image borders
/*	Parameter:	w	  -- Image width
/*				h     -- Image height
/*				pLine -- Detected lines 
/*				nLine -- Number of detected lines
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	AdjustLinePosition( int w, int	h, LINE *pLine, int nLine )
{
	int	i;
	CPoint	LeftTopCorner, RightTopCorner, LeftBottomCorner, RightBottomCorner;
	LeftTopCorner = CPoint(0, 0);
	RightTopCorner = CPoint(w-1, 0);
	LeftBottomCorner = CPoint(0, h-1);
	RightBottomCorner = CPoint(w-1, h-1);
	for( i=0; i<nLine; i++ )
	{
		CPoint	CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, LeftTopCorner, LeftBottomCorner );
		if( CrossPnt.y < 0 )
		{
			CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, LeftTopCorner, RightTopCorner );
			pLine[i].nType = 0;
		}
		else if( CrossPnt.y >= h )
		{
			CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, LeftBottomCorner, RightBottomCorner );
			pLine[i].nType = 0;
		}
		pLine[i].StPnt = CrossPnt;
		
		CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, RightTopCorner, RightBottomCorner );
		if( CrossPnt.y < 0 )
		{
			CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, LeftTopCorner, RightTopCorner );
			pLine[i].nType = 0;
		}
		else if( CrossPnt.y >= h )
		{
			CrossPnt = GetCrossPoint( pLine[i].StPnt, pLine[i].EdPnt, LeftBottomCorner, RightBottomCorner );
			pLine[i].nType = 0;
		}
		pLine[i].EdPnt = CrossPnt;
	}
	return 0;
}

