////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		RuleLine.cpp
// File Function:	API for rule line detection
//
//				Developed by: Yefeng Zheng
//			   First created: Nov. 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RuleLineAPI.h"
#include "ReadHMM.h"
#include "LineDetect.h"
#include "LineRemoval.h"

CImage	m_Img;

CRLLineDetection::CRLLineDetection()
{
	m_pLine = NULL;
	m_nLine = 0;
	m_bLine = false;
	m_HMM_Model.nMinLineGap = -1;
	m_HMM_Model.nMaxLineGap = -1;
	return;
}

CRLLineDetection::~CRLLineDetection()
{
	if( m_pLine != NULL )
		free( m_pLine );
	return;
}

/****************************************************************************************************
/*	Name:		RLSetImage
/*	Function:	Set image for rule line detection
/*	Parameter:	fnImg -- Image file name
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLSetImage( char *fnImg )
{
	if( m_Img.Read(fnImg) != 0 )
	{
		printf( "Can not read image %s!\n", fnImg );
		return -1;
	}

	return 0;
}

/******************************************************************************
/*	Name:		RLSetImage
/*	Function:	Set image from memory
/*	Parameter:	pImgData	-- Image data
/*				w			-- Image width
/*				h			-- Image height
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CRLLineDetection::RLSetImage( BYTE *pImgData, int w, int h )
{
	if( m_Img.Initialize( pImgData, w, (w+7)/8, h, DIB_1BIT, FALSE ) !=0 )
	{
		printf( "Can not initialize the image!\n" );
		return -1;
	}
	return 0;
}

/****************************************************************************************************
/*	Name:		RLSetHMMModel
/*	Function:	Set HMM model
/*	Parameter:	fnHMM_Model -- HMM model
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLSetHMMModel( char *fnHMM_Model )
{
	ReadHMM( fnHMM_Model, m_HMM_Model );
	return 0;
}

/****************************************************************************************************
/*	Name:		RLLineDetect
/*	Function:	Rule line detection
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLLineDetect( )
{
	if( m_Img.GethImg() == NULL )
	{	printf( "Please set image file name first!\n" );
		return -1;
	}
	if( m_HMM_Model.nMinLineGap == -1 && m_HMM_Model.nMaxLineGap == -1 )
	{
		printf( "Please set HMM model first!\n" );
		return -1;
	}
	if( m_pLine != NULL )
		free( m_pLine );

	DetectLine( m_Img, m_HMM_Model, TRUE, m_pLine, m_nLine, m_bLine );
	return 0;
}

/****************************************************************************************************
/*	Name:		RLLineExist
/*	Function:	Whether rule lines exist
/*	Parameter:	NULL
/*	Return:		true -- Exist
/***************************************************************************************************/
bool CRLLineDetection::RLLineExist( )
{
	return m_bLine;
}

/****************************************************************************************************
/*	Name:		RLGetLines
/*	Function:	Output detected lines
/*	Parameter:	pLine   -- Detected lines
/*				nLine   -- Number of lines
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLGetLines( LINE *&pLine, int &nLine )
{
	if( m_nLine == 0 )
	{
		pLine = NULL;
		nLine = 0;
		return -1;
	}
	nLine = m_nLine;
	pLine = (LINE*)malloc( sizeof(LINE)*m_nLine );
	for( int i=0; i<m_nLine; i++ )
		pLine[i] = m_pLine[i];
	return 0;
}

/****************************************************************************************************
/*	Name:		RLLineRemove
/*	Function:	Rule line removal
/*	Parameter:	
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLLineRemove( )
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image file name first!\n" );
		return -1;
	}
	if( m_pLine == NULL || m_nLine == 0 )
	{
		printf( "Please detect line first!\n" );
		return -1;
	}
	RemoveLineWithCharSeg( m_Img, m_pLine, m_nLine );
	return 0;
}

/****************************************************************************************************
/*	Name:		RLGetCleanImage
/*	Function:	Get cleaned image after rule line removal
/*	Parameter:	pCleanImg -- Cleaned image after line removal
/*				w         -- Image width
/*				h         -- Image height
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLGetCleanImage( BYTE *&pCleanImg, int &w, int &h)
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image file name first!\n" );
		return -1;
	}
	w = m_Img.GetWidth();
	h = m_Img.GetHeight();
	int wb = m_Img.GetLineSizeInBytes();
	BYTE *pImg = m_Img.LockRawImg();
	int ww = (w+7)/8;
	pCleanImg = (BYTE*)malloc( sizeof(BYTE)*ww*h);
	for( int i=0; i<h; i++ )
		memcpy( pCleanImg+i*ww, pImg+(h-1-i)*wb, ww );

/*	CImage Img;
	Img.Initialize( pCleanImg, w, ww, h, DIB_1BIT, FALSE );
	Img.Write( "RemovalImg.tif" );
*/	return 0;
}

/****************************************************************************************************
/*	Name:		RLSaveCleanImage
/*	Function:	Save cleaned image after rule line removal
/*	Parameter:	fnCleanImg -- Image file name
/*	Return:		0  -- Succeed
/***************************************************************************************************/
int CRLLineDetection::RLSaveCleanImage( char *fnCleanImg )
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image file name first!\n" );
		return -1;
	}
	m_Img.Write( fnCleanImg );
	return 0;
}
