////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		TrainHMM.cpp
// File Function:	Train a specified HMM model to detect background parallel lines
//
//				Developed by: Yefeng Zheng
//			   First created: Nov. 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ParalLine.h"
#include "Tools.h"
#include "math.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "DirLine.h"
#include "RuleLineAPI.h"
#include "LineDetect.h"
#include "TrainHMMDiscrete.h"

/****************************************************************************************************
/*	Name:		OutputHMM
/*	Function:	Output HMM model to file
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	OutputHMM( char *fnHMM, HMM_MODEL model )
{
	FILE *fp = fopen( fnHMM, "wt" );
	if( fp ==NULL )	
	{
		printf( "Can not open file to output HMM model!\n" );
		return -1;
	}
	int i;
	//Output text peak probalility
	fprintf( fp, "Text peak probability:\n" );
	for( i=0; i<MAXPEAK; i++ )
		fprintf( fp, "%5d: %5d\n", i, (int)model.TextPeakProb[i]);
	
	//Output line peak probability
	fprintf( fp, "Line peak probability:\n" );
	for( i=0; i<MAXPEAK; i++ )
		fprintf( fp, "%5d: %5d\n", i, (int)model.LinePeakProb[i]);

	//Output vertical line gap distribution
	fprintf( fp, "Vertical line gap:\n" );
	for( int nMin=0; nMin<MAXLINEGAP; nMin++ )
		if( model.LineGapDist[nMin] > 0 )
			break;
	for( int nMax=MAXLINEGAP-1; nMax>=0; nMax-- )
		if( model.LineGapDist[nMax] > 0 )
			break;

	fprintf( fp, "Min: %5d\tMax: %5d\n", nMin-50, nMax-50 );
	for( i=nMin; i<=nMax; i++ )
		fprintf( fp, "%5d  %5d\n", i-50, (int)model.LineGapDist[i]);
	fclose( fp );
	return 0;
}

/****************************************************************************************************
/*	Name:		TrainHMMDiscrete
/*	Function:	Train HMM model using discrete observation
/*	Parameter:	fnPeak  -- File of projection peak information
/*				fnGrt   -- File of groundtruthed lines
/*				bFilter -- Perform text filtering or not
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	TrainHMM( char *fnImg, char *fnGrt, BOOL bFilter, HMM_MODEL &model )
{
	CImage Img;
	if( Img.Read( fnImg ) != 0 )
	{
		printf( "Can not open image: %s to read!\n", fnImg );
		return -1;
	}

	int	w = Img.GetWidth();
	int	h = Img.GetHeight();

	//Read groundtruthed background parallel lines
	LINE	*pLine;
	int		nLine;
	if( ReadDetectResult( fnGrt, pLine, nLine ) != 0 )
	{
		printf( "Read ground truth failed: %s\n", fnGrt );
		return -1;
	}
	//Extend the end points of line to image borders
	AdjustLinePosition( w, h, pLine, nLine );
	//Sort lines from top to bottom
	SortLine( pLine, nLine );
	//Remove lines near image borders
	RemoveBorderLine( pLine, nLine );

	//Get average line gap
	double	nAverLineGap;
	GetLineGap( pLine, nLine, nAverLineGap );
	nAverLineGap = floor( nAverLineGap+0.5);

	//Get skew angle
	double	nSkewAngle;
	GetLineSkew( pLine, nLine, nSkewAngle );

	//Do skewed horizontal projection
	double	*Proj;
	int		nProj;
	if( bFilter )
	{
		CImage		FilteredImg;
		CDirLine	DirLine;
		DirLine.DSCCFiltering( Img, FilteredImg, TRUE );
		CImageTool::DoSkewHorProject( FilteredImg, nSkewAngle, Proj, nProj );
	}
	else
		CImageTool::DoSkewHorProject( Img, nSkewAngle, Proj, nProj );

	double	*ProjPeak;
	FindProjPeak( Proj, nProj, ProjPeak );
	Quantize( ProjPeak, nProj, w );

	//Get line projections
	int		dh = (int)fabs( w*tan(nSkewAngle) );
	int		i, j, nMaxPeak;
	for( i=0; i<nLine; i++ )
	{
		int nPos;
		if( nSkewAngle >= 0 )
			nPos = pLine[i].StPnt.y;
		else
			nPos = pLine[i].StPnt.y+dh;

		nMaxPeak = nPos;
		for( j=max(0, nPos-5); j<=min(nProj-1, nPos+5); j++ )
			if( ProjPeak[j] > ProjPeak[nMaxPeak] )
				nMaxPeak = j;
		if( ProjPeak[nMaxPeak] == 0 )
			int kk = 0;
		model.LinePeakProb[(int)ProjPeak[nMaxPeak]] ++;
		ProjPeak[nMaxPeak] = -1;
	}

	//Get text projections
	int n = 0;
	for( i=0; i<nProj; i++ )
	{
		if( ProjPeak[i] >= 0 )
			model.TextPeakProb[(int)ProjPeak[i]] ++;
		else
			n++;
	}

	//Get line gap distribution
	for( i=0; i<nLine-1; i++ )
	{
		int nVar =  (int)( ( (pLine[i+1].StPnt.y-pLine[i].StPnt.y) + 
							(pLine[i+1].EdPnt.y-pLine[i].EdPnt.y) ) / 2 
							- nAverLineGap + MAXLINEGAP/2
						);
		if( nVar < 0 || nVar >= MAXLINEGAP )
		{
			printf( "Line gap error!\n" );
			return -1;
		}
		model.LineGapDist[nVar]++;
	}

	free( pLine );
	free( Proj );
	free( ProjPeak );
	return -1;
}

/****************************************************************************************************
/*	Name:		TrainHMMDiscrete
/*	Function:	Train HMM model using discrete observation
/*	Parameter:	fnHMM   -- File name of HMM model
/*				bFilter -- Perform text filtering or not
/*	Return:		0  -- Succeed
/*				-1 -- Failed
/***************************************************************************************************/
int	TrainHMM( char *fnHMM, BOOL bFilter, HMM_MODEL &model )
{
	ZeroMemory( model.TextPeakProb, sizeof(int)*MAXPEAK );
	ZeroMemory( model.LinePeakProb, sizeof(int)*MAXPEAK );
	ZeroMemory( model.LineGapDist, sizeof(int)*MAXLINEGAP );

	char	buf[1000];
	CString cs;
	char	csGtDir[MAX_PATH];
	char	csGtSuffix[MAX_PATH];
	char	csImgDir[MAX_PATH];
	char	csImgSuffix[MAX_PATH];

	FILE	*fpTrain = fopen( fnHMM, "rt" );
	if( fpTrain == NULL )
	{
		printf( "Can not open %s to read!\n", fnHMM );
		return -1;
	}
	while( !feof( fpTrain ) )
	{
		fgets( buf, 1000, fpTrain );
		cs = CString( buf );
		cs.MakeUpper();
		if( cs.Find( "GROUNDTRUTH_DIR" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csGtDir, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "GT_SUFFIX" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csGtSuffix, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "IMG_DIR" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csImgDir, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "IMG_SUFFIX" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csImgSuffix, "%s", (LPCSTR)cs );
		}
	}
	rewind( fpTrain );
	while( !feof( fpTrain ) )
	{
		fgets( buf, 1000, fpTrain );
		if( strstr( buf, "---" ) != NULL )
			break;
	}

	char	fnTruth[MAX_PATH];
	char	fnImg[MAX_PATH];
	while( !feof(fpTrain) )
	{ 
		if( fgets( buf, 1000, fpTrain ) == NULL )
			break;
		if( strcmp( buf, "\n" ) == 0 )
			break;
		if( buf[strlen(buf)-1] == '\n' )
			buf[strlen(buf)-1] = '\0';
		sprintf( fnTruth, "%s\\%s%s", csGtDir, buf, csGtSuffix );
		sprintf( fnImg, "%s\\%s%s", csImgDir, buf, csImgSuffix );
		printf( "Training with image %s\n", fnImg );
		TrainHMM( fnImg, fnTruth, bFilter, model );
	}
	fclose( fpTrain );	
	return 0;
}
