////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		ExtractRunlengthFeature.cpp
// File Function:	Extract runlength histogram features and stroke complexity features
//
//				Developed by: Yefeng Zheng
//			   First created: May, 2001
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include "imageobj.h"
#include "imagetool.h"
#include "ExtractFeature.h"
#include "Tools.h"

extern int nAverWidth, nAverHeight, nCharSpace, nLineSpace;
extern FILE *fpTotalFeature;

/******************************************************************************
/*	Name:		ExtractStrokeComplexityFeature
/*	Function:	Extract stroke complexity histogram features
/*				Used in Chinese/English classification
/*	Parameter:	Line -- Char image, 1 byte/pixel
/*				w -- Image width
/*				h -- Image height
/*				BoundBox -- Boundbox of char image
/*				HChgFeature -- Horizontal stroke complexity histogram features
/*				VChgFeature -- Vertical stroke complexity histogram features
/*				Dimension -- Dimension
/*	Return:		0  -- Succeed
/*****************************************************************************/
int	ExtractStrokeComplexityFeature( BYTE *Line, int w, int h, CRect BoundBox, double *HChgFeature, double *VChgFeature, int Dimension )
{
	int		i, j;
	int	 nChange;
	ZeroMemory( HChgFeature, Dimension*sizeof(double) );
	ZeroMemory( VChgFeature, Dimension*sizeof(double) );
	int		wh = max( 1, w/h );
//Horizontal change numbers
	for( i=0; i<h; i++ )
	{
		nChange = 0;
		if( Line[i*w] == 1 )		nChange++;
		for( j=0; j<w-1; j++ )
			if( Line[i*w+j] != Line[i*w+j+1] )
				nChange++;
		if( Line[i*w+w-1] == 1 )	nChange++;
		if( nChange%2 != 0)
		{
			AfxMessageBox( "Error!" );
			return -1;
		}
		nChange /= 2;
		nChange /= wh; //Y.F. Zheng 02/13/2002
		if( nChange > Dimension )
			HChgFeature[Dimension-1]++;
		else if( nChange > 0 )
			HChgFeature[nChange-1]++;
	}
//Vertical change numbers
	for( i=0; i<w; i++ )
	{
		nChange = 0;
		if( Line[i] == 1 )			nChange++;
		for( j=0; j<h-1; j++ )
			if( Line[j*w+i] != Line[(j+1)*w+i] )
				nChange++;
		if( Line[(h-1)*w+i] == 1)	nChange++;
		if( nChange%2 != 0)
		{
			AfxMessageBox( "Error!" );
			return -1;
		}
		nChange /= 2;
		if( nChange > Dimension )
			VChgFeature[Dimension-1]++;
		else if( nChange > 0 )
			VChgFeature[nChange-1]++;
	}
	int	RealWidth = max( 16, BoundBox.right-BoundBox.left+1 );
	int	RealHeight = max( 12, BoundBox.bottom-BoundBox.top+1 );
	for( i=0; i<Dimension; i++ )
	{
		HChgFeature[i] /= RealHeight;
		VChgFeature[i] /= RealWidth;
	}		
	return 0;
}

/**************************************************************************************************
/*	Name:		ExtractStraightnessFeature
/*	Function:	Extract horizontal and vertical runlengths histogram features
/*				Used in both Chinese/English classification and Handwritten/Printed classification
/*	Parameter:	Line -- Char image, 1 byte/pixel
/*				w -- Image width
/*				h -- Image height
/*				BoundBox -- Boundbox of char image
/*				HFeature -- Horizontal runlengths histogram features
/*				VFeature -- Vertical runlengths histogram features
/*				Dimension -- Dimension
/*	Return:		0  -- Succeed
/**************************************************************************************************/
int	ExtractStraightnessFeature( BYTE  Line[], int w, int h, CRect BoundBox, double HFeature[], double VFeature[], int Dimension )
{
//Extract horizontal feature
	int		i, j, Total = 0, TotalPixel = 0;
	int		nSize = 1024;
	double	RawFeature[1024];
	ZeroMemory( RawFeature, nSize*sizeof(double));
	int		Start, End;
	int		RealWidth = BoundBox.right-BoundBox.left+1;
	int		RealHeight = BoundBox.bottom-BoundBox.top+1;
	double	delt = 1.0/(Dimension-1)/2;
	double	f1 = 1.0/sqrt(2*PI)/delt;
	for( i=0; i<h; i++ )
	{
		Start=0; 
		while( Start < w )
		{
			while( Start<w && Line[i*w+Start] == 0)	Start++;
			End = Start;
			if( Start >= w )	continue;
			while( End<w && Line[i*w+End] == 1 )		End++;
			RawFeature[End-Start]++;
			Total++;
			Start = End;
		}
	}
	if( Total == 0 )
		return -1;
/////////////////////////////////////////////
	ZeroMemory( HFeature, Dimension*sizeof(double) );
	double		Width = RealWidth/2.0;
	for( i=0; i<Dimension; i++ )
	{//Gauss smooth
		for( j=0; j<nSize; j++ )
		{
			double	f2 = (j*1.0/Width-i*1.0/(Dimension-1))/delt;
			double	factor = f1*exp( -0.5*f2*f2 );
			HFeature[i] += 1.0*RawFeature[j]*factor;
		}
	}

	for( i=0; i<Dimension; i++ )
		HFeature[i] /= RealHeight;

//Extract vertical straightness features
	ZeroMemory( RawFeature, nSize*sizeof(double) );
	Total = 0;
	for( i=0; i<w; i++ )
	{
		Start = 0;
		while( Start < h )
		{
			while( Start<h && Line[Start*w+i]== 0)	Start++;
			if( Start >= h )	break;
			End = Start;
			while( End<h && Line[End*w+i] == 1 )		End++;
			RawFeature[End-Start]++;
			Total++;
			Start = End;
		}
	}
	if( Total == 0 )	
		return -1;
//////////*//////////////////////////////////
	ZeroMemory( VFeature, Dimension*sizeof(double) );
	Width = RealHeight;
	for( i=0; i<Dimension; i++ )
	{//Gauss smooth
		for( j=0; j<nSize; j++ )
		{
			double	f2 = (j*1.0/Width-i*1.0/(Dimension-1))/delt;
			double	factor = f1*exp( -0.5*f2*f2 );
			VFeature[i] += 1.0*RawFeature[j]*factor;
		}
	}
//Normalization
	for( i=0; i<Dimension; i++ )
		VFeature[i] /= RealWidth;

	return 0;
}

/******************************************************************************
/*	Name:		ExtractSlantStraightnessFeature
/*	Function:	Extract 45 and 135 degree runlengths histogram features
/*				Used in Handwritten/Printed classification
/*	Parameter:	Line -- Char image, 1 byte/pixel
/*				w -- Image width
/*				h -- Image height
/*				BoundBox -- Boundbox of char image
/*				LT_RB_Feature -- 135 degree runlengths histogram features
/*				RT_LB_Feature -- 45 degree runlengths histogram features
/*				Dimension -- Dimension
/*	Return:		0  -- Succeed
/*****************************************************************************/
int	ExtractSlantStraightnessFeature( BYTE  Line[], int w, int h, RECT BoundBox, double LT_RB_Feature[], double RT_LB_Feature[], int Dimension )
{
//Extract left-top to right-bottom straightness feature;
	ZeroMemory( LT_RB_Feature, Dimension*sizeof(double) );
	ZeroMemory( RT_LB_Feature, Dimension*sizeof(double) );
	int		nSize = 1024;
	double	RawFeature[1024];
	ZeroMemory( RawFeature, nSize*sizeof(double) );
	int		i, j, k, Total, Start;
	int		Length = (int)(sqrt( (BoundBox.right-BoundBox.left+1)*(BoundBox.right-BoundBox.left+1) + (BoundBox.bottom-BoundBox.top+1)*(BoundBox.bottom-BoundBox.top+1) )/sqrt(2) );
	double	delt = 1.0/(Dimension-1)/2;
	double	f1 = 1.0/sqrt(2*PI)/delt;
	Total = 0;
	for( k=BoundBox.left; k<=BoundBox.right; k++ )
	{
		i=BoundBox.top;	j=k;
		Start = 0;
		while( i<=BoundBox.bottom && j<=BoundBox.right )
		{
			while( i<=BoundBox.bottom && j<=BoundBox.right && Line[i*w+j] != 1)
			{
				i++;	j++;
			}
			if( i>BoundBox.bottom || j>BoundBox.right )	break;
			Start = i;
			while( i<=BoundBox.bottom && j<=BoundBox.right && Line[i*w+j] == 1 )
			{
				i++;	j++;
			}
			if( i-Start < Length )
				RawFeature[i-Start]++;
			else
				RawFeature[Length-1]++;
			Total++;
		}
	}
	for( k=BoundBox.top+1; k<=BoundBox.bottom; k++ )
	{
		i=k;	j=BoundBox.left;
		while( i<=BoundBox.bottom && j<=BoundBox.right )
		{
			while( i<=BoundBox.bottom && j<=BoundBox.right && Line[i*w+j] != 1)
			{
				i++;	j++;
			}
			if( i>BoundBox.bottom || j>BoundBox.right )	break;
			Start = i;
			while( i<=BoundBox.bottom && j<=BoundBox.right && Line[i*w+j] == 1 )
			{
				i++;	j++;
			}
			if( i-Start < Length )
				RawFeature[i-Start]++;
			else
				RawFeature[Length-1]++;
			Total++;
		}
	}
	if( Total == 0 )
	{
		AfxMessageBox( "Slant Error!" );
		return -1;
	}

	double nWidth =  Length / 2;
	for( i=0; i<Dimension; i++ )
	{//Gauss smooth
		for( j=0; j<nSize; j++ )
		{
			double	f2 = (j*1.0/nWidth-i*1.0/(Dimension-1))/delt;
			double	factor = f1*exp( -0.5*f2*f2 );
			LT_RB_Feature[i] += 1.0*RawFeature[j]*factor;
		}
	}

	for( i=0; i<Dimension; i++ )
		LT_RB_Feature[i] /= Length;

//Extract right-top to left-bottom straightness features
	Total = 0;
	ZeroMemory( RawFeature, nSize*sizeof(double) );
	for( k=BoundBox.left; k<=BoundBox.right; k++ )
	{
		i=BoundBox.top;	j=k;
		Start = 0;
		while( i<=BoundBox.bottom && j>=BoundBox.left )
		{
			while( i<=BoundBox.bottom && j>=BoundBox.left && Line[i*w+j] != 1 )
			{
				i++;	j--;
			}
			if( i>BoundBox.bottom || j<BoundBox.left )	break;
			Start = i;
			while( i<=BoundBox.bottom && j>=BoundBox.left && Line[i*w+j] == 1 )
			{
				i++;	j--;
			}
			if( i-Start < Length )
				RawFeature[i-Start]++;
			else
				RawFeature[Length-1]++;
			Total++;
		}
	}
	for( k=BoundBox.top+1; k<=BoundBox.bottom; k++ )
	{
		i=k;	j=BoundBox.right;
		Start = 0;
		while( i<=BoundBox.bottom && j>=BoundBox.left )
		{
			while( i<h && j>=0 && Line[i*w+j] != 1 )
			{
				i++;	j--;
			}
			if( i>BoundBox.bottom || j<BoundBox.left )	break;
			Start = i;
			while( i<=BoundBox.bottom && j>=BoundBox.left && Line[i*w+j] == 1 )
			{
				i++;	j--;
			}
			if( i-Start < Length )
				RawFeature[i-Start]++;
			else
				RawFeature[Length-1]++;
			Total++;
		}
	}
	if( Total == 0 )	
	{
		AfxMessageBox( "Slant Error!" );
		return -1;
	}
	for( i=0; i<Dimension; i++ )
	{//Gauss smooth
		for( j=0; j<nSize; j++ )
		{
			double	f2 = (j*1.0/nWidth-i*1.0/(Dimension-1))/delt;
			double	factor = f1*exp( -0.5*f2*f2 );
			RT_LB_Feature[i] += 1.0*RawFeature[j]*factor;
		}
	}

	for( i=0; i<Dimension; i++ )
		RT_LB_Feature[i] /= Length ;
	return 0;
}

