////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		ExtractTextureFeature.cpp
// File Function:	Extract texture features. These features are used by David Doermann and Jian
//					Liang to evaluate the similarity between two binary image blocks
//
//				Developed by: Yefeng Zheng
//		Original ocpy is developed by: Jian Liang
//			   First created: May, 2001
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "img.h"
#include "math.h"
#include "Tools.h"
#include "define.h"

extern FILE	*fpTotalFeature;
extern int	nHandZone, nPrintZone;

double FDC_H[5], FDC_V[5], FDC_P[5], FDC_M[5];
double FPH[5*16];
double FDR_B_H[3], FDR_B_V[3], FDR_B_P[3], FDR_B_M[3];
double FDR_W_H[3], FDR_W_V[3], FDR_W_P[3], FDR_W_M[3];

int S_DC[] = {1, 2, 4, 8, 16 };
int S_PH[] = {1, 2, 4, 8, 16 };
int	s_dc = 4;
int	s_ph = 4;
int s_dr = 3; // step distances in byte unit.

/******************************************************************************
/*	Name:		NormalizeFDC
/*	Function:	Normalize bi-level co-occurence features
/*	Parameter:	FDC_H -- Horizontal bi-level co-occurence features
/*				FDC_V -- Vertical bi-level co-occurence features
/*				FDC_P -- Primary diagonal bi-level co-occurence features
/*				FDC_M -- Minor diagonal co-occurence features
/*	Return:		TRUE -- Succeed
/*****************************************************************************/
BOOL NormalizeFDC( double *FDC_H, double *FDC_V, double *FDC_P, double *FDC_M )
{
	double nSum;
	for( int j=0; j<s_dc; j++ )
	{
		nSum = FDC_H[j]+FDC_V[j]+FDC_P[j]+FDC_M[j];
		if( nSum != 0 )
		{
			FDC_H[j] /= nSum;
			FDC_V[j] /= nSum;
			FDC_P[j] /= nSum;
			FDC_M[j] /= nSum;
		}
	}
	return TRUE;
}

/******************************************************************************
/*	Name:		NormalizeFPH
/*	Function:	Normalize bi-level 2*2 gram features
/*	Parameter:	FPH   -- 2*2 gram features
/*				ones  -- Number of black pixels in the image
/*				zeros -- Number of white pixels in the image
/*	Return:		TRUE -- Succeed
/*****************************************************************************/
BOOL	NormalizeFPH( double *FPH, int ones, int zeros )
{
	int num1inbyte[] = {0,  1,  1,  2,  1,  2,  2,  3,  1,  2,  2,  3,  2,  3,  3,  4};
	int i;
	for( i=0; i<s_ph; i++ )
		FPH[i*16] = 0;

	double p = 1.0*ones/(ones + zeros);
	double nSum;
	for( i=0; i<s_ph; i++ )
	{
		nSum = 0;
		for( int j=0; j<16; j++ )
		{
			int b = num1inbyte[j];
			FPH[i*16+j] /= (pow(p, b)*pow(1-p, 4-b));
			nSum += FPH[i*16+j];
		}
		if( nSum != 0 )
		{
			for( j=0; j<16; j++ )
				FPH[i*16+j] /= nSum;
		}
	}
	return TRUE;
}

/******************************************************************************
/*	Name:		NormalizeFDR
/*	Function:	Normalize pseudo run lengths features
/*	Parameter:	FDR_H -- Horizontal pseudo run length features
/*				FDR_V -- Vertical pseudo run length features
/*				FDR_P -- Primary diagonal pseudo run length features
/*				FDR_M -- Minor diagonal pseudo run length features
/*	Return:		TRUE -- Succeed
/*****************************************************************************/
BOOL	NormalizeFDR( double *FDR_H, double *FDR_V, double *FDR_P, double *FDR_M )
{
	int i;
	for( i=0; i<s_dr; i++ )
	{
		FDR_H[i] *= pow(2, i);
		FDR_V[i] *= pow(2, i);
		FDR_P[i] *= pow(2, i);
		FDR_M[i] *= pow(2, i);
	}
	for( i=s_dr-1; i>0; i-- )
	{
		FDR_H[i] = FDR_H[i-1] - FDR_H[i];
		FDR_V[i] = FDR_V[i-1] - FDR_V[i];
		FDR_P[i] = FDR_P[i-1] - FDR_P[i];
		FDR_M[i] = FDR_M[i-1] - FDR_M[i];
	}

	double nSum = 0;
	for( i=1; i<s_dr; i++ )
		nSum += FDR_H[i]+FDR_V[i]+FDR_P[i]+FDR_M[i];
	if( nSum != 0 )
	{
		for( i=1; i<s_dr; i++ )
		{
			FDR_H[i] /= nSum;
			FDR_V[i] /= nSum;
			FDR_P[i] /= nSum;
			FDR_M[i] /= nSum;
		}
	}
	return TRUE;
}

/******************************************************************************
/*	Name:		ExtractTextureFeature
/*	Function:	Extract texture features
/*	Parameter:	img     -- Image
/*				Feature -- Feature extracted
/*				zone    -- Zone position in the image need to extract features
/*	Return:		TRUE -- Succeed
/*****************************************************************************/
int	ExtractTextureFeature(	BWIMG img, double *Feature )
{	
	int	ones, zeros;
	//Extract bi-level co-occurrence features
	if( jBWImgFDC (&img, FDC_H, FDC_V, FDC_P, FDC_M, S_DC, s_dc) != 0 )
	{
		printf( "Extract FDC feature error!" );
		return -1;
	}
	NormalizeFDC( FDC_H, FDC_V, FDC_P, FDC_M );

	//Extract bi-level 2*2 grams
	if( jBWImgFPH (&img, FPH, S_PH, s_ph, &ones, &zeros) != 0 )
	{
		printf( "Extract FPH feature error!" );
		return -1;
	}
	NormalizeFPH( FPH, ones, zeros );

	//Pseudo run-length
	if( jBWImgFDR (&img, FDR_B_H, FDR_B_V, FDR_B_P, FDR_B_M, s_dr, 0) != 0 )
	{
		printf( "Extract FDR black feature error!" );
		return -1;
	}
	NormalizeFDR( FDR_B_H, FDR_B_V, FDR_B_P, FDR_B_M );

	if( jBWImgFDR (&img, FDR_W_H, FDR_W_V, FDR_W_P, FDR_W_M, s_dr, 1) != 0 )
	{
		printf( "Extract FDR white feature error!" );
		return -1;
	}
	NormalizeFDR( FDR_W_H, FDR_W_V, FDR_W_P, FDR_W_M );

	int	nDimension = 0;
	for(int j=0; j<s_dc; j++ )
	{
		Feature[nDimension++] = FDC_H[j];
		Feature[nDimension++] = FDC_V[j];
		Feature[nDimension++] = FDC_P[j];
		Feature[nDimension++] = FDC_M[j];
	}
	for( j=0; j<s_ph; j++ )
	{//Remove the white
		for( int k=1; k<16; k++ )
			Feature[nDimension++] = FPH[j*16+k];
	}
	for( j=1; j<s_dr; j++ )
	{//Remove the first level
		Feature[nDimension++] = FDR_B_H[j];
		Feature[nDimension++] = FDR_B_V[j];
		Feature[nDimension++] = FDR_B_P[j];
		Feature[nDimension++] = FDR_B_M[j];
	}
	for( j=1; j<s_dr; j++ )
	{
		Feature[nDimension++] = FDR_W_H[j];
		Feature[nDimension++] = FDR_W_V[j];
		Feature[nDimension++] = FDR_W_P[j];
		Feature[nDimension++] = FDR_W_M[j];
	}
	return 0;
}


