////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		HCF.cpp
// File Function:	Highest Confidence First algorithm for MRF optimization
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PostProcess.h"
#include "define.h"
#include "tools.h"
#include "math.h"

/*******************************************************************************************
/*	Name:		AddCandidate
/*	Function:	Add candiate blocks into the candiate list
/*	Parameter:	FlipCandidate  -- Candidate list
/*				nPointer       -- Number of candidates in the list
/*				nConf		   -- Flip confidence
/*				nZoneType      -- The word type which the current candidate will flip to
/*				nIndex         -- Current candidate
/*	Return:		0 -- Succeed
/********************************************************************************************/
int AddCandidate( FLIP_CANDIDATE *FlipCandidate, int &nPointer, double nConf, int nZoneType, int nIndex )
{
	if( nZoneType >= 4 )
	{
		printf( "Bug!\n" );
		nZoneType = 3;
	}

	if( nPointer == -1 )
	{
		FlipCandidate[0].nConf = nConf;
		FlipCandidate[0].nFlippedType = nZoneType;
		FlipCandidate[0].nIndex = nIndex;
	}
	else
	{
		for(int j=0; j<=nPointer; j++ )
			if( nConf < FlipCandidate[j].nConf )	break;
		for(int k=nPointer+1; k>j; k-- )
			FlipCandidate[k] = FlipCandidate[k-1];
		FlipCandidate[j].nConf = nConf;
		FlipCandidate[j].nFlippedType = nZoneType;
		FlipCandidate[j].nIndex = nIndex;
	}
	nPointer ++;
	return 0;
}

/*******************************************************************************************
/*	Name:		HCF
/*	Function:	HCF optimization algorithm for MRF
/*	Parameter:	zone  -- Blocks
/*				nZone -- Number of blocks
/*				LeftIndex  -- Left neighboring index for Cp clique
/*				RightIndex -- Right neighboring index for Cp clique
/*				NearIndex  -- Neighboring index for Cn clique
/*				MRFModel   -- MRF model
/*	Return:		0 -- Succeed
/********************************************************************************************/
int	HCF( ZONE *zone, int nZone, int *LeftIndex, int *RightIndex, int **NearIndex, MRF_MODEL MRFModel )
{
	int		nMRFLabel, LabelLeft, LabelRight;
	int		NearLabel[4];
	int		nMRFSizeLevel;
	double	cpdf[3];
	int		nPointer, i, j, k;
	FLIP_CANDIDATE	*FlipCandidate;
	int	 nMaxConf = 0; 
	BOOL	bNeedProcess;
	BOOL	bNeedUpdate;
	int		nNear = 4;

	FlipCandidate = (FLIP_CANDIDATE*)malloc( sizeof( FLIP_CANDIDATE )*nZone );
	nPointer = -1;
	for( i=0; i<nZone; i++ )
	{
		if( NeedProcess( zone[i] ) == FALSE )		continue;
		nMRFSizeLevel = GetMRFSizeLevel( zone[i].nSize );
		nMRFLabel = ZoneType2MRFLabel( zone[i].nType );
		LabelLeft = GetMRFLabel( zone, LeftIndex[i] );
		LabelRight = GetMRFLabel( zone, RightIndex[i] );
		for( j=0; j<nNear; j++ )
			NearLabel[j] = GetMRFLabel( zone, NearIndex[j][i] );
		
		bNeedProcess = FALSE;
		if( (MRFModel.nCliqueType & CLIQUE_PRINT) != 0 )
		{
			if( LabelLeft != 3 || LabelRight != 3 )
				bNeedProcess = TRUE;
		}
		if( (MRFModel.nCliqueType & CLIQUE_NOISE) != 0 )
		{
			for( j=0; j<nNear; j++ )
			{
				if( NearLabel[j] != 3 )
					bNeedProcess = TRUE;
			}
		}
		if( bNeedProcess == FALSE )
			continue;

		GetEnergy( zone, nZone, i, LeftIndex, RightIndex, NearIndex, MRFModel, cpdf );

		nMaxConf = 0;
		for( j=1; j<3; j++ )
			if( cpdf[j] > cpdf[nMaxConf] )
				nMaxConf = j;

		if( nMRFLabel != nMaxConf )
		{
			double	nConf = cpdf[nMaxConf]-cpdf[nMRFLabel];
			AddCandidate( FlipCandidate, nPointer, nConf, MRFModel.nClassLabel[nMaxConf], i );
		}
	}

	int nIter = 0;
	while( nPointer >= 0 )
	{
		int nIndex = FlipCandidate[nPointer].nIndex;
		int nMRFSizeLevel = GetMRFSizeLevel( zone[nIndex].nSize );
		zone[nIndex].nType = FlipCandidate[nPointer].nFlippedType;
		if( zone[nIndex].nType == TYPE_OTHERS )	printf( "Wrong !!!!\n" );
		nPointer --;

		//There is a bug in the program. Using the following code to force the program to exit.
		nIter++;
		if( nIter > 2*nZone )
			break;

//Update the heap
		for( i=0; i<nZone; i++ )
		{
			bNeedUpdate = FALSE;
			if( (MRFModel.nCliqueType & CLIQUE_PRINT) != 0 )
			{
				if( LeftIndex[i] == nIndex || RightIndex[i] == nIndex )
					bNeedUpdate = TRUE;
			}
			if( (MRFModel.nCliqueType & CLIQUE_NOISE) != 0 )
			{
				for( j=0; j<nNear; j++ )
					if( NearIndex[j][i] == nIndex )
						bNeedUpdate = TRUE;
			}
			if( bNeedUpdate == FALSE )		continue;

			if( NeedProcess( zone[i] ) == FALSE )		continue;

			nMRFSizeLevel = GetMRFSizeLevel( zone[i].nSize );
			nMRFLabel = ZoneType2MRFLabel( zone[i].nType );
			LabelLeft = GetMRFLabel( zone, LeftIndex[i] );
			LabelRight = GetMRFLabel( zone, RightIndex[i] );
			for( j=0; j<nNear; j++ )
				NearLabel[j] = GetMRFLabel( zone, NearIndex[j][i] );
				
			GetEnergy( zone, nZone, i, LeftIndex, RightIndex, NearIndex, MRFModel, cpdf );

			//Already in the candidate list, remove it first
			for( j=0; j<=nPointer; j++ )
			{
				if( FlipCandidate[j].nIndex == i )
				{
					for( k=j; k<nPointer; k++ )
						FlipCandidate[k] = FlipCandidate[k+1];
					nPointer--;
					break;
				}
			}
			nMaxConf = 0;
			for( j=1; j<3; j++ )
				if( cpdf[j] > cpdf[nMaxConf] )
					nMaxConf = j;
			if( nMRFLabel != nMaxConf )
			{//Add it to the candidate list
				double	nConf = cpdf[nMaxConf] - cpdf[nMRFLabel];
				AddCandidate( FlipCandidate, nPointer, nConf, MRFModel.nClassLabel[nMaxConf], i );
			}
		}
	}
	free( FlipCandidate );
	return 0;
}
