////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		Ouput.cpp
// File Function:	Output segmentation, classification and final separated images
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "tools.h"
#include "Define.h"
#include "WordSegment.h"

/*******************************************************************************************
/*	Name:		OutputWordPos
/*	Function:	Ouput word segmentation and classification results
/*	Parameter:	fnWord      -- Output file name
/*				pWord       -- Segmented words
/*				nWord       -- Number of segmented words
/*				nClassLabel -- Class label
/*	Return:		0 -- Succeed
/********************************************************************************************/
int OutputWordPos( char *fnWord, ZONE *pWord, int nWord, int *nClassLabel, int nClassNum )
{
	FILE *fpWord = fopen( fnWord, "wt" );
	if( fpWord == NULL )
	{
		printf( "Can not open %s to write!", fnWord );
		return -1;
	}
	fprintf( fpWord, "Total Zones: %d\n", nWord );
	for( int i=0; i<nWord; i++ )
	{
		fprintf( fpWord, "%5d %5d %5d %5d    %d ", pWord[i].x, pWord[i].y, pWord[i].w, pWord[i].h, pWord[i].nType );

		fprintf( fpWord, "%9f ", pWord[i].nConf );
		for( int j=0; j<nClassNum; j++ )
			fprintf( fpWord, "%d: %9f ", nClassLabel[j], pWord[i].nClassConf[j] );
		fprintf( fpWord, "\n" );
	}
	fclose( fpWord);
	return 0;
}

/*******************************************************************************************
/*	Name:		OutputFeature
/*	Function:	Output extracted features to a file
/*	Parameter:	fnFeature  -- Output feature file name
/*				Feature    -- Extracted features
/*				nDimension -- Feature dimension
/*	Return:		0 -- Succeed
/********************************************************************************************/
int OutputFeature( char *fnFeature, double *Feature, int nDimension )
{
	FILE *fpFeature = fopen( fnFeature, "at" );
	for( int i=0; i<nDimension; i++ )
		//fprintf( fpFeature, "%9f ", Feature[i] );
		fprintf( fpFeature, "%9f\n", Feature[i] );
	fprintf( fpFeature, "\n" );
	fclose( fpFeature );
	return 0;
}

/**********************************************************************************************************
/*	Name:		ReadZone
/*	Function:	Read word segmentation and classification from file. Used for debug post processing module
/*	Parameter:	zone       -- Word blocks
/*				zone_count -- Number of word blocks
/*				fn         -- File name for word segmentation and classification results
/*	Return:		0 -- Succeed
/*********************************************************************************************************/
int ReadZone( ZONE *&zone, int &zone_count, char *fn )
{
	FILE *fp = fopen( fn, "rt" );
	if( fp == NULL )
	{
		printf( "Can not open %s to read!\n", fn );
		return -1;
	}
	
	char buf[1000];
	fgets( buf, 1000, fp );
	sscanf( buf, "Total Zones:%d", &zone_count );
	if( zone_count <= 0 )
	{
		printf( "%s: Zone number is less than 0!\n", fn );
		return -1;
	}

	zone = (ZONE*)malloc( sizeof( ZONE )*zone_count );

	if( zone == NULL )
	{
		printf( "Memory used up!\n" );
		return -1;
	}

	int		nClassNO[TYPE_TOTAL_CLASS];
	double	nClassConf[TYPE_TOTAL_CLASS];
	int i, j;
	for( i=0; i<zone_count; i++ )
	{
		zone[i].nConf = 0;
		for( j=0; j<TYPE_TOTAL_CLASS; j++ )	
		{
			nClassNO[j] = 0;
			nClassConf[j] = 0;
			zone[i].nClassConf[j] = 0;
		}

		if( feof(fp) )
		{
			printf( "Read file error!\n" );
			return -1;
		}
		fgets( buf, 1024, fp );
		sscanf( buf, "%d%d%d%d %d %lf %d:%lf %d:%lf %d:%lf\n", &zone[i].x, &zone[i].y, &zone[i].w, &zone[i].h, &zone[i].nType, &zone[i].nConf,
			&nClassNO[0], &nClassConf[0], &nClassNO[1], &nClassConf[1], &nClassNO[2], &nClassConf[2]);
		if( zone[i].x < 0 || zone[i].y < 0 || zone[i].w <= 0 || zone[i].h <= 0 )
		{
			printf( "Zone position error!\n" );
			return -1;
		}
		for( j=0; j<TYPE_TOTAL_CLASS; j++ )
		{
			if( nClassNO[j] == TYPE_PRINT_TEXT )
				zone[i].nClassConf[0] = nClassConf[j];
			if( nClassNO[j] == TYPE_HANDWRITING )
				zone[i].nClassConf[1] = nClassConf[j];
			if( nClassNO[j] == TYPE_NOISE )
				zone[i].nClassConf[2] = nClassConf[j];
		}

	}
	fclose( fp );
	return 0;
}

/**********************************************************************************************************
/*	Name:		SetZoneSize
/*	Function:	Set zone size
/*	Parameter:	zone  -- Word blocks
/*				nZone -- Number of word blocks
/*				cw	  -- Dominant character width
/*				ch    -- Dominant character height
/*	Return:		0 -- Succeed
/*********************************************************************************************************/
int	SetZoneSize( ZONE *zone, int nZone, int cw, int ch )
{
	for(int i=0; i<nZone; i++ )
	{
		if( zone[i].w <= 0.6*cw)
			zone[i].nSize = SIZE_TINY;
		else if( zone[i].w <= 3*cw )
			zone[i].nSize = SIZE_SMALL;
		else
			zone[i].nSize = SIZE_LARGE;
	}
	return 0;
}

/***********************************************************************************
/*	Name:		SetZoneConf
/*	Function:	Set zone classfication confidence for special cases
/*	Parameter:	zone      -- Zone
/*				nSize     -- zone size
/*				nClassNum -- Class number
/*	Return:		0 -- Succeed
/***********************************************************************************/
int	SetZoneConf( ZONE *zone, int nZone, int nClassNum )
{
	for(int i=0; i<nZone; i++ )
	{
		if( zone[i].nConf != 2 )		continue;
		if( zone[i].nSize == SIZE_TINY )
		{
			if( nClassNum == 3 )
			{
				zone[i].nConf = 0.7;
				zone[i].nClassConf[0] = 0.4;
				zone[i].nClassConf[1] = 0.4;
				zone[i].nClassConf[2] = 0.7;
			}
			else
			{
				zone[i].nConf = 0.7;
				zone[i].nClassConf[0] = 0.3;
				zone[i].nClassConf[1] = 0.0;
				zone[i].nClassConf[2] = 0.7;
			}
		}
		else
		{
			if( nClassNum == 3 )
			{
				zone[i].nConf = 0.95;
				zone[i].nClassConf[0] = 0.275;
				zone[i].nClassConf[1] = 0.275;
				zone[i].nClassConf[2] = 0.95;
			}
			else
			{
				zone[i].nConf = 0.95;
				zone[i].nClassConf[0] = 0.05;
				zone[i].nClassConf[1] = 0.00;
				zone[i].nClassConf[2] = 0.95;
			}
		}
	}
	return 0;
}