////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		PostProcess.cpp
// File Function:	MRF based post processing
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PostProcess.h"
#include "define.h"
#include "tools.h"
#include "math.h"

//Parameters for MRF model
double		power_print = 0.25;
double		power_noise = 0.48;
double		wp = 6;
double		wn = 7;
double		nMinConf = 0.02;

/*****************************************************************************
/*	Name:		SetMRFParamter
/*	Function:	Set parameters for MRF model
/*	Parameter:	zone     -- Zone
/*				nNewType -- New zone type flipped to
/*				FlipStat  -- Flip statistics
/*	Return:		0 -- Succeed
/*****************************************************************************/
int SetMRFParamter( int nClassNum )
{
	if( nClassNum == 3 )
	{
		power_print = 0.25;		//Power weight for Cp clique
		power_noise = 0.48;		//Power weight for Cn clique
		wp = 6;					//Weight for Cp clique
		wn = 7;					//Weight for Cn clique
		nMinConf = 0.05;		//Minimum classification confidence. 
								//Label will be forbidden to flip to the label with confidence less than this threshold
	}
	else
	{
		power_print = 1;
		power_noise = 0.25;
		wp = 3;			
		wn = 10;			
		nMinConf = 0;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		PostProcess
/*	Function:	Post process using MRF model
/*	Parameter:	SgZone	 -- Segmented blocks
/*				nSgZone  -- Number of segmented blocks
/*				MRFModel -- MRF model
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	PostProcess( ZONE *SgZone, int nSgZone, MRF_MODEL MRFModel )
{
	int i, nNear = 4;	
	int		*LeftIndex, *RightIndex;
	int		**NearIndex;
	LeftIndex = (int*)malloc( nSgZone*sizeof(int) );
	RightIndex = (int*)malloc( nSgZone*sizeof(int) );
	GetNearestZones4Print( SgZone, nSgZone, LeftIndex, RightIndex );

	NearIndex = (int**)malloc(sizeof(int*)*nNear );
	for( i=0; i<nNear; i++ )
		NearIndex[i] = (int*)malloc(sizeof(int) * nSgZone );

	GetNearestZones4Noise(SgZone, nSgZone, NearIndex );

//	ProcessTooHighZone( SgZone, nSgZone );
//	ProcessIsolatedZone( SgZone, nSgZone, LeftIndex, RightIndex, NearIndex );
	HCF( SgZone, nSgZone, LeftIndex, RightIndex, NearIndex, MRFModel );

	free( LeftIndex );
	free( RightIndex );
	for( i=0; i<nNear; i++ )
		free( NearIndex[i] );
	free( NearIndex );
	return 0;
}

/*****************************************************************************
/*	Name:		AllocateMRFModel
/*	Function:	Allocate memory for the MRF model
/*	Parameter:	MRFModel -- MRF model
/*	Return:		0 -- Succeed
/*****************************************************************************/
int AllocateMRFModel( MRF_MODEL &MRFModel )
{
	int i, j, k, l, m;
	MRFModel.nPatternP = (int****)malloc(sizeof(int***)*MRF_SIZE_LEVEL);
	for(i=0; i<MRF_SIZE_LEVEL; i++ )
	{
		MRFModel.nPatternP[i] = (int***)malloc(sizeof(int**)*TYPE_TOTAL_CLASS );
		for( j=0; j<TYPE_TOTAL_CLASS; j++ )
		{
			MRFModel.nPatternP[i][j] = (int**)malloc(sizeof(int*)*(TYPE_TOTAL_CLASS+1));
			for( k=0; k<TYPE_TOTAL_CLASS+1; k++ )
			{
				MRFModel.nPatternP[i][j][k] = (int*)malloc(sizeof(int)*(TYPE_TOTAL_CLASS+1));
				ZeroMemory( MRFModel.nPatternP[i][j][k], sizeof(int)*(TYPE_TOTAL_CLASS+1) );
			}
		}
	}

	MRFModel.nPatternN = (int******)malloc(sizeof(int*****)*MRF_SIZE_LEVEL);
	for(i=0; i<MRF_SIZE_LEVEL; i++ )
	{
		MRFModel.nPatternN[i] = (int*****)malloc(sizeof(int****)*TYPE_TOTAL_CLASS );
		for( j=0; j<TYPE_TOTAL_CLASS; j++ )
		{
			MRFModel.nPatternN[i][j] = (int****)malloc(sizeof(int***)*(TYPE_TOTAL_CLASS+1));
			for( k=0; k<TYPE_TOTAL_CLASS+1; k++ )
			{
				MRFModel.nPatternN[i][j][k] = (int***)malloc(sizeof(int**)*(TYPE_TOTAL_CLASS+1));
				for( l=0; l<TYPE_TOTAL_CLASS+1; l++ )
				{
					MRFModel.nPatternN[i][j][k][l] = (int**)malloc(sizeof(int*)*(TYPE_TOTAL_CLASS+1));
					for( m=0; m<TYPE_TOTAL_CLASS+1; m++ )
					{
						MRFModel.nPatternN[i][j][k][l][m] = (int*)malloc(sizeof(int)*(TYPE_TOTAL_CLASS+1));
						ZeroMemory( MRFModel.nPatternN[i][j][k][l][m], sizeof(int)*(TYPE_TOTAL_CLASS+1));
					}
				}
			}
		}
	}
	return 0;
}

/*****************************************************************************
/*	Name:		FreeMRFModel
/*	Function:	Free memory for the MRF model
/*	Parameter:	MRFModel -- MRF model
/*	Return:		0 -- Succeed
/*****************************************************************************/
int FreeMRFModel( MRF_MODEL &MRFModel )
{
	int i, j, k, l, m;
	if( MRFModel.nPatternP != NULL )
	{
		for(i=0; i<MRF_SIZE_LEVEL; i++ )
		{
			for( j=0; j<TYPE_TOTAL_CLASS; j++ )
			{
				for( k=0; k<TYPE_TOTAL_CLASS+1; k++ )
					free( MRFModel.nPatternP[i][j][k] );
				free( MRFModel.nPatternP[i][j] );
			}
			free( MRFModel.nPatternP[i] );
		}
		free( MRFModel.nPatternP );
	}

	if( MRFModel.nPatternN != NULL )
	{
		for(i=0; i<MRF_SIZE_LEVEL; i++ )
		{
			for( j=0; j<TYPE_TOTAL_CLASS; j++ )
			{
				for( k=0; k<TYPE_TOTAL_CLASS+1; k++ )
				{
					for( l=0; l<TYPE_TOTAL_CLASS+1; l++ )
					{
						for( m=0; m<TYPE_TOTAL_CLASS+1; m++ )
							free( MRFModel.nPatternN[i][j][k][l][m] );
						free( MRFModel.nPatternN[i][j][k][l] );
					}
					free( MRFModel.nPatternN[i][j][k] );
				}
				free( MRFModel.nPatternN[i][j] );
			}
			free( MRFModel.nPatternN[i] );
		}	
		free( MRFModel.nPatternN );
	}
	return 0;
}

/*****************************************************************************
/*	Name:		LoadMRFModel
/*	Function:	Load MRF model from file
/*	Parameter:	MRFModel    -- MRF model
/*				fnMRFModel  -- MRF model file name
/*				nCliqueType -- Clique type
/*	Return:		0 -- Succeed
/*****************************************************************************/
int LoadMRFModel( MRF_MODEL &MRFModel, FILE *fpMRFModel, int nCliqueType )
{
	int nMRFSizeLevel, nCount;
	int nMRFLabel, LabelLeft, LabelRight;
	int	NearLabel[4];
	int i, j;

	char	buf[1024], *p;
	while( !feof( fpMRFModel ) )
	{
		fgets( buf, 1024, fpMRFModel );
		if( nCliqueType == CLIQUE_PRINT && strstr( buf, "Clique Cp" ) != NULL )
			break;
		else if( nCliqueType == CLIQUE_NOISE && strstr( buf, "Clique Cn" ) != NULL )
			break;
	}

	int nTotalZone;
	fgets( buf, 1024, fpMRFModel );
	sscanf( buf, "Total %d zones:", &nTotalZone );
	fgets( buf, 1024, fpMRFModel );
	for( i=0; i<MRF_SIZE_LEVEL; i++ )
	{
		fgets( buf, 1024, fpMRFModel );
		sscanf( buf, "Size %d, Num=%d", &nMRFSizeLevel, &nCount );
		if( nMRFSizeLevel < 0 || nMRFSizeLevel >= MRF_SIZE_LEVEL )
		{
			printf( "Load MRF model failed! Size level error!\n" );
			return -1;
		}
		MRFModel.nZone[nMRFSizeLevel] = nCount;
		for( j=0; j<3; j++ )
		{
			fgets( buf, 1024, fpMRFModel );
			p = buf;
			while( *p == ' ' )	p++;
			sscanf( p, "Label %d:%d", &nMRFLabel, &nCount );
			if( nMRFLabel < 0 || nMRFLabel > 3 )
			{
				printf( "Load MRF model failed! Label error!\n" );
				return -1;
			}
			MRFModel.nLabel[nMRFSizeLevel][nMRFLabel] = nCount;
		}
	}

	fgets( buf, 1024, fpMRFModel );
	for( i=0; i<MRF_SIZE_LEVEL; i++ )
	{
		fgets( buf, 1024, fpMRFModel );
		fgets( buf, 1024, fpMRFModel );
		while( 1 )
		{
			nMRFLabel = -1;
			fgets( buf, 1024, fpMRFModel );
			if( buf[0] == '-' )	break;
			p = buf;
			while( *p == ' ' )	p++;
			if( nCliqueType == CLIQUE_PRINT )
			{
				LabelLeft = -1;
				LabelRight = -1;
				sscanf( p, "(%d,%d,%d) %d", &LabelLeft,  &nMRFLabel, &LabelRight, &nCount );
				if( nMRFLabel < 0 || nMRFLabel > 2 ||
					LabelLeft < 0 || LabelLeft > 3 ||
					LabelRight < 0 || LabelRight > 3 ||
					nCount < 0 )
				{
					printf( "File error!\n" );
					return -1;
				}
				MRFModel.nPatternP[i][nMRFLabel][LabelLeft][LabelRight] = nCount;
			}
			else if( nCliqueType == CLIQUE_NOISE )
			{
				NearLabel[0] = NearLabel[1] = NearLabel[2] = NearLabel[3] = -1;
				sscanf( p, "(%d,%d,%d,%d,%d) %d", &nMRFLabel, &NearLabel[0], &NearLabel[1], &NearLabel[2], &NearLabel[3], &nCount );
				if( nMRFLabel < 0 || nMRFLabel > 2 ||
					NearLabel[0] < 0 || NearLabel[0] >3 || NearLabel[1] < 0 || NearLabel[1] > 3 ||
					NearLabel[2] < 0 || NearLabel[2] > 3 || NearLabel[3] < 0 || NearLabel[3] > 3 ||
					nCount < 0 )
				{
					printf( "File error!\n" );
					return -1;
				}
				MRFModel.nPatternN[i][nMRFLabel][NearLabel[0]][NearLabel[1]][NearLabel[2]][NearLabel[3]] = nCount;
			}
		}
	}
	return 0;
}

/*****************************************************************************
/*	Name:		GetMRFLabel
/*	Function:	Convert from printed/handwritten/noise type to label type
/*				There are almost the same except type is from 1 to 4, and label
/*				is from 0 to 3. nMRFLabel for NULL block is 3.
/*	Parameter:	zone	 -- Blocks
/*				nCurZone -- Current zone
/*	Return:		nMRFLabel
/*****************************************************************************/
int GetMRFLabel( ZONE *zone, int nCurZone )
{
	if( nCurZone == -1 )
		return 3;
	else 
		return ZoneType2MRFLabel( zone[nCurZone].nType );
}

/*****************************************************************************
/*	Name:		ZoneType2MRFLabel
/*	Function:	Convert from printed/handwritten/noise type to label type
/*				There are almost the same except type is from 1 to 4, and label
/*				is from 0 to 3.
/*	Parameter:	nZoneType -- Zoen type
/*	Return:		nMRFLabel
/*****************************************************************************/
int ZoneType2MRFLabel( int nZoneType )
{
	int label;
	switch( nZoneType )
	{
	case TYPE_PRINT_TEXT:
		label = 0;
		break;
	case TYPE_HANDWRITING:
		label = 1;
		break;
	case TYPE_NOISE:
		label = 2;
		break;
	case TYPE_OTHERS:
		label = 2;
		break;
	default:
		printf( "Zone type %d error!\n", nZoneType );
		return -1;
	}
	return label;
}

/*****************************************************************************
/*	Name:		MRFLabel2ZoneType
/*	Function:	Convert from label to printed/handwritten/noise type.
/*				There are almost the same except type is from 1 to 4, and label
/*				is from 0 to 3.
/*	Parameter:	nLabel -- nMRFLabel
/*	Return:		Type
/*****************************************************************************/
int MRFLabel2ZoneType( int nLabel )
{
	int nZoneType;
	switch( nLabel )
	{
	case 0:
		nZoneType = TYPE_PRINT_TEXT;
		break;
	case 1:
		nZoneType = TYPE_HANDWRITING;
		break;
	case 2:
		nZoneType = TYPE_NOISE;
		break;
	default:
		printf( "Zone label error!\n" );
		return -1;
	}
	return nZoneType;
}

/*************************************************************************************
/*	Name:		GetMRFSizeLevel
/*	Function:	Convert block size to the size level of MRF model. Combining SIZE_TINY
/*				and SIZE_SMALL into one level.
/*	Parameter:	nSize -- Block size
/*	Return:		Size level in MRF model
/************************************************************************************/
int GetMRFSizeLevel( int nSize )
{
	int nMRFSizeLevel;
	switch( nSize )
	{
	case SIZE_TINY:
	case SIZE_SMALL:
		nMRFSizeLevel = 0;
		break;
	case SIZE_LARGE:
		nMRFSizeLevel = 1;
		break;
	default:
		printf( "Zone size error!\n" );
		return -1;
	}
	return nMRFSizeLevel;
}

/*****************************************************************************
/*	Name:		GetAPrioriProb
/*	Function:	Get a post priorio probability of each size level
/*	Parameter:	MRFModel -- MRF model
/*				nSize    -- Size level of MRF model
/*				nMRFLabel    -- Block label
/*	Return:		A post priori probability of each size level
/*****************************************************************************/
double	GetAPrioriProb( MRF_MODEL MRFModel, int nSize, int nMRFLabel )
{
	double Norm = 1;
	if( nMRFLabel != 3 )
		Norm *= 1.0*MRFModel.nLabel[nSize][nMRFLabel] / MRFModel.nZone[nSize];
	return Norm;
}

/*****************************************************************************
/*	Name:		GetEnergy
/*	Function:	Calculate clique potential for each block
/*	Parameter:	zone		-- Blocks
/*				nZone		-- Number of blocks
/*				nCurZone	-- Current block
/*				LeftIndex	-- Left neighboring block index for Cp clique
/*				RightIndex	-- Right neighboring block index for Cp clique
/*				NearIndex   -- Nearest neighboring block index for Cn clique
/*				MRFModel    -- MRF model
/*				cpdf        -- Conditional post priori probability
/*	Return:		0 -- Succeed
/*****************************************************************************/
int GetEnergy( ZONE *zone, int nZone, int nCurZone, int *LeftIndex, int *RightIndex, int **NearIndex, MRF_MODEL MRFModel, double	*cpdf )
{
	int		nMRFSizeLevel, nMRFLabel, LabelLeft, LabelRight;
	int		nNear = 4;
	int		NearLabel[4];
	int		i, k;

	nMRFSizeLevel = GetMRFSizeLevel( zone[nCurZone].nSize );

//Get identification confidence
	for( i=0; i<3; i++ )
	{
		if( zone[nCurZone].nClassConf[i] < nMinConf )
			cpdf[i] = -1e+10;
		else
			cpdf[i] = zone[nCurZone].nClassConf[i];
	}

	nMRFLabel = ZoneType2MRFLabel( zone[nCurZone].nType );
	
	double Norm = 1;
	double wcpdf = 0;

//Get clique potential for printex text clique
	if( ( MRFModel.nCliqueType & CLIQUE_PRINT ) != 0 )
	{
		LabelLeft = GetMRFLabel( zone, LeftIndex[nCurZone] );
		LabelRight = GetMRFLabel( zone, RightIndex[nCurZone] );
		if( LabelLeft != 3 || LabelRight != 3 )
		{
			Norm = 1;
			Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, LabelLeft );
			Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, LabelRight );

			for( i=0; i<3; i++ )
			{
				wcpdf = 1.0*MRFModel.nPatternP[nMRFSizeLevel][i][LabelLeft][LabelRight]/MRFModel.nZone[nMRFSizeLevel];
				if( MRFModel.nLabel[nMRFSizeLevel][i] == 0 )
					wcpdf = 0;
				else
					wcpdf /= pow( Norm * 1.0*MRFModel.nLabel[nMRFSizeLevel][i] / MRFModel.nZone[nMRFSizeLevel], power_print );
				cpdf[i] += wp * wcpdf;
			}

			for( k=0; k<nZone; k++ )
			{
				if( zone[k].nType == TYPE_OTHERS )		continue;
				if( LeftIndex[k] == nCurZone )
				{
					nMRFLabel = ZoneType2MRFLabel( zone[k].nType );
					LabelLeft = GetMRFLabel( zone, LeftIndex[k] );
					LabelRight = GetMRFLabel( zone, RightIndex[k] );
					nMRFSizeLevel = GetMRFSizeLevel( zone[k].nSize );

					Norm = 1;
					Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, nMRFLabel );
					Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, LabelRight );
					for( i=0; i<3; i++ )
					{
						wcpdf = 1.0*MRFModel.nPatternP[nMRFSizeLevel][nMRFLabel][i][LabelRight]/MRFModel.nZone[nMRFSizeLevel];
						if( MRFModel.nLabel[nMRFSizeLevel][i] == 0 )
							wcpdf = 0;
						else
							wcpdf /= pow(Norm * 1.0*MRFModel.nLabel[nMRFSizeLevel][i] / MRFModel.nZone[nMRFSizeLevel], power_print );
						cpdf[i] += wp * wcpdf;
					}
				}

				if( RightIndex[k] == nCurZone )
				{
					nMRFLabel = ZoneType2MRFLabel( zone[k].nType );
					LabelLeft = GetMRFLabel( zone, LeftIndex[k] );
					LabelRight = GetMRFLabel( zone, RightIndex[k] );
					nMRFSizeLevel = GetMRFSizeLevel( zone[k].nSize );

					Norm = 1;
					Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, nMRFLabel );
					Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, LabelLeft );

					for( i=0; i<3; i++ )
					{
						wcpdf = 1.0*MRFModel.nPatternP[nMRFSizeLevel][nMRFLabel][LabelLeft][i]/MRFModel.nZone[nMRFSizeLevel];
						if( MRFModel.nLabel[nMRFSizeLevel][i] == 0 )
							wcpdf = 0;
						else
							wcpdf /= pow(Norm * 1.0*MRFModel.nLabel[nMRFSizeLevel][i] / MRFModel.nZone[nMRFSizeLevel], power_print );
						cpdf[i] += wp * wcpdf;
					}
				}
			}
		}
	}
//Get clique potential for noise clique
	if( (MRFModel.nCliqueType & CLIQUE_NOISE ) != 0 )
	{
		Norm = 1;
		for( i=0; i<nNear; i++ )
		{
			NearLabel[i] = GetMRFLabel( zone, NearIndex[i][nCurZone] );
			Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, NearLabel[i] );
		}
		if( NearLabel[0] != 3 )
		{
			for( i=0; i<3; i++ )
			{
				wcpdf = 1.0*MRFModel.nPatternN[nMRFSizeLevel][i][NearLabel[0]][NearLabel[1]][NearLabel[2]][NearLabel[3]]/MRFModel.nZone[nMRFSizeLevel];
				if( MRFModel.nLabel[nMRFSizeLevel][i] == 0 )
					wcpdf = 0;
				else
					wcpdf /= pow( Norm * 1.0*MRFModel.nLabel[nMRFSizeLevel][i] / MRFModel.nZone[nMRFSizeLevel], power_noise );
				cpdf[i] += wn * wcpdf;
			}

			for( k=0; k<nZone; k++ )
			{
				if( zone[k].nType == TYPE_OTHERS )		continue;
				nMRFSizeLevel = GetMRFSizeLevel( zone[k].nSize );
				for( int nNeighbor=0; nNeighbor<nNear; nNeighbor++ )
				{
					if( NearIndex[nNeighbor][k] == nCurZone )
					{
						nMRFLabel = ZoneType2MRFLabel( zone[k].nType );
						Norm = GetAPrioriProb( MRFModel, nMRFSizeLevel, nMRFLabel );
						for( i=0; i<nNear; i++ )
						{
							NearLabel[i] = GetMRFLabel( zone, NearIndex[i][k] );
							if( i == nNeighbor )		continue;
							Norm *= GetAPrioriProb( MRFModel, nMRFSizeLevel, NearLabel[i] );
						}

						for( i=0; i<3; i++ )
						{
							switch( nNeighbor )
							{
							case 0:
								wcpdf = MRFModel.nPatternN[nMRFSizeLevel][nMRFLabel][i][NearLabel[1]][NearLabel[2]][NearLabel[3]];
								break;
							case 1:
								wcpdf = MRFModel.nPatternN[nMRFSizeLevel][nMRFLabel][NearLabel[0]][i][NearLabel[2]][NearLabel[3]];
								break;
							case 2:
								wcpdf = MRFModel.nPatternN[nMRFSizeLevel][nMRFLabel][NearLabel[0]][NearLabel[1]][i][NearLabel[3]];
								break;
							case 3:
								wcpdf = MRFModel.nPatternN[nMRFSizeLevel][nMRFLabel][NearLabel[0]][NearLabel[1]][NearLabel[2]][i];
								break;
							}
							wcpdf *= 1.0/MRFModel.nZone[nMRFSizeLevel];
							if( MRFModel.nLabel[nMRFSizeLevel][i] == 0 )
								wcpdf = 0;
							else
								wcpdf /= pow(Norm * 1.0*MRFModel.nLabel[nMRFSizeLevel][i] / MRFModel.nZone[nMRFSizeLevel], power_noise );
							cpdf[i] += wn * wcpdf;
						}
					}
				}
			}
		}
	}
	return 0;
}

/*****************************************************************************
/*	Name:		NeedProcess
/*	Function:	Need process current block
/*	Parameter:	zone		-- Block
/*	Return:		TRUE  -- Need process
/*				FALSE -- Do not need process
/*****************************************************************************/
BOOL NeedProcess( ZONE &zone )
{
	if( zone.nSize == SIZE_TINY )
		return TRUE;
	switch( zone.nType )
	{
	case TYPE_PRINT_TEXT:
		if( zone.nSize == SIZE_SMALL && zone.nConf >= 0.970 )	
			return FALSE;
		if( zone.nSize == SIZE_LARGE && zone.nConf >= 0.980 )
			return FALSE;
		break;
	case TYPE_HANDWRITING:
		if( zone.nConf >= 0.99 )	
			return FALSE;
		if( zone.nConf < 0.75 )
		{
			zone.nType = TYPE_PRINT_TEXT;
			zone.nConf = 1.00;
			zone.nSize = SIZE_LARGE;
		}
		break;
	case TYPE_NOISE:
		if( zone.nConf >= 0.999 )
			return FALSE;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


