////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		TrainMRFModel.cpp
// File Function:	API for MRF model training
//
//				Developed by: Yefeng Zheng
//			   First created: Nov. 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TextIDAPI.h"
#include "Tools.h"
#include "PostProcess.h"
#include "GetStat.h"

CTITrainMRFModel::CTITrainMRFModel()
{
	strcpy( m_fnTrainMRFConfig, "" );
	m_MRF_Model.nPatternP = NULL;
	m_MRF_Model.nPatternN = NULL;
	return;
}

CTITrainMRFModel::~CTITrainMRFModel()
{
	FreeMRFModel( m_MRF_Model );
	return;
}

/******************************************************************************
/*	Name:		TISetTrainConfig
/*	Function:	Set configure file for the training set
/*	Parameter:	fnTrainConfig -- Configure file
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTITrainMRFModel::TISetTrainConfig( char *fnTrainConfig )
{
	strcpy( m_fnTrainMRFConfig, fnTrainConfig );
	return 0;
}

/******************************************************************************
/*	Name:		TITrainMRFModel
/*	Function:	Train MRF model
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTITrainMRFModel::TITrainMRFModel( )
{
	AllocateMRFModel( m_MRF_Model );
	m_MRF_Model.nClassNum = 3;
	InitStatMeasure( m_MRF_Model );

	char	buf[1000];
	CString cs;
	char	csGtDir[MAX_PATH];
	char	csGtSuffix[MAX_PATH];

	FILE	*fpTrain = fopen( m_fnTrainMRFConfig, "rt" );
	if( fpTrain == NULL )
	{
		printf( "Can not open %s to read!\n", m_fnTrainMRFConfig );
		return -1;
	}
	while( !feof( fpTrain ) )
	{
		fgets( buf, 1000, fpTrain );
		cs = CString( buf );
		cs.MakeUpper();
		if( cs.Find( "GROUNDTRUTH_DIR" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csGtDir, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "GT_SUFFIX" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( csGtSuffix, "%s", (LPCSTR)cs );
		}
	}
	rewind( fpTrain );
	while( !feof( fpTrain ) )
	{
		fgets( buf, 1000, fpTrain );
		if( strstr( buf, "---" ) != NULL )
			break;
	}

	char	fnTruth[MAX_PATH];
	while( !feof(fpTrain) )
	{ 
		if( fgets( buf, 1000, fpTrain ) == NULL )
			break;
		if( strcmp( buf, "\n" ) == 0 )
			break;
		if( buf[strlen(buf)-1] == '\n' )
			buf[strlen(buf)-1] = '\0';
		sprintf( fnTruth, "%s\\%s%s", csGtDir, buf, csGtSuffix );
		GetStatistics( fnTruth, m_MRF_Model, CLIQUE_PRINT | CLIQUE_NOISE );
	}
	fclose( fpTrain );
	return 0;
}

/******************************************************************************
/*	Name:		TIOutputMRFModel
/*	Function:	Output the trained MRF model
/*	Parameter:	fnModel -- File name for the trained MRF model
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTITrainMRFModel::TIOutputMRFModel( char *fnModel )
{
	int i;
	int	nSize, Label, LabelLeft, LabelRight, LabelTop, LabelBottom;
	FILE *fpStat;
	fpStat = fopen( fnModel, "wt" );
	if( fpStat == NULL )
	{
		printf( "Can not open %s to write!\n" );
		return -1;
	}

	int nTotalZone = 0;
	for( i=0; i<MRF_SIZE_LEVEL; i++ )
		nTotalZone += m_MRF_Model.nZone[i];

	fprintf( fpStat, "Clique Cp\n" );
	fprintf( fpStat, "Total %d zones:\n", nTotalZone );
	fprintf( fpStat, "------------------------------------------------------\n" );
	for( nSize = 0; nSize<MRF_SIZE_LEVEL; nSize++ )
	{
		fprintf( fpStat, "Size %d, Num=%d (%6.4f)\n", nSize, m_MRF_Model.nZone[nSize], 1.0*m_MRF_Model.nZone[nSize]/nTotalZone );
		for( Label=0; Label<3; Label++ )
			fprintf( fpStat, "   Label %d:%d (%6.4f)\n", Label, m_MRF_Model.nLabel[nSize][Label], 1.0*m_MRF_Model.nLabel[nSize][Label]/m_MRF_Model.nZone[nSize] );
	}
	fprintf( fpStat, "------------------------------------------------------\n" );

	for( nSize = 0; nSize<MRF_SIZE_LEVEL; nSize++ )
	{
		fprintf( fpStat, "Size %d, Num=%d (%6.4f)\n", nSize, m_MRF_Model.nZone[nSize], 1.0*m_MRF_Model.nZone[nSize]/nTotalZone );
		fprintf( fpStat, "   (left, center, right)\n" );
		for( LabelLeft = 0; LabelLeft<4; LabelLeft++ )
		{
			for( LabelRight = 0; LabelRight<4; LabelRight++ )
				for( Label = 0; Label<3; Label++ )
					fprintf( fpStat, "    (%d, %d, %d)  %d\n", LabelLeft, Label, LabelRight, m_MRF_Model.nPatternP[nSize][Label][LabelLeft][LabelRight] );
		}
		fprintf( fpStat, "-------------------------------------------------------\n" );
	}

/////////////////////////////////////////////////////////////////////////////////////
	fprintf( fpStat, "\n\n\nClique Cn\n" );
	fprintf( fpStat, "Total %d zones:\n", nTotalZone );
	fprintf( fpStat, "------------------------------------------------------\n" );
	for( nSize = 0; nSize<MRF_SIZE_LEVEL; nSize++ )
	{
		fprintf( fpStat, "Size %d, Num=%d (%6.4f)\n", nSize, m_MRF_Model.nZone[nSize], 1.0*m_MRF_Model.nZone[nSize]/nTotalZone );
		for( Label=0; Label<3; Label++ )
			fprintf( fpStat, "   Label %d:%d (%6.4f)\n", Label, m_MRF_Model.nLabel[nSize][Label], 1.0*m_MRF_Model.nLabel[nSize][Label]/m_MRF_Model.nZone[nSize] );
	}
	fprintf( fpStat, "------------------------------------------------------\n" );
	for( nSize = 0; nSize<MRF_SIZE_LEVEL; nSize++ )
	{
		fprintf( fpStat, "Size %d, Num=%d (%6.4f)\n", nSize, m_MRF_Model.nZone[nSize], 1.0*m_MRF_Model.nZone[nSize]/nTotalZone );
		fprintf( fpStat, "   (center, N1, N2, N3, N4)\n" );
		for( Label = 0; Label<3; Label++ )
		{
			for( LabelLeft = 0; LabelLeft<4; LabelLeft++ )
			{
				for( LabelRight = 0; LabelRight<4; LabelRight++ )
					for( LabelTop = 0; LabelTop<4; LabelTop++ )
						for( LabelBottom=0; LabelBottom<4; LabelBottom++ )
								fprintf( fpStat, "    (%d, %d, %d, %d, %d)  %d\n",  Label, LabelLeft, LabelRight, LabelTop, LabelBottom,
									m_MRF_Model.nPatternN[nSize][Label][LabelLeft][LabelRight][LabelTop][LabelBottom] );
			}
		}
		fprintf( fpStat, "-------------------------------------------------------\n" );
	}
	fclose( fpStat );
	return 0;
}