////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		FisherClassify.cpp
// File Function:	Fisher classifer
//
//				Developed by: Yefeng Zheng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "define.h"
#include "tools.h"
#include <math.h>

#define	MAX_LINE_LEN	1024
char	buf[MAX_LINE_LEN];

/**************************************************************************************************
/*	Name:		InitFisherModel
/*	Function:	Initialize Fisher model
/*	Parameter:	model -- Fisher model
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int InitFisherModel( CLASSIFIER &model )
{
	model.nClassLabel = NULL;
	model.Eff		  = NULL;
	model.nMean1	  = NULL;
	model.nVar1		  = NULL;
	model.nMean2	  = NULL;
	model.nVar2		  = NULL;
	model.nClassLabel1= NULL;
	model.nClassLabel2= NULL;
	model.nTH		  = NULL;
	model.nDimension  = 0;
	model.nClassNum	  = 0;
	return 0;
}

/**************************************************************************************************
/*	Name:		FreeFisherModel
/*	Function:	Free Fisher model
/*	Parameter:	model -- Fisher model
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int	FreeFisherModel( CLASSIFIER &model )
{
	if( model.nClassLabel != NULL )
		free( model.nClassLabel );
	if( model.nClassLabel1 != NULL )
		free( model.nClassLabel1 );
	if( model.nClassLabel2 != NULL )
		free( model.nClassLabel2 );
	if( model.nMean1 != NULL )
		free( model.nMean1 );
	if( model.nMean2 != NULL )
		free( model.nMean2 );
	if( model.nTH != NULL )
		free( model.nTH );
	if( model.nVar1 != NULL )
		free( model.nVar1 );
	if( model.nVar2 != NULL )
		free( model.nVar2 );
	if( model.Eff != NULL )
	{
		for( int i=0; i<model.nClassNum*(model.nClassNum-1)/2; i++ )
			free( model.Eff[i] );
		free( model.Eff );
	}
	return 0;
}

/**************************************************************************************************
/*	Name:		LoadFisherModel
/*	Function:	Load Fisher model
/*	Parameter:	fnModel -- Fisher model file name
/*				model   -- Fisher model
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int LoadFisherModel( char *fnModel, CLASSIFIER &model )
{
	FILE *fpModel = fopen( fnModel, "rt" );
	if( fpModel == NULL )
	{
		fprintf( stderr, "can not open %s!\n", fnModel );
		return -1 ;
	}
//Get class number
	fgets( buf, MAX_LINE_LEN, fpModel );
	int	nClassNum = 2;
	int nClassifierType = 0;
	sscanf( buf, "Type=%d Class=%d\n", &nClassifierType, &nClassNum );
	if( nClassNum == 2 && nClassifierType == 0 )//Two classes classifier
		rewind( fpModel );
	else if( nClassifierType != 4 )
	{
		printf( "Type error!\n" );
		return -1;
	}
	if( nClassNum*(nClassNum-1)/2 >= 10 )
	{
		printf( "Too many classes!\n" );
		return -1;
	}
//Get Fisher coefficients dimension
	double	value;
	fgets( buf, MAX_LINE_LEN, fpModel );
	int	nDimension = 0;
	while( fscanf( fpModel, "%lf", &value ) == 1 )
		nDimension++;
	rewind( fpModel );
	if( nClassifierType == 4 )
		fgets( buf, MAX_LINE_LEN, fpModel );

//Read Fisher coefficients
	model.nClassNum = nClassNum;
	model.nDimension = nDimension;
	model.Eff = (double**)malloc( sizeof(double*)*nClassNum*(nClassNum-1)/2 );
	model.nTH = (double*)malloc( sizeof(double)*nClassNum*(nClassNum-1)/2 );
	model.nMean1 = (double*)malloc( sizeof(double)*nClassNum*(nClassNum-1)/2 );
	model.nVar1  = (double*)malloc( sizeof(double)*nClassNum*(nClassNum-1)/2 );
	model.nMean2 = (double*)malloc( sizeof(double)*nClassNum*(nClassNum-1)/2 );
	model.nVar2  = (double*)malloc( sizeof(double)*nClassNum*(nClassNum-1)/2 );
	model.nClassLabel1 = (int*)malloc( sizeof(int)*nClassNum*(nClassNum-1)/2 );
	model.nClassLabel2 = (int*)malloc( sizeof(int)*nClassNum*(nClassNum-1)/2 );
	model.nClassLabel = (int*)malloc( sizeof(int)*TYPE_TOTAL_CLASS );
	
	for( int i=0; i<TYPE_TOTAL_CLASS; i++ )
		model.nClassLabel[i] = i+1;

	int nClassNum2 = 0;
	int	nDimension2 = 0;
	for( i=0; i<nClassNum*(nClassNum-1)/2; i++ )
	{
		model.Eff[i] = (double*)malloc( nDimension*sizeof(double));
		if( model.Eff[i] == NULL )
		{
			fprintf(stderr, "Memory used up!\n" );
			return -1;
		}
		ZeroMemory( model.Eff[i], nDimension*sizeof( double ));
		nDimension2 = 0;

		fgets(buf, MAX_LINE_LEN, fpModel );
		while( fscanf( fpModel, "%lf", &value ) == 1 )
		{
			nDimension2++;
			if( nDimension2 > nDimension )				break;
			model.Eff[i][nDimension2-1] = value; 
		}
		if( nDimension2 != nDimension )
		{
  			printf( "Dimensions of feature file and fisher model do no agree!\n" );
			return -1;
		}
		fscanf( fpModel, "\n" );

//Read mean and variacne
		fscanf( fpModel, "TH=%lf\n", &model.nTH[i] );
		model.nMean1[i] = 0.0;
		model.nMean2[i] = 0.0;
		fscanf( fpModel, "Class1=%d Mean1=%lf Delta1=%lf\n", &model.nClassLabel1[i], &model.nMean1[i], &model.nVar1[i] );
		fscanf( fpModel, "Class2=%d Mean2=%lf Delta2=%lf\n", &model.nClassLabel2[i], &model.nMean2[i], &model.nVar2[i] );
		if( model.nVar1[i] < 0 || model.nVar2[i] < 0 )
		{
			fprintf( stderr, "Delta1<0 or Delta2<0   ERROR!!!\n" );
			return -1;
		}
	}
	fclose( fpModel );
	return 0;
}

/**************************************************************************************************
/*	Name:		FisherClassify
/*	Function:	Classification using Fisher classifier
/*	Parameter:	Feature -- Features
/*				word    -- Classification result are stored in this variable
/*				model   -- Fisher model
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int FisherClassify( double *Feature, WORD_ZONE &word, CLASSIFIER model )
{
	int		nWin[100];
	int		nClassNum = model.nClassNum;
	int		Dimension = model.nDimension;
	int		i, j, k;
	for( k=0; k<TYPE_TOTAL_CLASS; k++ )
	{
		word.nClassConf[k] = 0;
		nWin[k] = 0;
	}
	for( k=0; k<nClassNum*(nClassNum-1)/2; k++ )
	{
		double	y = 0;
		for( i=0; i<Dimension; i++ )
			y += Feature[i]*model.Eff[k][i];
		double	Conf1 = 1/model.nVar1[k]*exp( -0.5*(y-model.nMean1[k])*(y-model.nMean1[k])/model.nVar1[k]/model.nVar1[k] ) ;
		double	Conf2 = 1/model.nVar2[k]*exp( -0.5*(y-model.nMean2[k])*(y-model.nMean2[k])/model.nVar2[k]/model.nVar2[k] );
		if( Conf1 == 0 && Conf2 == 0 )
			Conf1 = Conf2 = 1;
		for( j=0; j<TYPE_TOTAL_CLASS; j++ )
			if( model.nClassLabel[j] == model.nClassLabel1[k] )
				break;
		if( j>= TYPE_TOTAL_CLASS )
		{
			printf( "Fisher classification error!\n" );
			return -1;
		}
		word.nClassConf[j] += Conf1/(Conf1+Conf2);
		if( Conf1 >= Conf2 )
			nWin[j] ++;

		for( j=0; j<TYPE_TOTAL_CLASS; j++ )
			if( model.nClassLabel[j] == model.nClassLabel2[k] )
				break;
		if( j>= TYPE_TOTAL_CLASS )
		{
			printf( "Fisher classification error!\n" );
			return -1;
		}
		word.nClassConf[j] += Conf2/(Conf1+Conf2);
		if( Conf2 > Conf1 )
			nWin[j] ++;
	}
	for( i=0; i<TYPE_TOTAL_CLASS; i++ )
		word.nClassConf[i] /= (nClassNum-1);

	int	nMaxWin = 0;
	for( i=0; i<TYPE_TOTAL_CLASS; i++ )
	{
		if( nWin[i] > nWin[nMaxWin] )
			nMaxWin = i;
	}
		
	int nMaxConf = nMaxWin;
	for( i=0; i<TYPE_TOTAL_CLASS; i++ )
	{
		if( nWin[i] == nWin[nMaxWin] )
		{		
			if( word.nClassConf[i] > word.nClassConf[nMaxConf] )
				nMaxConf = i;
		}
	}
	
	word.nType = model.nClassLabel[nMaxConf];
	word.nConf = word.nClassConf[nMaxConf];
	return 0;
}
