////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		ExtractGaborFeature.cpp
// File Function:	Extract Gabor features
//
//				Developed by: Yefeng Zheng
//		  The original copy is developed Xuewen Wang
//			   First created: Jan, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "math.h"
#include "imageobj.h"
#include "Gabor.h"
#include "imagetool.h"
#include "ExtractGaborFeature.h"

CGaborTrans   Gabor;
float	fWaveLength = 15;
float	pOrient[16];
int		nOrient = 16;
float	delta_x = (float)0.02;
float	delta_y = (float)0.05;
int		numFreq;
float	*pFreqReal = NULL, *pFreqImg = NULL;
BYTE	*ByteImg = NULL;
int		nImgSize = 64;
BYTE	ImgBlock[64*64];

/*****************************************************************************
/*	Name:		InitGabor
/*	Function:	Initialize 16 Gabor filters
/*	Parameter:	NULL
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	InitGabor( )
{
	for( int i=0; i<16; i++ )
		pOrient[i] = (float)(i*3.1415926/16);
	ByteImg = (BYTE*)malloc( nImgSize*nImgSize );
	fWaveLength = fWaveLength*nImgSize/64;
	delta_x = delta_x*64/nImgSize;
	delta_y = delta_y*64/nImgSize;
	
	Gabor.Init( nImgSize, fWaveLength, nOrient, pOrient, delta_x, delta_y, numFreq );
	pFreqReal = (float*)malloc( sizeof(float)*numFreq );
	pFreqImg = (float*)malloc( sizeof(float)*numFreq );
	return 0;
}

/*****************************************************************************
/*	Name:		FreeGabor
/*	Function:	Free memory used by Gabor filters
/*	Parameter:	NULL
/*	Return:		0 -- Succeed
/*****************************************************************************/
int FreeGabor( )
{
	Gabor.Free( );
	if( ByteImg != NULL )
		free( ByteImg );
	if( pFreqReal != NULL )
		free( pFreqReal );
	if( pFreqImg != NULL )
		free( pFreqImg );
	ByteImg = NULL;
	pFreqReal = NULL;
	pFreqImg = NULL;
	return 0;
}

/*****************************************************************************
/*	Name:		ExtractGaborFeature
/*	Function:	Extract Gabor features
/*	Parameter:	Img       -- Image
/*				fpFeature -- Feature file pointer
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	ExtractGaborFeature( CImage &Img, double *nAverVar )
{	
	int j;
	CRect	BoundBox;
	CImage  ImgBlock;
	float	nMean[16], nVar[16], nAverMean[16];
	int		nBlock;
		
	for( j=0; j<nOrient; j++ )
	{
		nAverMean[j] = 0;
		nAverVar[j] = 0;
	}

	nBlock = 0;
	int nTotalBlackPixels = 0;
	int	w = Img.GetWidth();
	int h = Img.GetHeight();
	for( int x1=0; x1<w; x1+=nImgSize/2 )
	{
		for( int y1=0; y1<h; y1+=nImgSize/2 )
		{
			int w1 = min( nImgSize, w-x1 );
			int h1 = min( nImgSize, h-y1 );
			CImageTool::CopyImage( Img, x1, y1, w1, h1, ImgBlock );
			int nBlackPixels = BitToByte( ImgBlock.LockRawImg(), ImgBlock.GetLineSizeInBytes(), w1, h1, ByteImg, nImgSize, nImgSize, BoundBox );
			ImgBlock.UnlockRawImg();
			
			Gabor.GaborTrans( ByteImg, pFreqReal, pFreqImg );
			GetGaborMeanVar( pFreqReal, nImgSize, nOrient, w1, h1, nMean, nVar );
			for( j=0; j<nOrient; j++ )
			{
				nAverMean[j] += nMean[j];
				nAverVar[j] += nVar[j]*nBlackPixels;
			}
			nBlock++;
			nTotalBlackPixels += nBlackPixels;
		}
	}

	for( j=0; j<nOrient; j++ )
	{
		nAverMean[j] /= nBlock;
		nAverVar[j] /= nTotalBlackPixels;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		ExtractGaborFeature
/*	Function:	Extract Gabor features
/*	Parameter:	Img       -- Image
/*				fpFeature -- Feature file pointer
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	ExtractGaborFeature( BYTE *ByteImg, int w, int h, double *nAverVar )
{	
	int j;
	float	nMean[16], nVar[16], nAverMean[16];
	int		nBlock;
		
	for( j=0; j<nOrient; j++ )
	{
		nAverMean[j] = 0;
		nAverVar[j] = 0;
	}

	nBlock = 0;
	int		nTotalBlackPixels = 0;
	int		nBlackPixels;
	for( int x1=0; x1<w; x1+=nImgSize/2 )
	{
		for( int y1=0; y1<h; y1+=nImgSize/2 )
		{
			int w1 = min( nImgSize, w-x1 );
			int h1 = min( nImgSize, h-y1 );
			
			ZeroMemory( ImgBlock, 64*64*sizeof(BYTE) );
			nBlackPixels = 0;
			for(int m=0; m<h1; m++ )
			{
				for(int n=0; n<w1; n++ )
				{
					ImgBlock[m*nImgSize+n] = ByteImg[(x1+m)*w+(y1+n)];
					if( ImgBlock[m*nImgSize+n] == 1 )
						nBlackPixels++;
				}
			}
			if( nBlackPixels < 5 )
			{
				for( j=0; j<nOrient; j++ )
				{
					nMean[j] = 0;
					nVar[j] = 0;
				}
			}
			else
			{
				Gabor.GaborTrans( ByteImg, pFreqReal, pFreqImg );
				GetGaborMeanVar( pFreqReal, nImgSize, nOrient, w1, h1, nMean, nVar );
			}
			for( j=0; j<nOrient; j++ )
			{
				nAverMean[j] += nMean[j];
				nAverVar[j] += nVar[j]*nBlackPixels;
			}
			nBlock++;
			nTotalBlackPixels += nBlackPixels;
		}
	}

	for( j=0; j<nOrient; j++ )
	{
		nAverMean[j] /= nBlock;
		nAverVar[j] /= nTotalBlackPixels;
	}
	return 0;
}

/***********************************************************************************
/*	Name:		GetGaborMeanVar
/*	Function:	Calculate the mean and variance of the image after Gabor filtering
/*	Parameter:	pFreqReal  -- Filtered images, there are nOrient images
/*				nImgSize   -- Filtered image size is nImgSize*nImgSize
/*				nOrient    -- Number of Gabor filters with different orientations
/*				w          -- Original image width
/*				h          -- Original image height
/*				Mean       -- Mean of each filtered image
/*				Var        -- Variance of each filtered image
/*	Return:		0 -- Succeed
/***********************************************************************************/
int	GetGaborMeanVar( float* pFreqReal, int nImgSize, int nOrient, int w, int h, float *nMean, float *nVar )
{
	int i, j, k;
	float *pGabor;
	for( i=0; i<nOrient; i++ )
	{
		pGabor = &pFreqReal[i*(nImgSize*nImgSize)];
		nMean[i] = 0;
		nVar[i] = 0;
		for( j=0; j<h; j++ )
			for( k=0; k<w; k++ )
				nMean[i] += pGabor[j*nImgSize+k];
		nMean[i] /= (w*h);

		for( j=0; j<h; j++ )
			for( k=0; k<w; k++ )
				nVar[i] += (pGabor[j*nImgSize+k]-nMean[i]) * (pGabor[j*nImgSize+k]-nMean[i]);
		nVar[i] = (float)sqrt( nVar[i]/(w*h) );
	}
	return 0;
}