////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		FisherTrain4.cpp
// File Function:	Fisher classifier training for multi-class case
//
//				Developed by: Zheng Yefeng
//			   First created: Jan, 2001
//			   Last modified: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "fishertrain.h"
#include "define.h"
#include "Tools.h"
#include "Output.h"
#include <math.h>
#include "ImageObj.h"
#include "ImageTool.h"
#include "ExtractFeature.h"

#define SKIP_CLASS\
	while(isspace(*p)) ++p;\
	while(!isspace(*p)) ++p;\
	while(isspace(*p)) ++p;

#define SKIP_ELEMENT\
	while(*p!=':') ++p;\
	++p;\
	while(isspace(*p)) ++p;\
	while(*p && !isspace(*p)) ++p;

/*****************************************************************************
/*	Name:		AllocateClassifier
/*	Function:	Allocate memeory for the Fisher classifier
/*	Parameter:	Classifier    -- Classifier
/*				nTotalClass   -- Number of class categories
/*				nDimension	  -- Dimension of the feature set
/*	Return:		0 -- Succeed
/*****************************************************************************/
int AllocateClassifier( CLASSIFIER &Classifier, int nTotalClass, int nDimension )
{
	Classifier.nClassNum = nTotalClass;
	Classifier.nDimension = nDimension;
	Classifier.Eff = (double**)malloc( sizeof(double*)*nTotalClass*(nTotalClass-1)/2 );
	for( int i=0; i<nTotalClass*(nTotalClass-1)/2; i++ )
		Classifier.Eff[i] = (double*)malloc( sizeof(double)*nDimension);

	Classifier.nTH = (double*)malloc( sizeof(double)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nMean1 = (double*)malloc( sizeof(double)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nVar1  = (double*)malloc( sizeof(double)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nMean2 = (double*)malloc( sizeof(double)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nVar2  = (double*)malloc( sizeof(double)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nClassLabel1 = (int*)malloc( sizeof(int)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nClassLabel2 = (int*)malloc( sizeof(int)*nTotalClass*(nTotalClass-1)/2 );
	Classifier.nClassLabel = (int*)malloc( sizeof(int)*TYPE_TOTAL_CLASS );
	Classifier.nClassLabel[0] = TYPE_PRINT_TEXT;
	Classifier.nClassLabel[1] = TYPE_HANDWRITING;
	Classifier.nClassLabel[2] = TYPE_NOISE;
	return 0;
}

/*****************************************************************************
/*	Name:		ParseTrainConfig
/*	Function:	Parse the configure of the training set
/*	Parameter:	fnTrain		-- Configure file
/*				TrainConfig	-- Configure after parsing
/*	Return:		0 -- Succeed
/*****************************************************************************/
int ParseTrainConfig( char *fnTrain, TRAIN_CONFIG &TrainConfig )
{
	int	nClass = 0;
	for( int i=0; i<TYPE_TOTAL_CLASS; i++ )
		TrainConfig.bTrain[i] = 0;
	strcpy( TrainConfig.csGtDir, "" );
	strcpy( TrainConfig.csGtSuffix, "" );
	strcpy( TrainConfig.csImgDir, "" );
	strcpy( TrainConfig.csImgSuffix, "" );

	FILE *fpTrain = fopen( fnTrain, "rt" );
	if( fpTrain == NULL )
	{
		printf( "Can not open %s file to read training configure!\n" );
		return -1;
	}

	char buf[1000];
	CString cs;
	while( !feof( fpTrain ) )
	{
		fgets( buf, 1000, fpTrain );
		cs = CString( buf );
		cs.MakeUpper();
		if( cs.Find( "GROUNDTRUTH_DIR" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( TrainConfig.csGtDir, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "IMG_DIR" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( TrainConfig.csImgDir, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "GT_SUFFIX" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( TrainConfig.csGtSuffix, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "IMG_SUFFIX" ) != -1 )
		{
			if( GetConfigInfo( cs ) == -1 )
				return -1;
			sprintf( TrainConfig.csImgSuffix, "%s", (LPCSTR)cs );
		}
		if( cs.Find( "CLASSES" ) != -1 )
		{
			if( cs.Find( "PRINT_TEXT" ) != -1 )
			{
				TrainConfig.bTrain[0] = 1;
				nClass ++;
			}
			if( cs.Find( "HANDWRITING" ) != -1 )
			{
				TrainConfig.bTrain[1] = 1;
				nClass++;
			}
			if( cs.Find( "NOISE" ) != -1 )
			{
				TrainConfig.bTrain[2] = 1;
				nClass++;
			}
		}
	}
	if( nClass <= 2 )
	{
		printf( "ERROR: The number of classes need to distinguish is less than 2!\n");
		return -1;
	}
	fclose( fpTrain );
	return 0;
}

/*****************************************************************************
/*	Name:		GetSampleNum
/*	Function:	Get the number of training samples for each class
/*	Parameter:	fnTrain		-- Configure file
/*				TrainConfig	-- Configure after parsing
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	GetSampleNum( char *fnTrain, TRAIN_CONFIG &TrainConfig )
{
	for( int i=0; i<TYPE_TOTAL_CLASS; i++ )
		TrainConfig.NumClass[i] = 0;

	FILE *fpTrain = fopen( fnTrain, "rt" );
	if( fpTrain == NULL )
		return -1;

	char buf[1000];
	//Find the beginning of the training list
	while( !feof(fpTrain) )
	{
		if( fgets( buf, 1000, fpTrain ) == NULL)
			break;
		if( strstr( buf, "---" ) != NULL )
			break;
	}

	//Read all groundtruth
	char	fnGt[MAX_PATH];
	ZONE	*pZone;
	int		nZone;
	while( !feof(fpTrain) )
	{
		if( fgets( buf, 1000, fpTrain) == NULL )
			break;
		if( strcmp( buf, "\n" ) == 0 )
			break;
		if( buf[strlen(buf)-1] == '\n' )
			buf[strlen(buf)-1] = '\0';
		sprintf( fnGt, "%s\\%s%s", TrainConfig.csGtDir, buf, TrainConfig.csGtSuffix );
		ReadZone( pZone, nZone, fnGt );
		for( i=0; i<nZone; i++ )
		{
			switch( pZone[i].nType )
			{
			case TYPE_PRINT_TEXT:
				TrainConfig.NumClass[0]++;
				break;
			case TYPE_HANDWRITING:
				TrainConfig.NumClass[1]++;
				break;
			case TYPE_NOISE:
				TrainConfig.NumClass[2]++;
				break;
			}
		}
		free( pZone );
	}
	fclose( fpTrain );
	return 0;
}

/*****************************************************************************
/*	Name:		ExtractTrainFeature
/*	Function:	Extract features on the training set
/*	Parameter:	fnTrain		-- Configure file
/*				TrainConfig	-- Configure after parsing
/*				Feature		-- Training features
/*	Return:		0 -- Succeed
/*****************************************************************************/
int ExtractTrainFeature( char *fnTrain, TRAIN_CONFIG &TrainConfig, double ***Feature)
{
	for( int i=0; i<TYPE_TOTAL_CLASS; i++ )
		TrainConfig.NumClass[i] = 0;

	FILE *fpTrain = fopen( fnTrain, "rt" );
	if( fpTrain == NULL )
		return -1;

	char buf[1000];
	//Find the beginning of the training list
	while( !feof(fpTrain) )
	{
		if( fgets( buf, 1000, fpTrain ) == NULL)
			break;
		if( strstr( buf, "---" ) != NULL )
			break;
	}

	//Read all groundtruth
	char	fnGt[MAX_PATH];
	char	fnImg[MAX_PATH];
	ZONE	*pZone;
	int		nZone;
	CImage	Img, ZoneImg;
	int		nExtract = 0;
	while( !feof(fpTrain) )
	{
		if( fgets( buf, 1000, fpTrain) == NULL )
			break;
		if( strcmp( buf, "\n" ) == 0 )
			break; 
		if( buf[strlen(buf)-1] == '\n' )
			buf[strlen(buf)-1] = '\0';
		sprintf( fnGt, "%s\\%s%s", TrainConfig.csGtDir, buf, TrainConfig.csGtSuffix );
		sprintf( fnImg, "%s\\%s%s", TrainConfig.csImgDir, buf, TrainConfig.csImgSuffix );
		if( Img.Read( fnImg ) != 0 )
		{
			printf( "Can not read image %s!\n" );
			return -1;
		}
		if( ReadZone( pZone, nZone, fnGt ) != 0 )
		{
			printf( "Can not read groundtruth %s!\n" );
			return -1;
		}

		for( i=0; i<nZone; i++ )
		{
			for( int j=0; j<TrainConfig.nTotalClass; j++ )
				if( pZone[i].nType == TrainConfig.ClassLabel[j] )
					break;
			if( j>=TrainConfig.nTotalClass )
				continue;
			if( rand() > TrainConfig.nRatio[j]*RAND_MAX )
				continue;
			CImageTool::CopyImage( Img, pZone[i].x, pZone[i].y, pZone[i].w, pZone[i].h, ZoneImg); 
			if( ::ExtractFeature( ZoneImg, Feature[j][TrainConfig.NumClass[j]] ) != 0 )
			{
				printf( "Extract feature failed!\n" );
				return -1;
			}
			TrainConfig.NumClass[j]++;
			nExtract ++;
			if( nExtract%100 == 1 )
				printf( "Extract features for zone: %d...\n", nExtract );
		}
		free( pZone );
	}
	fclose( fpTrain );
	return 0;
}

/*****************************************************************************
/*	Name:		ReadTextSamples
/*	Function:	Read samples from text file
/*	Parameter:	fn -- Sample file name
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	ReadTextSamples(char *fn, double ***&Feature, int &nTotalClass, int *NumClass, int *ClassLabel, int &nDimension)
{
	#define MAX_LINE_LEN 100000
	char buf[MAX_LINE_LEN];

	int	i, j, index, class_index ;
	char *p;
	double value;
	FILE *fp;
	BOOL bIndex;
	int nClass;
 
	fp = fopen( fn, "rt" );
	if( fp == NULL )
	{
		fprintf(stderr, "can not open %s to read!\n", fn );
		exit(1);
	}

//Get the dimension
	nDimension = 0;
	if( fgets(buf,MAX_LINE_LEN,fp) == NULL )
	{
		fprintf( stderr, "Reading error!\n" );
		exit( 1 );
	}
	p=buf;
	sscanf(p,"%d",&nClass);
	SKIP_CLASS
	bIndex = ( strchr(p, ':') != NULL );
	if( bIndex )
	{
		while(sscanf(p,"%d:%lf",&index, &value)==2)
		{
			nDimension = max(nDimension, index);
			SKIP_ELEMENT
		}
	}
	else
	{
		while(sscanf(p, "%lf", &value )==1)
		{
			nDimension++;
			while(!isspace(*p))
				++p;
			while(isspace(*p)) 
				++p;
		}
	}
	rewind(fp);	

//Get number of samples for each class
	nTotalClass = 0;
	int	nTotalSample = 0;
	while( !feof(fp) )
	{
		if( fgets(buf,MAX_LINE_LEN,fp) == NULL )
			break;
		nTotalSample ++;
		sscanf(buf,"%d",&nClass);
		for( i=0; i<nTotalClass; i++ )
			if( nClass == ClassLabel[i] )
				break;
		if( i<nTotalClass )
			NumClass[i] ++;
		else
		{
			ClassLabel[nTotalClass] = nClass;
			NumClass[nTotalClass] = 1;
			nTotalClass++;
		}
	}
	if( nTotalClass > 3 )
	{
		printf( "Too many classes!\n" );
		return -1;
	}
	rewind( fp );

//Allocate memory
	Feature = (double***)malloc(sizeof(double**)*nTotalClass );
	if( Feature == NULL )
		return -1;
	for( i=0; i<nTotalClass; i++ )
	{
		Feature[i] = (double**)malloc(sizeof(double*)*NumClass[i]);
		if( Feature[i] == NULL )
			return -1;
		for( j=0; j<NumClass[i]; j++ )
		{
			Feature[i][j] = (double*)malloc(sizeof(double)*nDimension);
			if( Feature[i][j] == NULL )
				return -1;
			ZeroMemory( Feature[i][j], sizeof(double)*nDimension);
		}
	}

//Reading samples
	ZeroMemory( NumClass, sizeof(int)*nTotalClass );
	nTotalSample = 0;
	while( fgets(buf,MAX_LINE_LEN,fp) != NULL )
 	{
		p=buf;
		sscanf(p,"%d",&nClass);
		SKIP_CLASS
		for( class_index = 0; class_index<nTotalClass; class_index++ )
			if( ClassLabel[class_index] == nClass )
				break;
		if( class_index >= nTotalClass )
			exit(1);

		bIndex = ( strchr(p, ':') != NULL );
		if( bIndex )
		{
			while(sscanf(p,"%d:%lf",&index,&value)==2)
			{
				Feature[class_index][NumClass[class_index]][index-1] = value;
				SKIP_ELEMENT;
			}
		}
		else
		{
			index = 0;
			while(sscanf(p, "%lf", &value )==1)
			{
				Feature[class_index][NumClass[class_index]][index] = value;
				index++;
				while(!isspace(*p))	++p;
				while(isspace(*p)) ++p;
			}
		}
		NumClass[class_index]++;
		nTotalSample++;
		if( nTotalSample%1000 == 0 )
			fprintf( stderr, "Reading No. %d ...\n", nTotalSample );
	}
	return 0;
}

/*****************************************************************************
/*	Name:		TrainFisher
/*	Function:	Train the Fisher classifier
/*	Parameter:	Feature		-- Training features
/*				NumClass	-- Number of training samples of each class
/*				Classifier  -- Fisher classifier
/*	Return:		0 -- Succeed
/*****************************************************************************/
int TrainFisher( double ***Feature, int *NumClass, CLASSIFIER &Classifier )
{
	int		i, j, k, m, n;
	int		nTotalClass = Classifier.nClassNum;
	int		Dimension = Classifier.nDimension;
	int		ClassLabel[10];
	memcpy( ClassLabel, Classifier.nClassLabel, sizeof(int)*nTotalClass );

//Calculate mean and variance for each class
	double **MeanFeature = (double**)malloc( sizeof(double*)*nTotalClass );
	double **VarFeature = (double**)malloc( sizeof(double*)*nTotalClass );
	double	TH = 0;
	for( i=0; i<nTotalClass; i++ )
	{
		fprintf(stderr, "Calculate mean of nClass %d samples...\n", ClassLabel[i]);
		MeanFeature[i] = (double*)malloc( Dimension*sizeof(double));
		Mean( Feature[i], NumClass[i], Dimension, MeanFeature[i] );
		fprintf(stderr, "Calculate variance matrix of nClass %d samples...\n", ClassLabel[i] );
		VarFeature[i] = (double*)malloc( Dimension*Dimension*sizeof(double));
		Var( Feature[i], NumClass[i], Dimension, VarFeature[i], MeanFeature[i] );
	}

//Calculate Fisher Efficients
	double *SumSMatrix = (double*)malloc(Dimension*Dimension*sizeof(double));
	double *Efficient = (double*)malloc( Dimension*sizeof(double) );
	k = 0;
	for( i=0; i<nTotalClass-1; i++ )
	{
		for( j=i+1; j<nTotalClass; j++ )
		{
			ZeroMemory( SumSMatrix, Dimension*Dimension*sizeof(double));
			for( m=0; m<Dimension; m++)
			{
				for( n=0; n<Dimension; n++)
					SumSMatrix[m*Dimension+n] = VarFeature[i][m*Dimension+n] + VarFeature[j][m*Dimension+n];
			}
			fprintf( stderr, "Invert the variance matrix ...\n" );
			if( Inv( SumSMatrix, Dimension ) != 0 )
			{
				fprintf(stderr, "Invert Matrix Error!" );
				exit(1);
			}
			fprintf(stderr, "Calculate Fisher Efficients...\n" );
			ZeroMemory( Efficient, Dimension*sizeof(double) );
			for( m=0; m<Dimension; m++ )
				for( n=0; n<Dimension; n++)
					Efficient[m] += SumSMatrix[m*Dimension+n]*( MeanFeature[i][n]-MeanFeature[j][n] );
			//Normalization	
			double Norm = 0;
			for( m=0; m<Dimension; m++ )
				Norm += Efficient[m]*Efficient[m];
			Norm = sqrt( Norm );

			for( m=0; m<Dimension; m++ )
				Efficient[m] /= Norm;

			double	ProjMean1, ProjMean2;
			double	ProjDelta1, ProjDelta2;

			fprintf( stderr, "Calculate mean and variance of projected samples ...\n" );
			ProjMean1 = 0;
			for( m=0; m<Dimension; m++ )
				ProjMean1 += Efficient[m]*MeanFeature[j][m];
			TH += ProjMean1;

			ProjDelta1=0;
			for( m=0; m<NumClass[j]; m++ )
			{	 
				double y = 0;
				for( n=0; n<Dimension; n++ )
					y += Efficient[n]*Feature[j][m][n];
				ProjDelta1 += (y-ProjMean1)*(y-ProjMean1);
			}
			ProjDelta1 = sqrt( ProjDelta1/(NumClass[j]-1) );

			ProjMean2 = 0;
			for( m=0; m<Dimension; m++ )
				ProjMean2 += Efficient[m]*MeanFeature[i][m];

			ProjDelta2=0;
			for( m=0; m< NumClass[i]; m++ )
			{	 
				double y = 0;
				for( n=0; n<Dimension; n++ )
					y += Efficient[n]*Feature[i][m][n];
				ProjDelta2 += (y-ProjMean2)*(y-ProjMean2);
			}
			ProjDelta2 = sqrt( ProjDelta2/(NumClass[i]-1) );
			TH = (ProjMean1+ProjMean2)/2;

			memcpy( Classifier.Eff[k], Efficient, sizeof(double)*Dimension );
			Classifier.nMean1[k] = ProjMean1;
			Classifier.nMean2[k] = ProjMean2;
			Classifier.nVar1[k] = ProjDelta1;
			Classifier.nVar2[k] = ProjDelta2;
			Classifier.nTH[k] = TH;
			Classifier.nClassLabel1[k] = ClassLabel[j];
			Classifier.nClassLabel2[k] = ClassLabel[i];
			k++;
		}
	}

//Free memory
	free( Efficient );
	for( i=0; i<nTotalClass; i++ )
		free( MeanFeature[i] );
	free( MeanFeature );
	for( i=0; i<nTotalClass; i++ )
		free( VarFeature[i] );
	free( VarFeature );
	free( SumSMatrix );

	int nTotalSample = 0;
	for( i=0; i<nTotalSample; i++ )
		nTotalSample += NumClass[i];

	printf("Total Sample: %d\n", nTotalSample );
	for( i=0; i<nTotalClass; i++ )
		fprintf(stderr, "    Class %d: %d\n", ClassLabel[i], NumClass[i] );
	return 0;	
}

