////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		GetNeighbor.cpp
// File Function:	Get neighboring words for MRF-based post processing
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PostProcess.h"
#include "define.h"
#include "tools.h"
#include "math.h"

/**************************************************************************************************
/*	Name:		GetNeighborNum
/*	Function:	Get neighoring words number
/*	Parameter:	zone	 -- Zone blocks
/*				nCurZone -- Current processing zone block
/*				nZone    -- Number of zone blocks
/*				TH_Ver	 --	Threshold for vertical distance
/*				TH_Hor   -- Threshold for horizontal distance
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetNeighborNum( ZONE *zone, int nCurZone, int nZone, int TH_Ver, int TH_Hor )
{
	int nNeighborNum = 0;
	ZONE LargeZone;
	LargeZone.x = max( 0, zone[nCurZone].x - TH_Hor );
	LargeZone.w = zone[nCurZone].x - LargeZone.x + zone[nCurZone].w + TH_Hor;
	LargeZone.y = max( 0, zone[nCurZone].y - TH_Ver );
	LargeZone.h = zone[nCurZone].y - LargeZone.y + zone[nCurZone].h + TH_Ver;

	for( int i=0; i<nZone; i++ )
	{
		if( i == nCurZone )			continue;
		if( zone[i].nType == TYPE_OTHERS )	continue;
		if( GetOverlapArea( zone[nCurZone], zone[i] ) > 0 )
			nNeighborNum ++;
	}
	return nNeighborNum;
}

/**************************************************************************************************
/*	Name:		GetLeftRightNeighbor
/*	Function:	Get left and right neighboring blocks
/*	Parameter:	zone	 -- Zone blocks
/*				nCurZone -- Current processing zone block
/*				nZone    -- Number of zone blocks
/*				LeftIndex  -- Left neigboring zone index
/*				LeftDist   -- Left distance to the nearset zone
/*				RightIndex -- Right neighboring zone index
/*				RightDist  -- Right distance to the nearest zone
/*				TH_Hor   -- Threshold for horizontal distance
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetLeftRightNeighbor( ZONE *zone, int nCurZone, int nZone, int &LeftIndex, int &LeftDist, int &RightIndex, int &RightDist, int TH_Hor )
{
	LeftDist = 10000;
	RightDist = 10000;
	LeftIndex = -1;
	RightIndex = -1;

	//Search LeftIndex and RightIndex
	for( int j=0; j<nZone; j++ )
	{
		if( j == nCurZone )	continue;
		//Filter some words
		if( zone[j].nSize == SIZE_TINY )	continue;
		if( zone[j].nType ==  TYPE_OTHERS )			continue;

		int MaxTop = max( zone[nCurZone].y, zone[j].y );
		int MinBottom = min( zone[nCurZone].y+zone[nCurZone].h, zone[j].y+zone[j].h );
		if( MaxTop >= MinBottom )//Not overlap in vertical
			continue;

		int TH_V = min( zone[nCurZone].h/3, zone[nCurZone].h/3 );
		if( MinBottom - MaxTop < TH_V )//Overlap rate is too small 
			continue;

		int MaxLeft = max( zone[nCurZone].x, zone[j].x );
		int MinRight = min( zone[nCurZone].x+zone[nCurZone].w, zone[j].x+zone[j].w );
		if( MaxLeft <= MinRight )//Overlap in horizontal
			continue;
		if( MaxLeft > MinRight+TH_Hor )//Too far away
			continue;

		int nDist = MaxLeft - MinRight + abs( zone[nCurZone].h-zone[j].h ) + 
					abs( (zone[nCurZone].y+zone[nCurZone].h)/2 - (zone[j].y+zone[j].h)/2 );
		if( zone[nCurZone].x > zone[j].x )//Left
		{
			if( LeftDist > nDist )
			{
				LeftDist = nDist;
				LeftIndex = j;
			}
		}
		else//Right
		{
			if( RightDist > nDist )
			{
				RightDist = nDist;
				RightIndex = j;
			}
		}
	}
	return 0;
}

/**************************************************************************************************
/*	Name:		GetTopBottomNeighbor
/*	Function:	Get top and bottom neighoring blocks
/*	Parameter:	zone	 -- Zone blocks
/*				nCurZone -- Current processing zone block
/*				nZone    -- Number of zone blocks
/*				TopIndex    -- Top neigboring zone index
/*				TopDist     -- Top distance to the nearset zone
/*				BottomIndex -- Bottom neighboring zone index
/*				BottomDist  -- Bottom distance to the nearest zone
/*				TH_Ver   -- Threshold for vertical distance
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetTopBottomNeighbor( ZONE *zone, int nCurZone, int nZone, int &TopIndex, int &TopDist, int &BottomIndex, int &BottomDist, int TH_Ver )
{
	TopDist = 10000;
	BottomDist = 10000;
	TopIndex = -1;
	BottomIndex = -1;

	//Search TopIndex and BottomIndex
	for( int j=0; j<nZone; j++ )
	{
		if( j == nCurZone )	
			continue;
		if( zone[j].nSize == SIZE_TINY )	
			continue;
		if( zone[j].nType == TYPE_OTHERS )
			continue;

		int MaxLeft = max( zone[nCurZone].x, zone[j].x );
		int MinRight = min( zone[nCurZone].x+zone[nCurZone].w, zone[j].x+zone[j].w );
		if( MaxLeft >= MinRight )//Not overlap in horizontal
			continue;

		int TH_H = min( zone[nCurZone].w/5, zone[nCurZone].w/5 );
		if( MinRight - MaxLeft < TH_H )//Overlap rate is too small 
			continue;

		int MaxTop = max( zone[nCurZone].y, zone[j].y );
		int MinBottom = min( zone[nCurZone].y+zone[nCurZone].h, zone[j].y+zone[j].h );
		if( MaxTop > MinBottom+TH_Ver )//Too far away
			continue;

		int nDist = abs( MaxTop - MinBottom ) + (MinRight-MaxLeft);
		if( zone[nCurZone].y > zone[j].y )//Top
		{
			if( TopDist > nDist )
			{
				TopDist = nDist;
				TopIndex = j;
			}
		}
		else//Bottom
		{
			if( BottomDist > nDist )
			{
				BottomDist = nDist;
				BottomIndex = j;
			}
		}
	}
	return 0;
}

/**************************************************************************************************
/*	Name:		GetNearestZones4Print
/*	Function:	Get nearest zone list for each zone. Used for clique of printed words.
/*	Parameter:	zone	 -- Zone blocks
/*				nZone    -- Number of zone blocks
/*				LeftIndex  -- Left neigboring zone index
/*				RightIndex -- Right neighboring zone index
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetNearestZones4Print( ZONE *zone, int nZone, int *LeftIndex, int *RightIndex )
{
	int TH_Hor = 60;
	int	LeftDist, RightDist;

	for( int i=0; i<nZone; i++ )
	{
		LeftIndex[i] = -1;
		RightIndex[i] = -1;

		GetLeftRightNeighbor( zone, i, nZone, LeftIndex[i], LeftDist, RightIndex[i], RightDist, TH_Hor );
	}
	return 0;
}

/**************************************************************************************************
/*	Name:		GetNearestZones4Noise
/*	Function:	Get nearest zone list for each zone. Used for clique of noise blocks
/*	Parameter:	zone	 -- Zone blocks
/*				nZone    -- Number of zone blocks
/*				NearIndex -- Index of nearest neighboring blocks
/*	Return:		0  -- Succeed
/*************************************************************************************************/
int GetNearestZones4Noise( ZONE *zone, int nZone, int **NearIndex )
{
	int TH_MaxDist = 10;
	int i, j, k, n;
	double nMinDist[4];
	double	Dist;
	for( i=0; i<nZone; i++ )
	{
		if( zone[i].nSize != SIZE_LARGE )
			TH_MaxDist = 5;
		else
			TH_MaxDist = 10;

		for(j=0; j<4; j++ )
		{
			NearIndex[j][i] = -1;
			nMinDist[j] = 1e+10;
		}
		for( j=0; j<nZone; j++ )
		{
			if( j==i )		continue;
			//Filter some words
			if( zone[j].nSize == SIZE_TINY )	continue;
			if( zone[j].nType == TYPE_OTHERS)				continue;

			int nMaxLeft = max( zone[i].x, zone[j].x );
			int nMinRight = min( zone[i].x+zone[i].w, zone[j].x + zone[j].w );
			int nMaxTop = max( zone[i].y, zone[j].y );
			int nMinBottom = min( zone[i].y+zone[i].h, zone[j].y+zone[j].h );
			if( nMaxLeft-nMinRight > TH_MaxDist )		continue;
			if( nMaxTop-nMinBottom > TH_MaxDist )		continue;
			Dist = max( nMaxLeft-nMinRight, nMaxTop-nMinBottom );
			for( k=0; k<4; k++ )
			{
				if( nMinDist[k] > Dist )	break;
			}
			if( k<4 )
			{
				for( n=3; n>k; n-- )
				{
					nMinDist[n] = nMinDist[n-1];
					NearIndex[n][i] = NearIndex[n-1][i];
				}
				NearIndex[k][i] = j;
				nMinDist[k] = Dist;
			}
		}
	}
	return 0;
}
