////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		GetStat.cpp
// File Function:	Get statistic information for post processing
//
//				Developed by: Yefeng Zheng
//			   First created: Jan, 2001
//			   Last modified: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PostProcess.h"
#include "define.h"
#include "tools.h"
#include "math.h"
#include "Output.h"

/*****************************************************************************
/*	Name:		InitStatMeasure
/*	Function:	Initialize statiscs information for post processing
/*	Parameter:	MRF_Model          -- Statistic information
/*	Return:		0 -- Succeed
/*****************************************************************************/
int InitStatMeasure( MRF_MODEL &MRF_Model )
{
	int nMRFSizeLevel;
	int nMRFLabel, LabelLeft, LabelRight, LabelTop, LabelBottom;

	for( nMRFSizeLevel = 0; nMRFSizeLevel<MRF_SIZE_LEVEL; nMRFSizeLevel++ )
	{
		MRF_Model.nZone[nMRFSizeLevel] = 0;
		for(  nMRFLabel = 0; nMRFLabel<TYPE_TOTAL_CLASS; nMRFLabel++ )
		{
			MRF_Model.nLabel[nMRFSizeLevel][nMRFLabel] = 0;
			for(  LabelLeft = 0; LabelLeft<TYPE_TOTAL_CLASS+1; LabelLeft++ )
				for(  LabelRight = 0; LabelRight<TYPE_TOTAL_CLASS+1; LabelRight++ )
				{
					MRF_Model.nPatternP[nMRFSizeLevel][nMRFLabel][LabelLeft][LabelRight] = 0;			
					for( LabelTop = 0; LabelTop<TYPE_TOTAL_CLASS+1; LabelTop++ )
						for( LabelBottom = 0; LabelBottom<TYPE_TOTAL_CLASS+1; LabelBottom++ )
							MRF_Model.nPatternN[nMRFSizeLevel][nMRFLabel][LabelLeft][LabelRight][LabelTop][LabelBottom] = 0;			
				}
		}
	}
	return 0;
}

/***********************************************************************************
/*	Name:		DeleteZoneType
/*	Function:	Delete some types of zones
/*	Parameter:	zone  -- Zone
/*				nZone -- Number of zones
/*				nType -- Zone type to be deleted
/*	Return:		0 -- Succeed
/***********************************************************************************/
int	DeleteZoneType( ZONE *zone, int &nZone, int nType )
{
	int i=0;
	while( i<nZone )
	{
		if( zone[i].nType == nType )
		{
			for( int j=i; j<nZone-1; j++ )
				zone[j] = zone[j+1];
			nZone--;
		}
		else
			i++;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		GetStatistics
/*	Function:	Get statiscs information for post processing
/*	Parameter:	fnTruth       -- Ground truth file
/*				MRF_Model          -- Statistic information
/*				nNeighborType -- Clique type
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	GetStatistics( char *fnTruth, MRF_MODEL &MRF_Model, int nNeighborType )
{	
	ZONE	*GtZone=NULL;
	int		nGtZone;
	int		i, j;

	if( ReadZone( GtZone, nGtZone, fnTruth ) != 0 )
		return -1;

	DeleteZoneType( GtZone, nGtZone, TYPE_OTHERS );
	if( MRF_Model.nClassNum == 2 )
		DeleteZoneType( GtZone, nGtZone, TYPE_HANDWRITING );

	SetZoneSize( GtZone, nGtZone, 20, 20 );
	
	int		*LeftIndex, *RightIndex;
	int		**NearIndex;
	int		NearLabel[4];

	LeftIndex = (int*)malloc( nGtZone*sizeof(int) );
	RightIndex = (int*)malloc( nGtZone*sizeof(int) );

	if( (nNeighborType & CLIQUE_PRINT) != 0 )
		GetNearestZones4Print( GtZone, nGtZone, LeftIndex, RightIndex);
	if( (nNeighborType & CLIQUE_NOISE) != 0 )
	{
		NearIndex = (int**)malloc( sizeof(int*) * 4 );
		for( i=0; i<4; i++ )
			NearIndex[i] = (int*)malloc( sizeof(int) * nGtZone );

		GetNearestZones4Noise( GtZone, nGtZone, NearIndex );
	}

	for( i=0; i<nGtZone; i++ )
	{
		int nMRFSizeLevel, nMRFLabel, LabelLeft, LabelRight;
		nMRFSizeLevel = GetMRFSizeLevel( GtZone[i].nSize );
		MRF_Model.nZone[nMRFSizeLevel] ++;

		nMRFLabel = ZoneType2MRFLabel( GtZone[i].nType );
		MRF_Model.nLabel[nMRFSizeLevel][nMRFLabel] ++;

		if( LeftIndex[i] >= 0 && LeftIndex[i] < nGtZone )
			LabelLeft = ZoneType2MRFLabel( GtZone[LeftIndex[i]].nType );
		else
			LabelLeft = TYPE_TOTAL_CLASS;

		if( RightIndex[i] >= 0 && RightIndex[i] < nGtZone )
			LabelRight = ZoneType2MRFLabel( GtZone[RightIndex[i]].nType );
		else
			LabelRight = TYPE_TOTAL_CLASS;

		if( (nNeighborType & CLIQUE_NOISE) != 0 )
		{
			for( j=0; j<4; j++ )
			{
				if( NearIndex[j][i] >= 0 && NearIndex[j][i] < nGtZone )
					NearLabel[j] = ZoneType2MRFLabel( GtZone[NearIndex[j][i]].nType );
				else
					NearLabel[j] = TYPE_TOTAL_CLASS;
			}
			MRF_Model.nPatternN[nMRFSizeLevel][nMRFLabel][NearLabel[0]][NearLabel[1]][NearLabel[2]][NearLabel[3]]++;
		}
		if( (nNeighborType & CLIQUE_PRINT) != 0 )
			MRF_Model.nPatternP[nMRFSizeLevel][nMRFLabel][LabelLeft][LabelRight]++;
	}
	free( LeftIndex );
	free( RightIndex );
	if( (nNeighborType & CLIQUE_NOISE) != 0 )
	{
		for( i=0; i<4; i++ )
			free( NearIndex[i] );
		free( NearIndex );
	}
	free( GtZone );
	return 0;
}