////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		LayerSeparate.cpp
// File Function:	API for layer separation
//
//				Developed by: Yefeng Zheng
//			   First created: Nov. 2003
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ImageObj.h"
#include "ImageTool.h"
#include "Tools.h"
#include "TextIDAPI.h"
#include "WordSegment.h"
#include "Classify.h"
#include "PostProcess.h"
#include "ExtractFeature.h"
#include "Output.h"

CWordSegment	WordSegment;	//Word segmentation class
CImage			m_Img;			//Image class

CTILayerSeparate::CTILayerSeparate()
{ 
	m_pWord = NULL;
	m_nWord = 0;
	InitFisherModel( m_Classifier );
	m_MRF_Model.nPatternP = NULL;
	m_MRF_Model.nPatternN = NULL;
	return; 
}

CTILayerSeparate::~CTILayerSeparate()
{
	if( m_pWord != NULL )
		free( m_pWord );
	m_pWord = NULL;
	m_nWord = 0;

	FreeFisherModel( m_Classifier );
	FreeMRFModel( m_MRF_Model );
	return;
}

/******************************************************************************
/*	Name:		TISetImage
/*	Function:	Set image
/*	Parameter:	fnImg -- Image file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetImage(char *fnImg)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	if( m_Img.Read( fnImg ) != 0 )
	{
		printf( "Read image %s failed!\n", fnImg );
		return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		TISetImage
/*	Function:	Set image from memory
/*	Parameter:	pImgData	-- Image data
/*				w			-- Image width
/*				h			-- Image height
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetImage( BYTE *pImgData, int w, int h )
{
	if( m_Img.Initialize( pImgData, w, (w+7)/8, h, DIB_1BIT, FALSE ) !=0 )
	{
		printf( "Can not initialize the image!\n" );
		return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		TISetMRFModel
/*	Function:	Set MRF model for post-processing
/*	Parameter:	fnMRFModel -- MRF model
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetMRFModel(char *fnMRFModel)
{
	FILE *fpMRFModel = fopen( fnMRFModel, "rt" );
	if( fpMRFModel == NULL )
	{
		printf( "Can not open MRF model: %s to read!\n" );
		return -1;
	}

	AllocateMRFModel( m_MRF_Model );
	m_MRF_Model.nCliqueType = CLIQUE_PRINT;
	LoadMRFModel( m_MRF_Model, fpMRFModel, CLIQUE_PRINT );
	rewind( fpMRFModel );
	LoadMRFModel( m_MRF_Model, fpMRFModel, CLIQUE_NOISE );
	m_MRF_Model.nCliqueType |= CLIQUE_NOISE;

	SetMRFParamter( 3 );
	m_MRF_Model.nClassNum = 3;
	m_MRF_Model.nClassLabel[0] = TYPE_PRINT_TEXT;
	m_MRF_Model.nClassLabel[1] = TYPE_HANDWRITING;
	m_MRF_Model.nClassLabel[2] = TYPE_NOISE;

	fclose( fpMRFModel );
	return 0;
}

/******************************************************************************
/*	Name:		TISetClassifier
/*	Function:	Set classifier
/*	Parameter:	fnClassifier -- Trained classifier file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetClassifier( char *fnClassifier )
{
	return LoadFisherModel( fnClassifier, m_Classifier );
}

/******************************************************************************
/*	Name:		TIWordSegmentation
/*	Function:	Word segmentation
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIWordSegmentation( )
{
	printf( "\tSegmentation into words...\n" );
	WordSegment.SegmentWord(m_Img);

	if( m_pWord != NULL )
		free( m_pWord );

	m_nWord = WordSegment.m_nWord;
	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*m_nWord );
	for( int i=0; i<WordSegment.m_nWord; i++ )
	{
		m_pWord[i].x = WordSegment.m_pWord[i].x;
		m_pWord[i].y = WordSegment.m_pWord[i].y;
		m_pWord[i].w = WordSegment.m_pWord[i].w;
		m_pWord[i].h = WordSegment.m_pWord[i].h;
		m_pWord[i].nType = WordSegment.m_pWord[i].nType;
		m_pWord[i].nIndex = i;
	}
//	TIOutputZone( "C:\\Layered_Out.zone" );
	return 0;
}

/******************************************************************************
/*	Name:		TIClassification
/*	Function:	Classification
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIClassification( )
{
	int i;
	CImage	WordImg;
	double	Feature[200];
	char	fnFeature[MAX_PATH];
	sprintf( fnFeature, "C:\\Feature.txt" );

	printf( "\tClassification " );
	double nScale = 20.0/WordSegment.est_para.cw;
	for( i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_OTHERS )//Filtered as noise
			continue;
		if( m_pWord[i].nType == TYPE_NOISE )//Filtered as noise
			continue;

		if( m_pWord[i].nIndex < 0 )
			CImageTool::CopyImage( m_Img, m_pWord[i].x, m_pWord[i].y, m_pWord[i].w, m_pWord[i].h, WordImg); 
		else
			WordSegment.OutputWordImage( WordImg, i );


		if( nScale < 0.7 || nScale > 1.5 )
		{
			BYTE *pWordImg = WordImg.LockRawImg();
			int	 nWordWb = WordImg.GetLineSizeInBytes();
			int  nWordWidth = WordImg.GetWidth();
			int  nWordHeight = WordImg.GetHeight();
			int	 nWordWidth2 = (int)(nWordWidth*nScale);
			int  nWordHeight2 = (int)(nWordHeight*nScale);
			nWordWidth2 = min( 512, nWordWidth2 );
			nWordHeight2 = min( 512, nWordHeight2 );

			CImage	WordImgScaled;
			WordImgScaled.Initialize( nWordWidth2, nWordHeight2, 1, 1 );
			int  nWordWb2 = WordImgScaled.GetLineSizeInBytes();
			BYTE	*pWordImgScaled = WordImgScaled.LockRawImg();
			CImageTool::Zoom( pWordImg, pWordImgScaled, nWordWb, nWordHeight, nWordWb2, nWordHeight2 );
			
			WordImg.UnlockRawImg();
			WordImgScaled.UnlockRawImg();
			WordImg.Initialize( WordImgScaled );
		}

		if( m_pWord[i].nIndex < 0 )
		{
			if( ExtractFeature( WordImg, Feature ) != 0 )
			{
				printf( "Extract feature failed!\n" );
				continue;
			}
		}
		else
		{
			if( ExtractFeature( WordImg, WordSegment, m_pWord[i].nIndex, Feature ) != 0 )
			{
				printf( "Extract feature failed!\n" );
				continue;
			}
		}
//		OutputFeature( fnFeature, Feature, 140 );
		FisherClassify( Feature, m_pWord[i], m_Classifier );
		if( i%50 == 0 )
			printf( "." );
	}
	printf( "\n" );
	return 0;
}

/******************************************************************************
/*	Name:		TIMRFPostProcessing
/*	Function:	MRF-based post processing
/*	Parameter:	NULL
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIMRFPostProcessing( )
{
	printf( "\tPost processing...\n" );
	ZONE *pZone;
	pZone = (ZONE*)malloc( sizeof(ZONE)*m_nWord );

	// Do not processing blocks labeled as others
	int	nZone = 0;
	for( int i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_OTHERS )
			continue;
		pZone[nZone].x		= m_pWord[i].x;
		pZone[nZone].y		= m_pWord[i].y;
		pZone[nZone].w		= m_pWord[i].w;
		pZone[nZone].h		= m_pWord[i].h;
		pZone[nZone].nType	= m_pWord[i].nType;
		pZone[nZone].nConf	= m_pWord[i].nConf;
		for( int j=0; j<TYPE_TOTAL_CLASS; j++ )
			pZone[nZone].nClassConf[j] = m_pWord[i].nClassConf[j];
		nZone++;
	}

	// MRF based post-processing
	SetZoneSize( pZone, nZone, WordSegment.est_para.cw, WordSegment.est_para.ch );
	PostProcess( pZone, nZone, m_MRF_Model );

	// Get the results back
	nZone=0;
	for( i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_OTHERS )
			continue;
		m_pWord[i].x		= pZone[nZone].x;
		m_pWord[i].y		= pZone[nZone].y;
		m_pWord[i].w		= pZone[nZone].w;
		m_pWord[i].h		= pZone[nZone].h;
		m_pWord[i].nType	= pZone[nZone].nType;
		nZone++;
	}	
	for( i=0; i<m_nWord; i++ )
	{
		int index = m_pWord[i].nIndex;
		if( index < 0 || index >= WordSegment.m_nWord )
			continue;
		WordSegment.m_pWord[index].nType = m_pWord[i].nType;
	}
	free( pZone );

	return 0;
}

/******************************************************************************
/*	Name:		TISetWord
/*	Function:	This function provides a way for users can use his/her own module
/*				to perform segmentation, classification of logo, figure, etc., and 
/*				then fill information to the structure of WORD_ZONE.
/*	Parameter:	pWord   -- Segmented word zone
/*				nWord   -- Number of word zone
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetWord(WORD_ZONE *pWord, int nWord)
{
	if( m_pWord != NULL )
		free( m_pWord );
	m_nWord = 0;

	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*nWord );
	for( int i=0; i<nWord; i++ )
	{
		m_pWord[i] = pWord[i];
		m_pWord[i].nIndex = -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		TISetWord
/*	Function:	This function provides a way for users can use his/her own module
/*				to perform segmentation, classification of logo, figure, etc., and 
/*				then fill information to the structure of WORD_ZONE.
/*				Similar to the previous function, but the custom's segmentation
/*				result is stored in ZONE file.
/*	Parameter:	fnSegment -- File name of the segmenation result
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISetWord(char *fnCustom)
{
	if( m_pWord != NULL )
		free( m_pWord );
	m_nWord = 0;

	ZONE *pZone;
	int  nZone;
	ReadZone( pZone, nZone, fnCustom );
	m_nWord = nZone;
	m_pWord = (WORD_ZONE*)malloc( sizeof(WORD_ZONE)*nZone);
	for(int i=0; i<nZone; i++ )
	{
		m_pWord[i].x = pZone[i].x;
		m_pWord[i].y = pZone[i].y;
		m_pWord[i].w = pZone[i].w;
		m_pWord[i].h = pZone[i].h;
		m_pWord[i].nType = pZone[i].nType;
	}
	free( pZone );
	return 0;
}

/******************************************************************************
/*	Name:		TIGetWord
/*	Function:	Get the segmented and classified word zones.
/*	Parameter:	pWord   -- Segmented word zone
/*				nWord   -- Number of word zone
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIGetWord(WORD_ZONE *&pWord, int &nWord )
{
	pWord = m_pWord;
	nWord = m_nWord;
	return 0;
}

/******************************************************************************
/*	Name:		TIOutputZone
/*	Function:	Output result to ZONE file
/*	Parameter:	fnZone -- Output file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIOutputZone( char *fnZone )
{
	FILE *fpZone = fopen( fnZone, "wt" );
	if( fpZone == NULL )
	{
		printf( "Can not open %s to output result!\n" );
		return -1;
	}
	fprintf( fpZone, "Total Zones: %d\n", m_nWord );
	for( int i=0; i<m_nWord; i++ )
	{
		if( m_pWord[i].nType == TYPE_PRINT_TEXT || m_pWord[i].nType == TYPE_HANDWRITING )
			fprintf( fpZone, "%5d %5d %5d %5d    %d ", m_pWord[i].x, m_pWord[i].y, m_pWord[i].w, m_pWord[i].h, m_pWord[i].nType );
		else
			fprintf( fpZone, "%5d %5d %5d %5d    %d ", m_pWord[i].x, m_pWord[i].y, m_pWord[i].w, m_pWord[i].h, TYPE_NOISE );
		fprintf( fpZone, "\n" );
	}
	fclose( fpZone);
	return 0;
}

/******************************************************************************
/*	Name:		TIGetLayer
/*	Function:	Get the specified content to an image
/*	Parameter:	pImgData	-- Image of the layer
/*				w			-- Image width
/*				h			-- Image height
/*				nType       -- Layer type, 1 -- Printed text, 2 -- Handwriting, 3 -- Noise
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TIGetLayer( BYTE *&pImgData, int &w, int &h, int nType )
{
	CImage	ImgOut;
	w = m_Img.GetWidth();
	h = m_Img.GetHeight();
	ImgOut.Initialize( w, h, 1, 1 );
	switch( nType )
	{
		case TYPE_PRINT_TEXT:
			WordSegment.OutputImage( ImgOut, TYPE_PRINT_TEXT );
			break;
		case TYPE_HANDWRITING:
			WordSegment.OutputImage( ImgOut, TYPE_HANDWRITING );
			break;
		default:
			WordSegment.OutputImage( ImgOut, TYPE_NOISE );
			break;
	}
	int wb = ImgOut.GetLineSizeInBytes();
	BYTE *pImg = ImgOut.LockRawImg();
	int ww = (w+7)/8;
	pImgData = (BYTE*)malloc( sizeof(BYTE)*ww*h);
	ZeroMemory( pImgData, sizeof(BYTE)*ww*h);
	for( int i=0; i<h; i++ )
		memcpy( pImgData+i*ww, pImg+(h-1-i)*wb, ww );

	return 0;
}

/******************************************************************************
/*	Name:		TISaveLayer
/*	Function:	Output the specified content to an image
/*	Parameter:	fnLayerImg	-- Image file name of the image
/*				nType		-- Layer type
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CTILayerSeparate::TISaveLayer( char *fnLayerImg, int nType )
{
	CImage	ImgOut;
	int w = m_Img.GetWidth();
	int h = m_Img.GetHeight();
	ImgOut.Initialize( w, h, 1, 1 );

	// Perservation the resolution information
	long	nXPelsPerInch, nYPelsPerInch;
	m_Img.GetResolution( nXPelsPerInch, nYPelsPerInch );
	ImgOut.SetResolution( nXPelsPerInch, nYPelsPerInch );

	switch( nType )
	{
		case TYPE_PRINT_TEXT:
			WordSegment.OutputImage( ImgOut, TYPE_PRINT_TEXT );
			break;
		case TYPE_HANDWRITING:
			WordSegment.OutputImage( ImgOut, TYPE_HANDWRITING );
			break;
		default:
			WordSegment.OutputImage( ImgOut, TYPE_NOISE );
			break;
	}
	ImgOut.Write( fnLayerImg );
	return 0;
}
