////////////////////////////////////////////////////////////////////////////////////////////////
// File NAME:		PostProcessSpecialCase.cpp
// File Function:	Process some special cases, such as isolated blocks or too high blocks
//
//				Developed by: Zheng Yefeng
//			   First created: June, 2002
//			University of Maryland, College Park
//					All Right Reserved
///////////////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PostProcess.h"
#include "define.h"
#include "tools.h"
#include "math.h"

/*****************************************************************************
/*	Name:		GetOverlapArea
/*	Function:	Get overlap area of two blocks
/*	Parameter:	zone1 -- Block one
/*				zone2 -- Block two
/*	Return:		Overlap area
/*****************************************************************************/
int GetOverlapArea( ZONE zone1, ZONE zone2 )
{
	int nMaxTop = max( zone1.y, zone2.y );
	int nMinBottom = min( zone1.y+zone1.h, zone2.y+zone2.h );
	if( nMaxTop >= nMinBottom )	
		return 0;

	int nMaxLeft = max( zone1.x, zone2.x );
	int nMinRight = min( zone1.x+zone1.w, zone2.x+zone2.w );
	if( nMaxLeft >= nMinRight )
		return 0;
	return (nMinBottom-nMaxTop)*(nMinRight-nMaxLeft);
}

/*****************************************************************************
/*	Name:		ProcessTooHighZone
/*	Function:	Process too high zones
/*	Parameter:	zone  -- Blocks
/*				nZone -- Number of blocks
/*	Return:		0 -- Succeed
/*****************************************************************************/
int ProcessTooHighZone( ZONE *zone, int nZone )
{
	double H_W_Print = 2;
	double H_W_Hand = 5;
	for( int i=0; i<nZone; i++ )
	{
		if( zone[i].nType == TYPE_NOISE )	continue;
		if( zone[i].h < 55 )				continue;//Too short
		if( zone[i].nType == TYPE_PRINT_TEXT && 1.0*zone[i].h/zone[i].w >= H_W_Print )
			zone[i].nType = TYPE_NOISE;
		if( zone[i].nType == TYPE_HANDWRITING && 1.0*zone[i].h/zone[i].w >= H_W_Hand )
			zone[i].nType = TYPE_NOISE;
	}
	return 0;
}

/*****************************************************************************
/*	Name:		ProcessIsolatedZone
/*	Function:	Process isolated zones
/*	Parameter:	zone		-- Blocks
/*				nZone		-- Number of blocks
/*				LeftIndex	-- Left neighboring block index for Cp clique
/*				RightIndex	-- Right neighboring block index for Cp clique
/*				NearIndex   -- Nearest neighboring block index for Cn clique
/*	Return:		0 -- Succeed
/*****************************************************************************/
int	ProcessIsolatedZone( ZONE *zone, int nZone, int *LeftIndex, int *RightIndex, int ** NearIndex )
{
	int i, j;
	for( i=0; i<nZone; i++ )
	{
		if( zone[i].nType == TYPE_OTHERS )		continue;
		if( LeftIndex[i] != -1 || RightIndex[i] != -1 )
			continue;//Find no neighboring nodes
		if( NearIndex[0][i] != -1 )
			continue;
		if( zone[i].nSize != SIZE_SMALL )	continue;//Do not process large nodes
		if( NeedProcess( zone[i] ) == FALSE )	continue;

		for( j=0; j<nZone; j++ )
		{
			if( j == i )					continue;
			if( zone[j].nType == TYPE_OTHERS )		continue;

			int nOverlap = GetOverlapArea( zone[i], zone[j] );
			if( nOverlap > 0 )
			{
				int nNeighborNum = GetNeighborNum( zone, i, nZone, 40, 60 ); 
				if( nNeighborNum <= 1 || nOverlap >= zone[i].w*zone[i].h*0.3 )
				{//Overlap with other nodes
					if( zone[i].nType != zone[j].nType )
					{
						zone[i].nType = zone[j].nType;
						if( zone[j].nType == TYPE_OTHERS )
							printf( "Here is a bug!\n" );
					}
					break;
				}
			}
		}
		if( j >= nZone )//No overlap with other nodes
		{
			BOOL	bNoise = TRUE;
			int nLeftNode, nLeftDist, nRightNode, nRightDist;
			int nTopNode, nTopDist, nBottomNode, nBottomDist;
			int hDist;
			if( zone[i].nType == TYPE_PRINT_TEXT )
			{
				GetLeftRightNeighbor( zone, i,nZone, nLeftNode, nLeftDist, nRightNode, nRightDist, 120 );
				if( nLeftNode != -1 )
				{
					if( zone[nLeftNode].nType == zone[i].nType )
					{
						hDist = abs( zone[i].h-zone[nLeftNode].h ) + abs( (zone[i].y+zone[i].h)/2 - (zone[nLeftNode].y+zone[nLeftNode].h)/2 );
						if( hDist <= 20 )	bNoise = FALSE;
					}
				}
				if( nRightNode != -1 )
				{
					if( zone[nRightNode].nType == zone[i].nType )
					{
						hDist = abs( zone[i].h-zone[nRightNode].h ) + abs( (zone[i].y+zone[i].h)/2 - (zone[nRightNode].y+zone[nRightNode].h)/2 );
						if( hDist <= 20 )	bNoise = FALSE;
					}
				}
			}
			if( zone[i].nType == TYPE_HANDWRITING )
			{
				GetLeftRightNeighbor( zone, i, nZone, nLeftNode, nLeftDist, nRightNode, nRightDist, 120 );
				GetTopBottomNeighbor( zone, i, nZone, nTopNode, nTopDist, nBottomNode, nBottomDist, 120 );
				int nHandNeighbor = 0;
				if( nLeftNode != -1 )
				{
					if( zone[nLeftNode].nType == zone[i].nType )	nHandNeighbor ++;
					else											nHandNeighbor --;
				}
				if( nRightNode != -1 )
				{
					if( zone[nRightNode].nType == zone[i].nType )	nHandNeighbor ++;
					else											nHandNeighbor --;
				}
				if( nTopNode != -1 )
				{
					if( zone[nTopNode].nType == zone[i].nType )		nHandNeighbor ++;
					else											nHandNeighbor --;
				}
				if( nBottomNode != -1 )
				{
					if( zone[nBottomNode].nType == zone[i].nType )	nHandNeighbor ++;
					else											nHandNeighbor --;
				}
				if( nHandNeighbor >= 1 )	bNoise = FALSE;
			}
			if( bNoise == TRUE && zone[i].nType != TYPE_NOISE )
				zone[i].nType = TYPE_NOISE;
		}
	}
	return 0;
}
