/////////////////////////////////////////////////////////////////////////
// File NAME:		ZoneSeg.cpp
// File Function:	API for zone segmentation
//
//				Developed by:Yefeng Zheng
//					All Right Reserved
/////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ZoneSegAPI.h"
#include "pset.h"
#include "ImageObj.h"

CImage	m_Img;

CZSZoneSeg::CZSZoneSeg()
{
	m_pZone = NULL;
	m_nZone = 0;
}

CZSZoneSeg::~CZSZoneSeg()
{
	if( m_pZone != NULL )
		free( m_pZone );
}

/******************************************************************************
/*	Name:		ZSSetImage
/*	Function:	Set image from file
/*	Parameter:	fnImg -- Image file name
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSSetImage( char *fnImg )
{
	if( m_Img.Read( fnImg ) != 0 )
	{
		printf( "Can not open %s image!\n", fnImg );
		return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		ZSSetImage
/*	Function:	Set image from memory
/*	Parameter:	pImgData	-- Image data
/*				w			-- Image width
/*				h			-- Image height
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSSetImage( BYTE *pImgData, int w, int h )
{
	if( m_Img.Initialize( pImgData, w, (w+7)/8, h, DIB_1BIT, FALSE ) !=0 )
	{
		printf( "Can not initialize the image!\n" );
		return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		ZSDocstrum
/*	Function:	Zone segmentation with the Docstrum algorithm
/*	Parameter:	para -- Parameters for the Docstrum algorithm
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSDocstrum( DOCSTRUM_PARAM *para )
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image first!\n" );
		return -1;
	}
	if( m_pZone != NULL )
		free( m_pZone );

	if( para == NULL )
	{
		DOCSTRUM_PARAM DefaultPara;
		DefaultPara.knn = 8;
		DefaultPara.lsize = 2;
		DefaultPara.hsize = 500;
		DefaultPara.hatol = 30;
		DefaultPara.vatol = 30;
		DefaultPara.scr = (float)0.4;
		DefaultPara.tcc = (float)2.578;
		DefaultPara.par = (float)2.345;
		DefaultPara.per = (float)0.6;
		DefaultPara.dis = 9;
		if( DocstrumFunction(m_Img, &DefaultPara, m_pZone, m_nZone ) != 0 )
			return -1;
	}
	else 
	{
		if( DocstrumFunction(m_Img, para, m_pZone, m_nZone ) != 0 )
			return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		ZSVoronoi
/*	Function:	Zone segmentation with the Voronoi algorithm
/*	Parameter:	para -- Parameters for the Voronoi algorithm
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSVoronoi( VORONOI_PARAM *para )
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image first!\n" );
		return -1;
	}
	if( m_pZone != NULL )
		free( m_pZone );

	if( para == NULL )
	{
		VORONOI_PARAM DefaultPara;
		DefaultPara.sr = 6;
		DefaultPara.nm = 6;
		DefaultPara.fr = (float)0.083;
		DefaultPara.ta = 200;

		DefaultPara.sw = 2;
		DefaultPara.Cw = 500;
		DefaultPara.Ch = 500;
		DefaultPara.Cr = 5;
		DefaultPara.Az = 50;
		DefaultPara.Al = 40000;
		DefaultPara.Br = 4;

		if( VoronoiFunction(m_Img, &DefaultPara, m_pZone, m_nZone) != 0 )
			return -1;
	}
	else
	{
		if( VoronoiFunction(m_Img, para, m_pZone, m_nZone) != 0 )
			return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		ZSXycut
/*	Function:	Zone segmentation with the XY-Cut algorithm
/*	Parameter:	para -- Parameters for the XY-Cut algorithm
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSXycut( XYCUT_PARAM *para )
{
	if( m_Img.GethImg() == NULL )
	{
		printf( "Please set image first!\n" );
		return -1;
	}
	if( m_pZone != NULL )
		free( m_pZone );

	if( para == NULL )
	{
		XYCUT_PARAM DefaultPara;
		DefaultPara.Tcx = 90;
		DefaultPara.Tcy = 80;
		DefaultPara.Tnx = 11;
		DefaultPara.Tny = 8;
		if( XycutFunction( m_Img, &DefaultPara, m_pZone, m_nZone ) != 0 )
			return -1;
	}
	else
	{
		if( XycutFunction( m_Img, para, m_pZone, m_nZone ) != 0 )
			return -1;
	}
	return 0;
}

/******************************************************************************
/*	Name:		ZSGetZoneSeg
/*	Function:	Get zone segmentation results
/*	Parameter:	pZone   -- Segmented zones (memory is allocated in function)
/*				nZone   -- Number of segmented zones
/*	Return:		0 -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSGetZoneSeg( ZONE *&pZone, int &nZone )
{
	nZone = m_nZone;
	pZone = (ZONE*)malloc( sizeof(ZONE)*m_nZone );
	if( pZone == NULL )
		return -1;
	for( int i=0; i<m_nZone; i++ )
		pZone[i] = m_pZone[i];
	return 0;
}

/******************************************************************************
/*	Name:		ZSOutputZone
/*	Function:	Output result to ZONE file
/*	Parameter:	fnZone -- Output file name
/*	Return:		0  -- Succeed
/*****************************************************************************/
int CZSZoneSeg::ZSOutputZone( char *fnZone )
{
	FILE *fpZone = fopen( fnZone, "wt" );
	if( fpZone == NULL )
	{
		printf( "Can not open %s to output result!\n" );
		return -1;
	}
	fprintf( fpZone, "Total Zones: %d\n", m_nZone );
	for( int i=0; i<m_nZone; i++ )
		fprintf( fpZone, "%5d %5d %5d %5d\n", m_pZone[i].x, m_pZone[i].y, m_pZone[i].w, m_pZone[i].h );
	fclose( fpZone);
	return 0;
}